FUNCTION delta_analytic()

! Computes analytical value of the phase shift for the square-well potential.

USE jm_constants
USE jm_globals, ONLY: scheme, potential, shift, projectile, v_light,  &
                      delta_analytic_error

IMPLICIT none

REAL(KIND=dp) :: x1(2,2), x2(2,2), y1(2,2), y2(2,2), il1(2,2), il2(2,2)
REAL(KIND=dp) :: ld, ka, kb, k1a, k1b
REAL(KIND=dp) :: sphbess, sphneum, sphbess_prim, sphneum_prim
REAL(KIND=dp) :: tan_delta_analytic, delta_analytic
REAL(KIND=dp) :: jka, jk1a, jpka, jkb, jpkb, jk1b, jpk1b, jpk1a
REAL(KIND=dp) :: nk1a, npk1a, nk1b, nkb, npk1b, npkb
REAL(KIND=dp) :: eps, epspr, k, kpr


IF (potential%type/='well' .OR. projectile%e<potential%v0) THEN
  delta_analytic_error = .true.
  RETURN
ENDIF

k = projectile%k 

SELECT CASE (scheme)
CASE('relativistic')

  kpr   = sqrt( (projectile%e-potential%v0) * ((projectile%e-potential%v0) + &
                two * v_light**2 ) ) / v_light
  eps   = sqrt( projectile%e / (projectile%e + two * v_light**2) )
  epspr = sqrt( (projectile%e-potential%v0) / ((projectile%e-potential%v0) + &
                two * v_light**2) )
  
  CALL formy (y1, potential%b, k,   eps)
  CALL formx (x1, potential%b, kpr, epspr)
  CALL formy (y2, potential%a, kpr, epspr)
  CALL formx (x2, potential%a, k,   eps)

  IF (delta_analytic_error) RETURN
  
  il1 = matmul(y2,x2)
  il2 = matmul(x1,il1)
  il1 = matmul(y1,il2)

  tan_delta_analytic = il1(2,1) / il1(1,1)
  
CASE('non-relativistic')

  kpr = sqrt( two * (projectile%e - potential%v0 ))

  ka  = k   * potential%a  ; kb  = k   * potential%b
  k1a = kpr * potential%a  ; k1b = kpr * potential%b

  ld = DBLE(projectile%l)

  IF (potential%a /= zero .AND. potential%b /= zero ) THEN

    jka  = sphbess(ld,ka)   ; jpka  = sphbess_prim(ld,ka)
    nk1a = sphneum(ld,k1a)  ; npk1a = sphneum_prim(ld,k1a)
    jk1a = sphbess(ld,k1a)  ; jpk1a = sphbess_prim(ld,k1a) 
    jk1b = sphbess(ld,k1b)  ; jpkb  = sphbess_prim(ld,kb)
    jkb  = sphbess(ld,kb)   ; npk1b = sphneum_prim(ld,k1b)
    nkb  = sphneum(ld,kb)   ; jpk1b = sphbess_prim(ld,k1b)
    nk1b = sphneum(ld,k1b)  ; npkb  = sphneum_prim(ld,kb)

    IF (delta_analytic_error) RETURN

    tan_delta_analytic =                                          & 
      ( k*jpkb*nk1b*(kpr*jka*jpk1a - k*jk1a*jpka) +               &
        k*jk1b*jpkb*(jpka*k*nk1a - kpr*jka*npk1a) +               &
        kpr*jkb*(jka*jpk1b*kpr*npk1a - k*jpk1b*jpka*nk1a +        &
        k*jk1a*jpka*npk1b - kpr*jka*jpk1a*npk1b)             ) /  &
      ( k*npkb*(k*jk1b*jpka*nk1a - k*jk1a*jpka*nk1b +             &
        kpr*jka*jpk1a*nk1b - kpr*jk1b*jka*npk1a) -                &
        kpr*nkb*(k*jpk1b*jpka*nk1a - kpr*jka*jpk1b*npk1a -        & 
        k*jk1a*jpka*npk1b + kpr*jka*jpk1a*npk1b)             )                

  ELSE IF ( potential%b /= zero ) THEN

    jk1b = sphbess(ld,k1b)  ; jpkb  = sphbess_prim(ld,kb)
    jkb  = sphbess(ld,kb)   ; npkb  = sphneum_prim(ld,kb)
    nkb  = sphneum(ld,kb)   ; jpk1b = sphbess_prim(ld,k1b)

    IF (delta_analytic_error) RETURN

    tan_delta_analytic = (k*jpkb*jk1b - kpr*jpk1b*jkb)   /  &
                         (k*npkb*jk1b - kpr*jpk1b*nkb)

  ENDIF

END SELECT

delta_analytic = atan(tan_delta_analytic)

IF (shift .AND. delta_analytic < zero) delta_analytic = delta_analytic + pi


RETURN
END FUNCTION delta_analytic

!***********************************************************************

SUBROUTINE formx(x,r,k,eps)
USE jm_constants
USE jm_globals, ONLY: projectile

IMPLICIT none

REAL(KIND=dp), INTENT(in)  :: r, k, eps
REAL(KIND=dp), INTENT(out) :: x(2,2)
REAL(KIND=dp)              :: kr, ld
REAL(KIND=dp)              :: sphneum, sphbess


kr = k * r
ld = DBLE(projectile%l)

x(1,1) =  sphbess(ld, kr)
x(1,2) = -sphneum(ld, kr)

IF (projectile%kappa > 0) THEN
  x(2,1) =  eps * sphbess(ld-one, kr)
  x(2,2) = -eps * sphneum(ld-one, kr)
ELSE
  x(2,1) = -eps * sphbess(ld+one, kr)
  x(2,2) =  eps * sphneum(ld+one, kr)
ENDIF

RETURN
END SUBROUTINE formx

!***********************************************************************

SUBROUTINE formy(y,r,k,eps)
USE jm_constants
USE jm_globals, ONLY: projectile

IMPLICIT none

REAL(KIND=dp), INTENT(in)  :: r, k, eps
REAL(KIND=dp), INTENT(out) :: y(2,2)

REAL(KIND=dp)              :: kr, ld
REAL(KIND=dp)              :: sphneum, sphbess


kr = k * r
ld = DBLE(projectile%l)

y(1,2) = sphneum(ld, kr)
y(2,2) = sphbess(ld, kr)

IF (projectile%kappa > 0) THEN
  y(1,1) = -eps * sphneum(ld-one, kr)
  y(2,1) = -eps * sphbess(ld-one, kr)
ELSE
  y(1,1) =  eps * sphneum(ld+one, kr)
  y(2,1) =  eps * sphbess(ld+one, kr)
ENDIF

RETURN
END SUBROUTINE formy

!***********************************************************************