SUBROUTINE fi2(m,n,r,fim,fin)

! Forms the basis function.
! This subroutine calculates fim=fi(m,r) and fin=fi(n,r) in one turn

USE jm_constants
USE jm_globals, ONLY: projectile, lambda, basis

IMPLICIT none

INTEGER, INTENT(IN)        :: m, n
REAL(KIND=dp), INTENT(IN)  :: r
REAL(KIND=dp), INTENT(OUT) :: fim, fin
REAL(KIND=dp)              :: lr, lr2, tmp
REAL(KIND=dp)              :: lagm, lagn

lr = lambda * r 

SELECT CASE (basis)
CASE('laguerre')

  CALL laguerre2(m, n, 2*projectile%l+one, lr, lagm, lagn)

  tmp = lr**(projectile%l+1) * dexp(-lr/two)
 
CASE('gauss')

  lr2 = lr*lr

  CALL laguerre2(m, n, projectile%l+half, lr2, lagm, lagn)

  tmp = lr**(projectile%l+1) * dexp(-lr2/two)

END SELECT

fim  = tmp * lagm
fin  = tmp * lagn

RETURN
END SUBROUTINE fi2

!***********************************************************************

SUBROUTINE psi2(m, n, r, psim, psin)

! Forms the basis function.
! This subroutine calculates psim=psi(m,r) and psin(n,r) in one turn

USE jm_constants
USE jm_globals, ONLY: lambda, basis, projectile

IMPLICIT none

INTEGER, INTENT(IN)        :: m, n
REAL(KIND=dp), INTENT(IN)  :: r
REAL(KIND=dp), INTENT(OUT) :: psim, psin
REAL(KIND=dp)              :: lr, lrl1, lr2
REAL(KIND=dp)              :: lagm, lagn, lagm1, lagn1, fim, fin
REAL(KIND=dp)              :: tmp1, tmp2, tmp3, tmp4

lr   = lambda * r
lrl1 = lr**(projectile%l+1)

SELECT CASE (basis)
CASE ('laguerre')
 
  !$OMP PARALLEL SECTIONS

  !$OMP SECTION
  CALL laguerre2(m, n, 2*projectile%l+one, lr, lagm, lagn)
  tmp2 = dexp(-lr/two) 

  !$OMP SECTION
  CALL laguerre2(m, n, 2*projectile%l+two, lr, lagm1, lagn1)
  tmp4 = lrl1 * lambda/two

  !$OMP END PARALLEL SECTIONS  


CASE('gauss')

  lr2 = lr * lr

  !$OMP PARALLEL SECTIONS

  !$OMP SECTION
  CALL laguerre2(m, n, projectile%l+half, lr2, lagm, lagn)
  tmp2 = dexp(-lr2/two)

  !$OMP SECTION
  CALL laguerre2(m, n, projectile%l+oneandhalf, lr2, lagm1, lagn1)
  tmp4 = lrl1 * lambda**2 * r

  !$OMP END PARALLEL SECTIONS
  
END SELECT

CALL fi2(m, n, r, fim, fin)

tmp1 = projectile%kappa / r
tmp3 = lambda**(projectile%l+1) * (projectile%l+1) * r**projectile%l

psim = tmp1 * fim  +  tmp2 *     &
     ( tmp3 * lagm  -  tmp4 * lagm  -  lrl1 * (lagm1 - lagm ) )

psin = tmp1 * fin  +  tmp2 *     &
     ( tmp3 * lagn  -  tmp4 * lagn  -  lrl1 * (lagn1 - lagn ) )

RETURN
END SUBROUTINE psi2

!***********************************************************************

FUNCTION s(n, last_call)

! Computes the coefficients ot the sine-like solution.
! If optional last_call is present, array storing values is deallocated.

USE jm_constants
USE jm_interfaces, ONLY: jm
USE jm_globals, ONLY: lambda, nend, basis, pass, projectile

IMPLICIT none

INTEGER, INTENT(IN)              :: n
INTEGER, OPTIONAL                :: last_call
INTEGER                          :: l, i, tend
REAL(KIND=dp)                    :: gamma, gegen, laguerre, fact
REAL(KIND=dp)                    :: k, sint, cost, eta
REAL(KIND=dp), ALLOCATABLE, SAVE :: t_s(:)
REAL(KIND=dp)                    :: s


IF (.NOT. allocated(t_s) .OR. pass == 2) THEN

  l = projectile%l
  k = projectile%k

  tend = nend
  IF (nend < 3) tend = 5
  IF (.NOT. allocated(t_s)) ALLOCATE(t_s(0:tend))
  sint = k / ( k*k/lambda + lambda/four)
  cost = (k**2 / lambda**2 - quarter)/(k**2 / lambda**2 + quarter)
  eta = k / lambda

  SELECT CASE(basis)
  CASE ('laguerre')
    t_s(0) = 2**l * gamma(l+one)/fact(2*l+1)*sint**(l+1)*gegen(0,l+1,cost)
    t_s(1) = 2**l * gamma(l+one)/fact(2*l+2)*sint**(l+1)*gegen(1,l+1,cost)

  CASE ('gauss')
    t_s(0) = sqrt(twopi)  / gamma(l+oneandhalf) *  &
       dexp(-eta**2 / two) * laguerre(0,l+half,eta**2) * eta**(l+1)
    t_s(1) =-sqrt(twopi)  / gamma(l+twoandhalf) *  &
       dexp(-eta**2 / two) * laguerre(1,l+half,eta**2) * eta**(l+1)

  END SELECT

  DO i = 2, tend     ! three-term recursion relation
    ! elements jm are non-relativistic here, so third argument is given  
    t_s(i) = -( jm(i-1,i-2,1)*t_s(i-2) + jm(i-1,i-1,1)*t_s(i-1) ) / jm(i-1,i,1)

  ENDDO

ENDIF

s = t_s(n)

IF (PRESENT(last_call)) DEALLOCATE(t_s)

RETURN
END FUNCTION s

!***********************************************************************

FUNCTION c(n, last_call)

! Computes the coefficients ot the cosine-like solution.
! If optional last_call is present, array storing values is deallocated.

USE jm_constants
USE jm_interfaces, ONLY: jm
USE jm_globals, ONLY: nend, lambda, basis, pass, projectile

IMPLICIT none

INTEGER, INTENT(IN)              :: n
INTEGER, OPTIONAL                :: last_call
INTEGER                          :: l, i, tend
REAL(KIND=dp)                    :: c
REAL(KIND=dp)                    :: fact, gamma, hyperg21, hyperg11
REAL(KIND=dp)                    :: k, sint, sin2t, eta
REAL(KIND=dp), ALLOCATABLE, SAVE :: t_c(:)

IF (.NOT. allocated(t_c) .OR. pass == 2) THEN

  l = projectile%l
  k = projectile%k

  tend = nend
  IF (tend<3) tend = 5
  IF (.NOT. allocated(t_c)) ALLOCATE(t_c(0:tend))
  sint = k / ( k*k/lambda + lambda/four)
  sin2t = lambda**2 / (four * k**2 + lambda**2)
  eta = k / lambda

  SELECT CASE(basis)
  CASE ('laguerre')
    t_c(0) = -(2)**l / sint**l / fact(2*l+1) *                          &
      hyperg21(-two*l-one, one, half-l, sin2t)/sqrt(pi)*gamma(l+half)
    t_c(1) = -(2)**l / sint**l / fact(2+2*l) *                          &
      hyperg21(-two*l-two, two, half-l, sin2t)/sqrt(pi)*gamma(l+half)

  CASE ('gauss')
    t_c(0) = sqrt(two/pi)*gamma(l+half) / gamma(l+oneandhalf)             &
      * dexp(-eta**2/two) * eta**(-l) * hyperg11(-l-half,half-l,eta**2)
    t_c(1) = sqrt(two/pi) * gamma(l+half) * (-1) / gamma(l+twoandhalf)    &
      * dexp(-eta**2/two) * eta**(-l) * hyperg11(-l-oneandhalf,half-l,eta**2)

  END SELECT
  
  DO i = 2, tend     ! three-term recursion relation
    ! elements jm are non-relativistic here, so third argument is given
    t_c(i) = -( jm(i-1,i-2,1)*t_c(i-2) + jm(i-1,i-1,1)*t_c(i-1) ) / jm(i-1,i,1)
  ENDDO

ENDIF

c = t_c(n)

IF (PRESENT(last_call)) DEALLOCATE(t_c)

RETURN
END FUNCTION c

!***********************************************************************

FUNCTION fifi(m,n)
USE jm_constants
USE jm_globals, ONLY: basis, lambda, projectile

IMPLICIT none

INTEGER, INTENT(IN) :: m, n
REAL(KIND=dp)       :: fifi
REAL(KIND=dp)       :: lfact, lgamma, fact_over_fact
INTEGER             :: kronecker


SELECT CASE(basis)
CASE ('laguerre')
  
  fifi = fact_over_fact(n+2*projectile%l+1, n) / lambda * (                 &
         2*(n+projectile%l+1)*kronecker(m,n) - n*kronecker(m,n-1) -         &
           (n+2*projectile%l+2)*kronecker(m,n+1) )

CASE ('gauss')
  
  fifi = dexp(lgamma(n+projectile%l+oneandhalf)-lfact(n)) / two * kronecker(m,n)
  
END SELECT

RETURN
END FUNCTION fifi

!***********************************************************************

FUNCTION psipsi(m,n)
USE jm_constants
USE jm_globals, ONLY: basis, lambda, projectile

IMPLICIT none

INTEGER, INTENT(IN) :: m, n
REAL(KIND=dp)       :: psipsi
REAL(KIND=dp)       :: lfact, lgamma, fact_over_fact
INTEGER             :: kronecker

SELECT CASE(basis)
CASE ('laguerre')
  
  psipsi = fact_over_fact(n+2*projectile%l+1, n) / four * (                   &
           2*(n+projectile%l+1)*(2*lambda-1)*kronecker(m,n) +                 &
              n*kronecker(m,n-1) + (n+2*projectile%l+2)*kronecker(m,n+1) )
  
CASE ('gauss')
  
  psipsi = lambda**2 *                                                        &
           dexp(lgamma(n+projectile%l+oneandhalf) - lfact(n)) / two * (       &
           (2*n+projectile%l+oneandhalf)*kronecker(m,n) + n*kronecker(m,n-1)  &
            + (n+projectile%l+oneandhalf)*kronecker(m,n+1) )
  
END SELECT

RETURN
END FUNCTION psipsi

!***********************************************************************
