SUBROUTINE control_calc_delta(fout)

! This subroutine controls process of computing the numerical values 
! of the phase shifts.

USE jm_constants
USE jm_interfaces, ONLY: s, c, vn
USE jm_globals, ONLY: projectile, scr, ntrunc, nstart, nend, pass, shift,  &
        estart, eend, estep, energies, delta_analytic_error, projectile_set_k

IMPLICIT none

CHARACTER(LEN=*), INTENT(IN) :: fout

REAL(KIND=dp) :: t
REAL(KIND=dp) :: delta, tan_delta, delta_analytic
REAL(KIND=dp) :: dummy
INTEGER       :: i, iend, unit


! Calculate phase shifts for <ntrunc> = <nstart>... <nend> 
! with fixed energy <e>

pass = 1

! Set some parameters, depending on e  
CALL projectile_set_k(projectile)

unit = 2
OPEN(unit, file=trim(fout), err=2)
CALL write_header(unit)

! Try to compute the analytical value
delta = delta_analytic()
IF (.NOT. delta_analytic_error) THEN
  IF (scr) PRINT '(/A,F20.16)', 'delta_analytic =',delta
  WRITE(unit,'(A19,F20.15)') '; ANALYTICAL RESULT'
  WRITE(unit,'(A10,F20.15)') '; delta = ', delta
  WRITE(unit,'(A1)') ';'
ENDIF

WRITE(unit,'(A44)') '; NUMERICAL RESULTS:  N   tan(delta)   delta'

IF (scr) PRINT '(/A)', &
  'Calculating numerical phase shifts for fixed projectile energy:'


DO ntrunc = nstart, nend
  
  tan_delta = t(ntrunc)            ! Compute tan(delta) and delta
  delta = atan(tan_delta)          ! for given ntrunc

  IF (shift .AND. delta < zero) delta = delta + pi

  ! Write results to file and (optionally) to screen  
  WRITE(unit,'(I5,F26.16,F26.16)') ntrunc, tan_delta, delta
  CALL FLUSH(unit)
  IF (scr) PRINT '(A4,I5,A10,F20.16)', 'n =', ntrunc, 'delta =', delta
 
ENDDO

CLOSE(unit)         

! Average results
CALL avrg(fout)

! Calculate root mean squared error
CALL rmse(fout)


IF (energies) THEN

  ! Calculate phase shifts for energies from range [<estart>, <eend>] 
  ! for <ntrunc> = <nend>
  ! (performed only if <energies> = <.true.>)

  pass = 2

  unit = 3
  OPEN(unit, file=trim(fout)//'.en', err=2)
  CALL write_header(unit)

  IF (scr) PRINT '(/A59,I5,A1)', &
    'Now calculating phase shifts for a given energy range; n = ', nend, ":"

  WRITE(unit,'(A46)',advance='no') & 
                    '; NUMERICAL RESULTS: energy  tan(delta)  delta'

  ! Test if analytical result is available and possibly complete the header
  delta = delta_analytic()
  IF (delta_analytic_error) THEN
    WRITE(unit,*)  ! Just terminate the record
  ELSE
    WRITE(unit,'(A18)') '  delta_analytical'
  ENDIF


  iend =  int( (eend - estart) / estep + 1.1_dp )

  ntrunc = nend

  DO i=1, iend

    projectile%e = estart + estep * (i-1)

    ! Set some parameters, depending on e
    CALL projectile_set_k(projectile)

    tan_delta = t(ntrunc)            ! Compute tan(delta) and delta
    delta = atan(tan_delta)          ! for ntrunc = nend

    IF (shift .AND. delta < zero) delta = delta + pi

    ! Write results to file and (optionally) to screen.
    WRITE(unit,'(F10.6,F26.16,F26.16)',advance='no')  &
          projectile%e, tan_delta, delta
    IF (scr) PRINT '(A9,F10.6,A9,F20.16)',            &
          'energy =',projectile%e,'delta =',delta

    ! Try to calculate analytical result
    delta = delta_analytic()
    IF (delta_analytic_error) THEN
      WRITE(unit,*)  ! Just terminate the record
    ELSE
      IF (scr) PRINT '(A28,F20.16)', 'delta_analytic = ', delta
      WRITE(unit,'(F26.16)') delta      
    ENDIF
    CALL FLUSH(unit)
  
  ENDDO

  CLOSE(unit)      

ENDIF


! Deallocating some arrays created during progrm execution, 
! it is indicated by calling functions with last, optional parameter.
dummy = s(1, 1)
dummy = c(1, 1)
dummy = vn(1,1, 1)

RETURN

2 CALL show_error('control_calc_delta','unable to open the output file')

END SUBROUTINE control_calc_delta
