FUNCTION g(x,y)

! Computes elements of the matrix approximating Green function.
! Elements of truncated potential are saved for particular ntrunc
! and used in calculations for ntrunc+1, thus only new elements of
! truncated potential need to be calculated, not the whole matrix.

USE jm_constants
USE jm_interfaces, ONLY: jm, vn
USE jm_globals, ONLY: ntrunc, nstart, nend, scheme, projectile,  &
                      pass, vn_is_loaded, eend, energies

IMPLICIT none

INTEGER, INTENT(IN)              :: x, y
INTEGER                          :: read_n_to_save_v
REAL(KIND=dp)                    :: g
REAL(KIND=dp), ALLOCATABLE       :: a(:,:), eigval(:), eigvec(:,:), v(:,:)
REAL(KIND=dp), ALLOCATABLE, SAVE :: a_saved(:,:), v_saved(:,:)
INTEGER                          :: dim, maxdim, i, j, ii
LOGICAL                          :: first_call_in_first_pass, &
                                    last_call_in_first_pass,  &
                                    last_call_in_second_pass 


! Matrices dimensions
SELECT CASE(scheme)
CASE ('relativistic')
  dim    = 2 * ntrunc
  maxdim = 2 * nend
CASE('non-relativistic')
  dim    = ntrunc
  maxdim = nend
END SELECT

! Allocate work matrices
ALLOCATE(v(0:dim-1, 0:dim-1), a(dim, dim), eigval(dim), eigvec(dim, dim))

! Some logical flags to simplify further conditions
IF (pass==1 .AND. ntrunc == nstart) THEN
  first_call_in_first_pass = .true.
ELSE
  first_call_in_first_pass = .false.
ENDIF
IF (pass==1 .AND. ntrunc == nend) THEN
  last_call_in_first_pass = .true.
ELSE
  last_call_in_first_pass = .false.
ENDIF
IF (pass==2 .AND. projectile%e == eend) THEN 
  last_call_in_second_pass = .true.
ELSE
  last_call_in_second_pass = .false.
ENDIF


IF ( first_call_in_first_pass .OR. pass == 2) THEN

  SELECT CASE (pass)
  CASE(1)

    ! Allocate matrices to store the elements computed for given ntrunc
    ! for use in the next calls of the subroutine.
    ! Matrix <a_saved> can be used only in the first pass with a fixed energy
    ! (it depends on energy).
    ! Matrix <v_saved> is used in both passes (it does not depend on energy).
    ALLOCATE(v_saved(0:maxdim-1,0:maxdim-1),a_saved(maxdim,maxdim))

    ! Calculate matrix v, only upper half (due to its symmetry)
    DO i=1,dim     
      DO j=1,i
        v(i-1,j-1) = vn(i-1,j-1)         
      ENDDO
    ENDDO

  CASE(2)

    ! Restore matrix v from v_saved (upper half),
    ! Matrix v is not saved between calls, so it has to be done in each call.
    !$OMP PARALLEL DO IF (dim>10)
    DO i=1,dim
       DO j=1,i
         v(i-1,j-1) = v_saved(i-1,j-1)
      ENDDO
    ENDDO
    !$OMP END PARALLEL DO    
  
  END SELECT


  ! Form matrix a = jm + v   (upper half)
  !$OMP PARALLEL DO IF (dim>10)
  DO i=1,dim     
    DO j=1,i
      a(i,j) = jm(i-1,j-1) + v(i-1,j-1)
    ENDDO
  ENDDO
  !$OMP END PARALLEL DO    

  ! Complete lower half of the matrix a, utilizing its symmetry
  !$OMP PARALLEL DO IF (dim>10)
  DO i=1,dim     
    DO j=1,i
      a(j,i) = a(i,j)
    ENDDO
  ENDDO
  !$OMP END PARALLEL DO    


ELSE  ! Use the elements stored in the matrices a_saved and v_saved

  SELECT CASE(scheme)
  CASE('relativistic')

    ! Restore saved elements of a_saved and v_saved "inside" the new a
    !$OMP PARALLEL DO IF (dim>10)
    DO i=2,dim-1
      DO j=2,dim-1
        a(i,j)     = a_saved(i-1,j-1)
        v(i-1,j-1) = v_saved(i-2,j-2)
      ENDDO
    ENDDO
    !$OMP END PARALLEL DO    


    ! Calculate potential elements for the added rows 
    ! Do not parallelize - 
    ! direct parallelization of basis function gave better results.
    DO i=1,dim
      v(dim-1,i-1) = vn(dim-1,i-1)
      v(0,i-1)     = vn(0,i-1)
    ENDDO

    ! Form matrix a for the added rows
    ! Do not parallelize due to utilizing symmetry in loop
    DO i=1,dim
      a(dim,i)     = jm(dim-1,i-1) + v(dim-1,i-1)
      a(i,dim)     = a(dim,i)
      v(i-1,dim-1) = v(dim-1,i-1)
      a(1,i)       = jm(0,i-1) + v(0,i-1)
      a(i,1)       = a(1,i)
      v(i-1,0)     = v(0,i-1)
    ENDDO


  CASE('non-relativistic')

    ! Restore saved elements of a_saved and v_saved
    !$OMP PARALLEL DO IF (dim>10)
    DO i=1,dim-1
      DO j=1,dim-1
        a(i,j)     = a_saved(i,j)
        v(i-1,j-1) = v_saved(i-1,j-1)
      ENDDO
    ENDDO
    !$OMP END PARALLEL DO

    ! Calculate elements of potential for the added row
    ! Parallelization gives wrong results
    DO i=1,dim
      v(dim-1,i-1) = vn(dim-1,i-1)
    ENDDO

    ! Calculate elements of matrix a for the added row    
    !$OMP PARALLEL DO IF (dim>50)
    DO i=1,dim
      a(dim,i) = jm(dim-1,i-1) + v(dim-1,i-1)
    ENDDO
    !$OMP END PARALLEL DO

    ! Complete the matrix a, utilizing its symmetry
    !$OMP PARALLEL DO IF (dim>50)
    DO i=1,dim
      a(i,dim) = a(dim,i)
    ENDDO
    !$OMP END PARALLEL DO


  END SELECT
  
ENDIF


IF (last_call_in_first_pass) THEN 

  ! a_saved is not used in 2nd pass, destroy it
  DEALLOCATE(a_saved)

  ! Save vn to use in 2nd pass (upper half)
  !$OMP PARALLEL DO IF (dim>10)
  DO i=1,dim
    DO j=1,i
      v_saved(i-1,j-1) = v(i-1,j-1)
    ENDDO
  ENDDO
  !$OMP END PARALLEL DO    
 
  ! Write v to file
  IF ((.NOT. vn_is_loaded) .OR. (vn_is_loaded .AND. nstart /= nend)) &
    CALL write_vn_to_file(dim, v)

  ! No second pass, destroy v_saved now
  IF (.NOT. energies)  DEALLOCATE(v_saved)


ELSE IF (pass==1) THEN

  ! First pass and not last call, save a and v in a_saved and v_saved
  !$OMP PARALLEL DO IF (dim>10)
  DO i=1,dim
    DO j=1,dim   !!! TODO: save and restore only upper half of the matrices 
      a_saved(i,j) = a(i,j)
      v_saved(i-1,j-1) = v(i-1,j-1)
    ENDDO
  ENDDO
  !$OMP END PARALLEL DO    

  ! Save v to file if requested in file with online parameters
  IF ( nstart /= ntrunc .AND. ntrunc == read_n_to_save_v() )   &
         CALL write_vn_to_file(dim, v)


ELSE IF (last_call_in_second_pass) THEN 

  ! v_saved is not needed anymore
  DEALLOCATE(v_saved)

ENDIF


! Look for the eigenvalues and eigenvectors of the matrix a.
CALL eigensys(dim, a, eigval, eigvec)


DEALLOCATE(a, v)


! Look for the matrix element <g>,
! used in the formula for tangent to the phase shift.
SELECT CASE(scheme)
CASE('relativistic')
  i  = 1
  ii = 1
CASE('non-relativistic')
  i  = x+1
  ii = y+1
END SELECT

! Form the matrix g
g = zero

!$OMP PARALLEL DO REDUCTION(+:g)
DO j=1,dim
  g = g + eigvec(i,j)*eigvec(ii,j)/eigval(j)
ENDDO
!$OMP END PARALLEL DO

DEALLOCATE(eigval, eigvec)


RETURN
END FUNCTION g

