SUBROUTINE convert_int_to_string(ival, chval)

! Converts integer number to a equivalent string
! Max digits of integer is set to 5 - it's enough for JMATRIX purposes

IMPLICIT none

INTEGER, INTENT(IN)           :: ival
CHARACTER(len=*), INTENT(OUT) :: chval

WRITE (chval,'(I5)') ival
      
END SUBROUTINE convert_int_to_string

!***********************************************************************      

SUBROUTINE rmse(fout)

! Calculates root mean squared error
! rms = sqrt( sum(x_analytical - xi)^2  / (n-1) )

USE jm_constants

IMPLICIT none

CHARACTER(LEN=*), INTENT(IN) :: fout
CHARACTER(LEN=100)           :: dummy = ""
REAL(KIND=dp)                :: analytic, sum, rms
REAL(KIND=dp)                :: delta(5000)
INTEGER                      :: ntrunc(5000)
INTEGER                      :: n, i, j, unit


unit = 4

! Reading data

OPEN(unit, file=fout, status="old", err=30)

DO WHILE (dummy /= "; ANALYTICAL RESULT ")      ! Read header
  READ(unit,'(A100)', END=50) dummy
ENDDO
READ(unit,'(A10,F16.20)') dummy, analytic       ! Read analytical result
READ(unit,'(A20)')        dummy
READ(unit,'(A20)')        dummy

DO i=1,5000   ! Read the proper data
  READ(unit,'(I5,2F26.16)',ERR=40, END=20) ntrunc(i),delta(i),delta(i)
ENDDO

20 CLOSE(unit)

! Amount of data points that has been read
n = i - 1


IF (n < 200) THEN
  CALL show_info('Not enough data points to calculate RMS')
  RETURN
ENDIF


! Calculate RMSE from delta(n/10) to delta(i), i = n/10 ... n

OPEN(unit,file=trim(fout)//".rms")

CALL write_header(unit)

DO j = n/10, n
  sum = zero
  DO i = n/20, j
    sum = sum + ( analytic-delta(i) )**2
  ENDDO
  rms = sqrt(sum / (j - n/20) )
  WRITE(unit,'(I5,F26.18)') ntrunc(j), rms
ENDDO

CLOSE(unit)

RETURN

30 CALL show_warning('rmse','error opening the input file')
RETURN

40 CALL show_warning('rmse','error reading data')
CLOSE(unit)
RETURN

50 CALL show_info('Analytical result not known. RMS will not be calculated')
CLOSE(unit)
RETURN

END SUBROUTINE rmse

!***********************************************************************      

SUBROUTINE avrg(fout)

! Averages results in ordinary way (arithmetic mean)

USE jm_constants

IMPLICIT none

CHARACTER(LEN=*), INTENT(IN) :: fout
CHARACTER(LEN=100)           :: dummy = ""
REAL(KIND=dp)                :: sum
REAL(KIND=dp)                :: delta(5000)
INTEGER                      :: ntrunc(5000)
INTEGER                      :: n, points_av, i, j, unit


unit = 4

! Reading data

OPEN(unit, file=fout, status="old", err=30)

! Read header
DO WHILE (dummy /= "; NUMERICAL RESULTS:  N   tan(delta)   delta") 
  READ(unit,'(A100)', END=50) dummy
ENDDO

DO i=1,5000   ! Read the proper data
  READ(unit,'(I5,2F26.16)',ERR=40, END=20) ntrunc(i),delta(i),delta(i)
ENDDO

20 CLOSE(unit)

! Amount of data points that has been read
n = i - 1


! Averaging each group of points_av deltas
points_av = 50

IF (n < points_av) THEN
  CALL show_info ('Not enough data points to perform averaging')
  RETURN
ENDIF

OPEN(unit,file=trim(fout)//".avg")

CALL write_header(unit)

DO i=points_av, n, points_av
  sum = zero
  DO j=i-points_av+1, i
    sum = sum + delta(j)
  ENDDO
  WRITE(unit,'(I5,F26.18)') ntrunc(i), sum / points_av
ENDDO

CLOSE(unit)

RETURN

30 CALL show_warning('avrg','error opening the input file')
RETURN

40 CALL show_warning('avrg','error reading data')
CLOSE(unit)
RETURN

50 CALL show_warning('avrg','bad input file')
CLOSE(unit)
RETURN

END SUBROUTINE avrg

!***********************************************************************      

SUBROUTINE write_header(unit)

! Writes the header of the output file.

USE jm_globals, ONLY: estart, eend, estep, projectile, lambda, basis, scheme, &
                      vol, potential, shift, nstart, nend, pass

IMPLICIT NONE

INTEGER, INTENT(IN) :: unit
INTEGER             :: val(8)

! Get the current date and time
CALL date_and_time(values=val)

WRITE(unit,'(A33)') '; OUTPUT FROM THE JMATRIX PROGRAM'
WRITE(unit,'(A8,I2,A1,I2,A1,I4,A8,I2,A1,I2)')  &
      '; date: ',val(3),'.',val(2),'.',val(1),'  time: ',val(5),'.',val(6)
WRITE(unit,'(A1)') ';'
IF (pass==1) THEN
  WRITE(unit,'(A13,F22.16)')   '; energy   = ', projectile%e
ELSE
  WRITE(unit,'(A13,F22.16)')   '; estart   = ', estart
  WRITE(unit,'(A13,F22.16)')   '; eend     = ', eend
  WRITE(unit,'(A13,F22.16)')   '; estep    = ', estep
ENDIF
WRITE(unit,'(A13,I5)')     '; l        = ', projectile%l
WRITE(unit,'(A13,I5)')     '; kappa    = ', projectile%kappa
WRITE(unit,'(A13,F22.16)') '; lambda   = ', lambda
WRITE(unit,'(A13,A)')      '; basis    = ', basis
WRITE(unit,'(A13,A)')      '; scheme   = ', scheme
WRITE(unit,'(A13,A)')      '; v_light  = ', vol
WRITE(unit,'(A13,L1)')     '; shift    = ', shift
IF (pass==1) THEN
  WRITE(unit,'(A13,I5)')   '; nstart   = ', nstart
  WRITE(unit,'(A13,I5)')   '; nend     = ', nend
ELSE
  WRITE(unit,'(A13,I5)')   '; ntrunc   = ', nend
ENDIF
WRITE(unit,'(A13,A)')      '; pot_type = ', potential%type
SELECT CASE (potential%type)
CASE ('well')
  WRITE(unit,'(A13,F22.16)') '; v0       = ', potential%v0
  WRITE(unit,'(A13,F22.16)') '; a        = ', potential%a
  WRITE(unit,'(A13,F22.16)') '; b        = ', potential%b
CASE ('coulomb')
  WRITE(unit,'(A13,F22.16)') '; r0       = ', potential%r0
  WRITE(unit,'(A13,F22.16)') '; z        = ', potential%z
  WRITE(unit,'(A13,F22.16)') '; alpha    = ', potential%alpha
CASE ('yukawa')
  WRITE(unit,'(A13,F22.16)') '; r0       = ', potential%r0
  WRITE(unit,'(A13,F22.16)') '; g        = ', potential%g
  WRITE(unit,'(A13,F22.16)') '; m        = ', potential%m
CASE ('user')
  WRITE(unit,'(A13,F22.16)') '; r0       = ', potential%r0
END SELECT

WRITE(unit,'(A1)') ';'

RETURN

END SUBROUTINE write_header

!***********************************************************************      

SUBROUTINE loadparams(fin)

! Reads input data from the file <fin>

USE jm_constants
USE jm_globals, ONLY: estart, eend, estep, projectile, lambda, basis, vol, scr,&
                      scheme, shift, nstart, nend, energies, potential, v_light

! Reads data from input file <fin>.

IMPLICIT none

CHARACTER (LEN=*), INTENT(IN) :: fin
CHARACTER (LEN=44)            :: item, eqsign, value_char
CHARACTER (LEN=80)            :: line
CHARACTER (LEN=10)            :: par_txt(100)
INTEGER                       :: par(100,2)
INTEGER                       :: i, unit
LOGICAL                       :: print_loaded

! Print (or not) loaded parameters to screen
print_loaded = .false.

PRINT '(A)', 'Loading parameters...'

unit = 1
OPEN(unit, file=fin, err=112, status='old')

! <par_txt> - parameter labels
! <par(i,1)> = 1 if i-th parameter is required
! <par(i,2)> = 1 if i-th parameter has been read from file
par = 0
par_txt(22) = 'e'        ; par(22,1) = 1 ! required
par_txt(21) = 'energies' ; par(21,1) = 0 ! default = .false.
par_txt(1)  = 'estart'   ; par(1,1)  = 0 ! required if <energies> = .true.
par_txt(2)  = 'eend'     ; par(2,1)  = 0 ! required if <energies> = .true.
par_txt(3)  = 'estep'    ; par(3,1)  = 0 ! default = (<eend>-<estart>)/10
par_txt(4)  = 'l'        ; par(4,1)  = 1 ! required
par_txt(5)  = 'kappa'    ; par(5,1)  = 0 ! required if <scheme>='relativistic'
par_txt(6)  = 'lambda'   ; par(6,1)  = 0 ! default = 1.0
par_txt(7)  = 'basis'    ; par(7,1)  = 1 ! required
par_txt(8)  = 'scheme'   ; par(8,1)  = 1 ! required
par_txt(9)  = 'v_light'  ; par(9,1)  = 0 ! default = 'finite'
par_txt(10) = 'pot_type' ; par(10,1) = 1 ! required
par_txt(11) = 'screen'   ; par(11,1) = 0 ! default = .true.
par_txt(12) = 'shift'    ; par(12,1) = 0 ! default = .false.
par_txt(13) = 'nstart'   ; par(13,1) = 0 ! default = 1
par_txt(14) = 'nend'     ; par(14,1) = 1 ! required
par_txt(15) = 'v0'       ; par(15,1) = 0 ! required if <pot_type>='well'
par_txt(16) = 'a'        ; par(16,1) = 0 ! required if <pot_type>='well'
par_txt(17) = 'b'        ; par(17,1) = 0 ! required if <pot_type>='well'
par_txt(18) = 'r0'       ; par(18,1) = 0 ! required if <pot_type>='coulomb' or <user>
par_txt(19) = 'z'        ; par(19,1) = 0 ! required if <pot_type>='coulomb'
par_txt(20) = 'alpha'    ; par(20,1) = 0 ! required if <pot_type>='coulomb'
par_txt(23) = 'g'        ; par(19,1) = 0 ! required if <pot_type>='yukawa'
par_txt(24) = 'm'        ; par(20,1) = 0 ! required if <pot_type>='yukawa'

DO

  READ (unit,'(A80)',end=555) line

  IF (index(line,';')==1 .OR. index(line,'#')==1 .OR. index(line,'%')==1  &
    .OR. index(line,'//')==1) CYCLE     ! line is a comment - read next

  ! Divide line into three parts, if blank line - cycle (end=...)
  READ (line,*,end=777) item, eqsign, value_char

  IF (index(eqsign,'=') /= 1) CYCLE     ! No '=' sign in line - read next


  SELECT CASE (item)

  CASE ('e')
    READ(value_char,*,err=111) projectile%e
    par(22,2) = 1
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(22)//' = ', projectile%e

  CASE ('estart')
    READ(value_char,*,err=111) estart
    par(1,2) = 1
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(1)//' = ', estart

  CASE ('eend')
    READ(value_char,*,err=111) eend
    par(2,2) = 1  
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(2)//' = ', eend

  CASE ('estep')
    READ(value_char,*,err=111) estep
    par(3,2) = 1
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(3)//' = ', estep

  CASE ('energies')
    READ(value_char,*,err=111) energies
    par(21,2) = 1
    IF (print_loaded) PRINT *, par_txt(21),' = ', energies

  CASE ('l')
    READ(value_char,*,err=111) projectile%l  
    par(4,2) = 1
    IF (print_loaded) PRINT '(A14,I5)', ' '//par_txt(4)//' = ', projectile%l

  CASE ('kappa')
    READ(value_char,*,err=111) projectile%kappa
    par(5,2) = 1
    IF (print_loaded) PRINT '(A14,I5)', ' '//par_txt(5)//' = ', projectile%kappa

  CASE ('lambda')
    READ(value_char,*,err=111) lambda
    par(6,2) = 1
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(6)//' = ', lambda

  CASE ('basis')
    basis = trim(value_char)
    par(7,2) = 1
    IF (print_loaded) PRINT *, par_txt(7),' = ', basis
  
  CASE ('scheme')
    scheme = trim(value_char)
    par(8,2) = 1
    IF (print_loaded) PRINT *, par_txt(8),' = ', scheme

  CASE ('v_light')
    vol = trim(value_char)
    par(9,2) = 1  
    IF (print_loaded) PRINT *, par_txt(9),' = ', vol

  CASE ('pot_type')
    potential%type = trim(value_char)
    par(10,2) = 1
    IF (print_loaded) PRINT *, par_txt(10),' = ', potential%type

  CASE ('screen')
    READ(value_char,*,err=111) scr
    par(11,2) = 1
    IF (print_loaded) PRINT *, par_txt(11),' = ', scr

  CASE ('shift')
    READ(value_char,*,err=111) shift
    par(12,2) = 1
    IF (print_loaded) PRINT *, par_txt(12),' = ', shift

  CASE ('nstart')
    READ(value_char,*,err=111) nstart
    par(13,2) = 1
    IF (print_loaded) PRINT '(A14,I5)', ' '//par_txt(13)//' = ', nstart

  CASE ('nend')
    READ(value_char,*,err=111) nend
    par(14,2) = 1
    IF (print_loaded) PRINT '(A14,I5)', ' '//par_txt(14)//' = ', nend

  CASE ('v0')
    READ(value_char,*,err=111) potential%v0
    par(15,2) = 1
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(15)//' = ', potential%v0

  CASE ('a')
    READ(value_char,*,err=111) potential%a
    par(16,2) = 1  
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(16)//' = ', potential%a

  CASE ('b')
    READ(value_char,*,err=111) potential%b
    par(17,2) = 1  
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(17)//' = ', potential%b

  CASE ('r0')
    READ(value_char,*,err=111) potential%r0
    par(18,2) = 1  
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(18)//' = ', potential%r0

  CASE ('z')
    READ(value_char,*,err=111) potential%z
    par(19,2) = 1
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(19)//' = ', potential%z

  CASE ('alpha')
    READ(value_char,*,err=111) potential%alpha
    par(20,2) = 1  
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(20)//' = ', potential%alpha


  CASE ('g')
    READ(value_char,*,err=111) potential%g
    par(20,2) = 1  
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(20)//' = ', potential%g

  CASE ('m')
    READ(value_char,*,err=111) potential%m
    par(20,2) = 1  
    IF (print_loaded) PRINT '(A14,F8.4)', ' '//par_txt(20)//' = ', potential%m
  
  CASE DEFAULT

    CALL show_warning('loadparams','unknown parameter <'//trim(item)// &
    '> = <'//trim(value_char)//'> '//'has been ignored')

  END SELECT

  777 CONTINUE

ENDDO

555 CLOSE(unit)

! Complete the requirements table
SELECT CASE(potential%type)
CASE('well')
  par(15,1) = 1    ! Require v0
  par(16,1) = 1    ! Require a
  par(17,1) = 1    ! Require b
CASE('coulomb')
  par(18,1) = 1    ! Require r0
  par(19,1) = 1    ! Require z
  par(20,1) = 1    ! Require alpha
CASE('yukawa')
  par(18,1) = 1    ! Require r0
  par(23,1) = 1    ! Require g
  par(24,1) = 1    ! Require m
END SELECT
IF (energies) THEN
  par(1,1) = 1     ! Require estart
  par(2,1) = 1     ! Require eend
ENDIF
IF (scheme == 'relativistic') par(5,1) = 1     ! Require kappa


! Determine if all required parameters have been read
DO i=1,22
  IF (par(i,1)==1 .AND. par(i,2) == 0) CALL show_error &
     ('loadparams', 'required parameter missing: <' // trim(par_txt(i)) // '>')
ENDDO


! Default values for not read parameters
DO i=1,22
  IF (par(i,1)==0 .AND. par(i,2) == 0) THEN 
    SELECT CASE(i)
    CASE(3)
      estep = (eend - estart) / ten
      IF (print_loaded) PRINT*, '(default) estep   = ', estep
    CASE(6)
      lambda = one
      IF (print_loaded) PRINT*, '(default) lambda  = ', lambda
    CASE(9)
      IF (scheme == 'relativistic') THEN
        vol = 'finite'
        IF (print_loaded) PRINT*, '(default) v_light = ',vol
      ENDIF
    CASE(11)
      scr = .true.
      IF (print_loaded) PRINT*, '(default) screen  = ', scr
    CASE(12)
      shift = .false.
      IF (print_loaded) PRINT*, '(default) shift   = ', shift
    CASE(13)
      nstart = 1
      IF (print_loaded) PRINT '(A20,I5)', ' (default) nstart  = ', nstart
    END SELECT
  ENDIF
ENDDO


! Check if parameter values are valid

IF (nstart < 1 .OR. nstart > nend) &
  CALL show_error('loadparams','incorrect values of <nstart> or <nend>')

IF (nend > 5000) &
  CALL show_error('loadparams','<nend> is to big')

IF (projectile%e <= zero) &
  CALL show_error('loadparams','<e> must be positive')

IF (energies .AND. estart > eend) &
  CALL show_error('loadparams','incorrect values of <estart> or <eend>')

IF (energies .AND. (estart <= zero .OR. eend <= zero)) &
  CALL show_error('loadparams','<estart> and <eend> must be positive')

IF (energies .AND. estep <= zero) &
  CALL show_error('loadparams','<estep> must be positive')

IF (scheme == 'relativistic' .AND. & 
  projectile%kappa /= projectile%l .AND. projectile%kappa /= -projectile%l-1) &
  CALL show_error('loadparams','<l> and <kappa> are not properly correlated')

IF (potential%type /= 'well' .AND. potential%type /= 'coulomb'      &
   .AND. potential%type /= 'yukawa' .AND. potential%type /= 'user') &
  CALL show_error('loadparams','<pot_type> must be <well>, <coulomb> or <user>')

IF (potential%type == 'well' .AND. potential%a >= potential%b) &
  CALL show_error('loadparams','incorrect values of <a> or <b>')

IF (potential%type == 'yukawa' .AND. potential%m <= zero) &
  CALL show_error('loadparams','<m> should be positive')

IF (basis /= 'laguerre' .AND. basis /= 'gauss') &
  CALL show_error('loadparams','<basis> must be <gauss> or <laguerre>')

IF (scheme /= 'relativistic' .AND. scheme /= 'non-relativistic') &
  CALL show_error &
       ('loadparams','<scheme> must be <relativistic> or <non-relativistic>')

IF (scheme == 'relativistic' .AND. vol /= 'finite' .AND. vol /= 'infinite') &
  CALL show_error('loadparams','<v_light> must be <finite> or <infinite>')

IF (scr .NEQV. .true. .AND. scr .NEQV. .false. ) &
  CALL show_error('loadparams','<scr> must be <.true.> or <.false.>')

IF (shift .NEQV. .true. .AND. shift .NEQV. .false. ) &
  CALL show_error('loadparams','<shift> must be <.true.> or <.false.>')

IF (lambda <= zero) CALL show_error('loadparams','<lambda> must be positive')


! Now some warning messages

IF (basis == 'gauss' .AND. (projectile%e > 25 .OR. eend > 25 .AND. energies)) &
  CALL show_warning('loadparams',                                             &
                    'results for e>25 may be inaccurate in gaussian basis')

IF (     (scheme == 'non-relativistic' .AND. nend > 3000)   &
    .OR. (scheme == 'relativistic'     .AND. nend > 2000) ) &
CALL show_warning('loadparams','be prepared for a long-running calculations')


IF (vol == 'finite') THEN
  v_light = 137.0359996287515_dp
ELSE
  v_light = 5000.0_dp
ENDIF


PRINT '(A)', 'Load complete.'

RETURN

111 CALL show_error('loadparams','incorrect value for <'//trim(item)//'>')

112 CALL show_error('loadparams','unable to open the input file')

END SUBROUTINE loadparams

!***********************************************************************      

SUBROUTINE show_error(where, what)

! Displays error message and terminates the program

IMPLICIT none

CHARACTER(len=*), INTENT(in) :: where, what

PRINT '(A)','Error ('//where //'): ' // what // '.'
STOP 'JMATRIX will terminate.'
      
END SUBROUTINE show_error

!***********************************************************************      

SUBROUTINE show_warning(where, what)

! Displays warning message

IMPLICIT none

CHARACTER(len=*), INTENT(in) :: where, what

PRINT *,'Warning ('//where //'): ' // what // '.'
      
END SUBROUTINE show_warning

!***********************************************************************      

SUBROUTINE show_info(what)

! Displays informational message

IMPLICIT none

CHARACTER(len=*), INTENT(in) :: what

PRINT *,'(Info: ' // what // '.)'
      
END SUBROUTINE show_info
