FUNCTION v(r)

! Contains formulas for the scattering potential.

USE jm_constants
USE jm_globals, ONLY: potential

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: r
REAL(KIND=dp)             :: v

SELECT CASE(potential%type)
CASE ('well')     

  ! Potential square-well
  IF ( r >= potential%a .OR. r <= potential%b ) THEN
    v =  potential%v0
  ELSE
    v = zero
  ENDIF
  
CASE ('coulomb')

  ! Coulomb-like potential
  v = - ( potential%z / r**potential%alpha )

CASE ('yukawa')

  ! Yukawa potential
  v = - potential%g**2 * ( EXP(-potential%m*r) / r )
  
!CASE ('l-j')
!
  ! Lennard-Jones potential
!  v = - 4*potential%eps *    &
!                ( (potential%sigma / r)**12 - (potential%sigma / r)**6 )
  
CASE ('user')

  ! User-specified potential included from file
  INCLUDE "user_potential"

END SELECT

RETURN
END FUNCTION v

!***********************************************************************

FUNCTION vn(i,j, last_call)

! Calculates the element of the truncated potential in selected basis.

USE jm_constants
USE jm_globals, ONLY: potential, scheme, ntrunc, nstart, pass, &
                vn_is_loaded, vn_file_match, ii, jj

IMPLICIT none

INTEGER, INTENT(IN)     :: i, j
INTEGER, OPTIONAL       :: last_call
INTEGER                 :: dim
REAL(KIND=dp)           :: vn
REAL(KIND=dp)           :: integral_inf, integral
REAL(KIND=dp), EXTERNAL :: fi_v_fi, psi_v_psi
REAL(KIND=dp), DIMENSION(:,:), ALLOCATABLE, SAVE :: vn_matrix

IF ((.NOT. vn_is_loaded) .AND. vn_file_match .AND.            &
     (pass == 2 .OR. (pass ==1 .AND. nstart == ntrunc))) THEN

  ! Try to load previously saved elements of truncated potential
  SELECT CASE(scheme)
  CASE('relativistic')
    dim = 2 * ntrunc
  CASE('non-relativistic')
    dim = ntrunc
  END SELECT

  ALLOCATE( vn_matrix(0:dim-1,0:dim-1) )
  CALL read_vn_from_file(dim, vn_matrix)
  IF (.NOT. vn_is_loaded) DEALLOCATE(vn_matrix)

ENDIF

! Use vn loaded from file, under condition that the file matches to ntrunc:
! second pass (so ntrunc is constant), or first pass for n=ntrunc
IF (vn_is_loaded .AND. ((pass==1 .AND. ntrunc==nstart) .OR. pass==2) ) THEN

  vn = vn_matrix(i,j)

ELSE

  SELECT CASE(scheme)
  CASE ('relativistic')
  
    IF ( (i < ntrunc) .AND. (j < ntrunc) ) THEN   ! The "++" area

      ii = ntrunc-i-1
      jj = ntrunc-j-1

      SELECT CASE(potential%type)
      CASE ('well')

        vn = integral(fi_v_fi, potential%a, potential%b)

      CASE ('coulomb', 'yukawa', 'user')

        IF (potential%r0 > 0) THEN      
          vn = integral(fi_v_fi, zero, potential%r0)        
        ELSE
          vn = integral_inf(fi_v_fi, zero)
        ENDIF

      END SELECT
    
    ELSE IF ((i >= ntrunc) .AND. (j >= ntrunc) ) THEN  ! The "--" area

      ii = i-ntrunc
      jj = j-ntrunc
    
      SELECT CASE(potential%type)
      CASE ('well')

        vn = integral(psi_v_psi, potential%a, potential%b)

      CASE ('coulomb', 'yukawa', 'user')

        IF (potential%r0 > 0) THEN 
          ! Integrate from 0 to r0     
          vn = integral(psi_v_psi, zero, potential%r0)
        ELSE
          ! Integrate from 0 to infinity
          vn = integral_inf(psi_v_psi, zero)
        ENDIF

      END SELECT
    
    ELSE    ! The "+-" and "-+" areas

      vn = zero
    
    ENDIF
  
  CASE ('non-relativistic')

    ii = i
    jj = j
  
    IF ( (i < ntrunc) .AND. (j < ntrunc) ) THEN
    
      SELECT CASE (potential%type)
      CASE ('well') 

        ! Square-well - integrate from a to b
        vn = integral(fi_v_fi, potential%a, potential%b)

      CASE ('coulomb', 'yukawa', 'user')

        IF (potential%r0 > 0) THEN
          ! Integrate from 0 to r0
          vn = integral(fi_v_fi, zero, potential%r0)   
        ELSE
          ! Integrate from 0 to infinity
          vn = integral(fi_v_fi, zero, potential%r0)   
        ENDIF

      END SELECT
    
    ELSE

      vn = zero

    ENDIF
  
  END SELECT

ENDIF ! (vn_is_loaded)

! Deallocated matrix containing potential readed from file
IF ( PRESENT(last_call) .AND. ALLOCATED(vn_matrix) ) DEALLOCATE(vn_matrix)


RETURN
END FUNCTION vn

!***********************************************************************

FUNCTION fi_v_fi(r)

! Forms function fi*v*fi to integrate.

USE jm_constants
USE JM_globals, ONLY: ii, jj

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: r
REAL(KIND=dp)             :: fi_v_fi
REAL(KIND=dp)             :: v, fim, fin


! Direct formula
!fi_v_fi = fi(ii,r) * v(r) * fi(jj,r)

! Computing fi(m,r) and fi(n,r) in one turn
CALL fi2(ii, jj, r, fim, fin)
fi_v_fi = fim * v(r) * fin

RETURN
END FUNCTION fi_v_fi

!***********************************************************************

FUNCTION psi_v_psi(r)

! Forms function psi*v*psi to integrate.

USE jm_constants
USE jm_globals, ONLY: ii, jj

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: r
REAL(KIND=dp)             :: psi_v_psi
REAL(KIND=dp)             :: v, psim, psin

! Direct formula
!psi_v_psi = psi(ii,r) * v(r) * psi(jj,r)

! Computing psi(m,r) and psi(n,r) in one turn
CALL psi2(ii, jj, r, psim, psin)
psi_v_psi = psim * v(r) * psin

RETURN
END FUNCTION psi_v_psi

!***********************************************************************

SUBROUTINE write_vn_to_file(dim,v)

! Save elements of the truncated potential to file.

USE jm_constants
USE jm_globals, ONLY: ntrunc, projectile, lambda, basis, scheme, potential

IMPLICIT none

INTEGER, INTENT(IN)       :: dim
REAL(KIND=dp), INTENT(IN) :: v(0:dim-1, 0:dim-1)
INTEGER           :: i, j
INTEGER           :: unit
CHARACTER(LEN=20) :: suffix


CALL convert_int_to_string(ntrunc, suffix)
unit = 4
OPEN(unit,file='saved.vn.'//trim(adjustl(suffix)))
CALL write_header(unit)
WRITE(unit,'(3I5,F5.2,A9,A1,A16,A1,A8)', advance='no') &
     ntrunc, projectile%l, projectile%kappa, lambda, basis, ' ', scheme,' ', &
     potential%type

SELECT CASE(potential%type)
CASE('well')
  WRITE(unit,'(3F22.16)') potential%v0, potential%a, potential%b
CASE('coulomb')
  WRITE(unit,'(3F22.16)') potential%z, potential%alpha, potential%r0
CASE('yukawa')
  WRITE(unit,'(3F22.16)') potential%g, potential%m, potential%r0
CASE('user')
  WRITE(unit,'(F22.16)') potential%r0
END SELECT


DO i=1,dim
  DO j=1,i                       ! Write elements of the truncated
    WRITE(unit,*) v(i-1,j-1)     ! potential to file to future use.
  ENDDO                          ! Only upper half is writen, due to symmetry.
ENDDO
CLOSE(unit)

END SUBROUTINE write_vn_to_file

!***********************************************************************

SUBROUTINE read_vn_from_file(dim,v)

! Reads previously saved elements of the truncated potential.

USE jm_constants
USE jm_globals, ONLY: ntrunc, projectile, lambda, basis, scheme, scr, pass, &
                      vn_is_loaded, vn_file_match, potential

IMPLICIT NONE

INTEGER,       INTENT(IN)  :: dim
REAL(KIND=dp), INTENT(OUT) :: v(0:dim-1,0:dim-1)
INTEGER                    :: nend_, l_, kappa_
REAL(KIND=dp)              :: lambda_, v0_, a_, b_, z_, alpha_, r0_, g_, m_
CHARACTER(LEN=8)           :: basis_, potentialtype_
CHARACTER(LEN=16)          :: scheme_
CHARACTER(LEN=80)          :: tmp
INTEGER                    :: unit, i, j

vn_is_loaded = .FALSE.
unit = 4

CALL convert_int_to_string(ntrunc, tmp)
OPEN(unit,file='saved.vn.'//trim(adjustl(tmp)),status='old',err=2)

DO
  READ(unit,'(A80)', err=2, end=2) tmp     ! Read and ignore the header 
  IF (index(tmp,";") == 0) EXIT            ! from the input file
ENDDO
BACKSPACE(unit)

! Testing if the <saved.vn> file match the current calculations
READ(unit,'(3I5,F5.2,A9,A1,A16,A1,A8)',err=2,end=2,advance='no')   &
           nend_,l_,kappa_,lambda_,basis_,tmp,scheme_,tmp,potentialtype_

SELECT CASE(potentialtype_)
CASE('well')
  READ(unit,'(3F22.16)') v0_, a_, b_
CASE('coulomb')
  READ(unit,'(3F22.16)') z_, alpha_, r0_
CASE('yukawa')
  READ(unit,'(3F22.16)') g_, m_, r0_
CASE('user')
  READ(unit,'(F22.16)') r0_
END SELECT


IF (ntrunc==nend_ .AND. projectile%l==l_ .AND. projectile%kappa==kappa_ &
   .AND. scheme==scheme_ .AND. lambda==lambda_ .AND. basis==basis_      &
   .AND. potential%type==potentialtype_                                 &
   .AND. ( (potential%type=='well' .AND. potential%v0==v0_ .AND.        &
            potential%a==a_  .AND. potential%b==b_)  .OR.               &
           (potential%type=='coulomb' .AND. potential%z==z_ .AND.       &
            potential%alpha==alpha_ .AND. potential%r0==r0_ )  .OR.     &
           (potential%type=='yukawa' .AND. potential%g==g_ .AND.        &
            potential%m==m_ .AND. potential%r0==r0_ )  .OR.              &
           (potential%type=='user' .AND. potential%r0==r0_) )       ) THEN


  DO i=1,dim
    DO j=1,i
      READ(unit,*,end=2) v(i-1,j-1)       ! Read elements of the truncated
    ENDDO                                 ! potential from file.
  ENDDO

  vn_is_loaded = .TRUE.
  IF(scr .AND. pass == 1) CALL show_info &
           ('Elements of the truncated potential have been read from file')
ELSE
  ! saved_vn doesn't match, set flag to avoid further checking
  vn_file_match = .false.
ENDIF

CLOSE(unit)
RETURN

! Invalid file - close it and don't check it anymore
2 CLOSE(unit) 
vn_file_match = .false.
RETURN

END SUBROUTINE read_vn_from_file

!***********************************************************************      

FUNCTION read_n_to_save_v()

IMPLICIT none

INTEGER read_n_to_save_v
INTEGER unit


unit = 4

OPEN(unit,file='jm.par',status='old',err=2)

READ(unit,*,err=3,end=3) read_n_to_save_v
CLOSE(unit)
RETURN

3 read_n_to_save_v = 0 
  CLOSE(unit)
2 RETURN

END FUNCTION read_n_to_save_v
