subroutine rs ( n, a, w, z, ierr )

!*******************************************************************************
!
!! RS computes eigenvalues and eigenvectors of real symmetric matrix.
!
!  Discussion:
!
!    This subroutine calls the recommended sequence of
!    subroutines from the eigensystem subroutine package (eispack)
!    to find the eigenvalues and eigenvectors (if desired)
!    of a double precision symmetric matrix.
!
!  Modified:
!
!    04 February 2003
!    06 December 2006 - adapation to the JMATRIX program
!
!  Reference:
!
!    James Wilkinson, Christian Reinsch,
!    Handbook for Automatic Computation,
!    Volume II, Linear Algebra, Part 2,
!    Springer Verlag, 1971.
!
!    Brian Smith, James Boyle, Jack Dongarra, Burton Garbow, Y Ikebe, 
!    V Klema, Cleve Moler,
!    Matrix Eigensystem Routines, EISPACK Guide,
!    Lecture Notes in Computer Science, Volume 6,
!    Springer Verlag, 1976.
!
!  Parameters:
!
!    Input, integer N, the order of the matrix.
!
!    Input, double precision A(N,N), the real symmetric matrix.
!
!    Input, integer MATZ, is zero if only eigenvalues are desired, and
!    nonzero if both eigenvalues and eigenvectors are desired.
!
!    Output, double precision W(N), the eigenvalues in ascending order.
!
!    Output, double precision Z(N,N), contains the eigenvectors, if MATZ is nonzero.
!
!    Output, integer IERR, is set equal to an error
!    completion code described in the documentation for TQLRAT and TQL2.  
!    The normal completion code is zero.
!
  implicit none

  integer n

  double precision a(n,n)
  double precision fv1(n)
  integer ierr
  double precision w(n)
  double precision z(n,n)


  call tred2 ( n, a, w, fv1, z )

  call tql2 ( n, w, fv1, z, ierr )


  return
end




subroutine tred2 ( n, a, d, e, z )

!*******************************************************************************
!
!! TRED2 transforms a real symmetric matrix to symmetric tridiagonal form.
!
!  Discussion:
!
!    This subroutine reduces a real symmetric matrix to a
!    symmetric tridiagonal matrix using and accumulating
!    orthogonal similarity transformations.
!
!    A and Z may coincide, in which case a single storage area is used
!    for the input of A and the output of Z.
!
!  Modified:
!
!    04 February 2003
!
!  Reference:
!
!    Martin, Christian Reinsch, James Wilkinson,
!    TRED2,
!    Numerische Mathematik, 
!    Volume 11, pages 181-195, 1968.
!
!    James Wilkinson, Christian Reinsch,
!    Handbook for Automatic Computation,
!    Volume II, Linear Algebra, Part 2,
!    Springer Verlag, 1971.
!
!    Brian Smith, James Boyle, Jack Dongarra, Burton Garbow, Y Ikebe, 
!    V Klema, Cleve Moler,
!    Matrix Eigensystem Routines, EISPACK Guide,
!    Lecture Notes in Computer Science, Volume 6,
!    Springer Verlag, 1976.
!
!  Parameters:
!
!    Input, integer N, the order of the matrix.
!
!    Input, double precision A(N,N), the real symmetric input matrix.  Only the
!    lower triangle of the matrix need be supplied.
!
!    Output, real D(N), the diagonal elements of the tridiagonal matrix.
!
!    Output, real E(N), contains the subdiagonal elements of the tridiagonal
!    matrix in E(2:N).  E(1) is set to zero.
!
!    Output, real Z(N,N), the orthogonal transformation matrix produced 
!    in the reduction.
!
  implicit none

  integer n

  double precision a(n,n)
  double precision d(n)
  double precision e(n)
  double precision f
  double precision g
  double precision h
  double precision hh
  integer i
  integer ii
  integer j
  integer k
  integer l
  double precision scale
  double precision z(n,n)

!   z(i:n,i) = a(i:n,i)    
!$OMP PARALLEL DO IF (n>50)
  do i = 1, n
    do j = i, n
      z(j,i) = a(j,i)
    enddo  
  end do
!$OMP END PARALLEL DO

!  d(1:n) = a(n,1:n)
!$OMP PARALLEL DO IF (n>50)
  do i=1, n
    d(i) = a(n,i)
  enddo
!$OMP END PARALLEL DO  

  do ii = 2, n

    i = n + 2 - ii
    l = i - 1
    h = 0.0D+00
    scale = 0.0D+00
!
!  Scale row.
!

!$OMP PARALLEL DO REDUCTION(+:scale) IF (l>50)
    do k = 1, l
      scale = scale + abs ( d(k) )
    end do
!$OMP END PARALLEL DO    

    if ( scale == 0.0D+00 ) then

      e(i) = d(l)

!$OMP PARALLEL DO IF (l>50)
      do j = 1, l
        d(j) = z(l,j)
        z(i,j) = 0.0D+00
        z(j,i) = 0.0D+00
      end do
!$OMP END PARALLEL DO

      go to 290

    end if

!   d(1:l) = d(1:l) / scale
!$OMP PARALLEL DO IF (l>50)
    DO j=1, l
      d(j) = d(j) / scale
    ENDDO
!$OMP END PARALLEL DO

!    h = h + dot_product ( d(1:l), d(1:l) )
!$OMP PARALLEL WORKSHARE IF (l>50)
     f = dot_product ( d(1:l), d(1:l) )
!$OMP END PARALLEL WORKSHARE
    h = h + f

    f = d(l)
    g = - sign ( sqrt ( h ), f )
    e(i) = scale * g
    h = h - f * g
    d(l) = f - g
!
!  Form A*U.
!
!    e(1:l) = 0.0D+00
!$OMP PARALLEL DO IF (l>50)
    do j=1,l
      e(j) = 0.0D+00
    enddo
!$OMP END PARALLEL DO

    do j = 1, l

      f = d(j)
      z(j,i) = f
      g = e(j) + z(j,j) * f

      ! Do not parallelize
      do k = j+1, l
        g = g + z(k,j) * d(k)
        e(k) = e(k) + z(k,j) * f
      end do

      e(j) = g

    end do
!
!  Form P.
!
!   e(1:l) = e(1:l) / h
!$OMP PARALLEL DO IF (l>50)
    DO j=1, l
      e(j) = e(j) / h
    ENDDO
!$OMP END PARALLEL DO

!$OMP PARALLEL WORKSHARE IF (l>50)
    f = dot_product ( e(1:l), d(1:l) )
!$OMP END PARALLEL WORKSHARE

    hh = 0.5D+00 * f / h
!
!  Form Q.
!
!    e(1:l) = e(1:l) - hh * d(1:l)
!$OMP PARALLEL DO IF(l>50)
    do j=1,l
      e(j) = e(j) - hh * d(j)
    enddo 
!$OMP END PARALLEL DO

!
!  Form reduced A.
!
    do j = 1, l

      f = d(j)
      g = e(j)

      ! Do not paralellize! Huge slow-down!
      ! TODO: Check why
      z(j:l,j) = z(j:l,j) - f * e(j:l) - g * d(j:l)

      d(j) = z(l,j)
      z(i,j) = 0.0D+00

    end do

290 continue

    d(i) = h


  end do
!
!  Accumulation of transformation matrices.
!
  do i = 2, n

    l = i - 1
    z(n,l) = z(l,l)
    z(l,l) = 1.0D+00
    h = d(i)
    if ( h /= 0.0D+00 ) then

!     d(1:l) = z(1:l,i) / h
!$OMP PARALLEL DO IF (l>50)
      DO j=1, l
        d(j) = z(j,i) / h
      ENDDO
!$OMP END PARALLEL DO

!$OMP PARALLEL DO PRIVATE(g) IF (l>50)
      do j = 1, l

        g = dot_product ( z(1:l,i), z(1:l,j) )

        !Do not parallelize alone
        do k = 1, l
          z(k,j) = z(k,j) - g * d(k)
        end do

      end do
!$OMP END PARALLEL DO

    end if

!$OMP PARALLEL DO IF (l>50)
    do j=1,l
      z(j,i) = 0.0D+00
    enddo
!$OMP END PARALLEL DO

  end do

!  d(1:n) = z(n,1:n)
!$OMP PARALLEL DO IF (n>50)
  do j=1,n
    d(j) = z(n,j)
  enddo
!$OMP END PARALLEL DO

!  z(n,1:n-1) = 0.0D+00
!$OMP PARALLEL DO IF (n>50)
  do j=1,n-1
    z(n,j) = 0.0D+00
  enddo
!$OMP END PARALLEL DO

  z(n,n) = 1.0D+00

  e(1) = 0.0D+00

  return
end


subroutine tql2 ( n, d, e, z, ierr )

!*******************************************************************************
!
!! TQL2 computes all eigenvalues/vectors, double precision symmetric tridiagonal matrix.
!
!  Discussion:
!
!    This subroutine finds the eigenvalues and eigenvectors of a symmetric 
!    tridiagonal matrix by the QL method.  The eigenvectors of a full 
!    symmetric matrix can also be found if TRED2 has been used to reduce this
!    full matrix to tridiagonal form.
!
!  Modified:
!
!    04 February 2003
!
!  Reference:
!
!    Bowdler, Martin, Christian Reinsch, James Wilkinson,
!    TQL2,
!    Numerische Mathematik,
!    Volume 11, pages 293-306, 1968.
!
!    James Wilkinson, Christian Reinsch,
!    Handbook for Automatic Computation,
!    Volume II, Linear Algebra, Part 2,
!    Springer Verlag, 1971.
!
!    Brian Smith, James Boyle, Jack Dongarra, Burton Garbow, Y Ikebe, 
!    V Klema, Cleve Moler,
!    Matrix Eigensystem Routines, EISPACK Guide,
!    Lecture Notes in Computer Science, Volume 6,
!    Springer Verlag, 1976.
!
!  Parameters:
!
!    Input, integer N, the order of the matrix.
!
!    Input/output, real D(N).  On input, the diagonal elements of the matrix.
!    On output, the eigenvalues in ascending order.  If an error exit is 
!    made, the eigenvalues are correct but unordered for indices 1,2,...,IERR-1.
!
!    Input/output, real E(N).  On input, E(2:N) contains the subdiagonal 
!    elements of the input matrix, and E(1) is arbitrary.
!    On output, E has been destroyed.
!
!    Input, real Z(N,N).  On input, the transformation matrix produced in 
!    the reduction by TRED2, if performed.  If the eigenvectors of the 
!    tridiagonal matrix are desired, Z must contain the identity matrix.
!    On output, Z contains the orthonormal eigenvectors of the symmetric
!    tridiagonal (or full) matrix.  If an error exit is made, Z contains 
!    the eigenvectors associated with the stored eigenvalues.
!
!    Output, integer IERR, error flag.
!    0, normal return,
!    J, if the J-th eigenvalue has not been determined after 
!    30 iterations.
!
  implicit none

  integer n

  double precision c
  double precision c2
  double precision c3
  double precision d(n)
  double precision dl1
  double precision e(n)
  double precision el1
  double precision f
  double precision g
  double precision h
  integer i
  integer ierr
  integer ii
  integer j
  integer k
  integer l
  integer l1
  integer l2
  integer m
  integer mml
  double precision p
  double precision pythag
  double precision r
  double precision s
  double precision s2
  double precision tst1
  double precision tst2
  double precision z(n,n)

  ierr = 0

  if ( n == 1 ) return

  do i = 2, n
    e(i-1) = e(i)
  end do

  f = 0.0D+00
  tst1 = 0.0D+00
  e(n) = 0.0D+00

  do l = 1, n

     j = 0
     h = abs ( d(l) ) + abs ( e(l) )
     tst1 = max ( tst1, h )
!
!  Look for a small sub-diagonal element.
!
     do m = l, n
       tst2 = tst1 + abs ( e(m) )
       if ( tst2 == tst1 ) exit
     end do

     if ( m == l ) go to 220

 130 continue

     if ( j >= 30 ) then
       ierr = l
       return
     end if

     j = j + 1
!
!  Form shift.
!
     l1 = l + 1
     l2 = l1 + 1
     g = d(l)
     p = ( d(l1) - g ) / ( 2.0D+00 * e(l) )
     r = pythag ( p, 1.0D+00 )
     d(l) = e(l) / ( p + sign ( r, p ) )
     d(l1) = e(l) * ( p + sign ( r, p ) )
     dl1 = d(l1)
     h = g - d(l)
!    d(l2:n) = d(l2:n) - h
!$OMP PARALLEL DO IF (n-l2 > 50)
     do k=l2,n
       d(k) = d(k) - h
     enddo
!$OMP END PARALLEL DO

     f = f + h
!
!  QL transformation.
!
     p = d(m)
     c = 1.0D+00
     c2 = c
     el1 = e(l1)
     s = 0.0D+00
     mml = m - l

     do ii = 1, mml

        c3 = c2
        c2 = c
        s2 = s
        i = m - ii
        g = c * e(i)
        h = c * p
        r = pythag ( p, e(i) )
        e(i+1) = s * r
        s = e(i) / r
        c = p / r
        p = c * d(i) - s * g
        d(i+1) = h + s * ( c * g + s * d(i) )
!
!  Form vector.
!
        ! Do not parallelize
        do k = 1, n
          h = z(k,i+1)
          z(k,i+1) = s * z(k,i) + c * h
          z(k,i) = c * z(k,i) - s * h
        end do
     end do

     p = - s * s2 * c3 * el1 * e(l) / dl1
     e(l) = s * p
     d(l) = c * p
     tst2 = tst1 + abs ( e(l) )

     if ( tst2 > tst1 ) then 
       go to 130
     end if

220  continue

     d(l) = d(l) + f

  end do

!  Order eigenvalues and eigenvectors - removed

  return

end


function pythag ( a, b )

!*******************************************************************************
!
!! PYTHAG computes SQRT ( A**2 + B**2 ) carefully.
!
!  Discussion:
!
!    The formula
!
!      PYTHAG = sqrt ( A**2 + B**2 )
!
!    is reasonably accurate, but can fail if, for example, A**2 is larger 
!    than the machine overflow.  The formula can lose most of its accuracy 
!    if the sum of the squares is very large or very small.
!
!  Modified:
!
!    04 February 2003
!
!  Reference:
!
!    James Wilkinson, Christian Reinsch,
!    Handbook for Automatic Computation,
!    Volume II, Linear Algebra, Part 2,
!    Springer Verlag, 1971.
!
!    Brian Smith, James Boyle, Jack Dongarra, Burton Garbow, Y Ikebe, 
!    V Klema, Cleve Moler,
!    Matrix Eigensystem Routines, EISPACK Guide,
!    Lecture Notes in Computer Science, Volume 6,
!    Springer Verlag, 1976.
!
!  Modified:
!
!    04 February 2003
!
!  Parameters:
!
!    Input, real A, B, the two legs of a right triangle.
!
!    Output, real PYTHAG, the length of the hypotenuse.
!
  implicit none

  double precision a
  double precision b
  double precision p
  double precision pythag
  double precision r
  double precision s
  double precision t
  double precision u

  p = max ( abs ( a ), abs ( b ) )

  if ( p /= 0.0D+00 ) then

    r = ( min ( abs ( a ), abs ( b ) ) / p )**2

    do

      t = 4.0D+00 + r

      if ( t == 4.0D+00 ) then
        exit
      end if

      s = r / t
      u = 1.0D+00 + 2.0D+00 * s
      p = u * p
      r = ( s / u )**2 * r

    end do

  end if

  pythag = p

  return
end
