FUNCTION laguerre(n,a,x)

! Computes associated Laguerre polynomials.

USE jm_constants

IMPLICIT none

INTEGER,       INTENT(IN)  :: n
REAL(KIND=dp), INTENT(IN)  :: a, x

REAL(KIND=dp)  :: laguerre
REAL(KIND=dp)  :: lim1, lim2
INTEGER        :: i

! Old formula:
! laguerre1 = exp(lgamma(n+a+1)-lfact(n))/gamma(a+1)*hyperg11(-dble(n),a+1,x)
! Hyperg11 fails for some arguments, not accaurate enough for big n
! Replaced by the recursive relation

! New formula: recursive relation: slower, but reliable
! http://functions.wolfram.com/Polynomials/LaguerreL3/03/ShowAll.html
! One may want to use tables to collect values, it is simplier, but slower.
! Using alternate RETURNs to speed-up calculations

SELECT CASE (n)

  CASE (0)
    laguerre = one
    RETURN

  CASE (1)
    laguerre = one + a - x
    RETURN

  CASE (2)
    laguerre = one + half*a*(three + a) - x*(two + a - half*x)
    RETURN

  CASE DEFAULT  ! n > 2, use recursive relation

    lim2 = one - x + a
    lim1 = one + half*a*(three + a) - x*(two + a - half*x)

    DO i=3,n
      laguerre = ( (two*i + a - one - x)*lim1 - (i + a - one)*lim2 )/i
      lim2 = lim1
      lim1 = laguerre
    ENDDO
    RETURN

END SELECT

RETURN
END FUNCTION laguerre

!***********************************************************************

SUBROUTINE laguerre2(m,n,a,x,lagm,lagn)

! Computes associated Laguerre polynomials.
! This function calculate lagm=L(m,a,x) and lagn=L(n,a,x) at once.
! Also see description of the laguerre() function above

USE jm_constants

IMPLICIT none

INTEGER,       INTENT(IN)  :: m, n
REAL(KIND=dp), INTENT(IN)  :: a, x
REAL(KIND=dp), INTENT(OUT) :: lagm, lagn
REAL(KIND=dp)              :: lim1, lim2
INTEGER                    :: i

lim2 = one - x + a
lim1 = one + half*a*(three + a) - x*(two + a - half*x)

SELECT CASE (n)
  CASE (0)
    lagn = one
  CASE (1) 
    lagn = lim2
  CASE (2) 
    lagn = lim1
END SELECT

SELECT CASE (m)
  CASE (0) 
    lagm = one
  CASE (1) 
    lagm = lim2
  CASE (2)
    lagm = lim1
END SELECT

IF (m < n) THEN

  DO i=3,n
    lagn = ( (two*i + a - one - x)*lim1 - (i + a - one)*lim2 )/i
    if (i == m) lagm = lagn
    lim2 = lim1
    lim1 = lagn
  ENDDO
  RETURN

ELSE IF (n < m) THEN

  DO i=3,m
    lagm = ( (two*i + a - one - x)*lim1 - (i + a - one)*lim2 )/i
    if (i == n) lagn = lagm
    lim2 = lim1
    lim1 = lagm
  ENDDO
  RETURN

ELSE   ! n = m

  DO i=3,m
    lagm = ( (two*i + a - one - x)*lim1 - (i + a - one)*lim2 )/i
    lim2 = lim1
    lim1 = lagm
  ENDDO
  lagn = lagm
  RETURN

ENDIF

RETURN
END SUBROUTINE laguerre2

!***********************************************************************

FUNCTION bessel(v,z)

! Computes Bessel function of the first kind.

USE jm_constants
USE jm_globals, ONLY: delta_analytic_error

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: v, z
REAL(KIND=dp)             :: res(100), v_frac
REAL(KIND=dp)             :: hyperg01, gamma
INTEGER                   :: nerr, v_int
REAL(KIND=dp)             :: bessel

v_int  = INT(v)
v_frac = v - v_int

IF (v < zero) THEN
  
  IF (v == INT(v)) THEN

    ! Call the external routine
    CALL rjbesl(z, -v_frac, -v_int+1, res, nerr)
    bessel = (-1)**INT(v) * res(-v_int+1)
    
  ELSE  ! Direct formula (may be wrong for huge <z> and <v>)

    bessel = (z/two)**INT(v) / gamma(v+one) * hyperg01(v+one,-quarter * z**2)
    
  ENDIF
  
ELSE    ! Positive <v> values

  ! Call the external routine
  CALL rjbesl(z, v_frac, v_int+1, res, nerr)
  bessel = res(v_int+1)
  
ENDIF

IF (nerr < 0) delta_analytic_error = .true.

RETURN
END FUNCTION bessel

!***********************************************************************

FUNCTION sphbess(v,z)

! Computes spherical Bessel function.

USE jm_constants
USE jm_globals, ONLY: delta_analytic_error

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: v, z
REAL(KIND=dp)             :: sphbess
REAL(KIND=dp)             :: bessel

IF (z == zero) THEN

  delta_analytic_error = .true.

ELSE

  sphbess = sqrt(pi/two/z) * bessel(v+half,z)

ENDIF

RETURN
END FUNCTION sphbess

!***********************************************************************

FUNCTION sphbess_prim(v,z)

! Computes derivative to the spherical Bessel function.

USE jm_constants

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: v, z
REAL(KIND=dp)             :: sphbess_prim
REAL(KIND=dp)             :: sphbess

sphbess_prim = (v*sphbess(v-1,z) - (v+1)*sphbess(v+1,z)) / (2*v+1)

END FUNCTION sphbess_prim

!***********************************************************************

FUNCTION neumann(v,z)

! Computes Neumann function (Bessel function of the second kind).

USE jm_constants
USE jm_globals, ONLY: delta_analytic_error

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: v, z
REAL(KIND=dp)             :: neumann
REAL(KIND=dp)             :: res(100)
REAL(KIND=dp)             :: v_frac
REAL(KIND=dp)             :: bessel
INTEGER                   :: v_int, nerr

v_int = INT(v)
v_frac = v - v_int

IF (v < zero) THEN
  
  IF (v == INT(v)) THEN  ! Negative integer vaules

    ! Call the external routine
    CALL rybesl(z, -v_frac, -v_int+1, res, nerr)
    neumann = (-1)**INT(v) * res(-v_int+1)
    
  ELSE    ! Direct formula

    neumann = one / sin(v*pi) * ( bessel(v,z)*cos(v*pi)-bessel(-v,z)   )
    
  ENDIF
  
ELSE

  ! Call the external routine
  CALL rybesl(z, v_frac, v_int+1, res, nerr)
  neumann = res(v_int+1)
  
ENDIF

IF (nerr < 0) delta_analytic_error = .true.

RETURN
END FUNCTION neumann

!***********************************************************************

FUNCTION sphneum(v,z)

! Computes spherical Neumann function

USE jm_constants
USE jm_globals, ONLY: delta_analytic_error

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: v, z
REAL(KIND=dp)             :: sphneum
REAL(KIND=dp)             :: neumann

IF (z == zero) THEN

  delta_analytic_error = .true.

ELSE

  sphneum = sqrt(pi/two/z) * neumann(v+half,z)

ENDIF

RETURN
END FUNCTION sphneum

!***********************************************************************

FUNCTION sphneum_prim(v,z)

! Computes derivative to the spherical Neumann function.

USE jm_constants

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: v, z
REAL(KIND=dp)             :: sphneum_prim
REAL(KIND=dp)             :: sphneum

sphneum_prim = (v*sphneum(v-1,z) - (v+1)*sphneum(v+1,z)) / (2*v+1)

END FUNCTION sphneum_prim

!***********************************************************************

FUNCTION kronecker(a,b)

! Computes the Kronecker delta

IMPLICIT none
                                   
INTEGER, INTENT(IN) :: a, b
INTEGER             :: kronecker

IF ( a == b ) THEN
  kronecker = 1
ELSE
  kronecker = 0
ENDIF

RETURN
END FUNCTION kronecker

!***********************************************************************

FUNCTION gamma(x)

! Computes the gamma function
! This is the interface to the external function only.

USE jm_constants

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: x
REAL(KIND=dp)             :: gamma, dgamma

gamma = dgamma(x)

IF (gamma == 1.d308) CALL show_error('gamma', 'improper argument')

RETURN
END FUNCTION gamma

!***********************************************************************

FUNCTION lgamma(x)

! Computes logarithm to the gamma function.
! Uses the external library routine

USE jm_constants

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: x
REAL(KIND=dp)             :: lgamma
REAL(KIND=dp)             :: dlgamma

lgamma = dlgamma(x)

RETURN
END FUNCTION lgamma

!***********************************************************************

FUNCTION gegen(n,l,x)

! Calculates the Gegenbauer polynomials for argument -1 < <x> < 1 
! for the integer <l>.

USE jm_constants

IMPLICIT none

INTEGER,       INTENT(IN) :: n, l
REAL(KIND=dp), INTENT(IN) :: x
REAL(KIND=dp)             :: gegen
REAL(KIND=dp)             :: gamma, lfact
INTEGER                   :: m

IF (dabs(x) > one) CALL show_error('gegen','improper argument')

gegen = zero

DO m=0,n
  gegen = gegen + exp(lfact(l+m-1)-lfact(m)) * cos((dble(n-2*m))*acos(x)) &
      * exp(lfact(l+n-m-1)-lfact(n-m)) / gamma(dble(l))**2
ENDDO

RETURN
END FUNCTION gegen

!***********************************************************************

FUNCTION hyperg11(a,c,z)

! Computes the hypergeometric function 1F1(a,c;z)

USE jm_constants

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: a, c, z
REAL(KIND=dp)             :: hyperg11
REAL(KIND=dp)             :: s, d, y
INTEGER                   :: n

s = one
y = one

DO n=1,500
  d = n * ( (c+n-one)**2 )
  y = (a+n-one) * (y/d)
  y = y * (c+n-one)
  y = y * z
  IF (s == s+y) EXIT
  s = s + y
ENDDO

IF (n == 500) CALL show_error('hyperg11','convergence failure')

hyperg11 = s

RETURN
END FUNCTION hyperg11

!***********************************************************************

FUNCTION hyperg21(a,b,c,z)

! Computes the hypergeometric function 2F1(a,b,c;z)

USE jm_constants

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: a, b, c, z
REAL(KIND=dp)             :: hyperg21
REAL(KIND=dp)             :: s, y
INTEGER                   :: n

s = one
y = one

DO n=0,500
  y = y*(a+n)/(c+n)*(b+n)*z/(n+1)
  IF (s == s+y ) EXIT
  s = s+y
ENDDO

IF (n == 500) CALL show_error('hyperg21','convergence failure')

hyperg21 = s

RETURN
END FUNCTION hyperg21

!***********************************************************************

FUNCTION hyperg01(c,z)

! Computes the hypergeometric function 0F1(c;z)

USE jm_constants

IMPLICIT none

REAL(KIND=dp), INTENT(IN) :: c, z
REAL(KIND=dp)             :: hyperg01
REAL(KIND=dp)             :: s, d, y
INTEGER                   :: n

s = one
y = one

DO n=1,500
  d = n * ( (c+n-one)**2 )
  y = y/d
  y = y * (c+n-one)
  y = y * z
  IF (s == s+y) EXIT
  s = s + y
ENDDO

IF (n == 500) CALL show_error('hyperg01','convergence failure')

hyperg01 = s

RETURN
END FUNCTION hyperg01

!***********************************************************************

FUNCTION fact_over_fact(m, n)

! Calculates (m!/n!) for m>n

USE jm_constants

IMPLICIT none

INTEGER, INTENT(in) :: m, n
REAL(KIND=dp)       :: fact_over_fact
INTEGER             :: i

IF (m<n) CALL show_error('fact_over_fact','improper arguments')

fact_over_fact = n + 1
DO i = n+2, m
  fact_over_fact = fact_over_fact * i
ENDDO

RETURN
END FUNCTION fact_over_fact

!***********************************************************************      

FUNCTION fact(n)

! Computes factorial of the natural number <n>. The output is real-type.
! Values are tabulated to speed-up calculations

USE jm_constants

IMPLICIT none

INTEGER, INTENT(IN) :: n
REAL(KIND=dp)       :: fact
REAL(KIND=dp), SAVE :: t_f(0:170)


IF (n < 0)   CALL show_error('fact','negative argument')
IF (n > 170) CALL show_error('fact','argument too big')

IF (t_f(0) /= one) THEN                  ! First call - tabulate values
  t_f(0)=   one
  t_f(1)=   one
  t_f(2)=   two
  t_f(3)=   six
  t_f(4)=   24.00000000000000000000000000000000000000D0
  t_f(5)=   120.0000000000000000000000000000000000000D0
  t_f(6)=   720.0000000000000000000000000000000000000D0
  t_f(7)=   5040.000000000000000000000000000000000000D0
  t_f(8)=   40320.00000000000000000000000000000000000D0
  t_f(9)=   362880.0000000000000000000000000000000000D0
  t_f(10)=  3.628800000000000000000000000000000000000D6
  t_f(11)=  3.991680000000000000000000000000000000000D7
  t_f(12)=  4.790016000000000000000000000000000000000D8
  t_f(13)=  6.227020800000000000000000000000000000000D9
  t_f(14)=  8.717829120000000000000000000000000000000D10
  t_f(15)=  1.307674368000000000000000000000000000000D12
  t_f(16)=  2.092278988800000000000000000000000000000D13
  t_f(17)=  3.556874280960000000000000000000000000000D14
  t_f(18)=  6.402373705728000000000000000000000000000D15
  t_f(19)=  1.216451004088320000000000000000000000000D17
  t_f(20)=  2.432902008176640000000000000000000000000D18
  t_f(21)=  5.109094217170944000000000000000000000000D19
  t_f(22)=  1.124000727777607680000000000000000000000D21
  t_f(23)=  2.585201673888497664000000000000000000000D22
  t_f(24)=  6.204484017332394393600000000000000000000D23
  t_f(25)=  1.551121004333098598400000000000000000000D25
  t_f(26)=  4.032914611266056355840000000000000000000D26
  t_f(27)=  1.088886945041835216076800000000000000000D28
  t_f(28)=  3.048883446117138605015040000000000000000D29
  t_f(29)=  8.841761993739701954543616000000000000000D30
  t_f(30)=  2.652528598121910586363084800000000000000D32
  t_f(31)=  8.222838654177922817725562880000000000000D33
  t_f(32)=  2.631308369336935301672180121600000000000D35
  t_f(33)=  8.683317618811886495518194401280000000000D36
  t_f(34)=  2.952327990396041408476186096435200000000D38
  t_f(35)=  1.033314796638614492966665133752320000000D40
  t_f(36)=  3.719933267899012174679994481508352000000D41
  t_f(37)=  1.376375309122634504631597958158090240000D43
  t_f(38)=  5.230226174666011117600072241000742912000D44
  t_f(39)=  2.039788208119744335864028173990289735680D46
  t_f(40)=  8.159152832478977343456112695961158942720D47
  t_f(41)=  3.345252661316380710817006205344075166515D49
  t_f(42)=  1.405006117752879898543142606244511569936D51
  t_f(43)=  6.041526306337383563735513206851399750726D52
  t_f(44)=  2.658271574788448768043625811014615890320D54
  t_f(45)=  1.196222208654801945619631614956577150644D56
  t_f(46)=  5.502622159812088949850305428800254892962D57
  t_f(47)=  2.586232415111681806429643551536119799692D59
  t_f(48)=  1.241391559253607267086228904737337503852D61
  t_f(49)=  6.082818640342675608722521633212953768876D62
  t_f(50)=  3.041409320171337804361260816606476884438D64
  t_f(51)=  1.551118753287382280224243016469303211063D66
  t_f(52)=  8.065817517094387857166063685640376697529D67
  t_f(53)=  4.274883284060025564298013753389399649690D69
  t_f(54)=  2.308436973392413804720927426830275810833D71
  t_f(55)=  1.269640335365827592596510084756651695958D73
  t_f(56)=  7.109985878048634518540456474637249497365D74
  t_f(57)=  4.052691950487721675568060190543232213498D76
  t_f(58)=  2.350561331282878571829474910515074683829D78
  t_f(59)=  1.386831185456898357379390197203894063459D80
  t_f(60)=  8.320987112741390144276341183223364380754D81
  t_f(61)=  5.075802138772247988008568121766252272260D83
  t_f(62)=  3.146997326038793752565312235495076408801D85
  t_f(63)=  1.982608315404440064116146708361898137545D87
  t_f(64)=  1.268869321858841641034333893351614808029D89
  t_f(65)=  8.247650592082470666723170306785496252186D90
  t_f(66)=  5.443449390774430640037292402478427526443D92
  t_f(67)=  3.647111091818868528824985909660546442717D94
  t_f(68)=  2.480035542436830599600990418569171581047D96
  t_f(69)=  1.711224524281413113724683388812728390923D98
  t_f(70)=  1.197857166996989179607278372168909873646D100
  t_f(71)=  8.504785885678623175211676442399260102886D101
  t_f(72)=  6.123445837688608686152407038527467274078D103
  t_f(73)=  4.470115461512684340891257138125051110077D105
  t_f(74)=  3.307885441519386412259530282212537821457D107
  t_f(75)=  2.480914081139539809194647711659403366093D109
  t_f(76)=  1.885494701666050254987932260861146558230D111
  t_f(77)=  1.451830920282858696340707840863082849837D113
  t_f(78)=  1.132428117820629783145752115873204622873D115
  t_f(79)=  8.946182130782975286851441715398316520698D116
  t_f(80)=  7.156945704626380229481153372318653216558D118
  t_f(81)=  5.797126020747367985879734231578109105412D120
  t_f(82)=  4.753643337012841748421382069894049466438D122
  t_f(83)=  3.945523969720658651189747118012061057144D124
  t_f(84)=  3.314240134565353266999387579130131288001D126
  t_f(85)=  2.817104114380550276949479442260611594801D128
  t_f(86)=  2.422709538367273238176552320344125971528D130
  t_f(87)=  2.107757298379527717213600518699389595230D132
  t_f(88)=  1.854826422573984391147968456455462843802D134
  t_f(89)=  1.650795516090846108121691926245361930984D136
  t_f(90)=  1.485715964481761497309522733620825737886D138
  t_f(91)=  1.352001527678402962551665687594951421476D140
  t_f(92)=  1.243841405464130725547532432587355307758D142
  t_f(93)=  1.156772507081641574759205162306240436215D144
  t_f(94)=  1.087366156656743080273652852567866010042D146
  t_f(95)=  1.032997848823905926259970209939472709540D148
  t_f(96)=  9.916779348709496892095714015418938011582D149
  t_f(97)=  9.619275968248211985332842594956369871234D151
  t_f(98)=  9.426890448883247745626185743057242473810D153
  t_f(99)=  9.332621544394415268169923885626670049072D155
  t_f(100)=  9.332621544394415268169923885626670049072D157
  t_f(101)=  9.425947759838359420851623124482936749562D159
  t_f(102)=  9.614466715035126609268655586972595484554D161
  t_f(103)=  9.902900716486180407546715254581773349090D163
  t_f(104)=  1.029901674514562762384858386476504428305D166
  t_f(105)=  1.081396758240290900504101305800329649721D168
  t_f(106)=  1.146280563734708354534347384148349428704D170
  t_f(107)=  1.226520203196137939351751701038733888713D172
  t_f(108)=  1.324641819451828974499891837121832599810D174
  t_f(109)=  1.443859583202493582204882102462797533793D176
  t_f(110)=  1.588245541522742940425370312709077287172D178
  t_f(111)=  1.762952551090244663872161047107075788761D180
  t_f(112)=  1.974506857221074023536820372759924883413D182
  t_f(113)=  2.231192748659813646596607021218715118256D184
  t_f(114)=  2.543559733472187557120132004189335234812D186
  t_f(115)=  2.925093693493015690688151804817735520034D188
  t_f(116)=  3.393108684451898201198256093588573203240D190
  t_f(117)=  3.969937160808720895401959629498630647790D192
  t_f(118)=  4.684525849754290656574312362808384164393D194
  t_f(119)=  5.574585761207605881323431711741977155627D196
  t_f(120)=  6.689502913449127057588118054090372586753D198
  t_f(121)=  8.094298525273443739681622845449350829971D200
  t_f(122)=  9.875044200833601362411579871448208012564D202
  t_f(123)=  1.214630436702532967576624324188129585545D205
  t_f(124)=  1.506141741511140879795014161993280686076D207
  t_f(125)=  1.882677176888926099743767702491600857595D209
  t_f(126)=  2.372173242880046885677147305139417080570D211
  t_f(127)=  3.012660018457659544809977077527059692324D213
  t_f(128)=  3.856204823625804217356770659234636406175D215
  t_f(129)=  4.974504222477287440390234150412680963966D217
  t_f(130)=  6.466855489220473672507304395536485253155D219
  t_f(131)=  8.471580690878820510984568758152795681634D221
  t_f(132)=  1.118248651196004307449963076076169029976D224
  t_f(133)=  1.487270706090685728908450891181304809868D226
  t_f(134)=  1.992942746161518876737324194182948445223D228
  t_f(135)=  2.690472707318050483595387662146980401050D230
  t_f(136)=  3.659042881952548657689727220519893345429D232
  t_f(137)=  5.012888748274991661034926292112253883237D234
  t_f(138)=  6.917786472619488492228198283114910358867D236
  t_f(139)=  9.615723196941089004197195613529725398826D238
  t_f(140)=  1.346201247571752460587607385894161555836D241
  t_f(141)=  1.898143759076170969428526414110767793728D243
  t_f(142)=  2.695364137888162776588507508037290267094D245
  t_f(143)=  3.854370717180072770521565736493325081944D247
  t_f(144)=  5.550293832739304789551054660550388118000D249
  t_f(145)=  8.047926057471991944849029257798062771100D251
  t_f(146)=  1.174997204390910823947958271638517164581D254
  t_f(147)=  1.727245890454638911203498659308620231933D256
  t_f(148)=  2.556323917872865588581178015776757943262D258
  t_f(149)=  3.808922637630569726985955243507369335460D260
  t_f(150)=  5.713383956445854590478932865261054003190D262
  t_f(151)=  8.627209774233240431623188626544191544816D264
  t_f(152)=  1.311335885683452545606724671234717114812D267
  t_f(153)=  2.006343905095682394778288746989117185662D269
  t_f(154)=  3.089769613847350887958564670363240465920D271
  t_f(155)=  4.789142901463393876335775239063022722176D273
  t_f(156)=  7.471062926282894447083809372938315446595D275
  t_f(157)=  1.172956879426414428192158071551315525115D278
  t_f(158)=  1.853271869493734796543609753051078529682D280
  t_f(159)=  2.946702272495038326504339507351214862195D282
  t_f(160)=  4.714723635992061322406943211761943779512D284
  t_f(161)=  7.590705053947218729075178570936729485014D286
  t_f(162)=  1.229694218739449434110178928491750176572D289
  t_f(163)=  2.004401576545302577599591653441552787813D291
  t_f(164)=  3.287218585534296227263330311644146572013D293
  t_f(165)=  5.423910666131588774984495014212841843822D295
  t_f(166)=  9.003691705778437366474261723593317460744D297
  t_f(167)=  1.503616514864999040201201707840084015944D300
  t_f(168)=  2.526075744973198387538018869171341146786D302
  t_f(169)=  4.269068009004705274939251888899566538069D304
  t_f(170)=  7.257415615307998967396728211129263114717D306

ENDIF

fact = t_f(n)

RETURN
END FUNCTION fact

!***********************************************************************

FUNCTION lfact(n)

! Computes logarithm to the factorial. 
! Values are tabulated.

USE jm_constants

IMPLICIT none

INTEGER, INTENT(IN) :: n
REAL(KIND=dp)       :: lfact
REAL(KIND=dp), SAVE :: t_lf(0:5003)


IF (n < 0)    CALL show_error('lfact','negative argument')
IF (n > 5003) CALL show_error('lfact','argument too big')

IF (t_lf(9) < ten) THEN               ! First call

  t_lf(0)=   zero
  t_lf(1)=   zero
  t_lf(2)=   0.6931471805599453094172321214581765680755_dp
  t_lf(3)=   1.791759469228055000812477358380702272723_dp
  t_lf(4)=   3.178053830347945619646941601297055408874_dp
  t_lf(5)=   4.787491742782045994247700934523243048400_dp
  t_lf(6)=   6.579251212010100995060178292903945321123_dp
  t_lf(7)=   8.525161361065414300165531036347125050760_dp
  t_lf(8)=   10.60460290274525022841722740072165475499_dp
  t_lf(9)=   12.80182748008146961120771787456670616428_dp
  t_lf(10)=  15.10441257307551529522570932925107037188_dp
  t_lf(11)=  17.50230784587388583928765290721619967170_dp
  t_lf(12)=  19.98721449566188614951736238705507851250_dp
  t_lf(13)=  22.55216385312342288557084982862039711731_dp
  t_lf(14)=  25.19122118273868150009343469352175341502_dp
  t_lf(15)=  27.89927138384089156608943926367046675919_dp
  t_lf(16)=  30.67186010608067280375836774950317303150_dp
  t_lf(17)=  33.50507345013688888400790236737629956708_dp
  t_lf(18)=  36.39544520803305357621562496267952754445_dp
  t_lf(19)=  39.33988418719949403622465239456738108169_dp
  t_lf(20)=  42.33561646075348502965987597070992185737_dp
  t_lf(21)=  45.38013889847690802616047395107562729165_dp
  t_lf(22)=  48.47118135183522387963964965049893315955_dp
  t_lf(23)=  51.60667556776437357044640248230912927799_dp
  t_lf(24)=  54.78472939811231919009334408360618468687_dp
  t_lf(25)=  58.00360522298051993929486275005855996592_dp
  t_lf(26)=  61.26170176100200198476558231308205513880_dp
  t_lf(27)=  64.55753862700633105895131802384963225274_dp
  t_lf(28)=  67.88974313718153498289113501020916511853_dp
  t_lf(29)=  71.25703896716800901007440704257107672402_dp
  t_lf(30)=  74.65823634883016438548764373417796663627_dp
  t_lf(31)=  78.09222355331531063141680805872032384672_dp
  t_lf(32)=  81.55795945611503717850296866601120668710_dp
  t_lf(33)=  85.05446701758151741396015748089886169157_dp
  t_lf(34)=  88.58082754219767880362692422023016479523_dp
  t_lf(35)=  92.13617560368709248333303629689953216439_dp
  t_lf(36)=  95.71969454214320248495799101366093670984_dp
  t_lf(37)=  99.33061245478742692932608668469238387374_dp
  t_lf(38)=  102.9681986145138126987523462380384139791_dp
  t_lf(39)=  106.6317602606434591262010789165262582885_dp
  t_lf(40)=  110.3206397147573954290535346141269756323_dp
  t_lf(41)=  114.0342117814617032329202979871643832206_dp
  t_lf(42)=  117.7718813997450715388381280889882652230_dp
  t_lf(43)=  121.5330815154386339623109706023341122586_dp
  t_lf(44)=  125.3172711493568951252073784232155946945_dp
  t_lf(45)=  129.1239336391272148825986282302868337433_dp
  t_lf(46)=  132.9525750356163098828226131835552064299_dp
  t_lf(47)=  136.8027226373263684696435638533273801388_dp
  t_lf(48)=  140.6739236482342593987077375760826121157_dp
  t_lf(49)=  144.5657439463448860089184430629689715750_dp
  t_lf(50)=  148.4777669517730320675371938508795234221_dp
  t_lf(51)=  152.4095925844973578391819737056751756623_dp
  t_lf(52)=  156.3608363030787851940699253901568474033_dp
  t_lf(53)=  160.3311282166309070282143945291859051737_dp
  t_lf(54)=  164.3201122631951814118173623614116588557_dp
  t_lf(55)=  168.3274454484276523304800652726029757950_dp
  t_lf(56)=  172.3527971391628015638371143804206852289_dp
  t_lf(57)=  176.3958484069973517152413870492310644708_dp
  t_lf(58)=  180.4562914175437710518418912030511526443_dp
  t_lf(59)=  184.5338288614494905024579415767708502684_dp
  t_lf(60)=  188.6281734236715911872884103898359167487_dp
  t_lf(61)=  192.7390472878449024360397994932615314951_dp
  t_lf(62)=  196.8661816728899939913861959392620652736_dp
  t_lf(63)=  201.0093163992815266792820391565502964125_dp
  t_lf(64)=  205.1681994826411985357854318852993558210_dp
  t_lf(65)=  209.3425867525368356464396786600908620653_dp
  t_lf(66)=  213.5322414945632611913140995964366936378_dp
  t_lf(67)=  217.7369341139542272509841715928004163884_dp
  t_lf(68)=  221.9564418191303339500681704535898960601_dp
  t_lf(69)=  226.1905483237275933322701685223226178832_dp
  t_lf(70)=  230.4390435657769523213935127204501618205_dp
  t_lf(71)=  234.7017234428182677427229672529631959172_dp
  t_lf(72)=  238.9783895618343230537651540911827770308_dp
  t_lf(73)=  243.2688490029827141828572629486213196017_dp
  t_lf(74)=  247.5729140961868839366425907411109433336_dp
  t_lf(75)=  251.8904022097231943772393546444858443173_dp
  t_lf(76)=  256.2211355500095254560828463192900509907_dp
  t_lf(77)=  260.5649409718632093052501426406983600202_dp
  t_lf(78)=  264.9216497985528010421161074406443808977_dp
  t_lf(79)=  269.2910976510198225362890529821257918199_dp
  t_lf(80)=  273.6731242856937041485587408011846857317_dp
  t_lf(81)=  278.0675734403661429141397217488747885503_dp
  t_lf(82)=  282.4742926876303960274237172433703727067_dp
  t_lf(83)=  286.8931332954269939508991894666617431598_dp
  t_lf(84)=  291.3239500942703075662342516899438017302_dp
  t_lf(85)=  295.7666013507606240210845456410431159053_dp
  t_lf(86)=  300.2209486470141317539746202758471395090_dp
  t_lf(87)=  304.6868567656687154725531375451315768191_dp
  t_lf(88)=  309.1641935801469219448667774874712358232_dp
  t_lf(89)=  313.6528299498790617831845930281410850426_dp
  t_lf(90)=  318.1526396202093268499930749566705006595_dp
  t_lf(91)=  322.6634991267261768911519151416789989939_dp
  t_lf(92)=  327.1852877037752172007931322164055482485_dp
  t_lf(93)=  331.7178871969284731381175417778704311636_dp
  t_lf(94)=  336.2611819791984770343557245691007814406_dp
  t_lf(95)=  340.8150588707990178689655113342148226173_dp
  t_lf(96)=  345.3794070622668541074469171784282311623_dp
  t_lf(97)=  349.9541180407702369295636388001321928762_dp
  t_lf(98)=  354.5390855194408088491915764084767289035_dp
  t_lf(99)=  359.1342053695753987760440104602869096126_dp
  t_lf(100)=  363.7393755555634901440799933696556380278_dp
  t_lf(101)=  368.3544960724047495949641916365686271847_dp
  t_lf(102)=  372.9794688856890206760262036128224559930_dp
  t_lf(103)=  377.6141978739186564467948059278758968137_dp
  t_lf(104)=  382.2585887730600291110999897338157451227_dp
  t_lf(105)=  386.9125491232175524822013470474076381965_dp
  t_lf(106)=  391.5759882173296196257630483078948725349_dp
  t_lf(107)=  396.2488170517915257990674471249181495365_dp
  t_lf(108)=  400.9309482789157454920876470786020797866_dp
  t_lf(109)=  405.6222961611448891924649635308112814376_dp
  t_lf(110)=  410.3227765269373054205448985634607749451_dp
  t_lf(111)=  415.0323067282496395563082394714147478136_dp
  t_lf(112)=  419.7508055995447340990825207006906338156_dp
  t_lf(113)=  424.4781934182570746676646521943066554876_dp
  t_lf(114)=  429.2143918666515701284861569845752112975_dp
  t_lf(115)=  433.9593239950148201938936691496115950555_dp
  t_lf(116)=  438.7129141861211848399114054248898597972_dp
  t_lf(117)=  443.4750881209189409587553833403002298113_dp
  t_lf(118)=  448.2457727453846057187886658354781040034_dp
  t_lf(119)=  453.0248962384961351041435531967944102686_dp
  t_lf(120)=  457.8123879812781810983912541313176533170_dp
  t_lf(121)=  462.6081785268749221865151412872479119167_dp
  t_lf(122)=  467.4121995716081787446837625121317032311_dp
  t_lf(123)=  472.2243839269805962399457711220916365241_dp
  t_lf(124)=  477.0446654925856331047093996895503468707_dp
  t_lf(125)=  481.8729792298879342285116776892289097893_dp
  t_lf(126)=  486.7092611368394122258247530279753174963_dp
  t_lf(127)=  491.5534482232980034988721938356916098911_dp
  t_lf(128)=  496.4054784872176206647928186858988458677_dp
  t_lf(129)=  501.2652908915792927796609064361672186079_dp
  t_lf(130)=  506.1328253420348751997323853324169014203_dp
  t_lf(131)=  511.0080226652360267438818093435864273134_dp
  t_lf(132)=  515.8908245878223975981734624013904354540_dp
  t_lf(133)=  520.7811737160441513632878425767214687209_dp
  t_lf(134)=  525.6790135159950627323751466945433680395_dp
  t_lf(135)=  530.5842882944334921811616417385371327930_dp
  t_lf(136)=  535.4969431801695441896628727207847890328_dp
  t_lf(137)=  540.4169241059976691049780628487646544234_dp
  t_lf(138)=  545.3441777911548737965972930389555528145_dp
  t_lf(139)=  550.2786517242855655537860779578310965970_dp
  t_lf(140)=  555.2202941468948698523266542774168171023_dp
  t_lf(141)=  560.1690540372730381305428501841115165158_dp
  t_lf(142)=  565.1248810948742988612895368380827271807_dp
  t_lf(143)=  570.0877257251342061414049678576131750853_dp
  t_lf(144)=  575.0575390247102067618643868172909327669_dp
  t_lf(145)=  580.0342727671307811636484181828790320119_dp
  t_lf(146)=  585.0178793888391176021577591617757511509_dp
  t_lf(147)=  590.0083119756178539037637098855846363148_dp
  t_lf(148)=  595.0055242493819689669662697995324366149_dp
  t_lf(149)=  600.0094705553274281079586980746365571027_dp
  t_lf(150)=  605.0201058494236838579726940994696346545_dp
  t_lf(151)=  610.0373856862386081867689303989541109448_dp
  t_lf(152)=  615.0612662070848845750296541952164941862_dp
  t_lf(153)=  620.0917041284773200380696792869346721311_dp
  t_lf(154)=  625.1286567308909491966542077298011577286_dp
  t_lf(155)=  630.1720818478101958171841313875697025786_dp
  t_lf(156)=  635.2219378550597328634673283089739000242_dp
  t_lf(157)=  640.2781836604080409208917735453773752730_dp
  t_lf(158)=  645.3407786934350077244819512083169627632_dp
  t_lf(159)=  650.4096828956552392500216655842685462382_dp
  t_lf(160)=  655.4848567108890661717085855247856167181_dp
  t_lf(161)=  660.5662610758735291676206911000389925662_dp
  t_lf(162)=  665.6538574111059132426189041691872719529_dp
  t_lf(163)=  670.7476076119126755766838365353080218295_dp
  t_lf(164)=  675.8474740397368739993850641512617825540_dp
  t_lf(165)=  680.9534195136374546094430122993756251980_dp
  t_lf(166)=  686.0654073019939978423357166441251722191_dp
  t_lf(167)=  691.1834011144107529495954674208699515849_dp
  t_lf(168)=  696.3073650938140118743477617656101867234_dp
  t_lf(169)=  701.4372638087370853464547366487408239330_dp
  t_lf(170)=  706.5730622457873471107222627212983146762_dp
  t_lf(171)=  711.7147258022900069535217806270312196228_dp
  t_lf(172)=  716.8622202791034599958290873832934197945_dp
  t_lf(173)=  722.0155118736012389427625877717344112637_dp
  t_lf(174)=  727.1745671728157679707583371624770251419_dp
  t_lf(175)=  732.3393531467392820250652085723725801506_dp
  t_lf(176)=  737.5098371417774338067960806361704157227_dp
  t_lf(177)=  742.6859868743512629488073762468126390514_dp
  t_lf(178)=  747.8677704246433480965424239089406648389_dp
  t_lf(179)=  753.0551562304841030927202283389620670145_dp
  t_lf(180)=  758.2481130813743134689459423889496591995_dp
  t_lf(181)=  763.4466101126401392157850294275826681267_dp
  t_lf(182)=  768.6506167997169345663611017340493430292_dp
  t_lf(183)=  773.8601029525583555065077360743974834802_dp
  t_lf(184)=  779.0750387101673411255661852705822093028_dp
  t_lf(185)=  784.2953945352456659445350402748398441063_dp
  t_lf(186)=  789.5211412089588671912766819577629035894_dp
  t_lf(187)=  794.7522498258134538155881601536011594249_dp
  t_lf(188)=  799.9886917886434030212435750662896862699_dp
  t_lf(189)=  805.2304388037030454005346635205004431135_dp
  t_lf(190)=  810.4774628758635315445616824070726608583_dp
  t_lf(191)=  815.7297363039101614174116323592750280049_dp
  t_lf(192)=  820.9872316759379429653102703249466131180_dp
  t_lf(193)=  826.2499218648428285171652013783296718822_dp
  t_lf(194)=  831.5177800239061566486991551214918101641_dp
  t_lf(195)=  836.7907795824699034507486471332058421131_dp
  t_lf(196)=  842.0688942417004206797938168630085547085_dp
  t_lf(197)=  847.3520979704384091865736141921713856835_dp
  t_lf(198)=  852.6403650011329444228432803654397429607_dp
  t_lf(199)=  857.9336698258574368182534016573082801626_dp
  t_lf(200)=  863.2319871924054734957066166881351851459_dp
  t_lf(201)=  868.5352921004645492467719339214214336011_dp
  t_lf(202)=  873.8435597978657540070733643097925993261_dp
  t_lf(203)=  879.1567657769075413393619890855976906612_dp
  t_lf(204)=  884.4748857707517577298412331833096960376_dp
  t_lf(205)=  889.7978957498901659083087558895732912655_dp
  t_lf(206)=  895.1257719186797469884945903260849086542_dp
  t_lf(207)=  900.4584907119451160620918336317401561819_dp
  t_lf(208)=  905.7960287916464340358142495591381810590_dp
  t_lf(209)=  911.1383630436112450398852205689911638961_dp
  t_lf(210)=  916.4854705743287137204038100040412335380_dp
  t_lf(211)=  921.8373287078047802161457662694955136560_dp
  t_lf(212)=  927.1939149824767926691246996514409245625_dp
  t_lf(213)=  932.5552071481862177818493994208764843639_dp
  t_lf(214)=  937.9211831632080692645710303593579379336_dp
  t_lf(215)=  943.2918211913357320626446322059299726087_dp
  t_lf(216)=  948.6670995990198970650820642810720794268_dp
  t_lf(217)=  954.0469969525603566161165813490576163669_dp
  t_lf(218)=  959.4314920153494456259111299227249945860_dp
  t_lf(219)=  964.8205637451659464463984840170860628616_dp
  t_lf(220)=  970.2141912915183079838956511711937329371_dp
  t_lf(221)=  975.6123539930360608001986732306321780775_dp
  t_lf(222)=  981.0150313749083402453792462600443275141_dp
  t_lf(223)=  986.4222031463684590040153514697674675408_dp
  t_lf(224)=  991.8338491982234988562068648205015301108_dp
  t_lf(225)=  997.2499496004279189881988739607989567992_dp
  t_lf(226)=  1002.670484599700204866198237575873155039_dp
  t_lf(227)=  1008.095434617181607541211619152706652084_dp
  t_lf(228)=  1013.524780246136048311450356064433384462_dp
  t_lf(229)=  1018.958502249690287959891591668365992780_dp
  t_lf(230)=  1024.396581558613483334716335954860553106_dp
  t_lf(231)=  1029.838999269135276875278877513191387840_dp
  t_lf(232)=  1035.285736640801586830713845909927829150_dp
  t_lf(233)=  1040.736775094367287396016070784596716628_dp
  t_lf(234)=  1046.192096209724988824277280821465263210_dp
  t_lf(235)=  1051.651681723869147785698990824463624559_dp
  t_lf(236)=  1057.115513528894757855149505441099675319_dp
  t_lf(237)=  1062.583573670029889040717696219503611946_dp
  t_lf(238)=  1068.055844343701363735489815702278094779_dp
  t_lf(239)=  1073.532307895632874402452979296540003876_dp
  t_lf(240)=  1079.012946818974865706117912352521423492_dp
  t_lf(241)=  1084.497743752465520701584323671232239227_dp
  t_lf(242)=  1089.986681478622207099125442948620674395_dp
  t_lf(243)=  1095.479742921962755556101669133233302918_dp
  t_lf(244)=  1100.976911147255957423687522479575270800_dp
  t_lf(245)=  1106.478169357800684408498987299687817899_dp
  t_lf(246)=  1111.983500893733047213178228031105927760_dp
  t_lf(247)=  1117.492889230361024409240742904559099902_dp
  t_lf(248)=  1123.006317976526006583421603593475986817_dp
  t_lf(249)=  1128.523770872990714198292321053689882975_dp
  t_lf(250)=  1134.045231790852960631511831174826622461_dp
  t_lf(251)=  1139.570684729984744517730416413834875703_dp
  t_lf(252)=  1145.100113817496167824460723874039459978_dp
  t_lf(253)=  1150.633503306223688059329420283814785396_dp
  t_lf(254)=  1156.170837573242224641794093212989254359_dp
  t_lf(255)=  1161.712101118400650788039632401011094239_dp
  t_lf(256)=  1167.257278562880213263377489372676506783_dp
  t_lf(257)=  1172.806354647775433061729283687437762235_dp
  t_lf(258)=  1178.359314232697050486014603559164311543_dp
  t_lf(259)=  1183.916142294396588235488051973638938437_dp
  t_lf(260)=  1189.476823925412115964976762991346797817_dp
  t_lf(261)=  1195.041344332734809374950525497553760832_dp
  t_lf(262)=  1200.609688836495906228517181630181463293_dp
  t_lf(263)=  1206.181842868673670779603107894386535384_dp
  t_lf(264)=  1211.757791971819986943311993073648720092_dp
  t_lf(265)=  1217.337521797806209152057221545903965502_dp
  t_lf(266)=  1222.921018106587908226588833842693175337_dp
  t_lf(267)=  1228.508266764988157756301894620285550261_dp
  t_lf(268)=  1234.099253745499014434806430859565626148_dp
  t_lf(269)=  1239.693965125100853541026384040060863266_dp
  t_lf(270)=  1245.292387084099228299230111205512804587_dp
  t_lf(271)=  1250.894505904978929199326981278718666626_dp
  t_lf(272)=  1256.500307971274926517245444382424499434_dp
  t_lf(273)=  1262.109779766459886249799529804355523473_dp
  t_lf(274)=  1267.722907872847956474531952053793565432_dp
  t_lf(275)=  1273.339678970514527767795414298211070011_dp
  t_lf(276)=  1278.960079836231677768831876609860144970_dp
  t_lf(277)=  1284.584097342419016268701274159384755809_dp
  t_lf(278)=  1290.211718456109653335307291199718476160_dp
  t_lf(279)=  1295.842930237931018964026945998105884779_dp
  t_lf(280)=  1301.477719841100268571984754439149781853_dp
  t_lf(281)=  1307.116074510434014337091392763842953794_dp
  t_lf(282)=  1312.757981581372127924724820791995829775_dp
  t_lf(283)=  1318.403428479015365549370807938790969980_dp
  t_lf(284)=  1324.052402717176571589534726714220357213_dp
  t_lf(285)=  1329.704891897445222115539758716256924094_dp
  t_lf(286)=  1335.360883708265074705072421857245548567_dp
  t_lf(287)=  1341.020365924024695814044537973726135885_dp
  t_lf(288)=  1346.683326404160641743921189054862070135_dp
  t_lf(289)=  1352.349753092273073904420258290608323206_dp
  t_lf(290)=  1358.019634015253593615621521777654599019_dp
  t_lf(291)=  1363.692957282425086129133488636281086437_dp
  t_lf(292)=  1369.369711084693367877060061736635982145_dp
  t_lf(293)=  1375.049883693710435182999901999448480335_dp
  t_lf(294)=  1380.733463461049116794023084844715542067_dp
  t_lf(295)=  1386.420438817388936619239894551661427330_dp
  t_lf(296)=  1392.110798271712996991859686587067404198_dp
  t_lf(297)=  1397.804530410515696610107365875800110612_dp
  t_lf(298)=  1403.501623897021101060517026272362407668_dp
  t_lf(299)=  1409.202067470411787487377266545737922391_dp
  t_lf(300)=  1414.905849945067988546808494692029176511_dp
  t_lf(301)=  1420.612960209816864275386689948818203276_dp
  t_lf(302)=  1426.323387227191733913600158369760856135_dp
  t_lf(303)=  1432.037120032701103055879601873596370996_dp
  t_lf(304)=  1437.754147734107324753557557791316930806_dp
  t_lf(305)=  1443.474459510714736376909706227968733192_dp
  t_lf(306)=  1449.198044612667117149366963441145087705_dp
  t_lf(307)=  1454.924892360254314302965426008427579989_dp
  t_lf(308)=  1460.654992143227888770967186572752242154_dp
  t_lf(309)=  1466.388333420125634233131034124728208679_dp
  t_lf(310)=  1472.124905717604826163078189903954930097_dp
  t_lf(311)=  1477.864698629784060297274185469387612374_dp
  t_lf(312)=  1483.607701817593542652974614512249986387_dp
  t_lf(313)=  1489.353905008133695869653379821507196741_dp
  t_lf(314)=  1495.103297994041949236495057179368848558_dp
  t_lf(315)=  1500.855870632867582298991659729883267336_dp
  t_lf(316)=  1506.611612846454494411999069514281031394_dp
  t_lf(317)=  1512.370514620331775026198831683579642704_dp
  t_lf(318)=  1518.132566003111951861155778180989402747_dp
  t_lf(319)=  1523.897757105896796432400993791316443652_dp
  t_lf(320)=  1529.666078101690568663505145853291690700_dp
  t_lf(321)=  1535.437519224820584528204789907237493406_dp
  t_lf(322)=  1541.212070770364992833534127603949045823_dp
  t_lf(323)=  1546.989723093587649373792689653710025895_dp
  t_lf(324)=  1552.770466609379978758208134844316481850_dp
  t_lf(325)=  1558.554291791709716243463140952334175734_dp
  t_lf(326)=  1564.341189173076423886945305439913102179_dp
  t_lf(327)=  1570.131149343973677278717867129044829534_dp
  t_lf(328)=  1575.924162952357821010836326866456766827_dp
  t_lf(329)=  1581.720220703123192902762630279672120266_dp
  t_lf(330)=  1587.519313357583718822237810549244139478_dp
  t_lf(331)=  1593.321431732960781723022427388820776284_dp
  t_lf(332)=  1599.126566701877270265332363855028499873_dp
  t_lf(333)=  1604.934709191857714092490949999904998447_dp
  t_lf(334)=  1610.745850184834414509167932898107954381_dp
  t_lf(335)=  1616.559980716659480943438764227697864771_dp
  t_lf(336)=  1622.377091876622685177608290693896276477_dp
  t_lf(337)=  1628.197174806975046902113157248544707690_dp
  t_lf(338)=  1634.020220702458065683637364253133521468_dp
  t_lf(339)=  1639.846220809838515943614740983672068844_dp
  t_lf(340)=  1645.675166427448723017299499177687736156_dp
  t_lf(341)=  1651.507048904732239807290607080195222666_dp
  t_lf(342)=  1657.341859641794844959507357107386304181_dp
  t_lf(343)=  1663.179590088960784874823415337715843370_dp
  t_lf(344)=  1669.020231746334183226547954215436220109_dp
  t_lf(345)=  1674.863776163365542983350711617395129572_dp
  t_lf(346)=  1680.710214938423267239701444127294297609_dp
  t_lf(347)=  1686.559539718370126432055120973680157144_dp
  t_lf(348)=  1692.411742198144600769468102485880947590_dp
  t_lf(349)=  1698.266814120347027932667584008379587299_dp
  t_lf(350)=  1704.124747274830487296391687539733318876_dp
  t_lf(351)=  1709.985533498296353106630910692066214595_dp
  t_lf(352)=  1715.849164673894450197779014877322226735_dp
  t_lf(353)=  1721.715632730827746965174734994510196737_dp
  t_lf(354)=  1727.584929643961521416603262726610596634_dp
  t_lf(355)=  1733.457047433436937212533476592349818370_dp
  t_lf(356)=  1739.331978164288967669685756375936020726_dp
  t_lf(357)=  1745.209713946068606746435888974174852696_dp
  t_lf(358)=  1751.090246932469307052030925525654431440_dp
  t_lf(359)=  1756.973569320957585925079018411259111417_dp
  t_lf(360)=  1762.859673352407741610721964582704880170_dp
  t_lf(361)=  1768.748551310740622530740019446480587245_dp
  t_lf(362)=  1774.640195522566393586996338606571772740_dp
  t_lf(363)=  1780.534598356831244366515470999424557044_dp
  t_lf(364)=  1786.431752224467985026508775427349408515_dp
  t_lf(365)=  1792.331649578050476530201643618014138725_dp
  t_lf(366)=  1798.234282911451842779765510079820455745_dp
  t_lf(367)=  1804.139644759506413051384942018978884910_dp
  t_lf(368)=  1810.047727697675343979860623336621787301_dp
  t_lf(369)=  1815.958524341715871166517877183504246298_dp
  t_lf(370)=  1821.872027347354141294903964309220057670_dp
  t_lf(371)=  1827.788229409961576434153786191692295170_dp
  t_lf(372)=  1833.707123264234722990312659996073531221_dp
  t_lf(373)=  1839.628701683878538531648394601175195359_dp
  t_lf(374)=  1845.552957481293070465377104918471627763_dp
  t_lf(375)=  1851.479883507263481280574628155072716386_dp
  t_lf(376)=  1857.409472650653375795647275189219419799_dp
  t_lf(377)=  1863.341717838101386558884034663146650009_dp
  t_lf(378)=  1869.276612033720974247592355238815583421_dp
  t_lf(379)=  1875.214148238803400595409591822469211329_dp
  t_lf(380)=  1881.154319491523832048853842830499605642_dp
  t_lf(381)=  1887.097118866650533013296528875138423741_dp
  t_lf(382)=  1893.042539475257108195563710948798967456_dp
  t_lf(383)=  1898.990574464437754185556765485951437708_dp
  t_lf(384)=  1904.941217017025481042872635573081199389_dp
  t_lf(385)=  1910.894460351313265266640691227715696058_dp
  t_lf(386)=  1916.850297720778096127912854402556931391_dp
  t_lf(387)=  1922.808722413807877934176187389747829836_dp
  t_lf(388)=  1928.769727753431151375127373254368144686_dp
  t_lf(389)=  1934.733307097049597667975993488000317337_dp
  t_lf(390)=  1940.699453836173289779442717621172525854_dp
  t_lf(391)=  1946.668161396158655550499005070855848508_dp
  t_lf(392)=  1952.639423235949118088961406922116737671_dp
  t_lf(393)=  1958.613232847818379324506076170208789269_dp
  t_lf(394)=  1964.589583757116313140703105620829796812_dp
  t_lf(395)=  1970.568469522017435009476810495537395374_dp
  t_lf(396)=  1976.549883733271915555163708790263929219_dp
  t_lf(397)=  1982.533820013959105968247140820440788503_dp
  t_lf(398)=  1988.520272019243543673074494233767502272_dp
  t_lf(399)=  1994.509233436133407129584119646021061244_dp
  t_lf(400)=  2000.500697983241389116454566798306142795_dp
  t_lf(401)=  2006.494659410547958302305245885655492526_dp
  t_lf(402)=  2012.491111499166979362787795240399917550_dp
  t_lf(403)=  2018.490048061113662344770447006507593365_dp
  t_lf(404)=  2024.491462939074812414489109516336935658_dp
  t_lf(405)=  2030.495350006181351554670849797253226116_dp
  t_lf(406)=  2036.501703165783084196376706694516494019_dp
  t_lf(407)=  2042.510516351225679184806745943513070483_dp
  t_lf(408)=  2048.521783525629840884703222162683252427_dp
  t_lf(409)=  2054.535498681672642630416744371811638515_dp
  t_lf(410)=  2060.551655841370996118301499199533410311_dp
  t_lf(411)=  2066.570249055867230725011934564435801406_dp
  t_lf(412)=  2072.591272405216757114615001122405595363_dp
  t_lf(413)=  2078.614719998177789870336404239568472716_dp
  t_lf(414)=  2084.640585972003104253350879666681896812_dp
  t_lf(415)=  2090.668864492233802551427111223199454905_dp
  t_lf(416)=  2096.699549752495065834566759272055656350_dp
  t_lf(417)=  2102.732635974293867283150789427853725837_dp
  t_lf(418)=  2108.768117406818623596638992559164885242_dp
  t_lf(419)=  2114.805988326740761327671926480850375632_dp
  t_lf(420)=  2120.846243038018175317607748037358621842_dp
  t_lf(421)=  2126.888875871700556735196464265755719316_dp
  t_lf(422)=  2132.933881185736568540355652652668176002_dp
  t_lf(423)=  2138.981253364782846509967093796285401120_dp
  t_lf(424)=  2145.030986820014804272363259299688988594_dp
  t_lf(425)=  2151.083075988939221101814312584014490409_dp
  t_lf(426)=  2157.137515335208591523956244474908226779_dp
  t_lf(427)=  2163.194299348437216077812986321777021254_dp
  t_lf(428)=  2169.253422544019012869951849381716651392_dp
  t_lf(429)=  2175.314879462947029841462525638169625001_dp
  t_lf(430)=  2181.378664671634637948953359606199836245_dp
  t_lf(431)=  2187.444772761738385736701026412524886498_dp
  t_lf(432)=  2193.513198349982496048555690609125169885_dp
  t_lf(433)=  2199.583936077984985895275975183097713854_dp
  t_lf(434)=  2205.656980612085390755727724372541427362_dp
  t_lf(435)=  2211.732326643174074848907000975052052312_dp
  t_lf(436)=  2217.809968886523109168118781670177607099_dp
  t_lf(437)=  2223.889902081618699318934561933875656755_dp
  t_lf(438)=  2229.972120991995145448839148149694901599_dp
  t_lf(439)=  2236.056620405070316798832360279437655299_dp
  t_lf(440)=  2242.143395131982623645746759555003501942_dp
  t_lf(441)=  2248.232440007429469638747955515734912811_dp
  t_lf(442)=  2254.323749889507167764468209696631534520_dp
  t_lf(443)=  2260.417319659552303402554216954476810223_dp
  t_lf(444)=  2266.513144221984528157152022105347136227_dp
  t_lf(445)=  2272.611218504150768370070596979243173086_dp
  t_lf(446)=  2278.711537456170832438123934310424489681_dp
  t_lf(447)=  2284.814096050784401270511607822451135874_dp
  t_lf(448)=  2290.918889283199386432120353294643375012_dp
  t_lf(449)=  2297.025912170941640725373875085469485013_dp
  t_lf(450)=  2303.135159753706006166783116347225088270_dp
  t_lf(451)=  2309.246627093208684514711823298227625158_dp
  t_lf(452)=  2315.360309273040915702128419034759999966_dp
  t_lf(453)=  2321.476201398523949722319900571167001961_dp
  t_lf(454)=  2327.594298596565297706750514269458675574_dp
  t_lf(455)=  2333.714596015516248122510113787693361548_dp
  t_lf(456)=  2339.837088825030634202166082820878270494_dp
  t_lf(457)=  2345.961772215924838901359214202031754029_dp
  t_lf(458)=  2352.088641400039023859217681927422538914_dp
  t_lf(459)=  2358.217691610099569013652952256063242563_dp
  t_lf(460)=  2364.348918099582709697894928664015979457_dp
  t_lf(461)=  2370.482316142579358216267653857290505997_dp
  t_lf(462)=  2376.617881033661097066247427537079517299_dp
  t_lf(463)=  2382.755608087747331140327274718338227241_dp
  t_lf(464)=  2388.895492639973586405179475236532845119_dp
  t_lf(465)=  2395.037530045560942717104644131223915673_dp
  t_lf(466)=  2401.181715679686588591824101127350979720_dp
  t_lf(467)=  2407.328044937355485903353632416085248706_dp
  t_lf(468)=  2413.476513233273132641032074574411971856_dp
  t_lf(469)=  2419.627116001719412005807499314218874337_dp
  t_lf(470)=  2425.779848696423516276646441438675412253_dp
  t_lf(471)=  2431.934706790439934025466131912001413206_dp
  t_lf(472)=  2438.091685776025489404333878650095640535_dp
  t_lf(473)=  2444.250781164517422371868664741406616870_dp
  t_lf(474)=  2450.411988486212498866854087641268730065_dp
  t_lf(475)=  2456.575303290247140076064633739608958881_dp
  t_lf(476)=  2462.740721144478560080253985343841618283_dp
  t_lf(477)=  2468.908237635366901297188944956715727462_dp
  t_lf(478)=  2475.077848367858357273569340672435813127_dp
  t_lf(479)=  2481.249548965269272508686018503835429741_dp
  t_lf(480)=  2487.423335069171209121768183681275025925_dp
  t_lf(481)=  2493.599202339276970302189766793871791694_dp
  t_lf(482)=  2499.777146453327570607073410234040783997_dp
  t_lf(483)=  2505.957163106980143294380761046216685550_dp
  t_lf(484)=  2512.139248013696775001339112445063297286_dp
  t_lf(485)=  2518.323396904634258198056593399993446639_dp
  t_lf(486)=  2524.509605528534751964450051706064251730_dp
  t_lf(487)=  2530.697869651617341755179962537314797235_dp
  t_lf(488)=  2536.888185057470488932183048005114941686_dp
  t_lf(489)=  2543.080547546945360957643225608158217267_dp
  t_lf(490)=  2549.274952938050033251871922549728940934_dp
  t_lf(491)=  2555.471397065844553830606112874744025587_dp
  t_lf(492)=  2561.669875782336861944702585727620312016_dp
  t_lf(493)=  2567.870384956379552052135392377855350157_dp
  t_lf(494)=  2574.072920473567474557615139372766698867_dp
  t_lf(495)=  2580.277478236136164859068332757803067216_dp
  t_lf(496)=  2586.484054162861092342666425568178130699_dp
  t_lf(497)=  2592.692644188957721069101232844134344525_dp
  t_lf(498)=  2598.903244265982373993389182425806417251_dp
  t_lf(499)=  2605.115850361733892658674273563790810274_dp
  t_lf(500)=  2611.330458460156084401311015806385726329_dp
  t_lf(501)=  2617.547064561240949199966011820053031399_dp
  t_lf(502)=  2623.765664680932678395601829180519461209_dp
  t_lf(503)=  2629.986254851032417602020590296155071240_dp
  t_lf(504)=  2636.208831119103786218168129877817832083_dp
  t_lf(505)=  2642.433389548379146043653087477957008880_dp
  t_lf(506)=  2648.659926217666611587939016009190510866_dp
  t_lf(507)=  2654.888437221257794751441236129243673781_dp
  t_lf(508)=  2661.118918668836276643323141179876319312_dp
  t_lf(509)=  2667.351366685386799386158005035936077960_dp
  t_lf(510)=  2673.585777411105170841820776345416094408_dp
  t_lf(511)=  2679.822147001308875276018237946297816708_dp
  t_lf(512)=  2686.060471626348383060773327039421405821_dp
  t_lf(513)=  2692.300747471519152594968090182076836472_dp
  t_lf(514)=  2698.542970736974317702737116618296268492_dp
  t_lf(515)=  2704.787137637638053848106478266575896952_dp
  t_lf(516)=  2711.033244403119616581809030259760622829_dp
  t_lf(517)=  2717.281287277628045712691924507497925837_dp
  t_lf(518)=  2723.531262519887528771582605043430729299_dp
  t_lf(519)=  2729.783166403053417409911350668794246473_dp
  t_lf(520)=  2736.036995214628890448817293807960282421_dp
  t_lf(521)=  2742.292745256382257366906189514478118885_dp
  t_lf(522)=  2748.550412844264896086297184142143258468_dp
  t_lf(523)=  2754.809994308329818987107327319056353761_dp
  t_lf(524)=  2761.071485992650861150091215573142232791_dp
  t_lf(525)=  2767.334884255242484895793332219960313504_dp
  t_lf(526)=  2773.600185467980194756296490605623562162_dp
  t_lf(527)=  2779.867386016521557082475189548039045908_dp
  t_lf(528)=  2786.136482300227818555601306848759407185_dp
  t_lf(529)=  2792.407470732086117937214812512379799422_dp
  t_lf(530)=  2798.680347738632285455377273106093221400_dp
  t_lf(531)=  2804.955109759874224288783813953657970433_dp
  t_lf(532)=  2811.231753249215868672732658371905356836_dp
  t_lf(533)=  2817.510274673381713212652909186508083030_dp
  t_lf(534)=  2823.790670512341908051783202085558634522_dp
  t_lf(535)=  2830.072937259237914599688360235808099163_dp
  t_lf(536)=  2836.357071420308716587610128596546351618_dp
  t_lf(537)=  2842.643069514817581275183178263490279498_dp
  t_lf(538)=  2848.930928074979365690820363565443693184_dp
  t_lf(539)=  2855.220643645888362845093012630295181943_dp
  t_lf(540)=  2861.512212785446682912713971917205299833_dp
  t_lf(541)=  2867.805632064293164434284157529789949969_dp
  t_lf(542)=  2874.100898065732810643798259724453988576_dp
  t_lf(543)=  2880.398007385666746082032592000009523208_dp
  t_lf(544)=  2886.696956632522688709368287225173532584_dp
  t_lf(545)=  2892.997742427185932784346363010608921875_dp
  t_lf(546)=  2899.300361402930837826317680553998122482_dp
  t_lf(547)=  2905.604810205352819031950413175595783235_dp
  t_lf(548)=  2911.911085492300834566100067546492001761_dp
  t_lf(549)=  2918.219183933810365197641947123762667917_dp
  t_lf(550)=  2924.529102212036881800322641489638349064_dp
  t_lf(551)=  2930.840837021189796287514940953888114207_dp
  t_lf(552)=  2937.154385067466891597968635386995365734_dp
  t_lf(553)=  2943.469743068989226397246933671919956386_dp
  t_lf(554)=  2949.786907755736510206533563342902743793_dp
  t_lf(555)=  2956.105875869482944716897663584082904301_dp
  t_lf(556)=  2962.426644163733527092920912745874801220_dp
  t_lf(557)=  2968.749209403660811114832339682719582984_dp
  t_lf(558)=  2975.073568366042122052969226602565168172_dp
  t_lf(559)=  2981.399717839197221212495556557476333812_dp
  t_lf(560)=  2987.727654622926416129870597119978407454_dp
  t_lf(561)=  2994.057375528449112445577320552739188994_dp
  t_lf(562)=  3000.388877378342803520101190998890537503_dp
  t_lf(563)=  3006.722157006482493902102027987508278485_dp
  t_lf(564)=  3013.057211257980552799152688137119331035_dp
  t_lf(565)=  3019.394036989126993742335578963961540346_dp
  t_lf(566)=  3025.732631067330176676398798232214857119_dp
  t_lf(567)=  3032.072990371057928747085131923348139667_dp
  t_lf(568)=  3038.415111789779080096666282820235703468_dp
  t_lf(569)=  3044.758992223905411017648467629619841947_dp
  t_lf(570)=  3051.104628584734006853070731753114585396_dp
  t_lf(571)=  3057.452017794390017069800697949054214560_dp
  t_lf(572)=  3063.801156785769814968750593211501015601_dp
  t_lf(573)=  3070.152042502484554532995788400625908452_dp
  t_lf(574)=  3076.504671898804120951385136638564672338_dp
  t_lf(575)=  3082.859041939601471391393408136827243736_dp
  t_lf(576)=  3089.215149600297362630687291339421354554_dp
  t_lf(577)=  3095.572991866805462192163866646692042961_dp
  t_lf(578)=  3101.932565735477839662080168003896472600_dp
  t_lf(579)=  3108.293868213050834905330344294202057069_dp
  t_lf(580)=  3114.656896316591299925948839902706509450_dp
  t_lf(581)=  3121.021647073443211154529664869441059633_dp
  t_lf(582)=  3127.388117521174648977458863849525723619_dp
  t_lf(583)=  3133.756304707525141355665276566519910689_dp
  t_lf(584)=  3140.126205690353368413009081788332982964_dp
  t_lf(585)=  3146.497817537585224906453819036969540618_dp
  t_lf(586)=  3152.871137327162237521810891421240215376_dp
  t_lf(587)=  3159.246162146990333970111191918809888054_dp
  t_lf(588)=  3165.622889094888960890551606885535126354_dp
  t_lf(589)=  3172.001315278540547596489798641965337102_dp
  t_lf(590)=  3178.381437815440312731123840470369398933_dp
  t_lf(591)=  3184.763253832846410929298883036454755613_dp
  t_lf(592)=  3191.146760467730416611335907193318909049_dp
  t_lf(593)=  3197.531954866728142063882526936969057689_dp
  t_lf(594)=  3203.918834186090786991547438347159940670_dp
  t_lf(595)=  3210.307395591636416751503085041702434575_dp
  t_lf(596)=  3216.697636258701766511329977559722908199_dp
  t_lf(597)=  3223.089553372094368598135344088513971106_dp
  t_lf(598)=  3229.483144126045000334412816483347662397_dp
  t_lf(599)=  3235.878405724160449683223015656229119084_dp
  t_lf(600)=  3242.275335379376596052071475923978549772_dp
  t_lf(601)=  3248.673930313911803631344556430557695904_dp
  t_lf(602)=  3255.074187759220624669339983808804899237_dp
  t_lf(603)=  3261.476104955947810111800546279013673397_dp
  t_lf(604)=  3267.879679153882625059431246821414502824_dp
  t_lf(605)=  3274.284907611913466522155893310570949063_dp
  t_lf(606)=  3280.691787597982780973852568935864640492_dp
  t_lf(607)=  3287.100316389042279236002541170711332259_dp
  t_lf(608)=  3293.510491271008446243097729209890068636_dp
  t_lf(609)=  3299.922309538718343266781599222617685676_dp
  t_lf(610)=  3306.335768495885700199550979780727664630_dp
  t_lf(611)=  3312.750865455057295522425417892065156944_dp
  t_lf(612)=  3319.167597737569621604299907226699688025_dp
  t_lf(613)=  3325.585962673505833003739639820275738962_dp
  t_lf(614)=  3332.005957601652975466755334509016407814_dp
  t_lf(615)=  3338.427579869459493336618102452202528746_dp
  t_lf(616)=  3344.850826832993013114037095137985367480_dp
  t_lf(617)=  3351.275695856898400928034666816595558304_dp
  t_lf(618)=  3357.702184314356091699615746490029701398_dp
  t_lf(619)=  3364.130289587040687801843781880179425128_dp
  t_lf(620)=  3370.560009065079825041208169780864323114_dp
  t_lf(621)=  3376.991340147013303806200658323442096346_dp
  t_lf(622)=  3383.424280239752483249813886010332955190_dp
  t_lf(623)=  3389.858826758539936393237054294445984139_dp
  t_lf(624)=  3396.294977126909364058354715458766534721_dp
  t_lf(625)=  3402.732728776645765556757752791671285279_dp
  t_lf(626)=  3409.172079147745864082853750222386672201_dp
  t_lf(627)=  3415.613025688378784778319966469162180743_dp
  t_lf(628)=  3422.055565854846983454578875948482009128_dp
  t_lf(629)=  3428.499697111547423979196506237386582827_dp
  t_lf(630)=  3434.945416930933002351110340909359178174_dp
  t_lf(631)=  3441.392722793474215508389078943024772257_dp
  t_lf(632)=  3447.841612187621072930813720848880712884_dp
  t_lf(633)=  3454.292082609765249117950922351257518706_dp
  t_lf(634)=  3460.744131564202475041567916642014306584_dp
  t_lf(635)=  3467.197756563095166689216116782956786618_dp
  t_lf(636)=  3473.652955126435288833590295401824723229_dp
  t_lf(637)=  3480.109724782007452179854488330276401294_dp
  t_lf(638)=  3486.568063065352242060516936062061618767_dp
  t_lf(639)=  3493.027967519729776864636881068419704273_dp
  t_lf(640)=  3499.489435696083494405158265251853127889_dp
  t_lf(641)=  3505.952465153004164444991671789854728075_dp
  t_lf(642)=  3512.417053456694125619108547965258707349_dp
  t_lf(643)=  3518.883198180931745006373116468806740436_dp
  t_lf(644)=  3525.350896907036098621119686286976469420_dp
  t_lf(645)=  3531.820147223831871110588533370471029800_dp
  t_lf(646)=  3538.290946727614472960264327541690186441_dp
  t_lf(647)=  3544.763293022115373524919396757347726853_dp
  t_lf(648)=  3551.237183718467648218752074069412359376_dp
  t_lf(649)=  3557.712616435171738213430068021097186299_dp
  t_lf(650)=  3564.189588798061421008102306250573056751_dp
  t_lf(651)=  3570.668098440269990250532068555481119396_dp
  t_lf(652)=  3577.148143002196643203431465164518222409_dp
  t_lf(653)=  3583.629720131473074264844416235716447491_dp
  t_lf(654)=  3590.112827482930272966034210046306351414_dp
  t_lf(655)=  3596.597462718565524884784393390702064947_dp
  t_lf(656)=  3603.083623507509613926320085249572178808_dp
  t_lf(657)=  3609.571307525994224438202684580855772788_dp
  t_lf(658)=  3616.060512457319541639546220115529302795_dp
  t_lf(659)=  3622.551235991822048858750093966349126896_dp
  t_lf(660)=  3629.043475826842520087642506357379322676_dp
  t_lf(661)=  3635.537229666694206373483195945858029952_dp
  t_lf(662)=  3642.032495222631214583685044906892843327_dp
  t_lf(663)=  3648.529270212817077091383312203253814172_dp
  t_lf(664)=  3655.027552362293510943110480790919714329_dp
  t_lf(665)=  3661.527339402949365082825620299476935236_dp
  t_lf(666)=  3668.028629073489754219401438565811610377_dp
  t_lf(667)=  3674.531419119405377937391463429983718101_dp
  t_lf(668)=  3681.035707292942023663485678449644850603_dp
  t_lf(669)=  3687.541491353070252113517028896290516334_dp
  t_lf(670)=  3694.048769065455263857205092347338603292_dp
  t_lf(671)=  3700.557538202426945650018425028729347338_dp
  t_lf(672)=  3707.067796542950095193605183616385935613_dp
  t_lf(673)=  3713.579541872594822998184094780742273819_dp
  t_lf(674)=  3720.092771983507130032106193456848881600_dp
  t_lf(675)=  3726.607484674379659855493447834068833993_dp
  t_lf(676)=  3733.123677750422623946434886960115824338_dp
  t_lf(677)=  3739.641349023334898939675578895436336366_dp
  t_lf(678)=  3746.160496311275294509070187747433060311_dp
  t_lf(679)=  3752.681117438833990636292262112580201754_dp
  t_lf(680)=  3759.203210237004143019394252428054045633_dp
  t_lf(681)=  3765.726772543153655385802879241810068383_dp
  t_lf(682)=  3772.251802200997117485211219265775731462_dp
  t_lf(683)=  3778.778297060567907548594064414585793416_dp
  t_lf(684)=  3785.306254978190458010228046563235051498_dp
  t_lf(685)=  3791.835673816452683300143996024441104528_dp
  t_lf(686)=  3798.366551444178568524877286376228820285_dp
  t_lf(687)=  3804.898885736400917864713767217083954307_dp
  t_lf(688)=  3811.432674574334261525855538216262507615_dp
  t_lf(689)=  3817.967915845347920096053494796856883990_dp
  t_lf(690)=  3824.504607442939225162273484320273970021_dp
  t_lf(691)=  3831.042747266706895058881242390096925197_dp
  t_lf(692)=  3837.582333222324564624649207021454269802_dp
  t_lf(693)=  3844.123363221514467856606993816707630241_dp
  t_lf(694)=  3850.665835182021272358377902784551666344_dp
  t_lf(695)=  3857.209747027586064490167447036653397766_dp
  t_lf(696)=  3863.755096687920484136997660670312364780_dp
  t_lf(697)=  3870.301882098681008021113958661222898904_dp
  t_lf(698)=  3876.850101201443380493730672305179715181_dp
  t_lf(699)=  3883.399751943677190750428142416771128364_dp
  t_lf(700)=  3889.950832278720595423569478069583036509_dp
  t_lf(701)=  3896.503340165755185514069826493611037082_dp
  t_lf(702)=  3903.057273569780996633726281767402109369_dp
  t_lf(703)=  3909.612630461591661538103404870321410070_dp
  t_lf(704)=  3916.169408817749703938668741177035598779_dp
  t_lf(705)=  3922.727606620561972591485696416956485832_dp
  t_lf(706)=  3929.287221858055214668298648655602632402_dp
  t_lf(707)=  3935.848252523951787424288199665958801289_dp
  t_lf(708)=  3942.410696617645507185133959519517377754_dp
  t_lf(709)=  3948.974552144177634684299156576316600857_dp
  t_lf(710)=  3955.539817114212995789646602563513999162_dp
  t_lf(711)=  3962.106489544016236666610038578840461493_dp
  t_lf(712)=  3968.674567455428212433179550483884840417_dp
  t_lf(713)=  3975.244048875842508369915467640237393746_dp
  t_lf(714)=  3981.814931838182092756082832359934402284_dp
  t_lf(715)=  3988.387214380876100410799022712691037828_dp
  t_lf(716)=  3994.960894547836746025811291385628793139_dp
  t_lf(717)=  4001.535970388436366384169700216813227940_dp
  t_lf(718)=  4008.112439957484590566635025223876084486_dp
  t_lf(719)=  4014.690301315205637255161511285762777988_dp
  t_lf(720)=  4021.269552527215738250221689578666723309_dp
  t_lf(721)=  4027.850191664500687326095644637163343859_dp
  t_lf(722)=  4034.432216803393513555530931622397227502_dp
  t_lf(723)=  4041.015626025552278242392588178030568941_dp
  t_lf(724)=  4047.600417417937994608066139459579931005_dp
  t_lf(725)=  4054.186589072792669384450930158394217889_dp
  t_lf(726)=  4060.774139087617465473387294672705178762_dp
  t_lf(727)=  4067.363065565150984839299747641548446651_dp
  t_lf(728)=  4073.953366613347670808710284190931474689_dp
  t_lf(729)=  4080.545040345356328957081755612466628917_dp
  t_lf(730)=  4087.138084879498765770191855924589535696_dp
  t_lf(731)=  4093.732498339248544273914233055808509267_dp
  t_lf(732)=  4100.328278853209855832895331639073002854_dp
  t_lf(733)=  4106.925424555096507325164480337858101160_dp
  t_lf(734)=  4113.523933583711022906201144398474706894_dp
  t_lf(735)=  4120.123804082923859582407854455509779697_dp
  t_lf(736)=  4126.725034201652735820300767894610858656_dp
  t_lf(737)=  4133.327622093842072424032783468939710707_dp
  t_lf(738)=  4139.931565918442544920107269437280346272_dp
  t_lf(739)=  4146.536863839390746694324428312360884608_dp
  t_lf(740)=  4153.143514025588962132127747559534872547_dp
  t_lf(741)=  4159.751514650885049019585507669910570394_dp
  t_lf(742)=  4166.360863894052429468252561673840984462_dp
  t_lf(743)=  4172.971559938770188633111180513475497346_dp
  t_lf(744)=  4179.583600973603280498687286439314909965_dp
  t_lf(745)=  4186.196985191982840014280474047645218092_dp
  t_lf(746)=  4192.811710792186600865033440774205058798_dp
  t_lf(747)=  4199.427775977319418171299403471341480661_dp
  t_lf(748)=  4206.045178955293895414445345910096089632_dp
  t_lf(749)=  4212.663917938811114892855097470562546363_dp
  t_lf(750)=  4219.283991145341471017469852828621811555_dp
  t_lf(751)=  4225.905396797105605761725366231488745483_dp
  t_lf(752)=  4232.528133121055445586215245387093625464_dp
  t_lf(753)=  4239.152198348855339163829075863024404410_dp
  t_lf(754)=  4245.777590716863295236483067458409811189_dp
  t_lf(755)=  4252.404308466112319939880063091120475118_dp
  t_lf(756)=  4259.032349842291852938005615788247585098_dp
  t_lf(757)=  4265.661713095729301714287941560140526650_dp
  t_lf(758)=  4272.292396481371673371522410265252331126_dp
  t_lf(759)=  4278.924398258767303297786351911950182249_dp
  t_lf(760)=  4285.557716692047680060647835041438753130_dp
  t_lf(761)=  4292.192350049909365897003213789295522749_dp
  t_lf(762)=  4298.828296605596012170863131955392517417_dp
  t_lf(763)=  4305.465554636880469176345801151044898797_dp
  t_lf(764)=  4312.104122426046989668030215346163619080_dp
  t_lf(765)=  4318.743998259873525505670999771107984665_dp
  t_lf(766)=  4325.385180429614116805081286429718631485_dp
  t_lf(767)=  4332.027667230981372991750824245003647714_dp
  t_lf(768)=  4338.671456964129045158483926453591585963_dp
  t_lf(769)=  4345.316547933634689133016283299512663608_dp
  t_lf(770)=  4351.962938448482418666201571075605336845_dp
  t_lf(771)=  4358.610626822045748155948610627289118001_dp
  t_lf(772)=  4365.259611372070524326638005923588529902_dp
  t_lf(773)=  4371.909890420657946288254157485375372393_dp
  t_lf(774)=  4378.561462294247673403934722594024447406_dp
  t_lf(775)=  4385.214325323601020399065405585019179896_dp
  t_lf(776)=  4391.868477843784239149433823571097671314_dp
  t_lf(777)=  4398.523918194151886590302517222494823912_dp
  t_lf(778)=  4405.180644718330278192568369577585173131_dp
  t_lf(779)=  4411.838655764201026456444160382510434257_dp
  t_lf(780)=  4418.497949683884663877328116637140819342_dp
  t_lf(781)=  4425.158524833724349842719514747618982738_dp
  t_lf(782)=  4431.820379574269660923193034318760481960_dp
  t_lf(783)=  4438.483512270260464024562042061889970680_dp
  t_lf(784)=  4445.147921290610871872441676034609036412_dp
  t_lf(785)=  4451.813605008393280304466880604238699300_dp
  t_lf(786)=  4458.480561800822486849428781973788927466_dp
  t_lf(787)=  4465.148790049239890076565359422273343258_dp
  t_lf(788)=  4471.818288139097769202179620994352527369_dp
  t_lf(789)=  4478.489054459943643444660792495480125164_dp
  t_lf(790)=  4485.161087405404710622851729491645900294_dp
  t_lf(791)=  4491.834385373172364496539213728705101696_dp
  t_lf(792)=  4498.508946764986790351643344144889812109_dp
  t_lf(793)=  4505.184769986621638336448220689880745460_dp
  t_lf(794)=  4511.861853447868774058948884841515781312_dp
  t_lf(795)=  4518.540195562523105959089358550693552426_dp
  t_lf(796)=  4525.219794748367488973333944085478442764_dp
  t_lf(797)=  4531.900649427157704012649037888838381339_dp
  t_lf(798)=  4538.582758024607512778575895422550116878_dp
  t_lf(799)=  4545.266118970373787445646380710195417123_dp
  t_lf(800)=  4551.950730698041714741934059983938675242_dp
  t_lf(801)=  4558.636591645110073963042365998453575871_dp
  t_lf(802)=  4565.323700252976588458310277207261102170_dp
  t_lf(803)=  4572.012054966923350131464329642664774040_dp
  t_lf(804)=  4578.701654236102316501364111118867375632_dp
  t_lf(805)=  4585.392496513520879871876976027346939119_dp
  t_lf(806)=  4592.084580256027508163276859914912791503_dp
  t_lf(807)=  4598.777903924297456960892058332330554325_dp
  t_lf(808)=  4605.472465982818552340027952963618073186_dp
  t_lf(809)=  4612.168264899877044029463205426304658365_dp
  t_lf(810)=  4618.865299147543528479062177828679125391_dp
  t_lf(811)=  4625.563567201658941400262558502408025873_dp
  t_lf(812)=  4632.263067541820619351385647521129470344_dp
  t_lf(813)=  4638.963798651368429942877762831257858088_dp
  t_lf(814)=  4645.665759017370970240725034201712611120_dp
  t_lf(815)=  4652.368947130611832949390725901059548636_dp
  t_lf(816)=  4659.073361485575939958704434241687907148_dp
  t_lf(817)=  4665.779000580435942842186304186921607721_dp
  t_lf(818)=  4672.485862917038689897317058517508170377_dp
  t_lf(819)=  4679.193947000891759321266389176361720120_dp
  t_lf(820)=  4685.903251341150058118568376125541668484_dp
  t_lf(821)=  4692.613774450602486340183324355383429782_dp
  t_lf(822)=  4699.325514845658666256310991841743997445_dp
  t_lf(823)=  4706.038471046335736068220864807189375589_dp
  t_lf(824)=  4712.752641576245207767241163486617346114_dp
  t_lf(825)=  4719.468024962579888751899870967957376398_dp
  t_lf(826)=  4726.184619736100866817038506206578430319_dp
  t_lf(827)=  4732.902424431124558131522844053461770757_dp
  t_lf(828)=  4739.621437585509817823954551602033371421_dp
  t_lf(829)=  4746.341657740645112798543990341875953405_dp
  t_lf(830)=  4753.063083441435756406037454019851688066_dp
  t_lf(831)=  4759.785713236291204597302096806298824610_dp
  t_lf(832)=  4766.509545677112413189858976976613202623_dp
  t_lf(833)=  4773.234579319279255880319217081372688618_dp
  t_lf(834)=  4779.960812721638002638320479358628934673_dp
  t_lf(835)=  4786.688244446488858120180989468599901678_dp
  t_lf(836)=  4793.416873059573559743086424721369237652_dp
  t_lf(837)=  4800.146697130063035063201324756679171976_dp
  t_lf(838)=  4806.877715230545118103651490799822838934_dp
  t_lf(839)=  4813.609925937012324280857319608222463964_dp
  t_lf(840)=  4820.343327828849683580210373286188886742_dp
  t_lf(841)=  4827.077919488822631634576917350912709953_dp
  t_lf(842)=  4833.813699503064958361582865700767983994_dp
  t_lf(843)=  4840.550666461066813818084749262383681640_dp
  t_lf(844)=  4847.288818955662770932661169770754314894_dp
  t_lf(845)=  4854.028155583019944779368903987111139743_dp
  t_lf(846)=  4860.768674942626168058397577252186541429_dp
  t_lf(847)=  4867.510375637278222451626817151559979758_dp
  t_lf(848)=  4874.253256273070125523440214776421743801_dp
  t_lf(849)=  4880.997315459381472839481447160139409711_dp
  t_lf(850)=  4887.742551808865834978349732565923088093_dp
  t_lf(851)=  4894.488963937439209113524581068764731376_dp
  t_lf(852)=  4901.236550464268524845083745081116644313_dp
  t_lf(853)=  4907.985310011760203963033544994655618432_dp
  t_lf(854)=  4914.735241205548773826307518962982589476_dp
  t_lf(855)=  4921.486342674485534043707796201941682062_dp
  t_lf(856)=  4928.238613050627276145263891383339488768_dp
  t_lf(857)=  4934.992050969225055934665933291632237353_dp
  t_lf(858)=  4941.746655068713018215593841669543387530_dp
  t_lf(859)=  4948.502423990697273586910818106715017099_dp
  t_lf(860)=  4955.259356379944827003818884196203404910_dp
  t_lf(861)=  4962.017450884372557804186245549606517933_dp
  t_lf(862)=  4968.776706155036250901351144477389744755_dp
  t_lf(863)=  4975.537120846119678846784746382603317785_dp
  t_lf(864)=  4982.298693614923734468056642700661777739_dp
  t_lf(865)=  4989.061423121855613789590902422328956848_dp
  t_lf(866)=  4995.825308030418048945728419117759677385_dp
  t_lf(867)=  5002.590347007198590797622733590428456161_dp
  t_lf(868)=  5009.356538721858940967491714901330346237_dp
  t_lf(869)=  5016.123881847124333005726604020776886459_dp
  t_lf(870)=  5022.892375058772962408323112744745687977_dp
  t_lf(871)=  5029.662017035625465204046672182674729332_dp
  t_lf(872)=  5036.432806459534444832675684999258460688_dp
  t_lf(873)=  5043.204742015374047037582897094807473811_dp
  t_lf(874)=  5049.977822391029582497815909479963700035_dp
  t_lf(875)=  5056.752046277387196926723540223085442683_dp
  t_lf(876)=  5063.527412368323588366045358560362864095_dp
  t_lf(877)=  5070.303919360695771406237386169968074829_dp
  t_lf(878)=  5077.081565954330888065647830421169005097_dp
  t_lf(879)=  5083.860350852016065062982915920904028992_dp
  t_lf(880)=  5090.640272759488317219314547317928052204_dp
  t_lf(881)=  5097.421330385424496727678800148612370636_dp
  t_lf(882)=  5104.203522441431288030097228230801958072_dp
  t_lf(883)=  5110.986847642035248043621825111755265627_dp
  t_lf(884)=  5117.771304704672891478759311414110063904_dp
  t_lf(885)=  5124.556892349680820995371366357978474872_dp
  t_lf(886)=  5131.343609300285901942874605737281927143_dp
  t_lf(887)=  5138.131454282595481433277654390094655900_dp
  t_lf(888)=  5144.920426025587651497292691662423158473_dp
  t_lf(889)=  5151.710523261101556075445485213582630598_dp
  t_lf(890)=  5158.501744723827741597781292208936844025_dp
  t_lf(891)=  5165.294089151298550907424216734592076143_dp
  t_lf(892)=  5172.087555283878560284894786187231569306_dp
  t_lf(893)=  5178.882141864755059331724764288891596552_dp
  t_lf(894)=  5185.677847639928573473529669922376419312_dp
  t_lf(895)=  5192.474671358203428844308233685624009128_dp
  t_lf(896)=  5199.272611771178359315334211279274424834_dp
  t_lf(897)=  5206.071667633237155433589696789572465262_dp
  t_lf(898)=  5212.871837701539355036260450701856751831_dp
  t_lf(899)=  5219.673120736010975309372887058761020647_dp
  t_lf(900)=  5226.475515499335286060199360441974800472_dp
  t_lf(901)=  5233.279020756943623974593364198876984778_dp
  t_lf(902)=  5240.083635277006247631939303271337698234_dp
  t_lf(903)=  5246.889357830423233051912743765049250704_dp
  t_lf(904)=  5253.696187190815409548746571623039802080_dp
  t_lf(905)=  5260.504122134515335670186417994898998647_dp
  t_lf(906)=  5267.313161440558314999795131652764177210_dp
  t_lf(907)=  5274.123303890673451602731100607399792100_dp
  t_lf(908)=  5280.934548269274744896578946427149642281_dp
  t_lf(909)=  5287.746893363452223730253635167907682847_dp
  t_lf(910)=  5294.560337962963119455430466807600545389_dp
  t_lf(911)=  5301.374880860223077776373818194185529191_dp
  t_lf(912)=  5308.190520850297409165447019348828614705_dp
  t_lf(913)=  5315.007256730892377632984435150085114458_dp
  t_lf(914)=  5321.825087302346527641594798652696774561_dp
  t_lf(915)=  5328.644011367622048956342192326271102651_dp
  t_lf(916)=  5335.464027732296179223617892173120064105_dp
  t_lf(917)=  5342.285135204552644072872668927732769727_dp
  t_lf(918)=  5349.107332595173134536725171377831649945_dp
  t_lf(919)=  5355.930618717528821586297789411887200299_dp
  t_lf(920)=  5362.754992387571907579956997941298113761_dp
  t_lf(921)=  5369.580452423827214424950705745503131750_dp
  t_lf(922)=  5376.406997647383808252740663060235834857_dp
  t_lf(923)=  5383.234626881886660410123605034314187559_dp
  t_lf(924)=  5390.063338953528344569520610835561375429_dp
  t_lf(925)=  5396.893132691040769763090225173045197872_dp
  t_lf(926)=  5403.724006925686949146587304475762084382_dp
  t_lf(927)=  5410.555960491252804300146397264660576612_dp
  t_lf(928)=  5417.388992224039004874415829904313371058_dp
  t_lf(929)=  5424.223100962852843391706626651465271483_dp
  t_lf(930)=  5431.058285549000145013049027667614518606_dp
  t_lf(931)=  5437.894544826277212083268760586388731602_dp
  t_lf(932)=  5444.731877640962803267405449703973972217_dp
  t_lf(933)=  5451.570282841810147092996690506329180198_dp
  t_lf(934)=  5458.409759280038989713943453916521625752_dp
  t_lf(935)=  5465.250305809327676712855691445586069227_dp
  t_lf(936)=  5472.091921285805268759951365725370968945_dp
  t_lf(937)=  5478.934604568043690947747706322536297687_dp
  t_lf(938)=  5485.778354517049915621940363183801376735_dp
  t_lf(939)=  5492.623169996258178530014373729981112793_dp
  t_lf(940)=  5499.469049871522228110270547975895827278_dp
  t_lf(941)=  5506.315993011107607745082081877377737569_dp
  t_lf(942)=  5513.163998285683970803319004472161915091_dp
  t_lf(943)=  5520.013064568317428297992520677009518798_dp
  t_lf(944)=  5526.863190734462928986277499536561922694_dp
  t_lf(945)=  5533.714375661956671740169347323998867177_dp
  t_lf(946)=  5540.566618231008550017121365536768020081_dp
  t_lf(947)=  5547.419917324194628261090494220069782086_dp
  t_lf(948)=  5554.274271826449650065493149241390071849_dp
  t_lf(949)=  5561.129680625059577930638745540393933024_dp
  t_lf(950)=  5567.986142609654164449266523760192338409_dp
  t_lf(951)=  5574.843656672199554754861531166413475423_dp
  t_lf(952)=  5581.702221706990920068468114892104311392_dp
  t_lf(953)=  5588.561836610645122180754131202698680066_dp
  t_lf(954)=  5595.422500282093408707106322937030965814_dp
  t_lf(955)=  5602.284211622574138954557032222459520600_dp
  t_lf(956)=  5609.146969535625540240354660059637782833_dp
  t_lf(957)=  5616.010772927078494502995120906887349443_dp
  t_lf(958)=  5622.875620705049355047529030859745142625_dp
  t_lf(959)=  5629.741511779932793267949573731168187745_dp
  t_lf(960)=  5636.608445064394675190448971030065960498_dp
  t_lf(961)=  5643.476419473364967682307299679150674918_dp
  t_lf(962)=  5650.345433924030674172146114913205617255_dp
  t_lf(963)=  5657.215487335828799728241004204073945666_dp
  t_lf(964)=  5664.086578630439345342541879765701114537_dp
  t_lf(965)=  5670.958706731778331268997570152310360941_dp
  t_lf(966)=  5677.831870565990849265722153085944439062_dp
  t_lf(967)=  5684.706069061444143591474538812253944780_dp
  t_lf(968)=  5691.581301148720720607850122332558733084_dp
  t_lf(969)=  5698.457565760611486839503929619242238862_dp
  t_lf(970)=  5705.334861832108915345638642695630564783_dp
  t_lf(971)=  5712.213188300400240256898256042953742210_dp
  t_lf(972)=  5719.092544104860679332708946470482723870_dp
  t_lf(973)=  5725.972928187046684395003084132801447382_dp
  t_lf(974)=  5732.854339490689219495150227085510169455_dp
  t_lf(975)=  5739.736776961687066671800478430450389043_dp
  t_lf(976)=  5746.620239548100159158220796019708710062_dp
  t_lf(977)=  5753.504726200142941898573781688030001170_dp
  t_lf(978)=  5760.390235870177759233451191412531453320_dp
  t_lf(979)=  5767.276767512708269615830950531166431839_dp
  t_lf(980)=  5774.164320084372887219476879594195332074_dp
  t_lf(981)=  5781.052892543938250302644686520249585134_dp
  t_lf(982)=  5787.942483852292716190796108966722846355_dp
  t_lf(983)=  5794.833092972439882742855446236178560791_dp
  t_lf(984)=  5801.724718869492136166369151210513023789_dp
  t_lf(985)=  5808.617360510664225047749707872902042403_dp
  t_lf(986)=  5815.511016865266860464599746644595257112_dp
  t_lf(987)=  5822.405686904700342047921295294733136255_dp
  t_lf(988)=  5829.301369602448209862818274411102661534_dp
  t_lf(989)=  5836.198063934070921977097869756258704688_dp
  t_lf(990)=  5843.095768877199557587968295262753249604_dp
  t_lf(991)=  5849.994483411529545577816745511523956376_dp
  t_lf(992)=  5856.894206518814418370832070443357196427_dp
  t_lf(993)=  5863.794937182859590963011932519856358939_dp
  t_lf(994)=  5870.696674389516164998863971917270749333_dp
  t_lf(995)=  5877.599417126674757768874852542365474174_dp
  t_lf(996)=  5884.503164384259356002580034245495723468_dp
  t_lf(997)=  5891.407915154221194332820757720199195980_dp
  t_lf(998)=  5898.313668430532658307523080979641765571_dp
  t_lf(999)=  5905.220423209181211826076912361440789849_dp
  t_lf(1000)= 5912.128178488163348878130886725493882472_dp
  t_lf(1001)=  5919.036933267478569463351670488467510106_dp
  t_lf(1002)=  5925.946686549123379571423898623592991745_dp
  t_lf(1003)=  5932.857437337085315102289483615185815904_dp
  t_lf(1004)=  5939.769184637336989607342533097110422282_dp
  t_lf(1005)=  5946.681927457830165733008609663622858376_dp
  t_lf(1006)=  5953.595664808489850248844602900716644976_dp
  t_lf(1007)=  5960.510395701208412542998099471633556284_dp
  t_lf(1008)=  5967.426119149839726468562871174754493695_dp
  t_lf(1009)=  5974.342834170193335425059992000134914410_dp
  t_lf(1010)=  5981.260539780028640559962181721732267775_dp
  t_lf(1011)=  5988.179234999049111975862293513303224692_dp
  t_lf(1012)=  5995.098918848896522829565454165994903247_dp
  t_lf(1013)=  6002.019590353145206210058265078683323405_dp
  t_lf(1014)=  6008.941248537296334682977717320194662888_dp
  t_lf(1015)=  6015.863892428772222389867101429225941862_dp
  t_lf(1016)=  6022.787521056910649591166238601316763961_dp
  t_lf(1017)=  6029.712133452959209542538860568777837043_dp
  t_lf(1018)=  6036.637728650069677594790956546295772259_dp
  t_lf(1019)=  6043.564305683292402408280422513454447746_dp
  t_lf(1020)=  6050.491863589570719173360425944392640762_dp
  t_lf(1021)=  6057.420401407735384729036572673375725286_dp
  t_lf(1022)=  6064.349918178499034472651266395715624154_dp
  t_lf(1023)=  6071.280412944450660954037619535145636369_dp
  t_lf(1024)=  6078.211884750050114048209940749727402050_dp
  t_lf(1025)=  6085.144332641622622601278222789217184917_dp
  t_lf(1026)=  6092.077755667353337444890218053330792137_dp
  t_lf(1027)=  6099.012152877281895675116651036377521664_dp
  t_lf(1028)=  6105.947523323297006092302909594055130251_dp
  t_lf(1029)=  6112.883866059131055699014213061307195145_dp
  t_lf(1030)=  6119.821180140354737153800806831045000173_dp
  t_lf(1031)=  6126.759464624371697079105211936489202582_dp
  t_lf(1032)=  6133.698718570413205122224996051132105026_dp
  t_lf(1033)=  6140.638941039532843668831960440771225319_dp
  t_lf(1034)=  6147.580131094601218109132086809966704896_dp
  t_lf(1035)=  6154.522287800300687557330089448848140063_dp
  t_lf(1036)=  6161.465410223120115925638002106239120093_dp
  t_lf(1037)=  6168.409497431349643254638925827537861375_dp
  t_lf(1038)=  6175.354548495075477202384903574359555116_dp
  t_lf(1039)=  6182.300562486174704595171899501517730551_dp
  t_lf(1040)=  6189.247538478310122943495074762141943068_dp
  t_lf(1041)=  6196.195475546925091827243996845450328307_dp
  t_lf(1042)=  6203.144372769238404054750124673426341339_dp
  t_lf(1043)=  6210.094229224239176500847905691973641556_dp
  t_lf(1044)=  6217.045043992681760529656132441096957707_dp
  t_lf(1045)=  6223.996816157080671908327862784176128184_dp
  t_lf(1046)=  6230.949544801705540118555238082547400045_dp
  t_lf(1047)=  6237.903229012576076973149964841968565459_dp
  t_lf(1048)=  6244.857867877457064445551085217512621057_dp
  t_lf(1049)=  6251.813460485853361620638977492472084792_dp
  t_lf(1050)=  6258.770005929004930675758326260748342073_dp
  t_lf(1051)=  6265.727503299881881801373110325279029322_dp
  t_lf(1052)=  6272.685951693179536971293500832400454548_dp
  t_lf(1053)=  6279.645350205313512472927969221655875972_dp
  t_lf(1054)=  6286.605697934414820108523900285529536286_dp
  t_lf(1055)=  6293.566993980324986978866615884210004043_dp
  t_lf(1056)=  6300.529237444591193761409965306388541888_dp
  t_lf(1057)=  6307.492427430461431395311554349316197908_dp
  t_lf(1058)=  6314.456563042879676086342292134394766713_dp
  t_lf(1059)=  6321.421643388481082545133257488505262420_dp
  t_lf(1060)=  6328.387667575587195372712950203676860966_dp
  t_lf(1061)=  6335.354634714201178507774821225716681517_dp
  t_lf(1062)=  6342.322543916003062650598594194739607118_dp
  t_lf(1063)=  6349.291394294345010579029322939338972848_dp
  t_lf(1064)=  6356.261184964246600272395399479044535819_dp
  t_lf(1065)=  6363.231915042390125759720858581706283260_dp
  t_lf(1066)=  6370.203583647115915609058341517767186021_dp
  t_lf(1067)=  6377.176189898417668975237006717436277035_dp
  t_lf(1068)=  6384.149732917937809123784531737945005095_dp
  t_lf(1069)=  6391.124211828962854349244162259616157519_dp
  t_lf(1070)=  6398.099625756418806206566552531323798728_dp
  t_lf(1071)=  6405.075973826866554974711930366485156403_dp
  t_lf(1072)=  6412.053255168497302272050930848681585426_dp
  t_lf(1073)=  6419.031468911128000743602298552074944195_dp
  t_lf(1074)=  6426.010614186196810740592580340477048643_dp
  t_lf(1075)=  6432.990690126758573913266941520275270958_dp
  t_lf(1076)=  6439.971695867480303638321358943686861212_dp
  t_lf(1077)=  6446.953630544636692202764697066214066894_dp
  t_lf(1078)=  6453.936493296105634666454578252523732221_dp
  t_lf(1079)=  6460.920283261363769325983537917380421279_dp
  t_lf(1080)=  6467.904999581482034703021729325748715737_dp
  t_lf(1081)=  6474.890641399121242980649432827331085564_dp
  t_lf(1082)=  6481.877207858527669811636850561373912269_dp
  t_lf(1083)=  6488.864698105528660423050150662072145048_dp
  t_lf(1084)=  6495.853111287528251941981484978194360223_dp
  t_lf(1085)=  6502.842446553502811867616761379406084803_dp
  t_lf(1086)=  6509.832703053996692615268325776419796003_dp
  t_lf(1087)=  6516.823879941117902058410419787765757858_dp
  t_lf(1088)=  6523.815976368533789995163347134387943802_dp
  t_lf(1089)=  6530.808991491466750466077724764163253811_dp
  t_lf(1090)=  6537.802924466689939850473032671056819670_dp
  t_lf(1091)=  6544.797774452523010668984928488167203750_dp
  t_lf(1092)=  6551.793540608827861020373478153014580926_dp
  t_lf(1093)=  6558.790222097004399581039591063582765061_dp
  t_lf(1094)=  6565.787818079986326096089555806638602382_dp
  t_lf(1095)=  6572.786327722236927291177669234225858297_dp
  t_lf(1096)=  6579.785750189744888134744555726580253392_dp
  t_lf(1097)=  6586.786084650020118380653903096146903178_dp
  t_lf(1098)=  6593.787330272089594321613014794875745902_dp
  t_lf(1099)=  6600.789486226493215684142812774722400880_dp
  t_lf(1100)=  6607.792551685279677596240739262056258595_dp
  t_lf(1101)=  6614.796525822002357559255416438137213465_dp
  t_lf(1102)=  6621.801407811715217355864948023845155176_dp
  t_lf(1103)=  6628.807196830968719826421403358829245453_dp
  t_lf(1104)=  6635.813892057805760446292329913394673549_dp
  t_lf(1105)=  6642.821492671757613637196111306059306329_dp
  t_lf(1106)=  6649.829997853839893745891641712442073549_dp
  t_lf(1107)=  6656.839406786548530623944140796247058251_dp
  t_lf(1108)=  6663.849718653855759742648002588688022226_dp
  t_lf(1109)=  6670.860932641206126777544373571939592436_dp
  t_lf(1110)=  6677.873047935512506597325705934577929512_dp
  t_lf(1111)=  6684.886063725152136592271847779456047969_dp
  t_lf(1112)=  6691.899979199962664277712329062706121456_dp
  t_lf(1113)=  6698.914793551238209108357396182100884660_dp
  t_lf(1114)=  6705.930505971725438439686055240403842993_dp
  t_lf(1115)=  6712.947115655619657572922919783353170659_dp
  t_lf(1116)=  6719.964621798560913820477038824656932415_dp
  t_lf(1117)=  6726.983023597630114529055120668402322273_dp
  t_lf(1118)=  6734.002320251345158997998682744771664481_dp
  t_lf(1119)=  6741.022510959657084230729662586795854324_dp
  t_lf(1120)=  6748.043594923946224457521935270756104533_dp
  t_lf(1121)=  6755.065571347018384368147013076363655695_dp
  t_lf(1122)=  6762.088439433101025993270968630582613803_dp
  t_lf(1123)=  6769.112198387839469173806338097542368147_dp
  t_lf(1124)=  6776.136847418293105557747440665151893224_dp
  t_lf(1125)=  6783.162385732931626064340209138675507552_dp
  t_lf(1126)=  6790.188812541631261755758278248751425102_dp
  t_lf(1127)=  6797.216127055671038056775736567447980680_dp
  t_lf(1128)=  6804.244328487729042263243628838517209798_dp
  t_lf(1129)=  6811.273416051878704280492012091831214316_dp
  t_lf(1130)=  6818.303388963585090533092135040131600196_dp
  t_lf(1131)=  6825.334246439701210987724139750981356111_dp
  t_lf(1132)=  6832.365987698464339231204591140692849452_dp
  t_lf(1133)=  6839.398611959492345546035137033711419572_dp
  t_lf(1134)=  6846.432118443780042926138702846302878689_dp
  t_lf(1135)=  6853.466506373695545975752843756362563373_dp
  t_lf(1136)=  6860.501774972976642634751226774708303742_dp
  t_lf(1137)=  6867.537923466727178673963708595284457355_dp
  t_lf(1138)=  6874.574951081413454904363125526126772401_dp
  t_lf(1139)=  6881.612857044860637044282732140363621687_dp
  t_lf(1140)=  6888.651640586249178189122228385316541705_dp
  t_lf(1141)=  6895.691300936111253828292513494880471311_dp
  t_lf(1142)=  6902.731837326327209354439711812278277043_dp
  t_lf(1143)=  6909.773248990122020010277643093839620848_dp
  t_lf(1144)=  6916.815535162061763218644770477744598456_dp
  t_lf(1145)=  6923.858695078050103241686765414903394413_dp
  t_lf(1146)=  6930.902727975324788115349192725486463833_dp
  t_lf(1147)=  6937.947633092454158805646452721060268207_dp
  t_lf(1148)=  6944.993409669333670533453033080457208661_dp
  t_lf(1149)=  6952.040056947182426214841332854532204618_dp
  t_lf(1150)=  6959.087574168539721964266836474252952583_dp
  t_lf(1151)=  6966.135960577261604608176248728674765162_dp
  t_lf(1152)=  6973.185215418517441156887364052727052547_dp
  t_lf(1153)=  6980.235337938786500182860946716573427429_dp
  t_lf(1154)=  6987.286327385854545053754754145302292404_dp
  t_lf(1155)=  6994.338183008810438968918055036859314778_dp
  t_lf(1156)=  7001.390904058042761748251588515521920985_dp
  t_lf(1157)=  7008.444489785236438322622891497757088810_dp
  t_lf(1158)=  7015.498939443369378875290299909520849846_dp
  t_lf(1159)=  7022.554252286709130584050716444834318130_dp
  t_lf(1160)=  7029.610427570809540914086444174796947079_dp
  t_lf(1161)=  7036.667464552507432411745022398910371229_dp
  t_lf(1162)=  7043.725362489919288949743079487103097980_dp
  t_lf(1163)=  7050.784120642437953374540747817891357732_dp
  t_lf(1164)=  7057.843738270729336506887178919434198287_dp
  t_lf(1165)=  7064.904214636729137446790163127329273405_dp
  t_lf(1166)=  7071.965549003639575134413807965781637043_dp
  t_lf(1167)=  7079.027740635926131118657673436336335399_dp
  t_lf(1168)=  7086.090788799314303485418710779607584242_dp
  t_lf(1169)=  7093.154692760786371897783814299795543338_dp
  t_lf(1170)=  7100.219451788578173700645783669890277559_dp
  t_lf(1171)=  7107.285065152175891042479016849830165195_dp
  t_lf(1172)=  7114.351532122312848967253321355559016521_dp
  t_lf(1173)=  7121.418851970966324429704854042164864880_dp
  t_lf(1174)=  7128.487023971354366187422386661192714125_dp
  t_lf(1175)=  7135.556047397932625523444855997417263113_dp
  t_lf(1176)=  7142.625921526391197753302503085600677982_dp
  t_lf(1177)=  7149.696645633651474470668845480589084283_dp
  t_lf(1178)=  7156.768218997863006486024269358477471599_dp
  t_lf(1179)=  7163.840640898400377412964183843492048901_dp
  t_lf(1180)=  7170.913910615860087857015457793354287301_dp
  t_lf(1181)=  7177.988027432057450162054268672123907703_dp
  t_lf(1182)=  7185.062990630023493669646543359667440951_dp
  t_lf(1183)=  7192.138799494001880446858870986241257890_dp
  t_lf(1184)=  7199.215453309445831438313127264563587894_dp
  t_lf(1185)=  7206.292951363015062998482077376193712161_dp
  t_lf(1186)=  7213.371292942572733760445929241302037368_dp
  t_lf(1187)=  7220.450477337182401797551199880094483550_dp
  t_lf(1188)=  7227.530503837104992034633343411743543100_dp
  t_lf(1189)=  7234.611371733795773865683378817142862294_dp
  t_lf(1190)=  7241.693080319901348935056257633143532767_dp
  t_lf(1191)=  7248.775628889256649039534934900242917755_dp
  t_lf(1192)=  7255.859016736881944108779059539721567947_dp
  t_lf(1193)=  7262.943243158979860221900891411895766728_dp
  t_lf(1194)=  7270.028307452932407618123490062145006203_dp
  t_lf(1195)=  7277.114208917298018659687412989633102939_dp
  t_lf(1196)=  7284.200946851808595705382117505924970798_dp
  t_lf(1197)=  7291.288520557366568853286988155101055475_dp
  t_lf(1198)=  7298.376929336041963511514419449440688730_dp
  t_lf(1199)=  7305.466172491069477755953679479615019680_dp
  t_lf(1200)=  7312.556249326845569434219371868822626936_dp
  t_lf(1201)=  7319.647159148925552975212209579040352673_dp
  t_lf(1202)=  7326.738901264020705863902522207077675373_dp
  t_lf(1203)=  7333.831474979995384741148446531349550808_dp
  t_lf(1204)=  7340.924879605864151088561106031054930710_dp
  t_lf(1205)=  7348.019114451788906458628276682991934084_dp
  t_lf(1206)=  7355.114178829076037210506071274658884812_dp
  t_lf(1207)=  7362.210072050173568712085060425045045444_dp
  t_lf(1208)=  7369.306793428668328969132993088904051439_dp
  t_lf(1209)=  7376.404342279283121642510890091934252959_dp
  t_lf(1210)=  7383.502717917873908414652768702548875766_dp
  t_lf(1211)=  7390.601919661427000666691621834433046965_dp
  t_lf(1212)=  7397.701946828056260427805529581184914963_dp
  t_lf(1213)=  7404.802798737000310558547930743513697403_dp
  t_lf(1214)=  7411.904474708619754130115135099818565738_dp
  t_lf(1215)=  7419.006974064394402961692120617657381901_dp
  t_lf(1216)=  7426.110296126920515278204540778294294846_dp
  t_lf(1217)=  7433.214440219908042450990672807792995873_dp
  t_lf(1218)=  7440.319405668177884784091774941978789480_dp
  t_lf(1219)=  7447.425191797659156309042996912818043369_dp
  t_lf(1220)=  7454.531797935386458551229609592386198891_dp
  t_lf(1221)=  7461.639223409497163231054894078305301060_dp
  t_lf(1222)=  7468.747467549228703863346564311100969941_dp
  t_lf(1223)=  7475.856529684915876218608097388627490521_dp
  t_lf(1224)=  7482.966409147988147609899818844720198170_dp
  t_lf(1225)=  7490.077105270966974969312042998058932909_dp
  t_lf(1226)=  7497.188617387463131678169007713093160414_dp
  t_lf(1227)=  7504.300944832174043115277775159144072205_dp
  t_lf(1228)=  7511.414086940881130887710701969342917625_dp
  t_lf(1229)=  7518.528043050447165708783520102689578677_dp
  t_lf(1230)=  7525.642812498813628888063520167333876178_dp
  t_lf(1231)=  7532.758394624998082398413798391006056216_dp
  t_lf(1232)=  7539.874788769091547485250023198247071518_dp
  t_lf(1233)=  7546.991994272255891783355703800071988317_dp
  t_lf(1234)=  7554.110010476721224906770507600140355710_dp
  t_lf(1235)=  7561.228836725783302477433781806819715492_dp
  t_lf(1236)=  7568.348472363800938558432093601712035153_dp
  t_lf(1237)=  7575.468916736193426457865318696219812535_dp
  t_lf(1238)=  7582.590169189437967869510586207827712833_dp
  t_lf(1239)=  7589.712229071067110316627234561913115891_dp
  t_lf(1240)=  7596.835095729666192865408854584056190446_dp
  t_lf(1241)=  7603.958768514870800074750498059367859552_dp
  t_lf(1242)=  7611.083246777364224149160218723403809353_dp
  t_lf(1243)=  7618.208529868874935261804293794984960324_dp
  t_lf(1244)=  7625.334617142174060014834753603333995737_dp
  t_lf(1245)=  7632.461507951072868004306230396774079534_dp
  t_lf(1246)=  7639.589201650420266457146630802345285051_dp
  t_lf(1247)=  7646.717697596100302907802745348053043692_dp
  t_lf(1248)=  7653.846995145029675882337638633831770842_dp
  t_lf(1249)=  7660.977093655155253557911520774560635094_dp
  t_lf(1250)=  7668.107992485451600365731790228923562221_dp
  t_lf(1251)=  7675.239690995918511505711065621644157412_dp
  t_lf(1252)=  7682.372188547578555341224295173817720902_dp
  t_lf(1253)=  7689.505484502474623642507452347282302807_dp
  t_lf(1254)=  7696.639578223667489647390900715515987917_dp
  t_lf(1255)=  7703.774469075233373908210245287750428798_dp
  t_lf(1256)=  7710.910156422261517893886386888528433751_dp
  t_lf(1257)=  7718.046639630851765316314566047136449845_dp
  t_lf(1258)=  7725.183918068112151150349428457499200113_dp
  t_lf(1259)=  7732.321991102156498316819570668781814036_dp
  t_lf(1260)=  7739.460858102102021998150637462212585950_dp
  t_lf(1261)=  7746.600518438066941556320846525481866269_dp
  t_lf(1262)=  7753.740971481168100023016816680605636921_dp
  t_lf(1263)=  7760.882216603518591132000778145925260099_dp
  t_lf(1264)=  7768.024253178225393863842652173239377293_dp
  t_lf(1265)=  7775.167080579387014473312107916240890351_dp
  t_lf(1266)=  7782.310698182091135969866541540075872742_dp
  t_lf(1267)=  7789.455105362412275021810981322152061399_dp
  t_lf(1268)=  7796.600301497409446254845207734367025844_dp
  t_lf(1269)=  7803.746285965123833915851894114906776667_dp
  t_lf(1270)=  7810.893058144576470872917326377307433270_dp
  t_lf(1271)=  7818.040617415765924922713254074887198069_dp
  t_lf(1272)=  7825.188963159665992376504664815213311248_dp
  t_lf(1273)=  7832.338094758223398896183764243464887536_dp
  t_lf(1274)=  7839.488011594355507551865189293374742168_dp
  t_lf(1275)=  7846.638713051948034072711487814622769687_dp
  t_lf(1276)=  7853.790198515852769262791167667866163729_dp
  t_lf(1277)=  7860.942467371885308553904157542056924446_dp
  t_lf(1278)=  7868.095519006822788667441334669873186520_dp
  t_lf(1279)=  7875.249352808401631357475865840319711227_dp
  t_lf(1280)=  7882.403968165315294207414482145211311412_dp
  t_lf(1281)=  7889.559364467212028452666469229002631592_dp
  t_lf(1282)=  7896.715541104692643801917107888462408346_dp
  t_lf(1283)=  7903.872497469308280229720549107981481318_dp
  t_lf(1284)=  7911.030232953558186713254657404843637160_dp
  t_lf(1285)=  7918.188746950887506886207211052831080252_dp
  t_lf(1286)=  7925.348038855685071582889011677837289906_dp
  t_lf(1287)=  7932.508108063281198245794933171212789220_dp
  t_lf(1288)=  7939.668953969945497169958735110840694773_dp
  t_lf(1289)=  7946.830575972884684557571583152101888859_dp
  t_lf(1290)=  7953.992973470240402356457662357054625807_dp
  t_lf(1291)=  7961.156145861087044856123043347097119192_dp
  t_lf(1292)=  7968.320092545429592015216069639774452401_dp
  t_lf(1293)=  7975.484812924201449494358981683022028360_dp
  t_lf(1294)=  7982.650306399262295368431283020137745340_dp
  t_lf(1295)=  7989.816572373395933492505490767838559873_dp
  t_lf(1296)=  7996.983610250308153495755400201361368964_dp
  t_lf(1297)=  8004.151419434624597377775835720191035243_dp
  t_lf(1298)=  8011.319999331888632681871061793334038735_dp
  t_lf(1299)=  8018.489349348559232219986591604229108409_dp
  t_lf(1300)=  8025.659468892008860324076061955163155429_dp
  t_lf(1301)=  8032.830357370521365598811144445061598753_dp
  t_lf(1302)=  8040.002014193289880150658138871427837966_dp
  t_lf(1303)=  8047.174438770414725268459949068966500744_dp
  t_lf(1304)=  8054.347630512901323530776577799461780325_dp
  t_lf(1305)=  8061.521588832658117315351099638894930979_dp
  t_lf(1306)=  8068.696313142494493686181282831551332629_dp
  t_lf(1307)=  8075.871802856118715633789636534043208483_dp
  t_lf(1308)=  8083.048057388135859644396662466091288975_dp
  t_lf(1309)=  8090.225076154045759573813493405372724540_dp
  t_lf(1310)=  8097.402858570240956801980908871226614640_dp
  t_lf(1311)=  8104.581404054004656644191934371847190001_dp
  t_lf(1312)=  8111.760712023508690995144858352175480429_dp
  t_lf(1313)=  8118.940781897811487182082544060653788191_dp
  t_lf(1314)=  8126.121613096856043003382375513395558739_dp
  t_lf(1315)=  8133.303205041467907929069061110826818469_dp
  t_lf(1316)=  8140.485557153353170439829828766958525044_dp
  t_lf(1317)=  8147.668668855096451481218286133623804449_dp
  t_lf(1318)=  8154.852539570158904009839392105901805118_dp
  t_lf(1319)=  8162.037168722876218608413594721543015548_dp
  t_lf(1320)=  8169.222555738456635146723239234031387897_dp
  t_lf(1321)=  8176.408700042978960465548841940196031469_dp
  t_lf(1322)=  8183.595601063390592060806763650132915313_dp
  t_lf(1323)=  8190.783258227505547745203204847786851887_dp
  t_lf(1324)=  8197.971670964002501264822285930279841829_dp
  t_lf(1325)=  8205.160838702422823848168273735761274879_dp
  t_lf(1326)=  8212.350760873168631665283773153580422292_dp
  t_lf(1327)=  8219.541436907500839174666923215067386489_dp
  t_lf(1328)=  8226.732866237537218335811323924191463214_dp
  t_lf(1329)=  8233.925048296250463665292576418959264622_dp
  t_lf(1330)=  8241.117982517466263114424948048974662097_dp
  t_lf(1331)=  8248.311668335861374746610778782870049996_dp
  t_lf(1332)=  8255.506105186961709192603829170662901705_dp
  t_lf(1333)=  8262.701292507140417862005836008551105951_dp
  t_lf(1334)=  8269.897229733615986889413092994181390243_dp
  t_lf(1335)=  8277.093916304450336793726913104999952807_dp
  t_lf(1336)=  8284.291351658546927829238360246119261877_dp
  t_lf(1337)=  8291.489535235648871007193662941764808753_dp
  t_lf(1338)=  8298.688466476337044766642245509868651053_dp
  t_lf(1339)=  8305.888144822028217273464335266487410478_dp
  t_lf(1340)=  8313.088569714973174326569630838993674004_dp
  t_lf(1341)=  8320.289740598254852850352549587942845901_dp
  t_lf(1342)=  8327.491656915786479952583114390791766516_dp
  t_lf(1343)=  8334.694318112309717527005594550146303973_dp
  t_lf(1344)=  8341.897723633392812380009585259261068816_dp
  t_lf(1345)=  8349.101872925428751860830297772982493574_dp
  t_lf(1346)=  8356.306765435633424974826441058797008347_dp
  t_lf(1347)=  8363.512400612043788959475208086545644053_dp
  t_lf(1348)=  8370.718777903516041302814538884110428402_dp
  t_lf(1349)=  8377.925896759723797184153020848511316036_dp
  t_lf(1350)=  8385.133756631156272316957507347189444997_dp
  t_lf(1351)=  8392.342356969116471173917791144015683491_dp
  t_lf(1352)=  8399.551697225719380574276462391520850405_dp
  t_lf(1353)=  8406.761776853890168613600414579445912998_dp
  t_lf(1354)=  8413.972595307362388916258338636224601593_dp
  t_lf(1355)=  8421.184152040676190190955968042656651272_dp
  t_lf(1356)=  8428.396446509176531069767809016111551785_dp
  t_lf(1357)=  8435.609478169011400211190612221641445527_dp
  t_lf(1358)=  8442.823246477130041647829918708246763539_dp
  t_lf(1359)=  8450.037750891281185359416645481576291238_dp
  t_lf(1360)=  8457.252990870011283051935867918508311686_dp
  t_lf(1361)=  8464.468965872662749123734709697039574665_dp
  t_lf(1362)=  8471.685675359372206799560568632253773982_dp
  t_lf(1363)=  8478.903118791068739413564791334387859297_dp
  t_lf(1364)=  8486.121295629472146822390363479811698943_dp
  t_lf(1365)=  8493.340205337091206929545208234968910622_dp
  t_lf(1366)=  8500.559847377221942302345285505237149144_dp
  t_lf(1367)=  8507.780221213945891862793861686978126292_dp
  t_lf(1368)=  8515.001326312128387633845075957085560943_dp
  t_lf(1369)=  8522.223162137416836522581267299148455270_dp
  t_lf(1370)=  8529.445728156239007121914448881812684869_dp
  t_lf(1371)=  8536.669023835801321512502825499888694108_dp
  t_lf(1372)=  8543.893048644087152046653347973134586433_dp
  t_lf(1373)=  8551.117802049855123096060988066550274076_dp
  t_lf(1374)=  8558.343283522637417745314701028863584666_dp
  t_lf(1375)=  8565.569492532738089413178922606507276884_dp
  t_lf(1376)=  8572.796428551231378383737925727144006760_dp
  t_lf(1377)=  8580.024091049960033229568441292707236114_dp
  t_lf(1378)=  8587.252479501533637109183629994759789058_dp
  t_lf(1379)=  8594.481593379326938921068780067365799762_dp
  t_lf(1380)=  8601.711432157478189296706001712241062361_dp
  t_lf(1381)=  8608.941995310887481415061694840392368544_dp
  t_lf(1382)=  8616.173282315215096621086685031673500288_dp
  t_lf(1383)=  8623.405292646879854830854655461870552533_dp
  t_lf(1384)=  8630.638025783057469706039852214686073706_dp
  t_lf(1385)=  8637.871481201678908580510009097436872185_dp
  t_lf(1386)=  8645.105658381428757121885028014148409191_dp
  t_lf(1387)=  8652.340556801743588710986164303474805300_dp
  t_lf(1388)=  8659.576175942810338522174305392777017970_dp
  t_lf(1389)=  8666.812515285564682287649397810958253617_dp
  t_lf(1390)=  8674.049574311689419728856174184518161607_dp
  t_lf(1391)=  8681.287352503612862638214060443106757213_dp
  t_lf(1392)=  8688.525849344507227594461506198223900795_dp
  t_lf(1393)=  8695.765064318287033294976980233535617727_dp
  t_lf(1394)=  8703.004996909607502488510510345904328419_dp
  t_lf(1395)=  8710.245646603862968491830924477517924678_dp
  t_lf(1396)=  8717.487012887185286273864870242932917524_dp
  t_lf(1397)=  8724.729095246442248090974254628614339218_dp
  t_lf(1398)=  8731.971893169236003657088956861663928970_dp
  t_lf(1399)=  8739.215406143901484832481527230786380614_dp
  t_lf(1400)=  8746.459633659504834815040095005056465327_dp
  t_lf(1401)=  8753.704575205841841817964871530713260018_dp
  t_lf(1402)=  8760.950230273436377217882452076199437159_dp
  t_lf(1403)=  8768.196598353538838157440594011435248024_dp
  t_lf(1404)=  8775.443678938124594586514281406684496879_dp
  t_lf(1405)=  8782.691471519892440726221679064603856459_dp
  t_lf(1406)=  8789.939975592263050940016034288981333728_dp
  t_lf(1407)=  8797.189190649377439996186704265710761913_dp
  t_lf(1408)=  8804.439116186095427706169272693883127189_dp
  t_lf(1409)=  8811.689751697994107923131173153475832454_dp
  t_lf(1410)=  8818.941096681366321885365360514854896076_dp
  t_lf(1411)=  8826.193150633219135889090367356019393064_dp
  t_lf(1412)=  8833.445913051272323275320551716123716203_dp
  t_lf(1413)=  8840.699383433956850715535487426372242861_dp
  t_lf(1414)=  8847.953561280413368780942270558186588315_dp
  t_lf(1415)=  8855.208446090490706780189017038207916160_dp
  t_lf(1416)=  8862.464037364744371850452009013224669193_dp
  t_lf(1417)=  8869.720334604435052286882812906999189449_dp
  t_lf(1418)=  8876.977337311527125095465242085256589120_dp
  t_lf(1419)=  8884.235044988687167754395273413490091160_dp
  t_lf(1420)=  8891.493457139282474169159951522145666033_dp
  t_lf(1421)=  8898.752573267379574806553929041393937098_dp
  t_lf(1422)=  8906.012392877742760992934597178178575997_dp
  t_lf(1423)=  8913.272915475832613362078760515606269466_dp
  t_lf(1424)=  8920.534140567804534438065504542108824958_dp
  t_lf(1425)=  8927.796067660507285338671295877371579478_dp
  t_lf(1426)=  8935.058696261481526584824445155182309375_dp
  t_lf(1427)=  8942.322025878958363001726852740775255214_dp
  t_lf(1428)=  8949.586056021857892697311449581930440320_dp
  t_lf(1429)=  8956.850786199787760103763941188459375500_dp
  t_lf(1430)=  8964.116215923041713067897363662674187612_dp
  t_lf(1431)=  8971.382344702598163976227568512470822497_dp
  t_lf(1432)=  8978.649172050118754900657069306866754376_dp
  t_lf(1433)=  8985.916697477946926750733709709036299961_dp
  t_lf(1434)=  8993.184920499106492418509350661678911330_dp
  t_lf(1435)=  9000.453840627300213902082226111385686288_dp
  t_lf(1436)=  9007.723457376908383393964783239906719402_dp
  t_lf(1437)=  9014.993770262987408320476706308228861720_dp
  t_lf(1438)=  9022.264778801268400318420424491573731790_dp
  t_lf(1439)=  9029.536482508155768135353725057374011764_dp
  t_lf(1440)=  9036.808880900725814439831135471736133653_dp
  t_lf(1441)=  9044.081973496725336528042503060870788846_dp
  t_lf(1442)=  9051.355759814570230913333690240825585964_dp
  t_lf(1443)=  9058.630239373344101785150518590344877438_dp
  t_lf(1444)=  9065.905411692796873324003037697036937648_dp
  t_lf(1445)=  9073.181276293343405859102866266009378359_dp
  t_lf(1446)=  9080.457832696062115855381754943100896367_dp
  t_lf(1447)=  9087.735080422693599716654654162817007806_dp
  t_lf(1448)=  9095.013018995639261391745437565824546437_dp
  t_lf(1449)=  9102.291647937959943770448033614922973695_dp
  t_lf(1450)=  9109.570966773374563856250056435195437147_dp
  t_lf(1451)=  9116.850975026258751702800103067874413391_dp
  t_lf(1452)=  9124.131672221643493101153699703643550831_dp
  t_lf(1453)=  9131.413057885213776004886435485991417453_dp
  t_lf(1454)=  9138.695131543307240680216120576292861911_dp
  t_lf(1455)=  9145.977892722912833568328846768145536969_dp
  t_lf(1456)=  9153.261340951669464847156615438986741575_dp
  t_lf(1457)=  9160.545475757864669679906730433301272495_dp
  t_lf(1458)=  9167.830296670433273137695433976294603291_dp
  t_lf(1459)=  9175.115803218956058783690293303644917253_dp
  t_lf(1460)=  9182.401994933658440906217625737226000600_dp
  t_lf(1461)=  9189.688871345409140388342781805399071810_dp
  t_lf(1462)=  9196.976431985718864201482391058076221949_dp
  t_lf(1463)=  9204.264676386738988510658714811372384516_dp
  t_lf(1464)=  9211.553604081260245379057045516095054671_dp
  t_lf(1465)=  9218.843214602711413059597645112133740501_dp
  t_lf(1466)=  9226.133507485158009861284025932377015375_dp
  t_lf(1467)=  9233.424482263300991578139448772342816661_dp
  t_lf(1468)=  9240.716138472475452468593344954417598963_dp
  t_lf(1469)=  9248.008475648649329773228963889598939240_dp
  t_lf(1470)=  9255.301493328422111758852906068092188611_dp
  t_lf(1471)=  9262.595191049023549276896324878278087454_dp
  t_lf(1472)=  9269.889568348312370824206470438837342981_dp
  t_lf(1473)=  9277.184624764775001094335906000774953339_dp
  t_lf(1474)=  9284.480359837524283007485153696561981957_dp
  t_lf(1475)=  9291.776773106298203207302722736734054860_dp
  t_lf(1476)=  9299.073864111458621012794440826532866994_dp
  t_lf(1477)=  9306.371632393990000813641749835430326842_dp
  t_lf(1478)=  9313.670077495498147897276140831969041745_dp
  t_lf(1479)=  9320.969198958208947696104192719126605591_dp
  t_lf(1480)=  9328.268996324967108443324744087758770098_dp
  t_lf(1481)=  9335.569469139234907225826570756597738838_dp
  t_lf(1482)=  9342.870616945090939422701562988431613253_dp
  t_lf(1483)=  9350.172439287228871517954797757829494172_dp
  t_lf(1484)=  9357.474935710956197276039083883218084809_dp
  t_lf(1485)=  9364.778105762192997268887522505176978862_dp
  t_lf(1486)=  9372.081948987470701743163373466269668314_dp
  t_lf(1487)=  9379.386464933930856816492050791233669457_dp
  t_lf(1488)=  9386.691653149323893991485388838531258645_dp
  t_lf(1489)=  9393.997513182007902976413425946721944991_dp
  t_lf(1490)=  9401.304044580947407801423845676510429686_dp
  t_lf(1491)=  9408.611246895712146219253898189389169217_dp
  t_lf(1492)=  9415.919119676475852379424097037407186491_dp
  t_lf(1493)=  9423.227662474015042764947250789811527689_dp
  t_lf(1494)=  9430.536874839707805380630445608406126119_dp
  t_lf(1495)=  9437.846756325532592182091445215007828482_dp
  t_lf(1496)=  9445.157306484067014734654619775220614022_dp
  t_lf(1497)=  9452.468524868486643091334956150127532749_dp
  t_lf(1498)=  9459.780411032563807879161939832052166049_dp
  t_lf(1499)=  9467.092964530666405583138134766353660317_dp
  t_lf(1500)=  9474.406184917756707017170122245871102076_dp
  t_lf(1501)=  9481.720071749390168971352095219240366535_dp
  t_lf(1502)=  9489.034624581714249025024840743565477032_dp
  t_lf(1503)=  9496.349842971467223515075081994155307807_dp
  t_lf(1504)=  9503.665726475977008648982193271218364356_dp
  t_lf(1505)=  9510.982274653159984752161147861233578761_dp
  t_lf(1506)=  9518.299487061519823639192210458622534275_dp
  t_lf(1507)=  9525.617363260146319098569344164567528965_dp
  t_lf(1508)=  9532.935902808714220480640567881411112312_dp
  t_lf(1509)=  9540.255105267482069378454574233969248048_dp
  t_lf(1510)=  9547.574970197291039391268801988138088546_dp
  t_lf(1511)=  9554.895497159563778960514851311172756974_dp
  t_lf(1512)=  9562.216685716303257268057636129758043522_dp
  t_lf(1513)=  9569.538535430091613186624986288301019277_dp
  t_lf(1514)=  9576.861045864089007272324544181652137397_dp
  t_lf(1515)=  9584.184216582032476789204747018713839898_dp
  t_lf(1516)=  9591.508047148234793755856447845283820942_dp
  t_lf(1517)=  9598.832537127583326004091306889352675694_dp
  t_lf(1518)=  9606.157686085538901239772480657508703385_dp
  t_lf(1519)=  9613.483493588134674095912350468937420055_dp
  t_lf(1520)=  9620.809959201974996168191065719884167504_dp
  t_lf(1521)=  9628.137082494234289023088531076859856123_dp
  t_lf(1522)=  9635.464863032655920168861141946174802310_dp
  t_lf(1523)=  9642.793300385551081979633069899860872116_dp
  t_lf(1524)=  9650.122394121797673562910220187416043352_dp
  t_lf(1525)=  9657.452143810839185560863127957294033377_dp
  t_lf(1526)=  9664.782549022683587875763029274404591326_dp
  t_lf(1527)=  9672.113609327902220309993138367386875679_dp
  t_lf(1528)=  9679.445324297628686111094784683963772530_dp
  t_lf(1529)=  9686.777693503557748412345513180804445612_dp
  t_lf(1530)=  9694.110716517944229559403529727206987764_dp
  t_lf(1531)=  9701.444392913601913313589981455454512140_dp
  t_lf(1532)=  9708.778722263902449922417500235523335529_dp
  t_lf(1533)=  9716.113704142774264048010207073327583534_dp
  t_lf(1534)=  9723.449338124701465544096977010070776331_dp
  t_lf(1535)=  9730.785623784722763072296198910579456254_dp
  t_lf(1536)=  9738.122560698430380548446533240625571072_dp
  t_lf(1537)=  9745.460148441968976409774274412016540447_dp
  t_lf(1538)=  9752.798386592034565693723863379395794661_dp
  t_lf(1539)=  9760.137274725873444919313871758973751017_dp
  t_lf(1540)=  9767.476812421281119761916391656524600822_dp
  t_lf(1541)=  9774.816999256601235512393216484698519691_dp
  t_lf(1542)=  9782.157834810724510311557488157840477415_dp
  t_lf(1543)=  9789.499318663087671150964616011004066198_dp
  t_lf(1544)=  9796.841450393672392631071243428761654667_dp
  t_lf(1545)=  9804.184229583004238467835850313963808831_dp
  t_lf(1546)=  9811.527655812151605738869233997208827891_dp
  t_lf(1547)=  9818.871728662724671860277608800090452761_dp
  t_lf(1548)=  9826.216447716874344285375406030197704342_dp
  t_lf(1549)=  9833.561812557291212916479042509531771960_dp
  t_lf(1550)=  9840.907822767204505221026957621984681018_dp
  t_lf(1551)=  9848.254477930381044043305097106644509731_dp
  t_lf(1552)=  9855.601777631124208103090747214181177717_dp
  t_lf(1553)=  9862.949721454272895172561196171061640465_dp
  t_lf(1554)=  9870.298308985200487922847121943916969631_dp
  t_lf(1555)=  9877.647539809813822431643876842575839547_dp
  t_lf(1556)=  9884.997413514552159343326961319124365334_dp
  t_lf(1557)=  9892.347929686386157673050952181410408212_dp
  t_lf(1558)=  9899.699087912816851246343975107793845906_dp
  t_lf(1559)=  9907.050887781874627765742488586347007304_dp
  t_lf(1560)=  9914.403328882118210496043676962435568957_dp
  t_lf(1561)=  9921.756410802633642559785134915601888713_dp
  t_lf(1562)=  9929.110133133033273834593765147538228678_dp
  t_lf(1563)=  9936.464495463454750444077906090978590846_dp
  t_lf(1564)=  9943.819497384560006833968657783578266636_dp
  t_lf(1565)=  9951.175138487534260425248182426061664630_dp
  t_lf(1566)=  9958.531418364085008836034422290649329917_dp
  t_lf(1567)=  9965.888336606441029664023202290374892515_dp
  t_lf(1568)=  9973.245892807351382821320068384552134814_dp
  t_lf(1569)=  9980.604086560084415413525456798387755813_dp
  t_lf(1570)=  9987.962917458426769154967893489475595269_dp
  t_lf(1571)=  9995.322385096682390312010889110480699459_dp
  t_lf(1572)=  10002.68248906967154216639002260148910419_dp
  t_lf(1573)=  10010.04322897272981999056739719898468140_dp
  t_lf(1574)=  10017.40460440170716852712120676892727376_dp
  t_lf(1575)=  10024.76661495296690196421856865266788018_dp
  t_lf(1576)=  10032.12926022338472639925006234620524085_dp
  t_lf(1577)=  10039.49253981034776478273456200138446485_dp
  t_lf(1578)=  10046.85645331175358433463296562397023921_dp
  t_lf(1579)=  10054.22100032600922642523930562611457410_dp
  t_lf(1580)=  10061.58618045203023891284747474373852580_dp
  t_lf(1581)=  10068.95199328923971093042141892307653525_dp
  t_lf(1582)=  10076.31843843756731011352613528159391322_dp
  t_lf(1583)=  10083.68551549744832226180616931939504561_dp
  t_lf(1584)=  10091.05322406982269342632753185703793259_dp
  t_lf(1585)=  10098.42156375613407441512805335956273154_dp
  t_lf(1586)=  10105.79053415832886770935016202601184146_dp
  t_lf(1587)=  10113.16013487885527678235891292655475940_dp
  t_lf(1588)=  10120.53036552066235781427680919964797182_dp
  t_lf(1589)=  10127.90122568719907379439554351992464860_dp
  t_lf(1590)=  10135.27271498241335100395324935056059628_dp
  t_lf(1591)=  10142.64483301075113787179418753493789048_dp
  t_lf(1592)=  10150.01757937715546619545600519118095738_dp
  t_lf(1593)=  10157.39095368706551472025779127566823212_dp
  t_lf(1594)=  10164.76495554641567506899011720048634726_dp
  t_lf(1595)=  10172.13958456163462001483609214403957581_dp
  t_lf(1596)=  10179.51484033964437409018018179920948792_dp
  t_lf(1597)=  10186.89072248785938652398913687166291222_dp
  t_lf(1598)=  10194.26723061418560650047685428076638904_dp
  t_lf(1599)=  10201.64436432701956073179235033229164093_dp
  t_lf(1600)=  10209.02212323524743333749726172749307562_dp
  t_lf(1601)=  10216.40050694824414802362640774562479498_dp
  t_lf(1602)=  10223.77951507587245255415194588159787218_dp
  t_lf(1603)=  10231.15914722848200550769853422897366022_dp
  t_lf(1604)=  10238.53940301690846531238367755923936309_dp
  t_lf(1605)=  10245.92028205247258155168408094641135344_dp
  t_lf(1606)=  10253.30178394697928853425536550327320188_dp
  t_lf(1607)=  10260.68390831271680112065891591053307400_dp
  t_lf(1608)=  10268.06665476245571279997592950819385216_dp
  t_lf(1609)=  10275.45002290944809600931492235156480252_dp
  t_lf(1610)=  10282.83401236742660468924501938150254257_dp
  t_lf(1611)=  10290.21862275060357906821331428536899616_dp
  t_lf(1612)=  10297.60385367367015266903043029439302511_dp
  t_lf(1613)=  10304.98970475179536153053414663403871226_dp
  t_lf(1614)=  10312.37617560062525563756657717291465165_dp
  t_lf(1615)=  10319.76326583628201255242589855590181936_dp
  t_lf(1616)=  10327.15097507536305324097902530864751479_dp
  t_lf(1617)=  10334.53930293494016008664691961042152925_dp
  t_lf(1618)=  10341.92824903255859708549940419456629100_dp
  t_lf(1619)=  10349.31781298623623221572141869137835851_dp
  t_lf(1620)=  10356.70799441446266197473762321521100210_dp
  t_lf(1621)=  10364.09879293619833807730710865087560371_dp
  t_lf(1622)=  10371.49020817087369630792472144606268076_dp
  t_lf(1623)=  10378.88223973838828752089015229556985660_dp
  t_lf(1624)=  10386.27488725910991078143047343574947764_dp
  t_lf(1625)=  10393.66815035387374864128623887699335917_dp
  t_lf(1626)=  10401.06202864398150454219558630857992348_dp
  t_lf(1627)=  10408.45652175120054234073499913126423629_dp
  t_lf(1628)=  10415.85162929776302794799950262317716589_dp
  t_lf(1629)=  10423.24735090636507307762908013565522623_dp
  t_lf(1630)=  10430.64368620016588109571200395646034031_dp
  t_lf(1631)=  10438.04063480278689496611958157457240752_dp
  t_lf(1632)=  10445.43819633831094728485052203665894260_dp
  t_lf(1633)=  10452.83637043128141239698672940098217282_dp
  t_lf(1634)=  10460.23515670670136058988583146767404996_dp
  t_lf(1635)=  10467.63455479003271435625915249003196495_dp
  t_lf(1636)=  10475.03456430719540672080713894207670418_dp
  t_lf(1637)=  10482.43518488456654162410744812267601497_dp
  t_lf(1638)=  10489.83641614897955635747401090298774129_dp
  t_lf(1639)=  10497.23825772772338604252838275605699107_dp
  t_lf(1640)=  10504.64070924854163014924760182669511601_dp
  t_lf(1641)=  10512.04377033963172104627557968521530246_dp
  t_lf(1642)=  10519.44744062964409457730776003651524033_dp
  t_lf(1643)=  10526.85171974768136265738139350008665531_dp
  t_lf(1644)=  10534.25660732329748788292629310790539954_dp
  t_lf(1645)=  10541.66210298649696014945335585434694062_dp
  t_lf(1646)=  10549.06820636773397527078046094125049533_dp
  t_lf(1647)=  10556.47491709791161559371758575544368719_dp
  t_lf(1648)=  10563.88223480838103260215511655632983428_dp
  t_lf(1649)=  10571.29015913094063150452137279590692253_dp
  t_lf(1650)=  10578.69868969783525779859731239870512939_dp
  t_lf(1651)=  10586.10782614175538580769824064798674038_dp
  t_lf(1652)=  10593.51756809583630918225410800806597087_dp
  t_lf(1653)=  10600.92791519365733336084165270923826172_dp
  t_lf(1654)=  10608.33886706924096998474322267757977873_dp
  t_lf(1655)=  10615.75042335705213326012859885038260317_dp
  t_lf(1656)=  10623.16258369199733826197753852041238041_dp
  t_lf(1657)=  10630.57534770942390117388206353729890867_dp
  t_lf(1658)=  10637.98871504511914145788873439859966722_dp
  t_lf(1659)=  10645.40268533530958594856227792044678358_dp
  t_lf(1660)=  10652.81725821666017486547297371988069481_dp
  t_lf(1661)=  10660.23243332627346973833115359733030040_dp
  t_lf(1662)=  10667.64821030168886323901302850523561351_dp
  t_lf(1663)=  10675.06458878088179091474283055222900665_dp
  t_lf(1664)=  10682.48156840226294481671694284400156124_dp
  t_lf(1665)=  10689.89914880467748901847628832210424745_dp
  t_lf(1666)=  10697.31732962740427701835376054832191001_dp
  t_lf(1667)=  10704.73611051015507102034390501050691605_dp
  t_lf(1668)=  10712.15549109307376308776239940922133868_dp
  t_lf(1669)=  10719.57547101673559816408313593433455623_dp
  t_lf(1670)=  10726.99604992214639895536087816576369980_dp
  t_lf(1671)=  10734.41722745074179266866755033953100727_dp
  t_lf(1672)=  10741.83900324438643960099021771375851981_dp
  t_lf(1673)=  10749.26137694537326357305873405146360864_dp
  t_lf(1674)=  10756.68434819642268420259086620823171953_dp
  t_lf(1675)=  10764.10791664068185101146245687104781756_dp
  t_lf(1676)=  10771.53208192172387936132985503564966109_dp
  t_lf(1677)=  10778.95684368354708821225143022748335243_dp
  t_lf(1678)=  10786.38220157057423969887449115734115403_dp
  t_lf(1679)=  10793.80815522765178051877335284658989272_dp
  t_lf(1680)=  10801.23470430004908512754363864601449207_dp
  t_lf(1681)=  10808.66184843345770073527716539208930724_dp
  t_lf(1682)=  10816.08958727399059409906094157827130702_dp
  t_lf(1683)=  10823.51792046818140010616291024795461427_dp
  t_lf(1684)=  10830.94684766298367214258609071926806488_dp
  t_lf(1685)=  10838.37636850577013424169171660714268373_dp
  t_lf(1686)=  10845.80648264433193500761083229021655794_dp
  t_lf(1687)=  10853.23718972687790330818259635237055341_dp
  t_lf(1688)=  10860.66848940203380573217624898219936323_dp
  t_lf(1689)=  10868.10038131884160580557233120773962992_dp
  t_lf(1690)=  10875.53286512675872496169729754555463133_dp
  t_lf(1691)=  10882.96594047565730526002414051811233409_dp
  t_lf(1692)=  10890.39960701582347384847004590464591235_dp
  t_lf(1693)=  10897.83386439795660916404042189977634919_dp
  t_lf(1694)=  10905.26871227316860886668689392060796409_dp
  t_lf(1695)=  10912.70415029298315950126502998437269910_dp
  t_lf(1696)=  10920.14017810933500788249565973069263971_dp
  t_lf(1697)=  10927.57679537456923419785167363896045785_dp
  t_lf(1698)=  10935.01400174144052682331013814413630033_dp
  t_lf(1699)=  10942.45179686311245884692743753349434429_dp
  t_lf(1700)=  10949.89018039315676629521295506073619925_dp
  t_lf(1701)=  10957.32915198555262805729453398879200750_dp
  t_lf(1702)=  10964.76871129468594750188661461309182735_dp
  t_lf(1703)=  10972.20885797534863578208952606582667185_dp
  t_lf(1704)=  10979.64959168273789682306592219963676135_dp
  t_lf(1705)=  10987.09091207245551398765778943537043550_dp
  t_lf(1706)=  10994.53281880050713841502482147036758619_dp
  t_lf(1707)=  11001.97531152330157902740225151667425037_dp
  t_lf(1708)=  11009.41838989765009420009345760645939798_dp
  t_lf(1709)=  11016.86205358076568508982981079669200084_dp
  t_lf(1710)=  11024.30630223026239061664732015710926999_dp
  t_lf(1711)=  11031.75113550415458409444664256319087922_dp
  t_lf(1712)=  11039.19655306085627150541996986604686249_dp
  t_lf(1713)=  11046.64255455918039141354518129890901736_dp
  t_lf(1714)=  11054.08913965833811651236445532865994252_dp
  t_lf(1715)=  11061.53630801793815680228127289221566934_dp
  t_lf(1716)=  11068.98405929798606439262641339158499609_dp
  t_lf(1717)=  11076.43239315888353992376014627637111178_dp
  t_lf(1718)=  11083.88130926142774060449435483500091792_dp
  t_lf(1719)=  11091.33080726681058986013479526104833648_dp
  t_lf(1720)=  11098.78088683661808858646009347199490086_dp
  t_lf(1721)=  11106.23154763282962800497041484634006131_dp
  t_lf(1722)=  11113.68278931781730411475500832120135090_dp
  t_lf(1723)=  11121.13461155434523373634402646487721853_dp
  t_lf(1724)=  11128.58701400556887214292615751411862192_dp
  t_lf(1725)=  11136.03999633503433227432967424930371902_dp
  t_lf(1726)=  11143.49355820667770552918050827597546862_dp
  t_lf(1727)=  11150.94769928482438413066689709034407317_dp
  t_lf(1728)=  11158.40241923418838506135602552986070969_dp
  t_lf(1729)=  11165.85771771987167556252389314675706156_dp
  t_lf(1730)=  11173.31359440736350019347538498988241724_dp
  t_lf(1731)=  11180.77004896253970944634720553407563135_dp
  t_lf(1732)=  11188.22708105166208991190195435096452845_dp
  t_lf(1733)=  11195.68469034137769599183717786324311415_dp
  t_lf(1734)=  11203.14287649871818315314872445737006949_dp
  t_lf(1735)=  11210.60163919109914272010316063698211666_dp
  t_lf(1736)=  11218.06097808631943819938937406934218331_dp
  t_lf(1737)=  11225.52089285256054313403479559657029348_dp
  t_lf(1738)=  11232.98138315838588048168691683747501027_dp
  t_lf(1739)=  11240.44244867274016351287596317827863114_dp
  t_lf(1740)=  11247.90408906494873822488970402370560923_dp
  t_lf(1741)=  11255.36630400471692726690644343461652942_dp
  t_lf(1742)=  11262.82909316212937537204723499400374735_dp
  t_lf(1743)=  11270.29245620764939629202330519766082324_dp
  t_lf(1744)=  11277.75639281211832123006955013570273116_dp
  t_lf(1745)=  11285.22090264675484876786979099142755851_dp
  t_lf(1746)=  11292.68598538315439628219423520843474820_dp
  t_lf(1747)=  11300.15164069328845284698429233770136959_dp
  t_lf(1748)=  11307.61786824950393361663453684431577238_dp
  t_lf(1749)=  11315.08466772452253568623619479823248515_dp
  t_lf(1750)=  11322.55203879144009542456105766281240437_dp
  t_lf(1751)=  11330.01998112372594727557919459573897173_dp
  t_lf(1752)=  11337.48849439522228402431824505447456971_dp
  t_lf(1753)=  11344.95757828014351852288642631669963055_dp
  t_lf(1754)=  11352.42723245307564687249568604776301785_dp
  t_lf(1755)=  11359.89745658897561305733566853332210121_dp
  t_lf(1756)=  11367.36825036317067502616334490598120804_dp
  t_lf(1757)=  11374.83961345135777221748728288843264101_dp
  t_lf(1758)=  11382.31154552960289452423960050962584148_dp
  t_lf(1759)=  11389.78404627434045269384266118078154724_dp
  t_lf(1760)=  11397.25711536237265015959152469926374702_dp
  t_lf(1761)=  11404.73075247086885629928707043375594540_dp
  t_lf(1762)=  11412.20495727736498111706855538589844040_dp
  t_lf(1763)=  11419.67972945976285134440816127228169502_dp
  t_lf(1764)=  11427.15506869632958795624382147592945903_dp
  t_lf(1765)=  11434.63097466569698509824030092634361667_dp
  t_lf(1766)=  11442.10744704686089042118212992875510079_dp
  t_lf(1767)=  11449.58448551918058681851556692210783724_dp
  t_lf(1768)=  11457.06208976237817556307028534592081209_dp
  t_lf(1769)=  11464.54025945653796083900494648170833844_dp
  t_lf(1770)=  11472.01899428210583566503423354703492598_dp
  t_lf(1771)=  11479.49829391988866920500828270025343112_dp
  t_lf(1772)=  11486.97815805105369546192875420101505996_dp
  t_lf(1773)=  11494.45858635712790335149904200402294235_dp
  t_lf(1774)=  11501.93957851999742815131932277829384767_dp
  t_lf(1775)=  11509.42113422190694432185029597725925705_dp
  t_lf(1776)=  11516.90325314545905969528256537104593619_dp
  t_lf(1777)=  11524.38593497361371102846165962470699688_dp
  t_lf(1778)=  11531.86917938968756091603168529732464558_dp
  t_lf(1779)=  11539.35298607735339605997355027789731992_dp
  t_lf(1780)=  11546.83735472063952689172658939470970991_dp
  t_lf(1781)=  11554.32228500392918854309526696425489391_dp
  t_lf(1782)=  11561.80777661195994316215542361136830260_dp
  t_lf(1783)=  11569.29382922982308357038727702052891908_dp
  t_lf(1784)=  11576.78044254296303825727507859462658881_dp
  t_lf(1785)=  11584.26761623717677770862597052609160842_dp
  t_lf(1786)=  11591.75534999861322206487318074920981224_dp
  t_lf(1787)=  11599.24364351377265010564123686365529501_dp
  t_lf(1788)=  11606.73249646950610955686337461859829434_dp
  t_lf(1789)=  11614.22190855301482871675376214391092751_dp
  t_lf(1790)=  11621.71187945184962939694955802861669389_dp
  t_lf(1791)=  11629.20240885391034117515016979433028250_dp
  t_lf(1792)=  11636.69349644744521695559337950943887478_dp
  t_lf(1793)=  11644.18514192105034983372025545352475395_dp
  t_lf(1794)=  11651.67734496366909126139297308528097095_dp
  t_lf(1795)=  11659.17010526459147050904182530411183857_dp
  t_lf(1796)=  11666.66342251345361542112981133785430171_dp
  t_lf(1797)=  11674.15729640023717446133525574765828410_dp
  t_lf(1798)=  11681.65172661526874004386492422602072948_dp
  t_lf(1799)=  11689.14671284921927314732207128422516466_dp
  t_lf(1800)=  11696.64225479310352920756577678889712105_dp
  t_lf(1801)=  11704.13835213827948528600980382090735130_dp
  t_lf(1802)=  11711.63500457644776850982103969926771217_dp
  t_lf(1803)=  11719.13221179965108578048936544276938401_dp
  t_lf(1804)=  11726.62997350027365474725253663668827403_dp
  t_lf(1805)=  11734.12828937104063604187135083369016875_dp
  t_lf(1806)=  11741.62715910501756677126202344885989778_dp
  t_lf(1807)=  11749.12658239560979526450429580930076017_dp
  t_lf(1808)=  11756.62655893656191707075535578874948812_dp
  t_lf(1809)=  11764.12708842195721220461116349588078798_dp
  t_lf(1810)=  11771.62817054621708363546824198919816111_dp
  t_lf(1811)=  11779.12980500410049701745041615840534057_dp
  t_lf(1812)=  11786.63199149070342165647636193772869570_dp
  t_lf(1813)=  11794.13472970145827271105516309564650232_dp
  t_lf(1814)=  11801.63801933213335462340836417174029378_dp
  t_lf(1815)=  11809.14186007883230577752825589781926572_dp
  t_lf(1816)=  11816.64625163799354438079333383902729247_dp
  t_lf(1817)=  11824.15119370638971556577303221231889951_dp
  t_lf(1818)=  11831.65668598112713970886495307453511665_dp
  t_lf(1819)=  11839.16272815964526196241888650943152018_dp
  t_lf(1820)=  11846.66931993971610299701295027058255929_dp
  t_lf(1821)=  11854.17646101944371095055816774235177677_dp
  t_lf(1822)=  11861.68415109726361458091875125039493714_dp
  t_lf(1823)=  11869.19238987194227761874626387185043485_dp
  t_lf(1824)=  11876.70117704257655431723669714795169694_dp
  t_lf(1825)=  11884.21051230859314619553032467184261479_dp
  t_lf(1826)=  11891.72039536974805997248497259455729111_dp
  t_lf(1827)=  11899.23082592612606668756408784420743385_dp
  t_lf(1828)=  11906.74180367814016200559168346827727052_dp
  t_lf(1829)=  11914.25332832653102770213689816653932536_dp
  t_lf(1830)=  11921.76539957236649432630152396157183002_dp
  t_lf(1831)=  11929.27801711704100503769443223339639160_dp
  t_lf(1832)=  11936.79118066227508061438736420170352962_dp
  t_lf(1833)=  11944.30488991011478562865704754996354764_dp
  t_lf(1834)=  11951.81914456293119578732905642603442983_dp
  t_lf(1835)=  11959.33394432341986643354924769841904663_dp
  t_lf(1836)=  11966.84928889460030220681898226997610342_dp
  t_lf(1837)=  11974.36517797981542785814067662468601209_dp
  t_lf(1838)=  11981.88161128273106021713052678019973901_dp
  t_lf(1839)=  11989.39858850733538130796550461069831565_dp
  t_lf(1840)=  11996.91610935793841261104194526156740568_dp
  t_lf(1841)=  12004.43417353917149046723322426921565750_dp
  t_lf(1842)=  12011.95278075598674262164416419487885206_dp
  t_lf(1843)=  12019.47193071365656590376991324847066731_dp
  t_lf(1844)=  12026.99162311777310504097710268466154698_dp
  t_lf(1845)=  12034.51185767424773260223511586477019362_dp
  t_lf(1846)=  12042.03263408931053006903528996030672289_dp
  t_lf(1847)=  12049.55395206950977003044582149187463833_dp
  t_lf(1848)=  12057.07581132171139949926005941458000277_dp
  t_lf(1849)=  12064.59821155309852434620574444127169684_dp
  t_lf(1850)=  12072.12115247117089484919259090021369585_dp
  t_lf(1851)=  12079.64463378374439235458540781574641238_dp
  t_lf(1852)=  12087.16865519895051704749971923992147546_dp
  t_lf(1853)=  12094.69321642523587682812657031000380364_dp
  t_lf(1854)=  12102.21831717136167729110289522036047244_dp
  t_lf(1855)=  12109.74395714640321280495347643605889758_dp
  t_lf(1856)=  12117.27013605974935868864014119716986859_dp
  t_lf(1857)=  12124.79685362110206448226342182424211803_dp
  t_lf(1858)=  12132.32410954047584830897145069285219502_dp
  t_lf(1859)=  12139.85190352819729232514036915394796153_dp
  t_lf(1860)=  12147.38023529490453925590000229155538522_dp
  t_lf(1861)=  12154.90910455154679001308798838421765428_dp
  t_lf(1862)=  12162.43851100938380239272495342445004385_dp
  t_lf(1863)=  12169.96845437998539084911268720395034278_dp
  t_lf(1864)=  12177.49893437523092734266660844299375997_dp
  t_lf(1865)=  12185.02995070730884325860310238132161174_dp
  t_lf(1866)=  12192.56150308871613239361157530513499629_dp
  t_lf(1867)=  12200.09359123225785500765029681003394626_dp
  t_lf(1868)=  12207.62621485104664293801429234168456839_dp
  t_lf(1869)=  12215.15937365850220577283270586272012304_dp
  t_lf(1870)=  12222.69306736835083808116217551324274308_dp
  t_lf(1871)=  12230.22729569462492769685185400993740402_dp
  t_lf(1872)=  12237.76205835166246505336476041118048031_dp
  t_lf(1873)=  12245.29735505410655356674917090786194294_dp
  t_lf(1874)=  12252.83318551690492106396274362648544825_dp
  t_lf(1875)=  12260.36954945530943225376102619631272451_dp
  t_lf(1876)=  12267.90644658487560223737091517903598012_dp
  t_lf(1877)=  12275.44387662146211105617852353016612360_dp
  t_lf(1878)=  12282.98183928123031927366976619780403623_dp
  t_lf(1879)=  12290.52033428064378458887079490782629116_dp
  t_lf(1880)=  12298.05936133646777947854420127519918221_dp
  t_lf(1881)=  12305.59892016576880986540566275889721645_dp
  t_lf(1882)=  12313.13901048591413480963442878183730331_dp
  t_lf(1883)=  12320.67963201457128722095973470577572016_dp
  t_lf(1884)=  12328.22078446970759558861388942201807425_dp
  t_lf(1885)=  12335.76246756958970672645140822917149210_dp
  t_lf(1886)=  12343.30468103278310953054215655547727238_dp
  t_lf(1887)=  12350.84742457815165974655503208130437178_dp
  t_lf(1888)=  12358.39069792485710574425724306231495224_dp
  t_lf(1889)=  12365.93450079235861529646273928014428904_dp
  t_lf(1890)=  12373.47883290041230335977181918903941009_dp
  t_lf(1891)=  12381.02369396907076085445237261700738205_dp
  t_lf(1892)=  12388.56908371868258444082162295123471152_dp
  t_lf(1893)=  12396.11500186989190728949560622182283131_dp
  t_lf(1894)=  12403.66144814363793084288196702658276988_dp
  t_lf(1895)=  12411.20842226115445756529996294346258543_dp
  t_lf(1896)=  12418.75592394396942467911985008624105176_dp
  t_lf(1897)=  12426.30395291390443888432207290289009353_dp
  t_lf(1898)=  12433.85250889307431205888490132335213127_dp
  t_lf(1899)=  12441.40159160388659793741734806265146280_dp
  t_lf(1900)=  12448.95120076904112976546235840390804475_dp
  t_lf(1901)=  12456.50133611152955892690339425099003030_dp
  t_lf(1902)=  12464.05199735463489454191563377866934388_dp
  t_lf(1903)=  12471.60318422193104403291107774507546465_dp
  t_lf(1904)=  12479.15489643728235465593489359222447719_dp
  t_lf(1905)=  12486.70713372484315599497833897008948293_dp
  t_lf(1906)=  12494.25989580905730341668158740214202817_dp
  t_lf(1907)=  12501.81318241465772248290773058901399007_dp
  t_lf(1908)=  12509.36699326666595431867715444480445239_dp
  t_lf(1909)=  12516.92132809039170193295937949990601896_dp
  t_lf(1910)=  12524.47618661143237748982732090679275031_dp
  t_lf(1911)=  12532.03156855567265052748675907216695408_dp
  t_lf(1912)=  12539.58747364928399712270161903080339288_dp
  t_lf(1913)=  12547.14390161872424999814343519338854819_dp
  t_lf(1914)=  12554.70085219073714957020112816209629137_dp
  t_lf(1915)=  12562.25832509235189593479494203247494926_dp
  t_lf(1916)=  12569.81632005088270178874608410679091901_dp
  t_lf(1917)=  12577.37483679392834628426127435007153022_dp
  t_lf(1918)=  12584.93387504937172981409904934295275191_dp
  t_lf(1919)=  12592.49343454537942972499227504175359460_dp
  t_lf(1920)=  12600.05351501040125695690890446210954392_dp
  t_lf(1921)=  12607.61411617316981360574057057359869213_dp
  t_lf(1922)=  12615.17523776270005140701613134414158312_dp
  t_lf(1923)=  12622.73687950828883113824478311906570901_dp
  t_lf(1924)=  12630.29904113951448293750083047457882791_dp
  t_lf(1925)=  12637.86172238623636753586964546243951222_dp
  t_lf(1926)=  12645.42492297859443840138176687476601720_dp
  t_lf(1927)=  12652.98864264700880479206948091757559850_dp
  t_lf(1928)=  12660.55288112217929571578758860066094394_dp
  t_lf(1929)=  12668.11763813508502479444740234113150273_dp
  t_lf(1930)=  12675.68291341698395603032032484919892570_dp
  t_lf(1931)=  12683.24870669941247047207464742787464431_dp
  t_lf(1932)=  12690.81501771418493377821646248296689900_dp
  t_lf(1933)=  12698.38184619339326467561281641539603771_dp
  t_lf(1934)=  12705.94919186940650431078243426316372000_dp
  t_lf(1935)=  12713.51705447487038649164652658358080608_dp
  t_lf(1936)=  12721.08543374270690881743934222534377096_dp
  t_lf(1937)=  12728.65432940611390469448525794201321005_dp
  t_lf(1938)=  12736.22374119856461623555629735015489239_dp
  t_lf(1939)=  12743.79366885380726804053104764312268296_dp
  t_lf(1940)=  12751.36411210586464185608299284096918545_dp
  t_lf(1941)=  12758.93507068903365211213330729354925157_dp
  t_lf(1942)=  12766.50654433788492233281015276233060556_dp
  t_lf(1943)=  12774.07853278726236241966349679105623992_dp
  t_lf(1944)=  12781.65103577228274680489141934004339815_dp
  t_lf(1945)=  12789.22405302833529347234079890690175844_dp
  t_lf(1946)=  12796.79758429108124384405216869067865852_dp
  t_lf(1947)=  12804.37162929645344353012540787928601115_dp
  t_lf(1948)=  12811.94618778065592393968978295345290979_dp
  t_lf(1949)=  12819.52125948016348475076867910583000691_dp
  t_lf(1950)=  12827.09684413172127723683616257222840306_dp
  t_lf(1951)=  12834.67294147234438844786929096382290290_dp
  t_lf(1952)=  12842.24955123931742624370684067453940049_dp
  t_lf(1953)=  12849.82667317019410517753184821636998884_dp
  t_lf(1954)=  12857.40430700279683322730206600614945652_dp
  t_lf(1955)=  12864.98245247521629937295911278905896681_dp
  t_lf(1956)=  12872.56110932581106201725375463501859553_dp
  t_lf(1957)=  12880.14027729320713824803138438195988988_dp
  t_lf(1958)=  12887.71995611629759393982837562205304497_dp
  t_lf(1959)=  12895.30014553424213469263657193017379576_dp
  t_lf(1960)=  12902.88084528646669760569973311466087256_dp
  t_lf(1961)=  12910.46205511266304388421229792472137750_dp
  t_lf(1962)=  12918.04377475278835227679733697223380712_dp
  t_lf(1963)=  12925.62600394706481334164706071328360202_dp
  t_lf(1964)=  12933.20874243597922453921571528121503981_dp
  t_lf(1965)=  12940.79198996028258614936114386253327905_dp
  t_lf(1966)=  12948.37574626098969801083771325344717005_dp
  t_lf(1967)=  12955.96001107937875708104970432158352172_dp
  t_lf(1968)=  12963.54478415699095581398064141737616129_dp
  t_lf(1969)=  12971.13006523563008135422038942536269276_dp
  t_lf(1970)=  12978.71585405736211554501817820920988794_dp
  t_lf(1971)=  12986.30215036451483574829602277741600763_dp
  t_lf(1972)=  12993.88895389967741647456329367056739891_dp
  t_lf(1973)=  13001.47626440570003182067945593660338917_dp
  t_lf(1974)=  13009.06408162569345871341823670819944488_dp
  t_lf(1975)=  13016.65240530302868095679270091613323108_dp
  t_lf(1976)=  13024.24123518133649408110691215396093293_dp
  t_lf(1977)=  13031.83057100450711099170603124170328274_dp
  t_lf(1978)=  13039.42041251668976841540285870831750246_dp
  t_lf(1979)=  13047.01075946229233414256495931290422758_dp
  t_lf(1980)=  13054.60161158598091506285261694085694906_dp
  t_lf(1981)=  13062.19296863267946599260395683109526900_dp
  t_lf(1982)=  13069.78483034756939929186963920132415234_dp
  t_lf(1983)=  13077.37719647608919526910557402672538532_dp
  t_lf(1984)=  13084.97006676393401337153813108001680194_dp
  t_lf(1985)=  13092.56344095705530415922232244341984886_dp
  t_lf(1986)=  13100.15731880166042206081941664137718794_dp
  t_lf(1987)=  13107.75170004421223890912640440209183564_dp
  t_lf(1988)=  13115.34658443142875825439567592096440260_dp
  t_lf(1989)=  13122.94197171028273045348918845424789915_dp
  t_lf(1990)=  13130.53786162800126853291730120080080056_dp
  t_lf(1991)=  13138.13425393206546482381833181739893878_dp
  t_lf(1992)=  13145.73114837021000836694074564198736465_dp
  t_lf(1993)=  13153.32854469042280308569572485246030548_dp
  t_lf(1994)=  13160.92644264094458672535368044862195456_dp
  t_lf(1995)=  13168.52484197026855055646406519410170117_dp
  t_lf(1996)=  13176.12374242713995984058362057500244733_dp
  t_lf(1997)=  13183.72314376055577505640394550440889080_dp
  t_lf(1998)=  13191.32304571976427388437500900766609165_dp
  t_lf(1999)=  13198.92344805426467394792694354328410057_dp
  t_lf(2000)=  13206.52435051380675630939815002879536976_dp
  t_lf(2001)=  13214.12575284839048971878342012989000319_dp
  t_lf(2002)=  13221.72765480826565561342143601432180739_dp
  t_lf(2003)=  13229.33005614393147386674664264869980677_dp
  t_lf(2004)=  13236.93295660613622928423610290528346498_dp
  t_lf(2005)=  13244.53635594587689884468754132585900235_dp
  t_lf(2006)=  13252.14025391439877968497035843891000308_dp
  t_lf(2007)=  13259.74465026319511782639695412247819451_dp
  t_lf(2008)=  13267.34954474400673764086723572586097746_dp
  t_lf(2009)=  13274.95493710882167205494470458578474451_dp
  t_lf(2010)=  13282.56082710987479349002801327375535717_dp
  t_lf(2011)=  13290.16721449964744553678736546740355450_dp
  t_lf(2012)=  13297.77409903086707536204059082595551767_dp
  t_lf(2013)=  13305.38148045650686684624916874426878742_dp
  t_lf(2014)=  13312.98935852978537444981989743664387529_dp
  t_lf(2015)=  13320.59773300416615780640330853597773875_dp
  t_lf(2016)=  13328.20660363335741704138531236055685273_dp
  t_lf(2017)=  13335.81597017131162881377392527435115309_dp
  t_lf(2018)=  13343.42583237222518307968827822118975038_dp
  t_lf(2019)=  13351.03618999053802057566243462246861429_dp
  t_lf(2020)=  13358.64704278093327101998185646552414422_dp
  t_lf(2021)=  13366.25839049833689203027564964864216496_dp
  t_lf(2022)=  13373.87023289791730875559299356167129845_dp
  t_lf(2023)=  13381.48256973508505422119741554086073125_dp
  t_lf(2024)=  13389.09540076549241038431780831501058637_dp
  t_lf(2025)=  13396.70872574503304989910030792915306447_dp
  t_lf(2026)=  13404.32254442984167858901035096329966120_dp
  t_lf(2027)=  13411.93685657629367862493941322850185555_dp
  t_lf(2028)=  13419.55166194100475240727609759147137160_dp
  t_lf(2029)=  13427.16696028083056715020638622286858533_dp
  t_lf(2030)=  13434.78275135286640016651300245335804088_dp
  t_lf(2031)=  13442.39903491444678485114893962560107861_dp
  t_lf(2032)=  13450.01581072314515736186530891915007727_dp
  t_lf(2033)=  13457.63307853677350399517873516806120781_dp
  t_lf(2034)=  13465.25083811338200925596858925698045746_dp
  t_lf(2035)=  13472.86908921125870461899938783920322157_dp
  t_lf(2036)=  13480.48783158892911798066871593817933335_dp
  t_lf(2037)=  13488.10706500515592379928603554024900050_dp
  t_lf(2038)=  13495.72678921893859392219273362886585255_dp
  t_lf(2039)=  13503.34700398951304909803873631670688170_dp
  t_lf(2040)=  13510.96770907635131117253597186910325129_dp
  t_lf(2041)=  13518.58890423916115596601390454707204514_dp
  t_lf(2042)=  13526.21058923788576683110728339751330623_dp
  t_lf(2043)=  13533.83276383270338888891115544819185469_dp
  t_lf(2044)=  13541.45542778402698394194308129198993013_dp
  t_lf(2045)=  13549.07858085250388606225736283434450385_dp
  t_lf(2046)=  13556.70222279901545785306094809523269263_dp
  t_lf(2047)=  13564.32635338467674738218551646771273797_dp
  t_lf(2048)=  13571.95097237083614578577506980375268022_dp
  t_lf(2049)=  13579.57607951907504554055316018948526788_dp
  t_lf(2050)=  13587.20167459120749940303867435043322732_dp
  t_lf(2051)=  13594.82775734927988001408386735668890524_dp
  t_lf(2052)=  13602.45432755557054016711309474226068902_dp
  t_lf(2053)=  13610.08138497258947373844543237529015672_dp
  t_lf(2054)=  13617.70892936307797727808909747979506282_dp
  t_lf(2055)=  13625.33696049000831225940029217792364155_dp
  t_lf(2056)=  13632.96547811658336798600378285705942671_dp
  t_lf(2057)=  13640.59448200623632515437720463086281184_dp
  t_lf(2058)=  13648.22397192263032007050574021957305331_dp
  t_lf(2059)=  13655.85394762965810951901846678444799901_dp
  t_lf(2060)=  13663.48440889144173628322229267564398060_dp
  t_lf(2061)=  13671.11535547233219531445401875342164033_dp
  t_lf(2062)=  13678.74678713690910054917565598032401931_dp
  t_lf(2063)=  13686.37870364998035237224271241766378932_dp
  t_lf(2064)=  13694.01110477658180572477972865376486833_dp
  t_lf(2065)=  13701.64399028197693885510189110415393333_dp
  t_lf(2066)=  13709.27735993165652271112608761525123019_dp
  t_lf(2067)=  13716.91121349133829097271928943276813227_dp
  t_lf(2068)=  13724.54555072696661072243664792342178841_dp
  t_lf(2069)=  13732.18037140471215375310618352290697400_dp
  t_lf(2070)=  13739.81567529097156851072141828324658574_dp
  t_lf(2071)=  13747.45146215236715267110776216734364093_dp
  t_lf(2072)=  13755.08773175574652634883290694619279752_dp
  t_lf(2073)=  13762.72448386818230593683591025293827841_dp
  t_lf(2074)=  13770.36171825697177857525406609569519626_dp
  t_lf(2075)=  13777.99943468963657724793105698543894199_dp
  t_lf(2076)=  13785.63763293392235650509426685371881230_dp
  t_lf(2077)=  13793.27631275779846881069350317462489226_dp
  t_lf(2078)=  13800.91547392945764151289773122324124426_dp
  t_lf(2079)=  13808.55511621731565443625076325541713040_dp
  t_lf(2080)=  13816.19523939001101809399117063749951949_dp
  t_lf(2081)=  13823.83584321640465251904599764586865131_dp
  t_lf(2082)=  13831.47692746557956671221215185063521312_dp
  t_lf(2083)=  13839.11849190684053870604362775138226926_dp
  t_lf(2084)=  13846.76053630971379624296698770081645886_dp
  t_lf(2085)=  13854.40306044394669806615177718984071599_dp
  t_lf(2086)=  13862.04606407950741582166679032984619277_dp
  t_lf(2087)=  13869.68954698658461657045732590967838286_dp
  t_lf(2088)=  13877.33350893558714590868278478025987558_dp
  t_lf(2089)=  13884.97794969714371169495815558351309503_dp
  t_lf(2090)=  13892.62286904210256838304711804805044207_dp
  t_lf(2091)=  13900.26826674153120195855866127588350190_dp
  t_lf(2092)=  13907.91414256671601547820326869571295033_dp
  t_lf(2093)=  13915.56049628916201521016886171253164478_dp
  t_lf(2094)=  13923.20732768059249737418082059341098676_dp
  t_lf(2095)=  13930.85463651294873547981451384832266479_dp
  t_lf(2096)=  13938.50242255838966826163286634532489696_dp
  t_lf(2097)=  13946.15068558929158820972558169383883585_dp
  t_lf(2098)=  13953.79942537824783069423070609025647615_dp
  t_lf(2099)=  13961.44864169806846368242327889873194499_dp
  t_lf(2100)=  13969.09833432177997804695985978846637884_dp
  t_lf(2101)=  13976.74850302262497846387175331863387529_dp
  t_lf(2102)=  13984.39914757406187489890376950462273910_dp
  t_lf(2103)=  13992.05026774976457468079936316557326538_dp
  t_lf(2104)=  13999.70186332362217516013698579415286718_dp
  t_lf(2105)=  14007.35393406973865695232646135577615229_dp
  t_lf(2106)=  14015.00647976243257776337816186648975028_dp
  t_lf(2107)=  14022.65950017623676679706170986672195678_dp
  t_lf(2108)=  14030.31299508589801974207487305205379366_dp
  t_lf(2109)=  14037.96696426637679433784724139189562006_dp
  t_lf(2110)=  14045.62140749284690651760718911203426439_dp
  t_lf(2111)=  14053.27632454069522712734452298682063763_dp
  t_lf(2112)=  14060.93171518552137921930510453045735204_dp
  t_lf(2113)=  14068.58757920313743591865760694337098429_dp
  t_lf(2114)=  14076.24391636956761886197642810775681688_dp
  t_lf(2115)=  14083.90072646104799720618862858460022964_dp
  t_lf(2116)=  14091.55800925402618720663659849113697501_dp
  t_lf(2117)=  14099.21576452516105236291197938093742919_dp
  t_lf(2118)=  14106.87399205132240413112017685650610147_dp
  t_lf(2119)=  14114.53269160959070320123859666419216995_dp
  t_lf(2120)=  14122.19186297725676133823552150082194506_dp
  t_lf(2121)=  14129.85150593182144378562031774810063965_dp
  t_lf(2122)=  14137.51162025099537223009942089159863677_dp
  t_lf(2123)=  14145.17220571269862832601629552294682483_dp
  t_lf(2124)=  14152.83326209506045777825730061342792700_dp
  t_lf(2125)=  14160.49478917641897498230911323097961646_dp
  t_lf(2126)=  14168.15678673532086822015707409703715876_dp
  t_lf(2127)=  14175.81925455052110541071751639075890756_dp
  t_lf(2128)=  14183.48219240098264041350082505192264710_dp
  t_lf(2129)=  14191.14560006587611988420564755588578210_dp
  t_lf(2130)=  14198.80947732457959068094833878000570610_dp
  t_lf(2131)=  14206.47382395667820781983537219616394538_dp
  t_lf(2132)=  14214.13863974196394297859008725368302471_dp
  t_lf(2133)=  14221.80392446043529354694876850593201275_dp
  t_lf(2134)=  14229.46967789229699222254466582705928033_dp
  t_lf(2135)=  14237.13589981795971715100216700715426245_dp
  t_lf(2136)=  14244.80259001803980260896692414912116708_dp
  t_lf(2137)=  14252.46974827335895022880131366285681250_dp
  t_lf(2138)=  14260.13737436494394076367817630598614150_dp
  t_lf(2139)=  14267.80546807402634639180933869926122053_dp
  t_lf(2140)=  14275.47402918204224355854896109242703831_dp
  t_lf(2141)=  14283.14305747063192635511528791637097247_dp
  t_lf(2142)=  14290.81255272163962043267789787299050671_dp
  t_lf(2143)=  14298.48251471711319745056105902854192146_dp
  t_lf(2144)=  14306.15294323930389005731729163219652705_dp
  t_lf(2145)=  14313.82383807066600740342872722187568830_dp
  t_lf(2146)=  14321.49519899385665118439732704672722363_dp
  t_lf(2147)=  14329.16702579173543221298848597223109884_dp
  t_lf(2148)=  14336.83931824736418751939599988209137986_dp
  t_lf(2149)=  14344.51207614400669797809981519281705187_dp
  t_lf(2150)=  14352.18529926512840646019140849407345075_dp
  t_lf(2151)=  14359.85898739439613650994506256218041126_dp
  t_lf(2152)=  14367.53314031567781154441671210705017808_dp
  t_lf(2153)=  14375.20775781304217457485542864573146431_dp
  t_lf(2154)=  14382.88283967075850844871599888971684656_dp
  t_lf(2155)=  14390.55838567329635661106442502926808445_dp
  t_lf(2156)=  14398.23439560532524438417153833703592634_dp
  t_lf(2157)=  14405.91086925171440076409326963584514555_dp
  t_lf(2158)=  14413.58780639753248073303946142216001118_dp
  t_lf(2159)=  14421.26520682804728808633643684774943011_dp
  t_lf(2160)=  14428.94307032872549877279186037757590113_dp
  t_lf(2161)=  14436.62139668523238474727373380191228235_dp
  t_lf(2162)=  14444.30018568343153833431866942495282875_dp
  t_lf(2163)=  14451.97943710938459710158786972037197500_dp
  t_lf(2164)=  14459.65915074935096924199251957587297828_dp
  t_lf(2165)=  14467.33932638978755946331356348307170989_dp
  t_lf(2166)=  14475.01996381734849538414409570522811923_dp
  t_lf(2167)=  14482.70106281888485443498583661235607951_dp
  t_lf(2168)=  14490.38262318144439126333440304993647125_dp
  t_lf(2169)=  14498.06464469227126564159130484249233839_dp
  t_lf(2170)=  14505.74712713880577087664381336516223954_dp
  t_lf(2171)=  14513.43007030868406271995705158347233751_dp
  t_lf(2172)=  14521.11347398973788877702584810194422528_dp
  t_lf(2173)=  14528.79733796999431841503708061401069064_dp
  t_lf(2174)=  14536.48166203767547316759640674681482906_dp
  t_lf(2175)=  14544.16644598119825763537644268255164165_dp
  t_lf(2176)=  14551.85168958917409088154660215063200416_dp
  t_lf(2177)=  14559.53739265040863832084795045950786617_dp
  t_lf(2178)=  14567.22355495390154410117956021074135275_dp
  t_lf(2179)=  14574.91017628884616397656597724792183474_dp
  t_lf(2180)=  14582.59725644462929867037851727627357716_dp
  t_lf(2181)=  14590.28479521083092772768621548203937076_dp
  t_lf(2182)=  14597.97279237722394385561534342060793140_dp
  t_lf(2183)=  14605.66124773377388775059948946535907619_dp
  t_lf(2184)=  14613.35016107063868341140527125166462994_dp
  t_lf(2185)=  14621.03953217816837393682181084858886723_dp
  t_lf(2186)=  14628.72936084690485780690515588061522794_dp
  t_lf(2187)=  14636.41964686758162564667187253907290787_dp
  t_lf(2188)=  14644.11039003112349747113906940358692176_dp
  t_lf(2189)=  14651.80159012864636041061113427342058826_dp
  t_lf(2190)=  14659.49324695145690691511647982246602074_dp
  t_lf(2191)=  14667.18536029105237343690059787516641082_dp
  t_lf(2192)=  14674.87792993912027958988471648897898249_dp
  t_lf(2193)=  14682.57095568753816778500233885712048176_dp
  t_lf(2194)=  14690.26443732837334334032891834814530812_dp
  t_lf(2195)=  14697.95837465388261506492288981111424946_dp
  t_lf(2196)=  14705.65276745651203631529923363130126875_dp
  t_lf(2197)=  14713.34761552889664652345969595599722456_dp
  t_lf(2198)=  14721.04291866386021319540672605730205611_dp
  t_lf(2199)=  14728.73867665441497437907111999300968012_dp
  t_lf(2200)=  14736.43488929376138160058627860180171440_dp
  t_lf(2201)=  14744.13155637528784326784489745885710571_dp
  t_lf(2202)=  14751.82867769257046854027680675639623715_dp
  t_lf(2203)=  14759.52625303937281166378957019421368552_dp
  t_lf(2204)=  14767.22428220964561676981633390137980380_dp
  t_lf(2205)=  14774.92276499752656313741828919533740231_dp
  t_lf(2206)=  14782.62170119734001091739197665177966915_dp
  t_lf(2207)=  14790.32109060359674731733451354102738907_dp
  t_lf(2208)=  14798.02093301099373324662267221705099373_dp
  t_lf(2209)=  14805.72122821441385042026457355659534115_dp
  t_lf(2210)=  14813.42197600892564892058558707071815050_dp
  t_lf(2211)=  14821.12317618978309521571284788196952825_dp
  t_lf(2212)=  14828.82482855242532063382561040981047204_dp
  t_lf(2213)=  14836.52693289247637029214145936695133285_dp
  t_lf(2214)=  14844.22948900574495247961119057221449412_dp
  t_lf(2215)=  14851.93249668822418849229795716328595746_dp
  t_lf(2216)=  14859.63595573609136292041905107718509801_dp
  t_lf(2217)=  14867.33986594570767438603145518918816205_dp
  t_lf(2218)=  14875.04422711361798673034505829389790882_dp
  t_lf(2219)=  14882.74903903655058064965017320663969986_dp
  t_lf(2220)=  14890.45430151141690577884873769073621351_dp
  t_lf(2221)=  14898.16001433531133322158130870769463163_dp
  t_lf(2222)=  14905.86617730551090852594468267403092666_dp
  t_lf(2223)=  14913.57279021947510510479768802132915021_dp
  t_lf(2224)=  14921.27985287484557809965540142603740026_dp
  t_lf(2225)=  14928.98736506944591868717473563315962476_dp
  t_lf(2226)=  14936.69532660128140882723703487401256454_dp
  t_lf(2227)=  14944.40373726853877645163599350305521696_dp
  t_lf(2228)=  14952.11259686958595109238188468281635186_dp
  t_lf(2229)=  14959.82190520297181994863574875937339045_dp
  t_lf(2230)=  14967.53166206742598439128984542378089469_dp
  t_lf(2231)=  14975.24186726185851690421331987729505252_dp
  t_lf(2232)=  14982.95252058535971846118467104005699084_dp
  t_lf(2233)=  14990.66362183719987633753523939382721148_dp
  t_lf(2234)=  14998.37517081682902235553055335903077790_dp
  t_lf(2235)=  15006.08716732387669156251898620428361174_dp
  t_lf(2236)=  15013.79961115815168134087978040211113051_dp
  t_lf(2237)=  15021.51250211964181094880509310700995819_dp
  t_lf(2238)=  15029.22584000851368149095330507049232460_dp
  t_lf(2239)=  15036.93962462511243631801341585132459605_dp
  t_lf(2240)=  15044.65385576996152185422292065674302282_dp
  t_lf(2241)=  15052.36853324376244885188412859080197039_dp
  t_lf(2242)=  15060.08365684739455407192643851786769812_dp
  t_lf(2243)=  15067.79922638191476238956463720118411831_dp
  t_lf(2244)=  15075.51524164855734932410582487686125299_dp
  t_lf(2245)=  15083.23170244873370399196010600091355063_dp
  t_lf(2246)=  15090.94860858403209248191270758933148154_dp
  t_lf(2247)=  15098.66595985621742165171770438672046398_dp
  t_lf(2248)=  15106.38375606723100334507603907578816562_dp
  t_lf(2249)=  15114.10199701919031902806302690579447570_dp
  t_lf(2250)=  15121.82068251438878484407302750077626659_dp
  t_lf(2251)=  15129.53981235529551708635145223558868692_dp
  t_lf(2252)=  15137.25938634455509808718675346712278104_dp
  t_lf(2253)=  15144.97940428498734252283751210691224067_dp
  t_lf(2254)=  15152.69986597958706413327220254706697281_dp
  t_lf(2255)=  15160.42077123152384285580166883129569734_dp
  t_lf(2256)=  15168.14211984414179237168679322382310303_dp
  t_lf(2257)=  15175.86391162095932806480627799828016494_dp
  t_lf(2258)=  15183.58614636566893539147189337305234602_dp
  t_lf(2259)=  15191.30882388213693866048096908590565067_dp
  t_lf(2260)=  15199.03194397440327022249832415566421312_dp
  t_lf(2261)=  15206.75550644668124006786223894886837292_dp
  t_lf(2262)=  15214.47951110335730583191147578117630541_dp
  t_lf(2263)=  15222.20395774899084320693274896315720519_dp
  t_lf(2264)=  15229.92884618831391675983043247432687510_dp
  t_lf(2265)=  15237.65417622623105115462267334396006473_dp
  t_lf(2266)=  15245.37994766781900277887045135843681142_dp
  t_lf(2267)=  15253.10616031832653177314849092859766406_dp
  t_lf(2268)=  15260.83281398317417446266928886264729974_dp
  t_lf(2269)=  15268.55990846795401619017387242630739121_dp
  t_lf(2270)=  15276.28744357842946454920524545782525246_dp
  t_lf(2271)=  15284.01541912053502301688281646664071972_dp
  t_lf(2272)=  15291.74383490037606498529843160644463665_dp
  t_lf(2273)=  15299.47269072422860819065695720137215818_dp
  t_lf(2274)=  15307.20198639853908953928667114340648836_dp
  t_lf(2275)=  15314.93172172992414032964702999486736197_dp
  t_lf(2276)=  15322.66189652517036186946367904716785359_dp
  t_lf(2277)=  15330.39251059123410148712286593078823041_dp
  t_lf(2278)=  15338.12356373524122893645970466648325627_dp
  t_lf(2279)=  15345.85505576448691319407701631885816107_dp
  t_lf(2280)=  15353.58698648643539964833374468526925766_dp
  t_lf(2281)=  15361.31935570871978767914421075008147447_dp
  t_lf(2282)=  15369.05216323914180862773172798110358064_dp
  t_lf(2283)=  15376.78540888567160415548235196588287597_dp
  t_lf(2284)=  15384.51909245644750499104678240473885827_dp
  t_lf(2285)=  15392.25321375977581006484067311911852944_dp
  t_lf(2286)=  15399.98777260413056603009583652213804981_dp
  t_lf(2287)=  15407.72276879815334716961705295701513359_dp
  t_lf(2288)=  15415.45820215065303568740141246237828776_dp
  t_lf(2289)=  15423.19407247060560238427932689495319485_dp
  t_lf(2290)=  15430.93037956715388771673855395357016737_dp
  t_lf(2291)=  15438.66712324960738323809477152741348990_dp
  t_lf(2292)=  15446.40430332744201342117443095945473589_dp
  t_lf(2293)=  15454.14191961029991786167780129649662611_dp
  t_lf(2294)=  15461.87997190798923386139229341352860706_dp
  t_lf(2295)=  15469.61846003048387939042832307539549834_dp
  t_lf(2296)=  15477.35738378792333642765213555625061537_dp
  t_lf(2297)=  15485.09674299061243467849217139995834077_dp
  t_lf(2298)=  15492.83653744902113566929770329549151329_dp
  t_lf(2299)=  15500.57676697378431721743061788330962543_dp
  t_lf(2300)=  15508.31743137570155827627335362448854996_dp
  t_lf(2301)=  15516.05853046573692415433813667669609190_dp
  t_lf(2302)=  15523.80006405501875210766478105257608104_dp
  t_lf(2303)=  15531.54203195483943730469643720923461421_dp
  t_lf(2304)=  15539.28443397665521916282478465474507817_dp
  t_lf(2305)=  15547.02726993208596805579826918124579235_dp
  t_lf(2306)=  15554.77053963291497239118908396655034380_dp
  t_lf(2307)=  15562.51424289108872605711668604939394715_dp
  t_lf(2308)=  15570.25837951871671623742772559958098869_dp
  t_lf(2309)=  15578.00294932807121159453434499537475667_dp
  t_lf(2310)=  15585.74795213158705081911487800838995561_dp
  t_lf(2311)=  15593.49338774186143154588304640283634268_dp
  t_lf(2312)=  15601.23925597165369963463381200295712546_dp
  t_lf(2313)=  15608.98555663388513881577609679156343232_dp
  t_lf(2314)=  15616.73228954163876069956463189525677671_dp
  t_lf(2315)=  15624.47945450815909514824523840974167429_dp
  t_lf(2316)=  15632.22705134685198101032987894296361190_dp
  t_lf(2317)=  15639.97507987128435721621984852598342843_dp
  t_lf(2318)=  15647.72353989518405423439749718275507329_dp
  t_lf(2319)=  15655.47243123243958588740889398146444148_dp
  t_lf(2320)=  15663.22175369709994152686185383288524700_dp
  t_lf(2321)=  15670.97150710337437856666575367630465642_dp
  t_lf(2322)=  15678.72169126563221537374156402187625713_dp
  t_lf(2323)=  15686.47230599840262451543251512059944241_dp
  t_lf(2324)=  15694.22335111637442636284780433025034573_dp
  t_lf(2325)=  15701.97482643439588304937373255816760392_dp
  t_lf(2326)=  15709.72673176747449278358863301041404024_dp
  t_lf(2327)=  15717.47906693077678451581992488200076102_dp
  t_lf(2328)=  15725.23183173962811295758358810500177815_dp
  t_lf(2329)=  15732.98502600951245395314831285085477007_dp
  t_lf(2330)=  15740.73864955607220020246852918020802176_dp
  t_lf(2331)=  15748.49270219510795733473246806852770006_dp
  t_lf(2332)=  15756.24718374257834033177334502843824027_dp
  t_lf(2333)=  15764.00209401459977030059369172046894180_dp
  t_lf(2334)=  15771.75743282744627159425478931248181673_dp
  t_lf(2335)=  15779.51319999754926928038507993444227335_dp
  t_lf(2336)=  15787.26939534149738695656334939917169876_dp
  t_lf(2337)=  15795.02601867603624491183438544101948559_dp
  t_lf(2338)=  15802.78306981806825863361672108266562126_dp
  t_lf(2339)=  15810.54054858465243765926397239509207210_dp
  t_lf(2340)=  15818.29845479300418477154317388664498289_dp
  t_lf(2341)=  15826.05678826049509553729540306347830990_dp
  t_lf(2342)=  15833.81554880465275818854586836487637410_dp
  t_lf(2343)=  15841.57473624316055384533251171227706320_dp
  t_lf(2344)=  15849.33435039385745707952404833946409110_dp
  t_lf(2345)=  15857.09439107473783681890023241249718121_dp
  t_lf(2346)=  15864.85485810395125759076899722056120614_dp
  t_lf(2347)=  15872.61575129980228110439697344202402389_dp
  t_lf(2348)=  15880.37707048075026817153173818251004970_dp
  t_lf(2349)=  15888.13881546540918096429599116256206413_dp
  t_lf(2350)=  15895.90098607254738560973569262024478968_dp
  t_lf(2351)=  15903.66358212108745512030603020753591586_dp
  t_lf(2352)=  15911.42660343010597265958090941717750730_dp
  t_lf(2353)=  15919.19004981883333514247348389737583483_dp
  t_lf(2354)=  15926.95392110665355716925705841382241770_dp
  t_lf(2355)=  15934.71821711310407529267750822037460629_dp
  t_lf(2356)=  15942.48293765787555261745016421972117017_dp
  t_lf(2357)=  15950.24808256081168373143591355173784065_dp
  t_lf(2358)=  15958.01365164190899996779306015821059452_dp
  t_lf(2359)=  15965.77964472131667499740327945630220547_dp
  t_lf(2360)=  15973.54606161933633075087178552762262043_dp
  t_lf(2361)=  15981.31290215642184366940360821302956193_dp
  t_lf(2362)=  15989.08016615317915128385965121325735890_dp
  t_lf(2363)=  15996.84785343036605912129797075000602922_dp
  t_lf(2364)=  16004.61596380889204793830747755900773903_dp
  t_lf(2365)=  16012.38449710981808128044302298354490301_dp
  t_lf(2366)=  16020.15345315435641336707258273157689652_dp
  t_lf(2367)=  16027.92283176387039730094899946962702002_dp
  t_lf(2368)=  16035.69263275987429360182048786940752659_dp
  t_lf(2369)=  16043.46285596403307906339584301627116353_dp
  t_lf(2370)=  16051.23350119816225593298202524935946436_dp
  t_lf(2371)=  16059.00456828422766141311352154909162367_dp
  t_lf(2372)=  16066.77605704434527748449460553565812544_dp
  t_lf(2373)=  16074.54796730078104104957733500963985255_dp
  t_lf(2374)=  16082.32029887595065439609983776989047530_dp
  t_lf(2375)=  16090.09305159241939597991114320145689175_dp
  t_lf(2376)=  16097.86622527290193152641051885456412787_dp
  t_lf(2377)=  16105.63981974026212544993096795148779928_dp
  t_lf(2378)=  16113.41383481751285259039823547834529504_dp
  t_lf(2379)=  16121.18827032781581026659835726025875409_dp
  t_lf(2380)=  16128.96312609448133064538846819771760113_dp
  t_lf(2381)=  16136.73840194096819342618726267597154035_dp
  t_lf(2382)=  16144.51409769088343884008317206452910191_dp
  t_lf(2383)=  16152.29021316798218096289999121687197230_dp
  t_lf(2384)=  16160.06674819616742134156134797780879906_dp
  t_lf(2385)=  16167.84370259948986293309706692390909588_dp
  t_lf(2386)=  16175.62107620214772435563613091754147123_dp
  t_lf(2387)=  16183.39886882848655445073259156349213747_dp
  t_lf(2388)=  16191.17708030299904715637242233519955351_dp
  t_lf(2389)=  16198.95571045032485669001094600148794412_dp
  t_lf(2390)=  16206.73475909525041304099210105043421742_dp
  t_lf(2391)=  16214.51422606270873777170244009071668170_dp
  t_lf(2392)=  16222.29411117777926012681437672846672613_dp
  t_lf(2393)=  16230.07441426568763344997481618521288643_dp
  t_lf(2394)=  16237.85513515180555190729691895584714767_dp
  t_lf(2395)=  16245.63627366165056751701435612047295193_dp
  t_lf(2396)=  16253.41782962088590748465901953627076175_dp
  t_lf(2397)=  16261.19980285532029184312475006083858770_dp
  t_lf(2398)=  16268.98219319090775139698124221247109522_dp
  t_lf(2399)=  16276.76500045374744597040387427111255463_dp
  t_lf(2400)=  16284.54822447008348295808679878177833845_dp
  t_lf(2401)=  16292.33186506630473617850820975555105737_dp
  t_lf(2402)=  16300.11592206894466502891827958722695967_dp
  t_lf(2403)=  16307.90039530468113394142183083866438601_dp
  t_lf(2404)=  16315.68528460033623213952937558815988528_dp
  t_lf(2405)=  16323.47058978287609369455171803398283868_dp
  t_lf(2406)=  16331.25631067941071788121487447971289069_dp
  t_lf(2407)=  16339.04244711719378983187361873536617275_dp
  t_lf(2408)=  16346.82899892362250148870351035652972922_dp
  t_lf(2409)=  16354.61596592623737285325280802885592679_dp
  t_lf(2410)=  16362.40334795272207353273721080225110673_dp
  t_lf(2411)=  16370.19114483090324458246190580282065718_dp
  t_lf(2412)=  16377.97935638875032064375693251594578447_dp
  t_lf(2413)=  16385.76798245437535237681340075554993041_dp
  t_lf(2414)=  16393.55702285603282918780962202739426761_dp
  t_lf(2415)=  16401.34647742211950224971773217279635680_dp
  t_lf(2416)=  16409.13634598117420781618289695811353936_dp
  t_lf(2417)=  16416.92662836187769082786870166923819023_dp
  t_lf(2418)=  16424.71732439305242881066383079372656831_dp
  t_lf(2419)=  16432.50843390366245606514664454048367353_dp
  t_lf(2420)=  16440.29995672281318814670575527255647290_dp
  t_lf(2421)=  16448.09189267975124663571619892689676143_dp
  t_lf(2422)=  16455.88424160386428419717228418023910920_dp
  t_lf(2423)=  16463.67700332468080992917968550674203675_dp
  t_lf(2424)=  16471.47017767187001499971082537495208132_dp
  t_lf(2425)=  16479.26376447524159857102906566310841831_dp
  t_lf(2426)=  16487.05776356474559401118869894689537732_dp
  t_lf(2427)=  16494.85217477047219539201919664650448820_dp
  t_lf(2428)=  16502.64699792265158427300363312426753311_dp
  t_lf(2429)=  16510.44223285165375677046266271409657237_dp
  t_lf(2430)=  16518.23787938798835091145688035339356510_dp
  t_lf(2431)=  16526.03393736230447427182184599079713954_dp
  t_lf(2432)=  16533.83040660539053189775149827289222906_dp
  t_lf(2433)=  16541.62728694817405451034712418354365524_dp
  t_lf(2434)=  16549.42457822172152699255048833450053284_dp
  t_lf(2435)=  16557.22228025723821715788115849897726598_dp
  t_lf(2436)=  16565.02039288606800480039949275462123616_dp
  t_lf(2437)=  16572.81891593969321102531817827214924867_dp
  t_lf(2438)=  16580.61784924973442785968663236444667912_dp
  t_lf(2439)=  16588.41719264795034814257399291149759257_dp
  t_lf(2440)=  16596.21694596623759569417783771252392466_dp
  t_lf(2441)=  16604.01710903663055576328718070148082930_dp
  t_lf(2442)=  16611.81768169130120575252969730885810804_dp
  t_lf(2443)=  16619.61866376255894622083453157479992747_dp
  t_lf(2444)=  16627.42005508285043216254343392905377292_dp
  t_lf(2445)=  16635.22185548475940456260437086532323614_dp
  t_lf(2446)=  16643.02406480100652222728313606430793329_dp
  t_lf(2447)=  16650.82668286444919388982987687409961880_dp
  t_lf(2448)=  16658.62970950808141059053883045165050302_dp
  t_lf(2449)=  16666.43314456503357833064094031767427115_dp
  t_lf(2450)=  16674.23698786857235099947039659247118246_dp
  t_lf(2451)=  16682.04123925210046357434751177462740873_dp
  t_lf(2452)=  16689.84589854915656559262170861111981281_dp
  t_lf(2453)=  16697.65096559341505489531975739880808213_dp
  t_lf(2454)=  16705.45644021868591164184575696631717049_dp
  t_lf(2455)=  16713.26232225891453259518070662455844279_dp
  t_lf(2456)=  16721.06861154818156567703086555621546477_dp
  t_lf(2457)=  16728.87530792070274479237544145199154022_dp
  t_lf(2458)=  16736.68241121082872492286549170679637784_dp
  t_lf(2459)=  16744.48992125304491748852725817480835897_dp
  t_lf(2460)=  16752.29783788197132597722449036091083304_dp
  t_lf(2461)=  16760.10616093236238184133564200974430616_dp
  t_lf(2462)=  16767.91489023910678066110315235487466276_dp
  t_lf(2463)=  16775.72402563722731857411334582163894976_dp
  t_lf(2464)=  16783.53356696188072897036680275033814163_dp
  t_lf(2465)=  16791.34351404835751945240036873379936741_dp
  t_lf(2466)=  16799.15386673208180905992328145708246078_dp
  t_lf(2467)=  16806.96462484861116575843120149936718696_dp
  t_lf(2468)=  16814.77578823363644419126323742089373092_dp
  t_lf(2469)=  16822.58735672298162369456835562326163477_dp
  t_lf(2470)=  16830.39933015260364657464886195139917112_dp
  t_lf(2471)=  16838.21170835859225664714993481203032085_dp
  t_lf(2472)=  16846.02449117716983803756547872838081708_dp
  t_lf(2473)=  16853.83767844469125424253185274602556843_dp
  t_lf(2474)=  16861.65126999764368745138230996199152238_dp
  t_lf(2475)=  16869.46526567264647812743626268025407837_dp
  t_lf(2476)=  16877.27966530645096484849876231332015524_dp
  t_lf(2477)=  16885.09446873594032440604685416335301590_dp
  t_lf(2478)=  16892.90967579812941216258073463889659552_dp
  t_lf(2479)=  16900.72528633016460266661890230629176544_dp
  t_lf(2480)=  16908.54130016932363052481775444989301656_dp
  t_lf(2481)=  16916.35771715301543153069733753369888270_dp
  t_lf(2482)=  16924.17453711877998404945621313046872838_dp
  t_lf(2483)=  16931.99175990428815065835965052423641413_dp
  t_lf(2484)=  16939.80938534734152004218660330973054050_dp
  t_lf(2485)=  16947.62741328587224914322216991891294196_dp
  t_lf(2486)=  16955.44584355794290556528347711195226950_dp
  t_lf(2487)=  16963.26467600174631023126816108871737719_dp
  t_lf(2488)=  16971.08391045560538029371585301852458917_dp
  t_lf(2489)=  16978.90354675797297229787430446158196860_dp
  t_lf(2490)=  16986.72358474743172559676301337648022897_dp
  t_lf(2491)=  16994.54402426269390601772843318528146045_dp
  t_lf(2492)=  17002.36486514260124977998606571231084253_dp
  t_lf(2493)=  17010.18610722612480766264595373568050478_dp
  t_lf(2494)=  17018.00775035236478942271930040284643999_dp
  t_lf(2495)=  17025.82979436055040846260515087405702065_dp
  t_lf(2496)=  17033.65223909003972674655727628129392437_dp
  t_lf(2497)=  17041.47508438031949996563260143609255071_dp
  t_lf(2498)=  17049.29833007100502295062371569827959153_dp
  t_lf(2499)=  17057.12197600183997533247920103996160323_dp
  t_lf(2500)=  17064.94602201269626744971670261578270693_dp
  t_lf(2501)=  17072.77046794357388650233485509224572926_dp
  t_lf(2502)=  17080.59531363460074295173136260642450102_dp
  t_lf(2503)=  17088.42055892603251716613571152831260372_dp
  t_lf(2504)=  17096.24620365825250631106617320194434378_dp
  t_lf(2505)=  17104.07224767177147148432192854883783649_dp
  t_lf(2506)=  17111.89869080722748509502231784376059496_dp
  t_lf(2507)=  17119.72553290538577848620638712777999273_dp
  t_lf(2508)=  17127.55277380713858980050706761747185441_dp
  t_lf(2509)=  17135.38041335350501208841548611242023178_dp
  t_lf(2510)=  17143.20845138563084165865206280611284922_dp
  t_lf(2511)=  17151.03688774478842667016220807834530925_dp
  t_lf(2512)=  17158.86572227237651596525558180058149077_dp
  t_lf(2513)=  17166.69495480992010814340902742962935048_dp
  t_lf(2514)=  17174.52458519907030087525443870969554314_dp
  t_lf(2515)=  17182.35461328160414045627395915855734082_dp
  t_lf(2516)=  17190.18503889942447159972605369037826765_dp
  t_lf(2517)=  17198.01586189455978746832712773570041839_dp
  t_lf(2518)=  17205.84708210916407994421450206844120888_dp
  t_lf(2519)=  17213.67869938551669013671768125033894649_dp
  t_lf(2520)=  17221.51071356602215912746598016522789498_dp
  t_lf(2521)=  17229.34312449321007895236169654874153193_dp
  t_lf(2522)=  17237.17593200973494381994913773346898882_dp
  t_lf(2523)=  17245.00913595837600156571092703511533773_dp
  t_lf(2524)=  17252.84273618203710534182412931169728495_dp
  t_lf(2525)=  17260.67673252374656554190984624506264939_dp
  t_lf(2526)=  17268.51112482665700196031103983184044914_dp
  t_lf(2527)=  17276.34591293404519618543444743905933594_dp
  t_lf(2528)=  17284.18109668931194422669355358783162970_dp
  t_lf(2529)=  17292.01667593598190937459068238636985305_dp
  t_lf(2530)=  17299.85265051770347529347737025082954268_dp
  t_lf(2531)=  17307.68902027824859934653327123866900148_dp
  t_lf(2532)=  17315.52578506151266615250493698396216044_dp
  t_lf(2533)=  17323.36294471151434137374689987693940747_dp
  t_lf(2534)=  17331.20049907239542573510857178047377269_dp
  t_lf(2535)=  17339.03844798842070927321155123375312325_dp
  t_lf(2536)=  17346.87679130397782581566300976742626426_dp
  t_lf(2537)=  17354.71552886357710768975190265449180892_dp
  t_lf(2538)=  17362.55466051185144066017582115648973631_dp
  t_lf(2539)=  17370.39418609355611909534737210449982642_dp
  t_lf(2540)=  17378.23410545356870136183003648835865959_dp
  t_lf(2541)=  17386.07441843688886544645452162465462363_dp
  t_lf(2542)=  17393.91512488863826480566768144369256500_dp
  t_lf(2543)=  17401.75622465406038444166713648694780663_dp
  t_lf(2544)=  17409.59771757852039720487577934873209638_dp
  t_lf(2545)=  17417.43960350750502032231140253801804267_dp
  t_lf(2546)=  17425.28188228662237215140773408772779553_dp
  t_lf(2547)=  17433.12455376160182915884421170836798714_dp
  t_lf(2548)=  17440.96761777829388312394286887973601834_dp
  t_lf(2549)=  17448.81107418266999856619174600755046714_dp
  t_lf(2550)=  17456.65492282082247039645527665025667123_dp
  t_lf(2551)=  17464.49916353896428179143313285388165695_dp
  t_lf(2552)=  17472.34379618342896229093004482858322756_dp
  t_lf(2553)=  17480.18882060067044611750013856834739655_dp
  t_lf(2554)=  17488.03423663726293071803036056399633383_dp
  t_lf(2555)=  17495.88004413990073552682858149810424377_dp
  t_lf(2556)=  17503.72624295539816094978299074737868242_dp
  t_lf(2557)=  17511.57283293068934756916041066231634351_dp
  t_lf(2558)=  17519.41981391282813556861217395422104478_dp
  t_lf(2559)=  17527.26718574898792437795721910468254461_dp
  t_lf(2560)=  17535.11494828646153253731306753103232136_dp
  t_lf(2561)=  17542.96310137266105778014635230176047094_dp
  t_lf(2562)=  17550.81164485511773733481557150700996769_dp
  t_lf(2563)=  17558.66057858148180844417973995964398447_dp
  t_lf(2564)=  17566.50990239952236910284761074056193779_dp
  t_lf(2565)=  17574.35961615712723901164313321644355608_dp
  t_lf(2566)=  17582.20971970230282074886380655742080562_dp
  t_lf(2567)=  17590.06021288317396115790957747477840845_dp
  t_lf(2568)=  17597.91109554798381295086091789309874086_dp
  t_lf(2569)=  17605.76236754509369652758570257570034975_dp
  t_lf(2570)=  17613.61402872298296200995548834514596941_dp
  t_lf(2571)=  17621.46607893024885149075277549036124370_dp
  t_lf(2572)=  17629.31851801560636149685180823682562992_dp
  t_lf(2573)=  17637.17134582788810566625644478465935759_dp
  t_lf(2574)=  17645.02456221604417763857959839949303347_dp
  t_lf(2575)=  17652.87816702914201415854971938099884957_dp
  t_lf(2576)=  17660.73216011636625839213075344208493169_dp
  t_lf(2577)=  17668.58654132701862345484297511617908696_dp
  t_lf(2578)=  17676.44131051051775615187305527889845762_dp
  t_lf(2579)=  17684.29646751639910092956267973183233104_dp
  t_lf(2580)=  17692.15201219431476403786599105824324455_dp
  t_lf(2581)=  17700.00794439403337790336707863127500538_dp
  t_lf(2582)=  17707.86426396543996571244969174277567533_dp
  t_lf(2583)=  17715.72097075853580620421229833310131406_dp
  t_lf(2584)=  17723.57806462343829867272255674723682384_dp
  t_lf(2585)=  17731.43554541038082817820621032820031448_dp
  t_lf(2586)=  17739.29341296971263096676635449290606701_dp
  t_lf(2587)=  17747.15166715189866009822996322633992282_dp
  t_lf(2588)=  17755.01030780751945128171949668491381637_dp
  t_lf(2589)=  17762.86933478727098891854834382704991510_dp
  t_lf(2590)=  17770.72874794196457235203978369620890620_dp
  t_lf(2591)=  17778.58854712252668232387007617650371835_dp
  t_lf(2592)=  17786.44873217999884763653721773148470401_dp
  t_lf(2593)=  17794.30930296553751202155781983037342640_dp
  t_lf(2594)=  17802.17025933041390121299548747066126924_dp
  t_lf(2595)=  17810.03160112601389022592499242925097406_dp
  t_lf(2596)=  17817.89332820383787083943745062385215412_dp
  t_lf(2597)=  17825.75544041550061928379262524976757135_dp
  t_lf(2598)=  17833.61793761273116413132538718212081759_dp
  t_lf(2599)=  17841.48081964737265439071427150754703538_dp
  t_lf(2600)=  17849.34408637138222780422097397993925897_dp
  t_lf(2601)=  17857.20773763683087934751053368953056345_dp
  t_lf(2602)=  17865.07177329590332993166284830088718334_dp
  t_lf(2603)=  17872.93619320089789530698706586075490227_dp
  t_lf(2604)=  17880.80099720422635516825129240857931805_dp
  t_lf(2605)=  17888.66618515841382246094094744832334215_dp
  t_lf(2606)=  17896.53175691609861288815998976732018150_dp
  t_lf(2607)=  17904.39771233003211461779012412368924743_dp
  t_lf(2608)=  17912.26405125307865818952398497564270357_dp
  t_lf(2609)=  17920.13077353821538662138917670113315005_dp
  t_lf(2610)=  17927.99787903853212571538093066202447728_dp
  t_lf(2611)=  17935.86536760723125456182201801056932114_dp
  t_lf(2612)=  17943.73323909762757624206943332468389936_dp
  t_lf(2613)=  17951.60149336314818872918823799953546642_dp
  t_lf(2614)=  17959.47013025733235598621382382348551884_dp
  t_lf(2615)=  17967.33914963383137926162472633362480178_dp
  t_lf(2616)=  17975.20855134640846858164898438713105884_dp
  t_lf(2617)=  17983.07833524893861443902790690658972589_dp
  t_lf(2618)=  17990.94850119540845967786196996732933802_dp
  t_lf(2619)=  17998.81904903991617157416442729980087685_dp
  t_lf(2620)=  18006.68997863667131411174907488711294352_dp
  t_lf(2621)=  18014.56128983999472045307946565403763598_dp
  t_lf(2622)=  18022.43298250431836560470772327611638790_dp
  t_lf(2623)=  18030.30505648418523927693195489288784969_dp
  t_lf(2624)=  18038.17751163424921893730211099467431668_dp
  t_lf(2625)=  18046.05034780927494305760498697471858504_dp
  t_lf(2626)=  18053.92356486413768455395990480465506937_dp
  t_lf(2627)=  18061.79716265382322441965745500819955063_dp
  t_lf(2628)=  18069.67114103342772555037451858239949774_dp
  t_lf(2629)=  18077.54549985815760676139962575462653614_dp
  t_lf(2630)=  18085.42023898332941699650354347351597244_dp
  t_lf(2631)=  18093.29535826436970972809081632004370888_dp
  t_lf(2632)=  18101.17085755681491754826881609763359202_dp
  t_lf(2633)=  18109.04673671631122695047168372434034201_dp
  t_lf(2634)=  18116.92299559861445330127737321246379799_dp
  t_lf(2635)=  18124.79963405958991600205683148810546937_dp
  t_lf(2636)=  18132.67665195521231384009516958184164661_dp
  t_lf(2637)=  18140.55404914156560052882550031849919621_dp
  t_lf(2638)=  18148.43182547484286043681693505559858321_dp
  t_lf(2639)=  18156.30998081134618450515904727296899315_dp
  t_lf(2640)=  18164.18851500748654635288592390691554206_dp
  t_lf(2641)=  18172.06742791978367857008373625767893938_dp
  t_lf(2642)=  18179.94671940486594919832657108530175952_dp
  t_lf(2643)=  18187.82638931947023839808606915290860366_dp
  t_lf(2644)=  18195.70643752044181530276122298430366407_dp
  t_lf(2645)=  18203.58686386473421505897548798115024395_dp
  t_lf(2646)=  18211.46766820940911605278916130026235709_dp
  t_lf(2647)=  18219.34885041163621732147578103012645647_dp
  t_lf(2648)=  18227.23041032869311615051209423407762298_dp
  t_lf(2649)=  18235.11234781796518585543193635195345624_dp
  t_lf(2650)=  18242.99466273694545374819515627889306586_dp
  t_lf(2651)=  18250.87735494323447928772351117556901089_dp
  t_lf(2652)=  18258.76042429454023241425624271484633487_dp
  t_lf(2653)=  18266.64387064867797206717883204194314251_dp
  t_lf(2654)=  18274.52769386357012488597921422488828328_dp
  t_lf(2655)=  18282.41189379724616409398651440567921995_dp
  t_lf(2656)=  18290.29647030784248856454814723626147324_dp
  t_lf(2657)=  18298.18142325360230206930189850348346508_dp
  t_lf(2658)=  18306.06675249287549270820038311970944305_dp
  t_lf(2659)=  18313.95245788411851252094604688695983765_dp
  t_lf(2660)=  18321.83853928589425727949565063843341169_dp
  t_lf(2661)=  18329.72499655687194646129394452816866615_dp
  t_lf(2662)=  18337.61182955582700340289700738352223062_dp
  t_lf(2663)=  18345.49903814164093563364649116215783775_dp
  t_lf(2664)=  18353.38662217330121538905677367140886603_dp
  t_lf(2665)=  18361.27458150990116030357778381923777986_dp
  t_lf(2666)=  18369.16291601063981428239702277858416068_dp
  t_lf(2667)=  18377.05162553482182855194506156666615850_dp
  t_lf(2668)=  18384.94070994185734288876955067375461936_dp
  t_lf(2669)=  18392.83016909126186702644353052803122115_dp
  t_lf(2670)=  18400.72000284265616224017458276030796028_dp
  t_lf(2671)=  18408.61021105576612310878211144054154709_dp
  t_lf(2672)=  18416.50079359042265945371079070311903273_dp
  t_lf(2673)=  18424.39175030656157845474896046569679055_dp
  t_lf(2674)=  18432.28308106422346694212149528280051399_dp
  t_lf(2675)=  18440.17478572355357386462741276627616628_dp
  t_lf(2676)=  18448.06686414480169293349322745583818514_dp
  t_lf(2677)=  18455.95931618832204544161379354018950739_dp
  t_lf(2678)=  18463.85214171457316325785311541826644339_dp
  t_lf(2679)=  18471.74534058411777199607833875684899634_dp
  t_lf(2680)=  18479.63891265762267435860086645081343643_dp
  t_lf(2681)=  18487.53285779585863365369927373140908642_dp
  t_lf(2682)=  18495.42717585970025748689942460181643488_dp
  t_lf(2683)=  18503.32186671012588162568791781357081997_dp
  t_lf(2684)=  18511.21693020821745403733571473787791715_dp
  t_lf(2685)=  18519.11236621516041909950952373804803267_dp
  t_lf(2686)=  18527.00817459224360198334923601886074670_dp
  t_lf(2687)=  18534.90435520085909320869042642124592077_dp
  t_lf(2688)=  18542.80090790250213337111164925181886218_dp
  t_lf(2689)=  18550.69783255877099804048697399210519860_dp
  t_lf(2690)=  18558.59512903136688283072491862728479992_dp
  t_lf(2691)=  18566.49279718209378864037564937450536606_dp
  t_lf(2692)=  18574.39083687285840706378902478177805740_dp
  t_lf(2693)=  18582.28924796567000597250676951566433018_dp
  t_lf(2694)=  18590.18803032264031526657276866487114245_dp
  t_lf(2695)=  18598.08718380598341279544617706294881885_dp
  t_lf(2696)=  18605.98670827801561044820273998197177977_dp
  t_lf(2697)=  18613.88660360115534041271042157579857429_dp
  t_lf(2698)=  18621.78686963792304160346613566165763849_dp
  t_lf(2699)=  18629.68750625094104625778106982777434260_dp
  t_lf(2700)=  18637.58851330293346670000278844791064813_dp
  t_lf(2701)=  18645.48989065672608227346299297638063786_dp
  t_lf(2702)=  18653.39163817524622643984050889466505292_dp
  t_lf(2703)=  18661.29375572152267404562975788848976293_dp
  t_lf(2704)=  18669.19624315868552875540566125745310641_dp
  t_lf(2705)=  18677.09910034996611065157660620326394419_dp
  t_lf(2706)=  18685.00232715869684400031779051264718335_dp
  t_lf(2707)=  18692.90592344831114518337794325223007334_dp
  t_lf(2708)=  18700.80988908234331079545309943046693851_dp
  t_lf(2709)=  18708.71422392442840590682178516110101668_dp
  t_lf(2710)=  18716.61892783830215249093664668899124293_dp
  t_lf(2711)=  18724.52400068780081801666823171752343501_dp
  t_lf(2712)=  18732.42944233686110420489730481243651210_dp
  t_lf(2713)=  18740.33525264952003594915275025486510828_dp
  t_lf(2714)=  18748.24143148991485039999278558185317859_dp
  t_lf(2715)=  18756.14797872228288621282787719063490086_dp
  t_lf(2716)=  18764.05489421096147295888441579869839544_dp
  t_lf(2717)=  18771.96217782038782069900887425011669688_dp
  t_lf(2718)=  18779.86982941509890972001283314490440115_dp
  t_lf(2719)=  18787.77784885973138043325992104727612493_dp
  t_lf(2720)=  18795.68623601902142343519637560566632194_dp
  t_lf(2721)=  18803.59499075780466972952758979722446254_dp
  t_lf(2722)=  18811.50411294101608111074366369721390208_dp
  t_lf(2723)=  18819.41360243368984070869763667428925446_dp
  t_lf(2724)=  18827.32345910095924369394072773096163035_dp
  t_lf(2725)=  18835.23368280805658814351956284962320728_dp
  t_lf(2726)=  18843.14427342031306606694101767321546916_dp
  t_lf(2727)=  18851.05523080315865459201095165089603543_dp
  t_lf(2728)=  18858.96655482212200731025375591777805165_dp
  t_lf(2729)=  18866.87824534283034578162028165885825850_dp
  t_lf(2730)=  18874.79030223100935119819235853547364674_dp
  t_lf(2731)=  18882.70272535248305620659275393287457863_dp
  t_lf(2732)=  18890.61551457317373688881006332460099372_dp
  t_lf(2733)=  18898.52866975910180490114865994810850323_dp
  t_lf(2734)=  18906.44219077638569977101446825130765694_dp
  t_lf(2735)=  18914.35607749124178135124796020613150533_dp
  t_lf(2736)=  18922.27032976998422243171640659769711655_dp
  t_lf(2737)=  18930.18494747902490150787804679082361894_dp
  t_lf(2738)=  18938.09993048487329570603147025434468983_dp
  t_lf(2739)=  18946.01527865413637386496413129252371529_dp
  t_lf(2740)=  18953.93099185351848977371454499664612146_dp
  t_lf(2741)=  18961.84706994982127556516333739321120787_dp
  t_lf(2742)=  18969.76351280994353526516894613274539368_dp
  t_lf(2743)=  18977.68032030088113849696438983976499240_dp
  t_lf(2744)=  18985.59749228972691434053214443446906129_dp
  t_lf(2745)=  18993.51502864367054534667478334496591509_dp
  t_lf(2746)=  19001.43292922999846170549965555983977930_dp
  t_lf(2747)=  19009.35119391609373556903649092924090964_dp
  t_lf(2748)=  19017.26982256943597552770743601301239680_dp
  t_lf(2749)=  19025.18881505760122124036963610121389101_dp
  t_lf(2750)=  19033.10817124826183821765108980031575979_dp
  t_lf(2751)=  19041.02789100918641275830111179185099112_dp
  t_lf(2752)=  19048.94797420823964703827734703394589756_dp
  t_lf(2753)=  19056.86842071338225435229188579440901229_dp
  t_lf(2754)=  19064.78923039267085450753963348143041821_dp
  t_lf(2755)=  19072.71040311425786936933269227890637099_dp
  t_lf(2756)=  19080.63193874639141855836511310241710050_dp
  t_lf(2757)=  19088.55383715741521529933297637339517656_dp
  t_lf(2758)=  19096.47609821576846242063535856745936383_dp
  t_lf(2759)=  19104.39872178998574850488233843267157026_dp
  t_lf(2760)=  19112.32170774869694418993679219900500943_dp
  t_lf(2761)=  19120.24505596062709862021732101597839197_dp
  t_lf(2762)=  19128.16876629459633604799024626558787472_dp
  t_lf(2763)=  19136.09283861951975258437919930671541842_dp
  t_lf(2764)=  19144.01727280440731309982142161945472673_dp
  t_lf(2765)=  19151.94206871836374827370047923760550502_dp
  t_lf(2766)=  19159.86722623058845179288568178926073383_dp
  t_lf(2767)=  19167.79274521037537769890908141325330253_dp
  t_lf(2768)=  19175.71862552711293788351151028752700027_dp
  t_lf(2769)=  19183.64486705028389973228969749852787868_dp
  t_lf(2770)=  19191.57146964946528391617708650273685373_dp
  t_lf(2771)=  19199.49843319432826233049155348673073014_dp
  t_lf(2772)=  19207.42575755463805618128380452490044371_dp
  t_lf(2773)=  19215.35344260025383421872080556839231505_dp
  t_lf(2774)=  19223.28148820112861111723917397917688772_dp
  t_lf(2775)=  19231.20989422730914600220403355358323587_dp
  t_lf(2776)=  19239.13866054893584112280940676434362511_dp
  t_lf(2777)=  19247.06778703624264067095678829333970574_dp
  t_lf(2778)=  19254.99727355955692974584911283297911796_dp
  t_lf(2779)=  19262.92711998929943346403789760659915697_dp
  t_lf(2780)=  19270.85732619598411621466190610161724153_dp
  t_lf(2781)=  19278.78789205021808105961624412743825946_dp
  t_lf(2782)=  19286.71881742270146927839136250748503164_dp
  t_lf(2783)=  19294.65010218422736005732200249522548635_dp
  t_lf(2784)=  19302.58174620568167032298668037180080651_dp
  t_lf(2785)=  19310.51374935804305471949886664187177591_dp
  t_lf(2786)=  19318.44611151238280572943157279864166941_dp
  t_lf(2787)=  19326.37883253986475393811761478271609554_dp
  t_lf(2788)=  19334.31191231174516844106837701654298280_dp
  t_lf(2789)=  19342.24535069937265739425445426064619831_dp
  t_lf(2790)=  19350.17914757418806870699210051371797113_dp
  t_lf(2791)=  19358.11330280772439087717996476984935102_dp
  t_lf(2792)=  19366.04781627160665396863114265672252044_dp
  t_lf(2793)=  19373.98268783755183073024612081241925914_dp
  t_lf(2794)=  19381.91791737736873785677273731955885740_dp
  t_lf(2795)=  19389.85350476295793739089982660769621068_dp
  t_lf(2796)=  19397.78944986631163826643176096220397700_dp
  t_lf(2797)=  19405.72575255951359799229164314413620329_dp
  t_lf(2798)=  19413.66241271473902447710144563471683151_dp
  t_lf(2799)=  19421.59943020425447799408793167399456519_dp
  t_lf(2800)=  19429.53680490041777328606373156972282647_dp
  t_lf(2801)=  19437.47453667567788181023348471351464707_dp
  t_lf(2802)=  19445.41262540257483412257549336062961833_dp
  t_lf(2803)=  19453.35107095373962240154986851120343661_dp
  t_lf(2804)=  19461.28987320189410311088468117814779032_dp
  t_lf(2805)=  19469.22903201985089980119216394413475950_dp
  t_lf(2806)=  19477.16854728051330605016753800082874693_dp
  t_lf(2807)=  19485.10841885687518854112356983162127639_dp
  t_lf(2808)=  19493.04864662202089027961448934832870181_dp
  t_lf(2809)=  19500.98923044912513394790342762638681735_dp
  t_lf(2810)=  19508.93017021145292539702805740576435350_dp
  t_lf(2811)=  19516.87146578235945727621964323985220795_dp
  t_lf(2812)=  19524.81311703529001279943123058568786178_dp
  t_lf(2813)=  19532.75512384377986964873122423975373510_dp
  t_lf(2814)=  19540.69748608145420401431912633794133986_dp
  t_lf(2815)=  19548.64020362202799477092072265978526848_dp
  t_lf(2816)=  19556.58327633930592779032052320941581032_dp
  t_lf(2817)=  19564.52670410718230038978977899251807209_dp
  t_lf(2818)=  19572.47048679964092591616891157356895392_dp
  t_lf(2819)=  19580.41462429075503846536370538438960531_dp
  t_lf(2820)=  19588.35911645468719773701512486722684549_dp
  t_lf(2821)=  19596.30396316568919402410312937677770104_dp
  t_lf(2822)=  19604.24916429810195333724536833940037460_dp
  t_lf(2823)=  19612.19471972635544266345214747780481059_dp
  t_lf(2824)=  19620.14062932496857535909956395936731030_dp
  t_lf(2825)=  19628.08689296854911667688321411943570725_dp
  t_lf(2826)=  19636.03351053179358942651538195114241048_dp
  t_lf(2827)=  19643.98048188948717976892911984386879523_dp
  t_lf(2828)=  19651.92780691650364314375313509714131725_dp
  t_lf(2829)=  19659.87548548780521032982189653891144666_dp
  t_lf(2830)=  19667.82351747844249363848587514039095107_dp
  t_lf(2831)=  19675.77190276355439323948733084738292510_dp
  t_lf(2832)=  19683.72064121836800361916755494385785470_dp
  t_lf(2833)=  19691.66973271819852017077197313085241896_dp
  t_lf(2834)=  19699.61917713844914591662000914608511579_dp
  t_lf(2835)=  19707.56897435461099836190710217044458597_dp
  t_lf(2836)=  19715.51912424226301647990676347016016221_dp
  t_lf(2837)=  19723.46962667707186782834104871144521047_dp
  t_lf(2838)=  19731.42048153479185579668831216113688908_dp
  t_lf(2839)=  19739.37168869126482698419759755575479498_dp
  t_lf(2840)=  19747.32324802242007870837950778586854642_dp
  t_lf(2841)=  19755.27515940427426664374388170609283413_dp
  t_lf(2842)=  19763.22742271293131259055509134679928176_dp
  t_lf(2843)=  19771.18003782458231237337625657511830780_dp
  t_lf(2844)=  19779.13300461550544386917415683336112327_dp
  t_lf(2845)=  19787.08632296206587516475710097597144938_dp
  t_lf(2846)=  19795.03999274071567284331849643485731942_dp
  t_lf(2847)=  19802.99401382799371039985933797078370630_dp
  t_lf(2848)=  19810.94838610052557678526331411874443836_dp
  t_lf(2849)=  19818.90310943502348507879870611118419455_dp
  t_lf(2850)=  19826.85818370828618128882172956790512564_dp
  t_lf(2851)=  19834.81360879719885328145644357875423870_dp
  t_lf(2852)=  19842.76938457873303983702682497802314517_dp
  t_lf(2853)=  19850.72551092994653983401707762116680788_dp
  t_lf(2854)=  19858.68198772798332156033671732821793029_dp
  t_lf(2855)=  19866.63881485007343215166744285738374709_dp
  t_lf(2856)=  19874.59599217353290715666927181999710877_dp
  t_lf(2857)=  19882.55351957576368022882388784747861749_dp
  t_lf(2858)=  19890.51139693425349294469361157546572924_dp
  t_lf(2859)=  19898.46962412657580474837487312298262362_dp
  t_lf(2860)=  19906.42820103038970302192552771865561230_dp
  t_lf(2861)=  19914.38712752343981328154581896470721064_dp
  t_lf(2862)=  19922.34640348355620949929325593596202209_dp
  t_lf(2863)=  19930.30602878865432455011213088853358791_dp
  t_lf(2864)=  19938.26600331673486078395886380438769635_dp
  t_lf(2865)=  19946.22632694588370072280481832673877684_dp
  t_lf(2866)=  19954.18699955427181788229869085036649900_dp
  t_lf(2867)=  19962.14802102015518771787103062356428745_dp
  t_lf(2868)=  19970.10939122187469869506390369766507539_dp
  t_lf(2869)=  19978.07111003785606348386916742903740522_dp
  t_lf(2870)=  19986.03317734660973027685927500020235674_dp
  t_lf(2871)=  19993.99559302673079423089498108437444906_dp
  t_lf(2872)=  20001.95835695689890903219477033435365874_dp
  t_lf(2873)=  20009.92146901587819858455127983535742249_dp
  t_lf(2874)=  20017.88492908251716882048043502513774137_dp
  t_lf(2875)=  20025.84873703574861963508946585662650041_dp
  t_lf(2876)=  20033.81289275458955694245041616142954705_dp
  t_lf(2877)=  20041.77739611814110485426620426977511787_dp
  t_lf(2878)=  20049.74224700558841798061673695703357441_dp
  t_lf(2879)=  20057.70744529620059385257302172167237910_dp
  t_lf(2880)=  20065.67299086933058546646766425749267756_dp
  t_lf(2881)=  20073.63888360441511394961057876720224734_dp
  t_lf(2882)=  20081.60512338097458134723917847779507910_dp
  t_lf(2883)=  20089.57171007861298353049275236380231923_dp
  t_lf(2884)=  20097.53864357701782322520117166521529292_dp
  t_lf(2885)=  20105.50592375596002316127850630571216896_dp
  t_lf(2886)=  20113.47355049529383934251256677668963700_dp
  t_lf(2887)=  20121.44152367495677443654182145644467892_dp
  t_lf(2888)=  20129.40984317496949128481157268459491570_dp
  t_lf(2889)=  20137.37850887543572653230170721238576981_dp
  t_lf(2890)=  20145.34752065654220437681876790281638709_dp
  t_lf(2891)=  20153.31687839855855043764552376342244418_dp
  t_lf(2892)=  20161.28658198183720574334164456197213875_dp
  t_lf(2893)=  20169.25663128681334083848951440414234014_dp
  t_lf(2894)=  20177.22702619400477000917964574531662815_dp
  t_lf(2895)=  20185.19776658401186562703058136884840026_dp
  t_lf(2896)=  20193.16885233751747261153859689331411546_dp
  t_lf(2897)=  20201.14028333528682301055294037428893412_dp
  t_lf(2898)=  20209.11205945816745069867276854484553794_dp
  t_lf(2899)=  20217.08418058708910619336236119613399657_dp
  t_lf(2900)=  20225.05664660306367158858161613786463659_dp
  t_lf(2901)=  20233.02945738718507560572924710109666802_dp
  t_lf(2902)=  20241.00261282062920876169652585523382083_dp
  t_lf(2903)=  20248.97611278465383865382982771033674276_dp
  t_lf(2904)=  20256.94995716059852536160065646756405677_dp
  t_lf(2905)=  20264.92414582988453696478224076752479459_dp
  t_lf(2906)=  20272.89867867401476517793220867133083779_dp
  t_lf(2907)=  20280.87355557457364110098126119493686927_dp
  t_lf(2908)=  20288.84877641322705108572817840669649029_dp
  t_lf(2909)=  20296.82434107172225271804190359368740516_dp
  t_lf(2910)=  20304.80024943188779091557186190699825679_dp
  t_lf(2911)=  20312.77650137563341414076807981254869847_dp
  t_lf(2912)=  20320.75309678494999072901308060484807965_dp
  t_lf(2913)=  20328.73003554190942533166793918909378278_dp
  t_lf(2914)=  20336.70731752866457547383528630486649027_dp
  t_lf(2915)=  20344.68494262744916822664245835508686498_dp
  t_lf(2916)=  20352.66291072057771699384839401953837234_dp
  t_lf(2917)=  20360.64122169044543841257828287580805515_dp
  t_lf(2918)=  20368.61987541952816936799037432461654568_dp
  t_lf(2919)=  20376.59887179038228412167975722385779490_dp
  t_lf(2920)=  20384.57821068564461155362432177889705481_dp
  t_lf(2921)=  20392.55789198803235251747851541842354333_dp
  t_lf(2922)=  20400.53791558034299730902090360805479111_dp
  t_lf(2923)=  20408.51828134545424324756194482056764919_dp
  t_lf(2924)=  20416.49898916632391237011878619470297590_dp
  t_lf(2925)=  20424.48003892598986923816428277656572119_dp
  t_lf(2926)=  20432.46143050756993885675783865132006033_dp
  t_lf(2927)=  20440.44316379426182470586606274075297233_dp
  t_lf(2928)=  20448.42523866934302688368162556693381906_dp
  t_lf(2929)=  20456.40765501617076036174909586620871982_dp
  t_lf(2930)=  20464.39041271818187335170692758370558222_dp
  t_lf(2931)=  20472.37351165889276578345515848894939903_dp
  t_lf(2932)=  20480.35695172189930789455877143065085047_dp
  t_lf(2933)=  20488.34073279087675893069705809577952059_dp
  t_lf(2934)=  20496.32485474957968595696971305720349845_dp
  t_lf(2935)=  20504.30931748184188277987077288799935876_dp
  t_lf(2936)=  20512.29412087157628897974190119153231763_dp
  t_lf(2937)=  20520.27926480277490905351690554708982186_dp
  t_lf(2938)=  20528.26474915950873166756975660372933870_dp
  t_lf(2939)=  20536.25057382592764902047876287257137214_dp
  t_lf(2940)=  20544.23673868626037631551993717252279808_dp
  t_lf(2941)=  20552.22324362481437134270297217883691608_dp
  t_lf(2942)=  20560.21008852597575417016362311048099149_dp
  t_lf(2943)=  20568.19727327420922694472667527345777026_dp
  t_lf(2944)=  20576.18479775405799380145405295547520235_dp
  t_lf(2945)=  20584.17266185014368088199300404513160074_dp
  t_lf(2946)=  20592.16086544716625646153967172852738767_dp
  t_lf(2947)=  20600.14940842990395118423374070036766487_dp
  t_lf(2948)=  20608.13829068321317840680022051761287006_dp
  t_lf(2949)=  20616.12751209202845465025480302399111020_dp
  t_lf(2950)=  20624.11707254136232015948960418562135967_dp
  t_lf(2951)=  20632.10697191630525957055647320402017532_dp
  t_lf(2952)=  20640.09721010202562268546542341527716402_dp
  t_lf(2953)=  20648.08778698376954535431611024558007432_dp
  t_lf(2954)=  20656.07870244686087046458065137593571074_dp
  t_lf(2955)=  20664.06995637670106903735645327524725506_dp
  t_lf(2956)=  20672.06154865876916143040807639324414653_dp
  t_lf(2957)=  20680.05347917862163864781753856548232303_dp
  t_lf(2958)=  20688.04574782189238375606282257409806344_dp
  t_lf(2959)=  20696.03835447429259340634471933255842986_dp
  t_lf(2960)=  20704.03129902161069946298250282264877093_dp
  t_lf(2961)=  20712.02458134971229073769929670970917578_dp
  t_lf(2962)=  20720.01820134454003482961835550000632109_dp
  t_lf(2963)=  20728.01215889211360007079184518442691765_dp
  t_lf(2964)=  20736.00645387852957757708406953771896863_dp
  t_lf(2965)=  20744.00108618996140340423144861459530491_dp
  t_lf(2966)=  20751.99605571265928080890191550545136240_dp
  t_lf(2967)=  20759.99136233295010261457675608752993126_dp
  t_lf(2968)=  20767.98700593723737368207827433437669846_dp
  t_lf(2969)=  20775.98298641200113348456702272965195077_dp
  t_lf(2970)=  20783.97930364379787878683269347306902139_dp
  t_lf(2971)=  20791.97595751926048642870312146868527986_dp
  t_lf(2972)=  20799.97294792509813621239620455123614588_dp
  t_lf(2973)=  20807.97027474809623389363990003692937835_dp
  t_lf(2974)=  20815.96793787511633427638580948335155607_dp
  t_lf(2975)=  20823.96593719309606441094221551112023761_dp
  t_lf(2976)=  20831.96427258904904689535278567987600337_dp
  t_lf(2977)=  20839.96294395006482327984750872537393029_dp
  t_lf(2978)=  20847.96195116330877757419277795502279320_dp
  t_lf(2979)=  20855.96129411602205985776788526844448142_dp
  t_lf(2980)=  20863.96097269552150999219553711969114268_dp
  t_lf(2981)=  20871.96098678919958143635435077086213402_dp
  t_lf(2982)=  20879.96133628452426516360163540519905012_dp
  t_lf(2983)=  20887.96202106903901368103510807349037891_dp
  t_lf(2984)=  20895.96304103036266515062253904296657275_dp
  t_lf(2985)=  20903.96439605618936761202866490498382329_dp
  t_lf(2986)=  20911.96608603428850330696905077884634106_dp
  t_lf(2987)=  20919.96811085250461310492092512626169349_dp
  t_lf(2988)=  20927.97047039875732103002135206631446848_dp
  t_lf(2989)=  20935.97316456104125888898344665662644269_dp
  t_lf(2990)=  20943.97619322742599099986167838468632162_dp
  t_lf(2991)=  20951.97955628605593902149764709631231984_dp
  t_lf(2992)=  20959.98325362515030688347805377798328194_dp
  t_lf(2993)=  20967.98728513300300581643692600845923210_dp
  t_lf(2994)=  20975.99165069798257948253449450482432740_dp
  t_lf(2995)=  20983.99635020853212920594545301093197173_dp
  t_lf(2996)=  20992.00138355316923930318966881431478159_dp
  t_lf(2997)=  21000.00675062048590251313874543303633158_dp
  t_lf(2998)=  21008.01245129914844552653217248879600241_dp
  t_lf(2999)=  21016.01848547789745461483713047893539675_dp
  t_lf(3000)=  21024.02485304554770135828635007991101507_dp
  t_lf(3001)=  21032.03155389098806847292875676036330854_dp
  t_lf(3002)=  21040.03858790318147573652796185519074957_dp
  t_lf(3003)=  21048.04595497116480601314399085508690291_dp
  t_lf(3004)=  21056.05365498404883137623396850087018997_dp
  t_lf(3005)=  21064.06168783101813933010780834067552374_dp
  t_lf(3006)=  21072.07005340133105912957528171272353109_dp
  t_lf(3007)=  21080.07875158431958819762116765896985001_dp
  t_lf(3008)=  21088.08778226938931864094551105749108313_dp
  t_lf(3009)=  21096.09714534601936386320634128600643299_dp
  t_lf(3010)=  21104.10684070376228527580252799747982397_dp
  t_lf(3011)=  21112.11686823224401910603477410330538864_dp
  t_lf(3012)=  21120.12722782116380330248306882215252072_dp
  t_lf(3013)=  21128.13791936029410453743924566513224273_dp
  t_lf(3014)=  21136.14894273948054530623361149253541399_dp
  t_lf(3015)=  21144.16029784864183112329493329597037579_dp
  t_lf(3016)=  21152.17198457776967781478338913427213571_dp
  t_lf(3017)=  21160.18400281692873890763640868404036569_dp
  t_lf(3018)=  21168.19635245625653311486764715805667799_dp
  t_lf(3019)=  21176.20903338596337191695965389909433186_dp
  t_lf(3020)=  21184.22204549633228723919111377472134892_dp
  t_lf(3021)=  21192.23538867771895922473985558256078593_dp
  t_lf(3022)=  21200.24906282055164410340313702705363093_dp
  t_lf(3023)=  21208.26306781533110215577703045000866711_dp
  t_lf(3024)=  21216.27740355263052577273704739005213022_dp
  t_lf(3025)=  21224.29206992309546761006245321243476410_dp
  t_lf(3026)=  21232.30706681744376883804703549243591643_dp
  t_lf(3027)=  21240.32239412646548748593940155473886285_dp
  t_lf(3028)=  21248.33805174102282688105619156954815753_dp
  t_lf(3029)=  21256.35403955205006418241190388578236362_dp
  t_lf(3030)=  21264.37035745055347900870933884430224269_dp
  t_lf(3031)=  21272.38700532761128216053497616171796639_dp
  t_lf(3032)=  21280.40398307437354443660390910974612400_dp
  t_lf(3033)=  21288.42129058206212554389926613823960662_dp
  t_lf(3034)=  21296.43892774197060310155135730376663794_dp
  t_lf(3035)=  21304.45689444546420173830208887183951735_dp
  t_lf(3036)=  21312.47519058397972228340049476145372161_dp
  t_lf(3037)=  21320.49381604902547105077553809735302840_dp
  t_lf(3038)=  21328.51277073218118921633264003023992163_dp
  t_lf(3039)=  21336.53205452509798228822069617985086750_dp
  t_lf(3040)=  21344.55166731949824966991664355225579151_dp
  t_lf(3041)=  21352.57160900717561431597494258275746693_dp
  t_lf(3042)=  21360.59187947999485248028964006119133212_dp
  t_lf(3043)=  21368.61247862989182355671697910908586083_dp
  t_lf(3044)=  21376.63340634887340001190682209985898358_dp
  t_lf(3045)=  21384.65466252901739741019145144581278826_dp
  t_lf(3046)=  21392.67624706247250453038061152095703464_dp
  t_lf(3047)=  21400.69815984145821357431195264844677478_dp
  t_lf(3048)=  21408.72040075826475046700633505746012258_dp
  t_lf(3049)=  21416.74296970525300524827774700846319180_dp
  t_lf(3050)=  21424.76586657485446255564788689979935840_dp
  t_lf(3051)=  21432.78909125957113219841575410418295718_dp
  t_lf(3052)=  21440.81264365197547982273288754275169170_dp
  t_lf(3053)=  21448.83652364471035766753518458861057283_dp
  t_lf(3054)=  21456.86073113048893541118252580305103375_dp
  t_lf(3055)=  21464.88526600209463110865772324761471371_dp
  t_lf(3056)=  21472.91012815238104221917660168564978713_dp
  t_lf(3057)=  21480.93531747427187672406131288973098832_dp
  t_lf(3058)=  21488.96083386076088433472927350802983797_dp
  t_lf(3059)=  21496.98667720491178779065040651517106735_dp
  t_lf(3060)=  21505.01284739985821424712565518303178607_dp
  t_lf(3061)=  21513.03934433880362675274002675606336835_dp
  t_lf(3062)=  21521.06616791502125581634371060576906929_dp
  t_lf(3063)=  21529.09331802185403106341510257167467952_dp
  t_lf(3064)=  21537.12079455271451298165985347320167948_dp
  t_lf(3065)=  21545.14859740108482475570034540000391805_dp
  t_lf(3066)=  21553.17672646051658419071028435926634263_dp
  t_lf(3067)=  21561.20518162463083572484938217888927012_dp
  t_lf(3068)=  21569.23396278711798253035338423709063949_dp
  t_lf(3069)=  21577.26306984173771870313498261344317741_dp
  t_lf(3070)=  21585.29250268231896154075143663541003390_dp
  t_lf(3071)=  21593.32226120275978390859500452973285152_dp
  t_lf(3072)=  21601.35234529702734669416257098123714290_dp
  t_lf(3073)=  21609.38275485915783134926113585442307633_dp
  t_lf(3074)=  21617.41348978325637252000610914727222228_dp
  t_lf(3075)=  21625.44454996349699076446963642368453085_dp
  t_lf(3076)=  21633.47593529412252535783645751252196163_dp
  t_lf(3077)=  21641.50764566944456718492507916902809709_dp
  t_lf(3078)=  21649.53968098384339171993231967006423002_dp
  t_lf(3079)=  21657.57204113176789209325955995979449866_dp
  t_lf(3080)=  21665.60472600773551224527931197880352503_dp
  t_lf(3081)=  21673.63773550633218016690099019877278026_dp
  t_lf(3082)=  21681.67106952221224122679504714840487570_dp
  t_lf(3083)=  21689.70472795009839158513490785589172415_dp
  t_lf(3084)=  21697.73871068478161169371641165049185845_dp
  t_lf(3085)=  21705.77301762112109988231474266232823691_dp
  t_lf(3086)=  21713.80764865404420603113910263695000226_dp
  t_lf(3087)=  21721.84260367854636532924565134112459286_dp
  t_lf(3088)=  21729.87788258969103211876951088034035790_dp
  t_lf(3089)=  21737.91348528260961382483690067819976183_dp
  t_lf(3090)=  21745.94941165250140497101873968486009257_dp
  t_lf(3091)=  21753.98566159463352128018732158751839381_dp
  t_lf(3092)=  21762.02223500434083386063793739222158944_dp
  t_lf(3093)=  21770.05913177702590347733758773458831755_dp
  t_lf(3094)=  21778.09635180815891490816319465892811899_dp
  t_lf(3095)=  21786.13389499327761138499198938230403036_dp
  t_lf(3096)=  21794.17176122798722911950701873386945851_dp
  t_lf(3097)=  21802.20995040796043191358097853187806192_dp
  t_lf(3098)=  21810.24846242893724585410184713267030611_dp
  t_lf(3099)=  21818.28729718672499409210405675923195210_dp
  t_lf(3100)=  21826.32645457719823170606920399314303773_dp
  t_lf(3101)=  21834.36593449629868064926056399443149316_dp
  t_lf(3102)=  21842.40573684003516478095593560054949844_dp
  t_lf(3103)=  21850.44586150448354498144360644993468705_dp
  t_lf(3104)=  21858.48630838578665435064648867892953160_dp
  t_lf(3105)=  21866.52707738015423349023973655473349248_dp
  t_lf(3106)=  21874.56816838386286586912741763307213179_dp
  t_lf(3107)=  21882.60958129325591327214406866889935949_dp
  t_lf(3108)=  21890.65131600474345133184722656321286523_dp
  t_lf(3109)=  21898.69337241480220514326728310046380983_dp
  t_lf(3110)=  21906.73575041997548496148127012058085631_dp
  t_lf(3111)=  21914.77844991687312198187743907880212330_dp
  t_lf(3112)=  21922.82147080217140420297775567680882565_dp
  t_lf(3113)=  21930.86481297261301237168568640156909516_dp
  t_lf(3114)=  21938.90847632500695601082690938531331461_dp
  t_lf(3115)=  21946.95246075622850952885083700265253119_dp
  t_lf(3116)=  21954.99676616321914841156109205049414546_dp
  t_lf(3117)=  21963.04139244298648549574333321457484660_dp
  t_lf(3118)=  21971.08633949260420732455907881458618456_dp
  t_lf(3119)=  21979.13160720921201058457443053947550148_dp
  t_lf(3120)=  21987.17719549001553862429285103702223970_dp
  t_lf(3121)=  21995.22310423228631805406140080867074471_dp
  t_lf(3122)=  22003.26933333336169542722009088329524103_dp
  t_lf(3123)=  22011.31588269064477400236425820352615198_dp
  t_lf(3124)=  22019.36275220160435058659012055692066851_dp
  t_lf(3125)=  22027.40994176377485245959391722305160671_dp
  t_lf(3126)=  22035.45745127475627437849529028795014544_dp
  t_lf(3127)=  22043.50528063221411566325580980069890084_dp
  t_lf(3128)=  22051.55342973387931736256379361475675320_dp
  t_lf(3129)=  22059.60189847754819950005681987022657912_dp
  t_lf(3130)=  22067.65068676108239840075357663416815368_dp
  t_lf(3131)=  22075.69979448240880409756693922562350005_dp
  t_lf(3132)=  22083.74922153951949781777041121166934190_dp
  t_lf(3133)=  22091.79896783047168954929030997194547624_dp
  t_lf(3134)=  22099.84903325338765568669632209312921541_dp
  t_lf(3135)=  22107.89941770645467675676329767313091159_dp
  t_lf(3136)=  22115.95012108792497522347739588876633046_dp
  t_lf(3137)=  22124.00114329611565337235993691170664371_dp
  t_lf(3138)=  22132.05248422940863127398255744700044128_dp
  t_lf(3139)=  22140.10414378625058482654750881778483088_dp
  t_lf(3140)=  22148.15612186515288387740717763033084691_dp
  t_lf(3141)=  22156.20841836469153042339714962667453803_dp
  t_lf(3142)=  22164.26103318350709688985737736913781879_dp
  t_lf(3143)=  22172.31396622030466448821625190340710852_dp
  t_lf(3144)=  22180.36721737385376165201261751587368982_dp
  t_lf(3145)=  22188.42078654298830255123100713800445116_dp
  t_lf(3146)=  22196.47467362660652568482561385695820493_dp
  t_lf(3147)=  22204.52887852367093255130875136884019437_dp
  t_lf(3148)=  22212.58340113320822639727979306024096330_dp
  t_lf(3149)=  22220.63824135430925104377081572637685976_dp
  t_lf(3150)=  22228.69339908612892979028540973157564274_dp
  t_lf(3151)=  22236.74887422788620439640735269136570425_dp
  t_lf(3152)=  22244.80466667886397414085607850636124150_dp
  t_lf(3153)=  22252.86077633840903495786610780781445445_dp
  t_lf(3154)=  22260.91720310593201865076783958445185501_dp
  t_lf(3155)=  22268.97394688090733218264733695134363673_dp
  t_lf(3156)=  22277.03100756287309704396297269538758767_dp
  t_lf(3157)=  22285.08838505143108869699703238983330428_dp
  t_lf(3158)=  22293.14607924624667609702060451343581574_dp
  t_lf(3159)=  22301.20409004704876129005031813961376287_dp
  t_lf(3160)=  22309.26241735362971908707571937869589114_dp
  t_lf(3161)=  22317.32106106584533681463630786326700439_dp
  t_lf(3162)=  22325.38002108361475414162748416406319041_dp
  t_lf(3163)=  22333.43929730692040298221488811310489521_dp
  t_lf(3164)=  22341.49888963580794747473683659308044975_dp
  t_lf(3165)=  22349.55879797038622403647479742868344321_dp
  t_lf(3166)=  22357.61902221082718149417206358794275217_dp
  t_lf(3167)=  22365.67956225736582129018101897083830219_dp
  t_lf(3168)=  22373.74041801030013776411961362993936574_dp
  t_lf(3169)=  22381.80158936999105850991789233470238609_dp
  t_lf(3170)=  22389.86307623686238480813564595868536161_dp
  t_lf(3171)=  22397.92487851140073213343248023853554333_dp
  t_lf(3172)=  22405.98699609415547073707182102644282982_dp
  t_lf(3173)=  22414.04942888573866630434059923507546272_dp
  t_lf(3174)=  22422.11217678682502068676658225707655723_dp
  t_lf(3175)=  22430.17523969815181270901554173124522491_dp
  t_lf(3176)=  22438.23861752051883905035067012579661389_dp
  t_lf(3177)=  22446.30231015478835520053688071682963531_dp
  t_lf(3178)=  22454.36631750188501649007284715856448865_dp
  t_lf(3179)=  22462.43063946279581919463385997227587102_dp
  t_lf(3180)=  22470.49527593857004171360879792436999527_dp
  t_lf(3181)=  22478.56022683031918582261473242195881775_dp
  t_lf(3182)=  22486.62549203921691799987290272779428852_dp
  t_lf(3183)=  22494.69107146649901082633001898675663478_dp
  t_lf(3184)=  22502.75696501346328445940906876445787825_dp
  t_lf(3185)=  22510.82317258146954818027402102613574395_dp
  t_lf(3186)=  22518.88969407193954201449303923208119526_dp
  t_lf(3187)=  22526.95652938635687842598503249556897256_dp
  t_lf(3188)=  22535.02367842626698408413459054184526427_dp
  t_lf(3189)=  22543.09114109327704170396056452336715570_dp
  t_lf(3190)=  22551.15891728905593195922377158837856081_dp
  t_lf(3191)=  22559.22700691533417546835951646823735463_dp
  t_lf(3192)=  22567.29540987390387485312083824486544331_dp
  t_lf(3193)=  22575.36412606661865686981860488446124134_dp
  t_lf(3194)=  22583.43315539539361461304479207837284221_dp
  t_lf(3195)=  22591.50249776220524979176549641795711536_dp
  t_lf(3196)=  22599.57215306909141507767044594851876874_dp
  t_lf(3197)=  22607.64212121815125652566598369920450864_dp
  t_lf(3198)=  22615.71240211154515606639871187218793711_dp
  t_lf(3199)=  22623.78299565149467407069719599678460037_dp
  t_lf(3200)=  22631.85390174028249198581933951344421163_dp
  t_lf(3201)=  22639.92512028025235504339324995003582834_dp
  t_lf(3202)=  22647.99665117380901503893962808962572427_dp
  t_lf(3203)=  22656.06849432341817318286392130619767225_dp
  t_lf(3204)=  22664.14064963160642302280669156362892602_dp
  t_lf(3205)=  22672.21311700096119343724085743485671384_dp
  t_lf(3206)=  22680.28589633413069170020467790369067846_dp
  t_lf(3207)=  22688.35898753382384661705955366228435658_dp
  t_lf(3208)=  22696.43239050281025173116192911400823602_dp
  t_lf(3209)=  22704.50610514392010860133878533549972302_dp
  t_lf(3210)=  22712.58013136004417015005642084412988994_dp
  t_lf(3211)=  22720.65446905413368408217242315914838069_dp
  t_lf(3212)=  22728.72911812920033637416093983746840569_dp
  t_lf(3213)=  22736.80407848831619483370156290955228907_dp
  t_lf(3214)=  22744.87935003461365272952234543827033777_dp
  t_lf(3215)=  22752.95493267128537249138767327504455849_dp
  t_lf(3216)=  22761.03082630158422948012191899416351322_dp
  t_lf(3217)=  22769.10703082882325582756000844897183268_dp
  t_lf(3218)=  22777.18354615637558434631623341380095960_dp
  t_lf(3219)=  22785.26037218767439250926284635411684407_dp
  t_lf(3220)=  22793.33750882621284649861017550551276070_dp
  t_lf(3221)=  22801.41495597554404532448020014096529447_dp
  t_lf(3222)=  22809.49271353928096501286572716628992462_dp
  t_lf(3223)=  22817.57078142109640286286751100706755211_dp
  t_lf(3224)=  22825.64915952472292177310185913754975763_dp
  t_lf(3225)=  22833.72784775395279463717146555427050565_dp
  t_lf(3226)=  22841.80684601263794880809241401537436936_dp
  t_lf(3227)=  22849.88615420468991063157049195209207563_dp
  t_lf(3228)=  22857.96577223407975004802015461242619159_dp
  t_lf(3229)=  22866.04570000483802526321967722002303832_dp
  t_lf(3230)=  22874.12593742105472748749623072446838260_dp
  t_lf(3231)=  22882.20648438687922574333481408391811398_dp
  t_lf(3232)=  22890.28734080652021174130517295812198598_dp
  t_lf(3233)=  22898.36850658424564482420103120057665850_dp
  t_lf(3234)=  22906.44998162438269697928615762380884953_dp
  t_lf(3235)=  22914.53176583131769791854198617269257758_dp
  t_lf(3236)=  22922.61385910949608022681170287829551590_dp
  t_lf(3237)=  22930.69626136342232457773590778007473066_dp
  t_lf(3238)=  22938.77897249765990501737515439834497474_dp
  t_lf(3239)=  22946.86199241683123431541486331286379325_dp
  t_lf(3240)=  22954.94532102561760938384829995815461341_dp
  t_lf(3241)=  22963.02895822875915676303349988285650308_dp
  t_lf(3242)=  22971.11290393105477817502021743997928126_dp
  t_lf(3243)=  22979.19715803736209614404316617848417679_dp
  t_lf(3244)=  22987.28172045259739968407801109512943041_dp
  t_lf(3245)=  22995.36659108173559005335676438004044497_dp
  t_lf(3246)=  23003.45176982981012657573942735100579738_dp
  t_lf(3247)=  23011.53725660191297252883891192108129106_dp
  t_lf(3248)=  23019.62305130319454109879646518271908867_dp
  t_lf(3249)=  23027.70915383886364140160501052033984715_dp
  t_lf(3250)=  23035.79556411418742457087800808304190524_dp
  t_lf(3251)=  23043.88228203449132991196162646099937588_dp
  t_lf(3252)=  23051.96930750515903112228820601404411676_dp
  t_lf(3253)=  23060.05664043163238257786918249995585718_dp
  t_lf(3254)=  23068.14428071941136568582582744409834656_dp
  t_lf(3255)=  23076.23222827405403530285634908223259684_dp
  t_lf(3256)=  23084.32048300117646621953808469560370301_dp
  t_lf(3257)=  23092.40904480645269971036370174272518663_dp
  t_lf(3258)=  23100.49791359561469014941051137666142354_dp
  t_lf(3259)=  23108.58708927445225169154218372101887059_dp
  t_lf(3260)=  23116.67657174881300501904233966328216124_dp
  t_lf(3261)=  23124.76636092460232415357967891155064880_dp
  t_lf(3262)=  23132.85645670778328333340448865112089258_dp
  t_lf(3263)=  23140.94685900437660395567656133169446443_dp
  t_lf(3264)=  23149.03756772046060158382473391523917607_dp
  t_lf(3265)=  23157.12858276217113301983844431966358880_dp
  t_lf(3266)=  23165.21990403570154344139188380544499558_dp
  t_lf(3267)=  23173.31153144730261360370150667214283129_dp
  t_lf(3268)=  23181.40346490328250710601784086029288500_dp
  t_lf(3269)=  23189.49570431000671772265272489247033372_dp
  t_lf(3270)=  23197.58824957389801679844327803628642528_dp
  t_lf(3271)=  23205.68110060143640070855409163269691894_dp
  t_lf(3272)=  23213.77425729915903838251931020619983473_dp
  t_lf(3273)=  23221.86771957366021889242645126023274451_dp
  t_lf(3274)=  23229.96148733159129910514399256229023188_dp
  t_lf(3275)=  23238.05556047966065139849493523991213305_dp
  t_lf(3276)=  23246.14993892463361144127873014168203593_dp
  t_lf(3277)=  23254.24462257333242603704413366765996921_dp
  t_lf(3278)=  23262.33961133263620103151573764218739557_dp
  t_lf(3279)=  23270.43490510948084928357709578967810541_dp
  t_lf(3280)=  23278.53050381085903869971354698177440691_dp
  t_lf(3281)=  23286.62640734382014033181801265303059221_dp
  t_lf(3282)=  23294.72261561547017653826322263300895523_dp
  t_lf(3283)=  23302.81912853297176920814400011625903744_dp
  t_lf(3284)=  23310.91594600354408804859341258901715187_dp
  t_lf(3285)=  23319.01306793446279893507677125352693349_dp
  t_lf(3286)=  23327.11049423306001232456763683855652504_dp
  t_lf(3287)=  23335.20822480672423173151016465888537005_dp
  t_lf(3288)=  23343.30625956290030226647229638816229085_dp
  t_lf(3289)=  23351.40459840908935923739448023950293487_dp
  t_lf(3290)=  23359.50324125284877681333877510740265252_dp
  t_lf(3291)=  23367.60218800179211675064336771389182801_dp
  t_lf(3292)=  23375.70143856358907718138770492225355929_dp
  t_lf(3293)=  23383.80099284596544146407361613395485567_dp
  t_lf(3294)=  23391.90085075670302709642797306960622410_dp
  t_lf(3295)=  23400.00101220363963469023260625365223584_dp
  t_lf(3296)=  23408.10147709466899700808736917599655950_dp
  t_lf(3297)=  23416.20224533774072806201241239276574019_dp
  t_lf(3298)=  23424.30331684086027227379590075380100500_dp
  t_lf(3299)=  23432.40469151208885369699357750711847583_dp
  t_lf(3300)=  23440.50636925954342530048674923137485925_dp
  t_lf(3301)=  23448.60834999139661831350543538719505120_dp
  t_lf(3302)=  23456.71063361587669163202359575793483876_dp
  t_lf(3303)=  23464.81322004126748128643351817093831934_dp
  t_lf(3304)=  23472.91610917590834997040661765247572640_dp
  t_lf(3305)=  23481.01930092819413663084806657418062131_dp
  t_lf(3306)=  23489.12279520657510611885284339681108873_dp
  t_lf(3307)=  23497.22659191955689890157095530940051158_dp
  t_lf(3308)=  23505.33069097570048083488975739920020515_dp
  t_lf(3309)=  23513.43509228362209299684145797110682113_dp
  t_lf(3310)=  23521.53979575199320158164406626536782215_dp
  t_lf(3311)=  23529.64480128954044785428420510012197821_dp
  t_lf(3312)=  23537.75010880504559816555037689160993201_dp
  t_lf(3313)=  23545.85571820734549402742543708053129096_dp
  t_lf(3314)=  23553.96162940533200224874719421887599580_dp
  t_lf(3315)=  23562.06784230795196513104622084846315428_dp
  t_lf(3316)=  23570.17435682420715072447012383122208940_dp
  t_lf(3317)=  23578.28117286315420314370368697278772361_dp
  t_lf(3318)=  23586.38829033390459294379446261609301654_dp
  t_lf(3319)=  23594.49570914562456755579355237116013330_dp
  t_lf(3320)=  23602.60342920753510178212148029205222110_dp
  t_lf(3321)=  23610.71145042891184835156922461277973150_dp
  t_lf(3322)=  23618.81977271908508853384463661168751366_dp
  t_lf(3323)=  23626.92839598743968281357463728930754345_dp
  t_lf(3324)=  23635.03732014341502162367374431867103313_dp
  t_lf(3325)=  23643.14654509650497613798964316045444168_dp
  t_lf(3326)=  23651.25607075625784912313667732890601139_dp
  t_lf(3327)=  23659.36589703227632584942829354908010730_dp
  t_lf(3328)=  23667.47602383421742506081963796231083845_dp
  t_lf(3329)=  23675.58645107179245000377165961589699612_dp
  t_lf(3330)=  23683.69717865476693951494823721545785890_dp
  t_lf(3331)=  23691.80820649296061916765800452516230222_dp
  t_lf(3332)=  23699.91953449624735247695270887283814135_dp
  t_lf(3333)=  23708.03116257455509216329409595463878551_dp
  t_lf(3334)=  23716.14309063786583147470147253828196812_dp
  t_lf(3335)=  23724.25531859621555556729225673568026348_dp
  t_lf(3336)=  23732.36784635969419294412798325585286267_dp
  t_lf(3337)=  23740.48067383844556695227838845813807048_dp
  t_lf(3338)=  23748.59380094266734733801635710470946460_dp
  t_lf(3339)=  23756.70722758261100186005666946102675351_dp
  t_lf(3340)=  23764.82095366858174796075164381391407365_dp
  t_lf(3341)=  23772.93497911093850449515692557024064771_dp
  t_lf(3342)=  23781.04930382009384351788082986546613174_dp
  t_lf(3343)=  23789.16392770651394212763080005138824540_dp
  t_lf(3344)=  23797.27885068071853436937069954707393451_dp
  t_lf(3345)=  23805.39407265328086319400280932694578788_dp
  t_lf(3346)=  23813.50959353482763247548855778610905327_dp
  t_lf(3347)=  23821.62541323603895908532216386601561847_dp
  t_lf(3348)=  23829.74153166764832502427152814424190593_dp
  t_lf(3349)=  23837.85794874044252961130086009127786344_dp
  t_lf(3350)=  23845.97466436526164172958968287555213804_dp
  t_lf(3351)=  23854.09167845299895212956300995622005360_dp
  t_lf(3352)=  23862.20899091460092578884764024228007369_dp
  t_lf(3353)=  23870.32660166106715432906967081712287004_dp
  t_lf(3354)=  23878.44451060345030848940847813041473795_dp
  t_lf(3355)=  23886.56271765285609065682257014503166964_dp
  t_lf(3356)=  23894.68122272044318745286286319634764780_dp
  t_lf(3357)=  23902.80002571742322237698908827529436335_dp
  t_lf(3358)=  23910.91912655506070850630518208600127861_dp
  t_lf(3359)=  23919.03852514467300125162966855424442450_dp
  t_lf(3360)=  23927.15822139763025116981718647512720041_dp
  t_lf(3361)=  23935.27821522535535683224746868813240129_dp
  t_lf(3362)=  23943.39850653932391774939822755566539303_dp
  t_lf(3363)=  23951.51909525106418735141855059819546990_dp
  t_lf(3364)=  23959.63998127215702602461955890583564625_dp
  t_lf(3365)=  23967.76116451423585420379922940341817209_dp
  t_lf(3366)=  23975.88264488898660552031843019455965590_dp
  t_lf(3367)=  23984.00442230814768000584536605059960140_dp
  t_lf(3368)=  23992.12649668350989735168577864337122858_dp
  t_lf(3369)=  24000.24886792691645022361639334725350863_dp
  t_lf(3370)=  24008.37153595026285763213925135658630405_dp
  t_lf(3371)=  24016.49450066549691835807471248003233170_dp
  t_lf(3372)=  24024.61776198461866443341106028456438248_dp
  t_lf(3373)=  24032.74131981968031467732878726916592934_dp
  t_lf(3374)=  24040.86517408278622828731778345277810137_dp
  t_lf(3375)=  24048.98932468609285848530579716322424140_dp
  t_lf(3376)=  24057.11377154180870621871668191451122779_dp
  t_lf(3377)=  24065.23851456219427391637708805975884938_dp
  t_lf(3378)=  24073.36355365956201929919040240675729263_dp
  t_lf(3379)=  24081.48888874627630924549688318350885149_dp
  t_lf(3380)=  24089.61451973475337371103908164278202948_dp
  t_lf(3381)=  24097.74044653746125970345178519840111076_dp
  t_lf(3382)=  24105.86666906691978531119586029241699009_dp
  t_lf(3383)=  24113.99318723570049378685551620215865382_dp
  t_lf(3384)=  24122.12000095642660768471865371015040865_dp
  t_lf(3385)=  24130.24711014177298305256010497869710831_dp
  t_lf(3386)=  24138.37451470446606367754771309528572172_dp
  t_lf(3387)=  24146.50221455728383538619134158552225195_dp
  t_lf(3388)=  24154.63020961305578039825504572781204341_dp
  t_lf(3389)=  24162.75849978466283173455277874909084713_dp
  t_lf(3390)=  24170.88708498503732767854814693431375872_dp
  t_lf(3391)=  24179.01596512716296629167886834479757626_dp
  t_lf(3392)=  24187.14514012407475998232673021257569344_dp
  t_lf(3393)=  24195.27460988885899012835398016034797506_dp
  t_lf(3394)=  24203.40437433465316175312722619007396977_dp
  t_lf(3395)=  24211.53443337464595825495005988844729885_dp
  t_lf(3396)=  24219.66478692207719618982575651508131790_dp
  t_lf(3397)=  24227.79543489023778010747154456990857585_dp
  t_lf(3398)=  24235.92637719246965744050607608072479639_dp
  t_lf(3399)=  24244.05761374216577344673186721066589221_dp
  t_lf(3400)=  24252.18914445277002620443461685936592373_dp
  t_lf(3401)=  24260.32096923777722166062144872127517945_dp
  t_lf(3402)=  24268.45308801073302873212025977078916427_dp
  t_lf(3403)=  24276.58550068523393445946249536711794231_dp
  t_lf(3404)=  24284.71820717492719921347180811287593873_dp
  t_lf(3405)=  24292.85120739351081195448119425985814912_dp
  t_lf(3406)=  24300.98450125473344554410133783405117018_dp
  t_lf(3407)=  24309.11808867239441210946302875026624005_dp
  t_lf(3408)=  24317.25196956034361845985665700553450612_dp
  t_lf(3409)=  24325.38614383248152155569192058022823136_dp
  t_lf(3410)=  24333.52061140275908402970101993742008207_dp
  t_lf(3411)=  24341.65537218517772976030874699491876139_dp
  t_lf(3412)=  24349.79042609378929949709301115137408865_dp
  t_lf(3413)=  24357.92577304269600653825947937847748713_dp
  t_lf(3414)=  24366.06141294605039246005414154624232788_dp
  t_lf(3415)=  24374.19734571805528289803774602827857747_dp
  t_lf(3416)=  24382.33357127296374338014618423952190165_dp
  t_lf(3417)=  24390.47008952507903521146103609068127664_dp
  t_lf(3418)=  24398.60690038875457141061462140237205607_dp
  t_lf(3419)=  24406.74400377839387269775403510814244676_dp
  t_lf(3420)=  24414.88139960845052353398877659001789248_dp
  t_lf(3421)=  24423.01908779342812821224671573341570406_dp
  t_lf(3422)=  24431.15706824788026699946327026095548986_dp
  t_lf(3423)=  24439.29534088641045233002880060744194517_dp
  t_lf(3424)=  24447.43390562367208505041936003175610501_dp
  t_lf(3425)=  24455.57276237436841071493606882618834568_dp
  t_lf(3426)=  24463.71191105325247593247851238050867712_dp
  t_lf(3427)=  24471.85135157512708476427769348742299372_dp
  t_lf(3428)=  24479.99108385484475517251419963863209544_dp
  t_lf(3429)=  24488.13110780730767551974737615711596495_dp
  t_lf(3430)=  24496.27142334746766111908142584212986835_dp
  t_lf(3431)=  24504.41203039032611083499448536934058463_dp
  t_lf(3432)=  24512.55292885093396373475685799016808794_dp
  t_lf(3433)=  24520.69411864439165579036471111234631154_dp
  t_lf(3434)=  24528.83559968584907663091567611859060380_dp
  t_lf(3435)=  24536.97737189050552634535291629267192546_dp
  t_lf(3436)=  24545.11943517360967233550435697275990816_dp
  t_lf(3437)=  24553.26178945045950621934390004121817255_dp
  t_lf(3438)=  24561.40443463640230078440157258872376767_dp
  t_lf(3439)=  24569.54737064683456699124968705924463014_dp
  t_lf(3440)=  24577.69059739720201102699221739164937108_dp
  t_lf(3441)=  24585.83411480299949140868472262414570116_dp
  t_lf(3442)=  24593.97792277977097613661227611994903819_dp
  t_lf(3443)=  24602.12202124310949989735298500717137784_dp
  t_lf(3444)=  24610.26641010865712131655481060349084400_dp
  t_lf(3445)=  24618.41108929210488026135352651731140801_dp
  t_lf(3446)=  24626.55605870919275519235977678244545221_dp
  t_lf(3447)=  24634.70131827570962056514332179344297387_dp
  t_lf(3448)=  24642.84686790749320428114268496414255383_dp
  t_lf(3449)=  24650.99270752043004518792853793342147419_dp
  t_lf(3450)=  24659.13883703045545062874928679006474786_dp
  t_lf(3451)=  24667.28525635355345404128744618374296573_dp
  t_lf(3452)=  24675.43196540575677260555551233187289190_dp
  t_lf(3453)=  24683.57896410314676494086016982321723018_dp
  t_lf(3454)=  24691.72625236185338885176379075904401130_dp
  t_lf(3455)=  24699.87383009805515912297230816209315411_dp
  t_lf(3456)=  24708.02169722797910536307866872306796720_dp
  t_lf(3457)=  24716.16985366790072989709119284446119867_dp
  t_lf(3458)=  24724.31829933414396570767629258281572711_dp
  t_lf(3459)=  24732.46703414308113442504512048358462769_dp
  t_lf(3460)=  24740.61605801113290436541384444816815994_dp
  t_lf(3461)=  24748.76537085476824861796736567103873070_dp
  t_lf(3462)=  24756.91497259050440318025641833669012138_dp
  t_lf(3463)=  24765.06486313490682514195811117203378187_dp
  t_lf(3464)=  24773.21504240458915091693009211038085554_dp
  t_lf(3465)=  24781.36551031621315452348863823886040362_dp
  t_lf(3466)=  24789.51626678648870591284109387259716588_dp
  t_lf(3467)=  24797.66731173217372934560320002676978404_dp
  t_lf(3468)=  24805.81864507007416181633197874235491595_dp
  t_lf(3469)=  24813.97026671704391152600495566349481724_dp
  t_lf(3470)=  24822.12217658998481640237662396456504098_dp
  t_lf(3471)=  24830.27437460584660266814317218372273451_dp
  t_lf(3472)=  24838.42686068162684345684661773754097772_dp
  t_lf(3473)=  24846.57963473437091747644960686883565013_dp
  t_lf(3474)=  24854.73269668117196772051226051752193687_dp
  t_lf(3475)=  24862.88604643917086022690256410284985593_dp
  t_lf(3476)=  24871.03968392555614288397191746521274929_dp
  t_lf(3477)=  24879.19360905756400428412757923744874328_dp
  t_lf(3478)=  24887.34782175247823262473385769971054072_dp
  t_lf(3479)=  24895.50232192763017465627401771910993427_dp
  t_lf(3480)=  24903.65710950039869467770499068599508893_dp
  t_lf(3481)=  24911.81218438821013357893709143343448418_dp
  t_lf(3482)=  24919.96754650853826793037106296580358094_dp
  t_lf(3483)=  24928.12319577890426911942488642683953079_dp
  t_lf(3484)=  24936.27913211687666253398291010768492528_dp
  t_lf(3485)=  24944.43535544007128679269996743182164705_dp
  t_lf(3486)=  24952.59186566615125302209326975693689950_dp
  t_lf(3487)=  24960.74866271282690418035497550420064011_dp
  t_lf(3488)=  24968.90574649785577442781845256370072460_dp
  t_lf(3489)=  24977.06311693904254854401136613141151006_dp
  t_lf(3490)=  24985.22077395423902139122883910859451398_dp
  t_lf(3491)=  24993.37871746134405742456004693947566568_dp
  t_lf(3492)=  25001.53694737830355024830172327794103194_dp
  t_lf(3493)=  25009.69546362311038221869216715936860469_dp
  t_lf(3494)=  25017.85426611380438409289945641009340265_dp
  t_lf(3495)=  25026.01335476847229472419768585491100347_dp
  t_lf(3496)=  25034.17272950524772080326516248298358283_dp
  t_lf(3497)=  25042.33239024231109664553860310504413855_dp
  t_lf(3498)=  25050.49233689788964402455749318041902790_dp
  t_lf(3499)=  25058.65256939025733205123287841162511101_dp
  t_lf(3500)=  25066.81308763773483709897497339766320680_dp
  t_lf(3501)=  25074.97389155868950277461408410514043086_dp
  t_lf(3502)=  25083.13498107153529993504945315952517478_dp
  t_lf(3503)=  25091.29635609473278674956074897768355366_dp
  t_lf(3504)=  25099.45801654678906880771703155787732821_dp
  t_lf(3505)=  25107.61996234625775927281813931513151643_dp
  t_lf(3506)=  25115.78219341173893908080355269881475383_dp
  t_lf(3507)=  25123.94470966187911718456390145592523863_dp
  t_lf(3508)=  25132.10751101537119084359039330844680250_dp
  t_lf(3509)=  25140.27059739095440595889755249673897271_dp
  t_lf(3510)=  25148.43396870741431745315476710375623264_dp
  t_lf(3511)=  25156.59762488358274969596225431745814944_dp
  t_lf(3512)=  25164.76156583833775697420716281157543285_dp
  t_lf(3513)=  25172.92579149060358400743564122843784619_dp
  t_lf(3514)=  25181.09030175935062650817681133234745573_dp
  t_lf(3515)=  25189.25509656359539178715469376849294407_dp
  t_lf(3516)=  25197.42017582240045940332424351114432110_dp
  t_lf(3517)=  25205.58553945487444185866776101633776246_dp
  t_lf(3518)=  25213.75118738017194533768805380895164478_dp
  t_lf(3519)=  25221.91711951749353049153483173248001885_dp
  t_lf(3520)=  25230.08333578608567326670092737242039519_dp
  t_lf(3521)=  25238.24983610524072577822504123149918496_dp
  t_lf(3522)=  25246.41662039429687722733781908744955991_dp
  t_lf(3523)=  25254.58368857263811486348817660222074055_dp
  t_lf(3524)=  25262.75104055969418499068689367582141212_dp
  t_lf(3525)=  25270.91867627494055401810460824896848681_dp
  t_lf(3526)=  25279.08659563789836955486144625680991800_dp
  t_lf(3527)=  25287.25479856813442154894563122169963459_dp
  t_lf(3528)=  25295.42328498526110347019852354680557516_dp
  t_lf(3529)=  25303.59205480893637353730364593470870779_dp
  t_lf(3530)=  25311.76110795886371598871735750658104200_dp
  t_lf(3531)=  25319.93044435479210239747894513849197400_dp
  t_lf(3532)=  25328.10006391651595302983800626236163470_dp
  t_lf(3533)=  25336.26996656387509824763710290053279816_dp
  t_lf(3534)=  25344.44015221675473995438777201534371118_dp
  t_lf(3535)=  25352.61062079508541308497808235892606771_dp
  t_lf(3536)=  25360.78137221884294713895003290419721912_dp
  t_lf(3537)=  25368.95240640804842775728519262613432213_dp
  t_lf(3538)=  25377.12372328276815834263708588338002505_dp
  t_lf(3539)=  25385.29532276311362172294893192450282918_dp
  t_lf(3540)=  25393.46720476924144185839545111128759328_dp
  t_lf(3541)=  25401.63936922135334559158755431372874690_dp
  t_lf(3542)=  25409.81181603969612444097883558840542862_dp
  t_lf(3543)=  25417.98454514456159643741289170409757472_dp
  t_lf(3544)=  25426.15755645628656800375059532631738013_dp
  t_lf(3545)=  25434.33084989525279587751655171634279087_dp
  t_lf(3546)=  25442.50442538188694907650407164080884983_dp
  t_lf(3547)=  25450.67828283666057090727809582539771969_dp
  t_lf(3548)=  25458.85242218009004101651560872112680158_dp
  t_lf(3549)=  25467.02684333273653748512318158462314422_dp
  t_lf(3550)=  25475.20154621520599896507138690504673017_dp
  t_lf(3551)=  25483.37653074814908685888592804043951069_dp
  t_lf(3552)=  25491.55179685226114754173542955568436640_dp
  t_lf(3553)=  25499.72734444828217462605593518341047577_dp
  t_lf(3554)=  25507.90317345699677126865226155852971303_dp
  t_lf(3555)=  25516.07928379923411252021645690708236300_dp
  t_lf(3556)=  25524.25567539586790771720371470115818826_dp
  t_lf(3557)=  25532.43234816781636291600619292428813279_dp
  t_lf(3558)=  25540.60930203604214336936529002631898370_dp
  t_lf(3559)=  25548.78653692155233604496302888383329765_dp
  t_lf(3560)=  25556.96405274539841218613330012210386421_dp
  t_lf(3561)=  25565.14184942867618991463381599781883610_dp
  t_lf(3562)=  25573.31992689252579687541972568882219666_dp
  t_lf(3563)=  25581.49828505813163292335994228832513504_dp
  t_lf(3564)=  25589.67692384672233285183733105689672029_dp
  t_lf(3565)=  25597.85584317957072916317400754647546126_dp
  t_lf(3566)=  25606.03504297799381488082309307709425432_dp
  t_lf(3567)=  25614.21452316335270640326837371941609922_dp
  t_lf(3568)=  25622.39428365705260639957340741497194551_dp
  t_lf(3569)=  25630.57432438054276674652172215160916300_dp
  t_lf(3570)=  25638.75464525531645150728984620453235918_dp
  t_lf(3571)=  25646.93524620291089995159500935387945368_dp
  t_lf(3572)=  25655.11612714490728961725945169845583406_dp
  t_lf(3573)=  25663.29728800293069941313337420247774476_dp
  t_lf(3574)=  25671.47872869865007276331866243838140410_dp
  t_lf(3575)=  25679.66044915377818079263561212436422728_dp
  t_lf(3576)=  25687.84244929007158555327498200076540318_dp
  t_lf(3577)=  25696.02472902933060329257779634509030521_dp
  t_lf(3578)=  25704.20728829339926776188541599186111494_dp
  t_lf(3579)=  25712.39012700416529356640249310095783943_dp
  t_lf(3580)=  25720.57324508356003955601552110712178238_dp
  t_lf(3581)=  25728.75664245355847225700978728424859747_dp
  t_lf(3582)=  25736.94031903617912934462763117142036265_dp
  t_lf(3583)=  25745.12427475348408315641100773373466628_dp
  t_lf(3584)=  25753.30850952757890424627144957030143512_dp
  t_lf(3585)=  25761.49302328061262497923061773471205756_dp
  t_lf(3586)=  25769.67781593477770316677472580025611331_dp
  t_lf(3587)=  25777.86288741230998574276621668358351996_dp
  t_lf(3588)=  25786.04823763548867247985616643679791353_dp
  t_lf(3589)=  25794.23386652663627974634098372953332240_dp
  t_lf(3590)=  25802.41977400811860430340706806982236659_dp
  t_lf(3591)=  25810.60596000234468714270718395592097697_dp
  t_lf(3592)=  25818.79242443176677736421240211112161668_dp
  t_lf(3593)=  25826.97916721888029609428355273037280348_dp
  t_lf(3594)=  25835.16618828622380044390622926163496244_dp
  t_lf(3595)=  25843.35348755637894750703347465674784358_dp
  t_lf(3596)=  25851.54106495197045839898037525656846554_dp
  t_lf(3597)=  25859.72892039566608233481488052366532219_dp
  t_lf(3598)=  25867.91705381017656074768925970332793394_dp
  t_lf(3599)=  25876.10546511825559144705669918047324631_dp
  t_lf(3600)=  25884.29415424269979281671763680660337927_dp
  t_lf(3601)=  25892.48312110634866805264052179769330872_dp
  t_lf(3602)=  25900.67236563208456944050178095116171553_dp
  t_lf(3603)=  25908.86188774283266267288986389830196697_dp
  t_lf(3604)=  25917.05168736156089120611833189812050441_dp
  t_lf(3605)=  25925.24176441127994065659304628984331260_dp
  t_lf(3606)=  25933.43211881504320323667860415480316100_dp
  t_lf(3607)=  25941.62275049594674223000925999440547243_dp
  t_lf(3608)=  25949.81365937712925650618966330978253902_dp
  t_lf(3609)=  25958.00484538177204507483083287097694017_dp
  t_lf(3610)=  25966.19630843309897167886687918943701145_dp
  t_lf(3611)=  25974.38804845437642942709807725765068281_dp
  t_lf(3612)=  25982.58006536891330546590598199427858842_dp
  t_lf(3613)=  25990.77235910006094569008636903256172080_dp
  t_lf(3614)=  25998.96492957121311949274587351446075975_dp
  t_lf(3615)=  26007.15777670580598455420828940332028883_dp
  t_lf(3616)=  26015.35090042731805166987658150422719334_dp
  t_lf(3617)=  26023.54430065927014961699675188403342783_dp
  t_lf(3618)=  26031.73797732522539006026979171262290427_dp
  t_lf(3619)=  26039.93193034878913249625803870380338701_dp
  t_lf(3620)=  26048.12615965360894923653234931857893671_dp
  t_lf(3621)=  26056.32066516337459042950658370588762304_dp
  t_lf(3622)=  26064.51544680181794912090598999655297906_dp
  t_lf(3623)=  26072.71050449271302635281616303557510753_dp
  t_lf(3624)=  26080.90583815987589630125934093635663923_dp
  t_lf(3625)=  26089.10144772716467145224489096839711376_dp
  t_lf(3626)=  26097.29733311847946781624092424777309695_dp
  t_lf(3627)=  26105.49349425776237018101406648772582417_dp
  t_lf(3628)=  26113.68993106899739740278449968527779220_dp
  t_lf(3629)=  26121.88664347621046773564347706936801288_dp
  t_lf(3630)=  26130.08363140346936419918060091690516139_dp
  t_lf(3631)=  26138.28089477488369998426823995576024892_dp
  t_lf(3632)=  26146.47843351460488389695055001842645224_dp
  t_lf(3633)=  26154.67624754682608584038464838723314914_dp
  t_lf(3634)=  26162.87433679578220233478157888198293275_dp
  t_lf(3635)=  26171.07270118574982207529479118405238828_dp
  t_lf(3636)=  26179.27134064104719152780394416772678198_dp
  t_lf(3637)=  26187.47025508603418056254192912019169345_dp
  t_lf(3638)=  26195.66944444511224812551309467654627356_dp
  t_lf(3639)=  26203.86890864272440794765074107579758170_dp
  t_lf(3640)=  26212.06864760335519429166203695840679738_dp
  t_lf(3641)=  26220.26866125153062773650859737594856349_dp
  t_lf(3642)=  26228.46894951181818099947104696917595753_dp
  t_lf(3643)=  26236.66951230882674479574597739261636330_dp
  t_lf(3644)=  26244.87034956720659373552379302211770023_dp
  t_lf(3645)=  26253.07146121164935225849602377687904210_dp
  t_lf(3646)=  26261.27284716688796060574076851979271639_dp
  t_lf(3647)=  26269.47450735769664082893501696975373258_dp
  t_lf(3648)=  26277.67644170889086283684268236731317123_dp
  t_lf(3649)=  26285.87865014532731047902726128102042804_dp
  t_lf(3650)=  26294.08113259190384766673812092636952246_dp
  t_lf(3651)=  26302.28388897355948453091949819279074919_dp
  t_lf(3652)=  26310.48691921527434361729137823696360208_dp
  t_lf(3653)=  26318.69022324206962611845150400322209262_dp
  t_lf(3654)=  26326.89380097900757814294785137433041194_dp
  t_lf(3655)=  26335.09765235119145702127098783877557315_dp
  t_lf(3656)=  26343.30177728376549764871581558430358639_dp
  t_lf(3657)=  26351.50617570191487886506228279206536598_dp
  t_lf(3658)=  26359.71084753086568987102472961178559738_dp
  t_lf(3659)=  26367.91579269588489668141961784616899221_dp
  t_lf(3660)=  26376.12101112228030861500147576265967344_dp
  t_lf(3661)=  26384.32650273540054482091697168301594846_dp
  t_lf(3662)=  26392.53226746063500084172711207629868661_dp
  t_lf(3663)=  26400.73830522341381521294764179914031449_dp
  t_lf(3664)=  26408.94461594920783609905780588890562908_dp
  t_lf(3665)=  26417.15119956352858796592771392091691502_dp
  t_lf(3666)=  26425.35805599192823828961462939063510961_dp
  t_lf(3667)=  26433.56518515999956430147858787590602191_dp
  t_lf(3668)=  26441.77258699337591976956782887343508067_dp
  t_lf(3669)=  26449.98026141773120181622460718788412695_dp
  t_lf(3670)=  26458.18820835877981777186203058172692033_dp
  t_lf(3671)=  26466.39642774227665206486265106959300146_dp
  t_lf(3672)=  26474.60491949401703314754961776260823482_dp
  t_lf(3673)=  26482.81368353983670045818127953653980380_dp
  t_lf(3674)=  26491.02271980561177141892020601270788904_dp
  t_lf(3675)=  26499.23202821725870846972767540296914948_dp
  t_lf(3676)=  26507.44160870073428613813475767994105297_dp
  t_lf(3677)=  26515.65146118203555814484120129135285785_dp
  t_lf(3678)=  26523.86158558719982454509341124330961106_dp
  t_lf(3679)=  26532.07198184230459890579288583167006987_dp
  t_lf(3680)=  26540.28264987346757551828655860399733647_dp
  t_lf(3681)=  26548.49358960684659664679057128697077396_dp
  t_lf(3682)=  26556.70480096863961981239908241607720235_dp
  t_lf(3683)=  26564.91628388508468511262979525615540635_dp
  t_lf(3684)=  26573.12803828245988257645796730327677748_dp
  t_lf(3685)=  26581.34006408708331955479074221083181717_dp
  t_lf(3686)=  26589.55236122531308814633372338588180899_dp
  t_lf(3687)=  26597.76492962354723265880178675615099574_dp
  t_lf(3688)=  26605.97776920822371710542620831380005199_dp
  t_lf(3689)=  26614.19087990582039273671025999965871546_dp
  t_lf(3690)=  26622.40426164285496560738550530122553867_dp
  t_lf(3691)=  26630.61791434588496417852110359978646582_dp
  t_lf(3692)=  26638.83183794150770695473850981678117166_dp
  t_lf(3693)=  26647.04603235636027015648403327737587740_dp
  t_lf(3694)=  26655.26049751711945542731179693040196941_dp
  t_lf(3695)=  26663.47523335050175757612971513870869538_dp
  t_lf(3696)=  26671.69023978326333235436118518287223639_dp
  t_lf(3697)=  26679.90551674219996426797526440441851395_dp
  t_lf(3698)=  26688.12106415414703442433818155256838458_dp
  t_lf(3699)=  26696.33688194597948841383910739131582709_dp
  t_lf(3700)=  26704.55297004461180422624318597171600267_dp
  t_lf(3701)=  26712.76932837699796020172490417790356543_dp
  t_lf(3702)=  26720.98595687013140301653495321489445853_dp
  t_lf(3703)=  26729.20285545104501570325381162195803050_dp
  t_lf(3704)=  26737.42002404681108570558535516759127014_dp
  t_lf(3705)=  26745.63746258454127296764387461119315563_dp
  t_lf(3706)=  26753.85517099138657805768795780273366038_dp
  t_lf(3707)=  26762.07314919453731032625476793534722090_dp
  t_lf(3708)=  26770.29139712122305609864832496716206626_dp
  t_lf(3709)=  26778.50991469871264690173547228811160717_dp
  t_lf(3710)=  26786.72870185431412772500328562526820888_dp
  t_lf(3711)=  26794.94775851537472531583175595669851197_dp
  t_lf(3712)=  26803.16708460928081650893565283926765955_dp
  t_lf(3713)=  26811.38668006345789658992954905052124418_dp
  t_lf(3714)=  26819.60654480537054769297006179905167018_dp
  t_lf(3715)=  26827.82667876252240723242943997191237071_dp
  t_lf(3716)=  26836.04708186245613636855470096198062427_dp
  t_lf(3717)=  26844.26775403275338850706659455298855303_dp
  t_lf(3718)=  26852.48869520103477783265274513554249611_dp
  t_lf(3719)=  26860.70990529495984787630939718413285059_dp
  t_lf(3720)=  26868.93138424222704011648626244319845085_dp
  t_lf(3721)=  26877.15313197057366261398904065004968034_dp
  t_lf(3722)=  26885.37514840777585868059425886417012597_dp
  t_lf(3723)=  26893.59743348164857558133114757640432078_dp
  t_lf(3724)=  26901.81998712004553327038534473809488691_dp
  t_lf(3725)=  26910.04280925085919316057929167965138268_dp
  t_lf(3726)=  26918.26589980202072692638425758060985818_dp
  t_lf(3727)=  26926.48925870149998534041900170927030383_dp
  t_lf(3728)=  26934.71288587730546714339015506977189758_dp
  t_lf(3729)=  26942.93678125748428794742947537827557426_dp
  t_lf(3730)=  26951.16094477012214917278320143806160260_dp
  t_lf(3731)=  26959.38537634334330701780880499610750853_dp
  t_lf(3732)=  26967.61007590531054146223451004137906964_dp
  t_lf(3733)=  26975.83504338422512530363702124793377067_dp
  t_lf(3734)=  26984.06027870832679322709297487429089721_dp
  t_lf(3735)=  26992.28578180589371090795969690465350671_dp
  t_lf(3736)=  27000.51155260524244414774092455776230540_dp
  t_lf(3737)=  27008.73759103472792804299321849570674172_dp
  t_lf(3738)=  27016.96389702274343618722886413820047295_dp
  t_lf(3739)=  27025.19047049772054990577113142793215593_dp
  t_lf(3740)=  27033.41731138812912752351783319991295254_dp
  t_lf(3741)=  27041.64441962247727366556919298254323689_dp
  t_lf(3742)=  27049.87179512931130859067610360069607440_dp
  t_lf(3743)=  27058.09943783721573755746492836174675891_dp
  t_lf(3744)=  27066.32734767481322022339506688444801176_dp
  t_lf(3745)=  27074.55552457076454007640557777814065613_dp
  t_lf(3746)=  27082.78396845376857389920722039628029533_dp
  t_lf(3747)=  27091.01267925256226126617634777393168529_dp
  t_lf(3748)=  27099.24165689592057407280715261401336716_dp
  t_lf(3749)=  27107.47090131265648609767883781194431316_dp
  t_lf(3750)=  27115.70041243162094259689435250322976599_dp
  t_lf(3751)=  27123.93019018170282993094740398370238180_dp
  t_lf(3752)=  27132.16023449182894522397452508788381398_dp
  t_lf(3753)=  27140.39054529096396605534904571752693488_dp
  t_lf(3754)=  27148.62112250811042018357388619011525493_dp
  t_lf(3755)=  27156.85196607230865530243015892620837650_dp
  t_lf(3756)=  27165.08307591263680882933863371530446569_dp
  t_lf(3757)=  27173.31445195821077772589119039261603737_dp
  t_lf(3758)=  27181.54609413818418835050945122409646886_dp
  t_lf(3759)=  27189.77800238174836634318785363448357647_dp
  t_lf(3760)=  27198.01017661813230654227849212331464409_dp
  t_lf(3761)=  27206.24261677660264293327512629708517483_dp
  t_lf(3762)=  27214.47532278646361862955381990224138564_dp
  t_lf(3763)=  27222.70829457705705588502774357378347751_dp
  t_lf(3764)=  27230.94153207776232613867374171818174094_dp
  t_lf(3765)=  27239.17503521799632009088833152733870752_dp
  t_lf(3766)=  27247.40880392721341781163086957273530094_dp
  t_lf(3767)=  27255.64283813490545888031168875594410408_dp
  t_lf(3768)=  27263.87713777060171255738307559364463474_dp
  t_lf(3769)=  27272.11170276386884798759102489239946589_dp
  t_lf(3770)=  27280.34653304431090443484577582101106031_dp
  t_lf(3771)=  27288.58162854156926154866920021654160211_dp
  t_lf(3772)=  27296.81698918532260966217718066430555895_dp
  t_lf(3773)=  27305.05261490528692012155518247260022744_dp
  t_lf(3774)=  27313.28850563121541564698529011988550341_dp
  t_lf(3775)=  27321.52466129289854072498304508582235498_dp
  t_lf(3776)=  27329.76108182016393203210248818829111201_dp
  t_lf(3777)=  27337.99776714287638888996787563649625164_dp
  t_lf(3778)=  27346.23471719093784375159060397578376500_dp
  t_lf(3779)=  27354.47193189428733271892994494310972623_dp
  t_lf(3780)=  27362.70941118290096609165625697346302385_dp
  t_lf(3781)=  27370.94715498679189894707540569721941459_dp
  t_lf(3782)=  27379.18516323601030175117319124664556311_dp
  t_lf(3783)=  27387.42343586064333100073864554683736914_dp
  t_lf(3784)=  27395.66197279081509989652512800252287518_dp
  t_lf(3785)=  27403.90077395668664904740821310764200437_dp
  t_lf(3786)=  27412.13983928845591720549942849968830072_dp
  t_lf(3787)=  27420.37916871635771203217496685571613059_dp
  t_lf(3788)=  27428.61876217066368089497855978193424573_dp
  t_lf(3789)=  27436.85861958168228169535776648417639461_dp
  t_lf(3790)=  27445.09874087975875372719299452251438673_dp
  t_lf(3791)=  27453.33912599527508856607863435011065329_dp
  t_lf(3792)=  27461.57977485865000098931575361434729619_dp
  t_lf(3793)=  27469.82068740033889992657586135756766899_dp
  t_lf(3794)=  27478.06186355083385944119531629567488733_dp
  t_lf(3795)=  27486.30330324066358974206001727559892609_dp
  t_lf(3796)=  27494.54500640039340822604007781751914040_dp
  t_lf(3797)=  27502.78697296062521055093425033496056909_dp
  t_lf(3798)=  27511.02920285199744173888392919571807719_dp
  t_lf(3799)=  27519.27169600518506731021662523924951468_dp
  t_lf(3800)=  27527.51445235089954444767886770196427320_dp
  t_lf(3801)=  27535.75747181988879319101855272096298757_dp
  t_lf(3802)=  27544.00075434293716766187682068950314968_dp
  t_lf(3803)=  27552.24429985086542731894960772401899105_dp
  t_lf(3804)=  27560.48810827453070824337907937315648120_dp
  t_lf(3805)=  27568.73217954482649445433521745423943846_dp
  t_lf(3806)=  27576.97651359268258925474789354210373580_dp
  t_lf(3807)=  27585.22111034906508660714982515956601233_dp
  t_lf(3808)=  27593.46596974497634253959087312817320143_dp
  t_lf(3809)=  27601.71109171145494658158420083255101823_dp
  t_lf(3810)=  27609.95647617957569323004487833187420053_dp
  t_lf(3811)=  27618.20212308044955344518157631795908852_dp
  t_lf(3812)=  27626.44803234522364617630205687146981033_dp
  t_lf(3813)=  27634.69420390508120991749322980597210083_dp
  t_lf(3814)=  27642.94063769124157429313660511430223931_dp
  t_lf(3815)=  27651.18733363496013167322003364318080833_dp
  t_lf(3816)=  27659.43429166752830881840668962042944721_dp
  t_lf(3817)=  27667.68151172027353855482231004478043604_dp
  t_lf(3818)=  27675.92899372455923147852176722134017918_dp
  t_lf(3819)=  27684.17673761178474768959611188651428118_dp
  t_lf(3820)=  27692.42474331338536855588128541485918910_dp
  t_lf(3821)=  27700.67301076083226850622976053712802457_dp
  t_lf(3822)=  27708.92153988563248685330643082396040491_dp
  t_lf(3823)=  27717.17033061932889964587012990346136312_dp
  t_lf(3824)=  27725.41938289350019155050222198355597849_dp
  t_lf(3825)=  27733.66869663976082776274376574172653172_dp
  t_lf(3826)=  27741.91827178976102594760281402576986359_dp
  t_lf(3827)=  27750.16810827518672820939347207978555984_dp
  t_lf(3828)=  27758.41820602775957309086839716995147959_dp
  t_lf(3829)=  27766.66856497923686760160648253499232007_dp
  t_lf(3830)=  27774.91918506141155927561752852682915453_dp
  t_lf(3831)=  27783.17006620611220825812576363794244095_dp
  t_lf(3832)=  27791.42120834520295942149413783371658727_dp
  t_lf(3833)=  27799.67261141058351451025137022068894583_dp
  t_lf(3834)=  27807.92427533418910431518379258542773361_dp
  t_lf(3835)=  27816.17620004799046087645408973393893748_dp
  t_lf(3836)=  27824.42838548399378971570909684827833574_dp
  t_lf(3837)=  27832.68083157424074209713887325564738615_dp
  t_lf(3838)=  27840.93353825080838731744933107590640541_dp
  t_lf(3839)=  27849.18650544580918502471075617637017442_dp
  t_lf(3840)=  27857.43973309139095756604461771818430031_dp
  t_lf(3841)=  27865.69322111973686236411112132673927579_dp
  t_lf(3842)=  27873.94696946306536432236001955968660057_dp
  t_lf(3843)=  27882.20097805363020825900725188040044646_dp
  t_lf(3844)=  27890.45524682372039136970004477240151401_dp
  t_lf(3845)=  27898.70977570566013571883316095154877930_dp
  t_lf(3846)=  27906.96456463180886075947904484793108176_dp
  t_lf(3847)=  27915.21961353456115588189466963857475688_dp
  t_lf(3848)=  27923.47492234634675299056794911554605236_dp
  t_lf(3849)=  27931.73049099963049910976663557198765103_dp
  t_lf(3850)=  27939.98631942691232901755268268130651191_dp
  t_lf(3851)=  27948.24240756072723790822511003234417150_dp
  t_lf(3852)=  27956.49875533364525408315446356612885305_dp
  t_lf(3853)=  27964.75536267827141166997202363794892923_dp
  t_lf(3854)=  27973.01222952724572337007696980221668710_dp
  t_lf(3855)=  27981.26935581324315323442476868712665589_dp
  t_lf(3856)=  27989.52674146897358946756010849167017790_dp
  t_lf(3857)=  27997.78438642718181725985776069936312277_dp
  t_lf(3858)=  28006.04229062064749164793480656129185813_dp
  t_lf(3859)=  28014.30045398218511040319772275599848171_dp
  t_lf(3860)=  28022.55887644464398694848787738552408125_dp
  t_lf(3861)=  28030.81755794090822330278904411582210627_dp
  t_lf(3862)=  28039.07649840389668305396059881595600145_dp
  t_lf(3863)=  28047.33569776656296435946011949421875840_dp
  t_lf(3864)=  28055.59515596189537297501916667076918965_dp
  t_lf(3865)=  28063.85487292291689531123607756578221978_dp
  t_lf(3866)=  28072.11484858268517151804966361966953507_dp
  t_lf(3867)=  28080.37508287429246859705775689785325486_dp
  t_lf(3868)=  28088.63557573086565354164460686707911371_dp
  t_lf(3869)=  28096.89632708556616650488118486354671406_dp
  t_lf(3870)=  28105.15733687158999399516250930542197671_dp
  t_lf(3871)=  28113.41860502216764209954616033378974709_dp
  t_lf(3872)=  28121.68013147056410973475620809701088853_dp
  t_lf(3873)=  28129.94191615007886192581683432397590762_dp
  t_lf(3874)=  28138.20395899404580311227998216210352328_dp
  t_lf(3875)=  28146.46625993583325048201142448632444341_dp
  t_lf(3876)=  28154.72881890884390733249969601592430232_dp
  t_lf(3877)=  28162.99163584651483645965238960720111132_dp
  t_lf(3878)=  28171.25471068231743357404437202162707846_dp
  t_lf(3879)=  28179.51804334975740074458252930179718012_dp
  t_lf(3880)=  28187.78163378237471986955170662110185918_dp
  t_lf(3881)=  28196.04548191374362617500656210798265081_dp
  t_lf(3882)=  28204.30958767747258174047410868202089350_dp
  t_lf(3883)=  28212.57395100720424905193177237717399280_dp
  t_lf(3884)=  28220.83857183661546458202584996741352337_dp
  t_lf(3885)=  28229.10345009941721239749530295203686360_dp
  t_lf(3886)=  28237.36858572935459779376587910222067453_dp
  t_lf(3887)=  28245.63397866020682095667960681716150785_dp
  t_lf(3888)=  28253.89962882578715065132476148760684265_dp
  t_lf(3889)=  28262.16553615994289793793145691691912328_dp
  t_lf(3890)=  28270.43170059655538991479806860523566014_dp
  t_lf(3891)=  28278.69812206953994348821374936098785212_dp
  t_lf(3892)=  28286.96480051284583916934235126622292877_dp
  t_lf(3893)=  28295.23173586045629489803312148802866362_dp
  t_lf(3894)=  28303.49892804638843989352359279809419282_dp
  t_lf(3895)=  28311.76637700469328853200014293624564158_dp
  t_lf(3896)=  28320.03408266945571425098175013187071679_dp
  t_lf(3897)=  28328.30204497479442348049252517968831217_dp
  t_lf(3898)=  28336.57026385486192960098865345352358586_dp
  t_lf(3899)=  28344.83873924384452692800543313381154735_dp
  t_lf(3900)=  28353.10747107596226472349014872166812008_dp
  t_lf(3901)=  28361.37645928546892123378657161473166424_dp
  t_lf(3902)=  28369.64570380665197775423693212778434065_dp
  t_lf(3903)=  28377.91520457383259272036725985460530968_dp
  t_lf(3904)=  28386.18496152136557582562204168677998383_dp
  t_lf(3905)=  28394.45497458363936216561419913048433487_dp
  t_lf(3906)=  28402.72524369507598640885643879377309978_dp
  t_lf(3907)=  28410.99576879013105699394008205481769364_dp
  t_lf(3908)=  28419.26654980329373035312753196605533789_dp
  t_lf(3909)=  28427.53758666908668516232458740051652637_dp
  t_lf(3910)=  28435.80887932206609661739886630488421323_dp
  t_lf(3911)=  28444.08042769682161073681065168929531363_dp
  t_lf(3912)=  28452.35223172797631869052252565671311891_dp
  t_lf(3913)=  28460.62429135018673115515420835506746428_dp
  t_lf(3914)=  28468.89660649814275269534907022346693521_dp
  t_lf(3915)=  28477.16917710656765617131883729982261157_dp
  t_lf(3916)=  28485.44200311021805717253306066137394322_dp
  t_lf(3917)=  28493.71508444388388847751997228206201444_dp
  t_lf(3918)=  28501.98842104238837453974540071164094179_dp
  t_lf(3919)=  28510.26201284058800599953647101104128613_dp
  t_lf(3920)=  28518.53585977337251422201686431698653950_dp
  t_lf(3921)=  28526.80996177566484586102046325640094106_dp
  t_lf(3922)=  28535.08431878242113744895026018791962256_dp
  t_lf(3923)=  28543.35893072863069001254945591400550030_dp
  t_lf(3924)=  28551.63379754931594371455172708297610649_dp
  t_lf(3925)=  28559.90891917953245252117769098583195702_dp
  t_lf(3926)=  28568.18429555436885889544464684833992848_dp
  t_lf(3927)=  28576.45992660894686851625672302454388975_dp
  t_lf(3928)=  28584.73581227842122502324260971393350411_dp
  t_lf(3929)=  28593.01195249797968478730810695108327212_dp
  t_lf(3930)=  28601.28834720284299170687076765385968792_dp
  t_lf(3931)=  28609.56499632826485202974396546446658153_dp
  t_lf(3932)=  28617.84189980953190920063776697683864911_dp
  t_lf(3933)=  28626.11905758196371873424403771438175017_dp
  t_lf(3934)=  28634.39646958091272311387326090397627841_dp
  t_lf(3935)=  28642.67413574176422671561059768568688184_dp
  t_lf(3936)=  28650.95205599993637075795876690293769797_dp
  t_lf(3937)=  28659.23023029088010827693537203519634758_dp
  t_lf(3938)=  28667.50865855007917912659235216464105562_dp
  t_lf(3939)=  28675.78734071305008500492528311004188909_dp
  t_lf(3940)=  28684.06627671534206450514030401534726084_dp
  t_lf(3941)=  28692.34546649253706819224649374740818155_dp
  t_lf(3942)=  28700.62490998024973370494157043707247781_dp
  t_lf(3943)=  28708.90460711412736088275883639071615859_dp
  t_lf(3944)=  28717.18455782984988691844333940532572644_dp
  t_lf(3945)=  28725.46476206312986153552527023967951187_dp
  t_lf(3946)=  28733.74521974971242219105866462717367871_dp
  t_lf(3947)=  28742.02593082537526930349352676257461933_dp
  t_lf(3948)=  28750.30689522592864150564953965562885161_dp
  t_lf(3949)=  28758.58811288721529092275957611919866089_dp
  t_lf(3950)=  28766.86958374511045847555127244859062366_dp
  t_lf(3951)=  28775.15130773552184920833497505217842877_dp
  t_lf(3952)=  28783.43328479438960764206641841146430718_dp
  t_lf(3953)=  28791.71551485768629315235254078154772756_dp
  t_lf(3954)=  28799.99799786141685537236889199074825393_dp
  t_lf(3955)=  28808.28073374161860962065713556103364297_dp
  t_lf(3956)=  28816.56372243436121235377119514910603926_dp
  t_lf(3957)=  28824.84696387574663664374064300166977540_dp
  t_lf(3958)=  28833.13045800190914768031997572771467708_dp
  t_lf(3959)=  28841.41420474901527829799247021576940125_dp
  t_lf(3960)=  28849.69820405326380452769735996518029930_dp
  t_lf(3961)=  28857.98245585088572117324911945772231331_dp
  t_lf(3962)=  28866.26696007814421741241769146941880982_dp
  t_lf(3963)=  28874.55171667133465242263853941250597909_dp
  t_lf(3964)=  28882.83672556678453103132145390419303900_dp
  t_lf(3965)=  28891.12198670085347939072708978241015999_dp
  t_lf(3966)=  28899.40750000993322067738025672926956954_dp
  t_lf(3967)=  28907.69326543044755081598903352065911584_dp
  t_lf(3968)=  28915.97928289885231422783882269540870903_dp
  t_lf(3969)=  28924.26555235163537960363050912998517130_dp
  t_lf(3970)=  28932.55207372531661570073193261484639480_dp
  t_lf(3971)=  28940.83884695644786716481193105658723117_dp
  t_lf(3972)=  28949.12587198161293037582625737600274682_dp
  t_lf(3973)=  28957.41314873742752931832471953634452381_dp
  t_lf(3974)=  28965.70067716053929147604893941851734808_dp
  t_lf(3975)=  28973.98845718762772375079017246092130683_dp
  t_lf(3976)=  28982.27648875540418840547667610125205036_dp
  t_lf(3977)=  28990.56477180061187903146016109599341967_dp
  t_lf(3978)=  28998.85330626002579653997090575073509278_dp
  t_lf(3979)=  29007.14209207045272517771115897098628037_dp
  t_lf(3980)=  29015.43112916873120856655650383899735835_dp
  t_lf(3981)=  29023.72041749173152576733489913740684825_dp
  t_lf(3982)=  29032.00995697635566736765316187546316304_dp
  t_lf(3983)=  29040.29974755953731159374069942829048125_dp
  t_lf(3984)=  29048.58978917824180044628034537433708368_dp
  t_lf(3985)=  29056.88008176946611586019619851092320990_dp
  t_lf(3986)=  29065.17062527023885588836840984285432730_dp
  t_lf(3987)=  29073.46141961762021090924490757454465441_dp
  t_lf(3988)=  29081.75246474870193985832009529216448006_dp
  t_lf(3989)=  29090.04376060060734648345060359914255021_dp
  t_lf(3990)=  29098.33530711049125562397822046608047339_dp
  t_lf(3991)=  29106.62710421553998951363017047492828428_dp
  t_lf(3992)=  29114.91915185297134410716695797728720701_dp
  t_lf(3993)=  29123.21144996003456543074803394810512061_dp
  t_lf(3994)=  29131.50399847401032595598559099896974065_dp
  t_lf(3995)=  29139.79679733221070099765683561984122854_dp
  t_lf(3996)=  29148.08984647197914513504513124455660595_dp
  t_lf(3997)=  29156.38314583069046865688045018393941485_dp
  t_lf(3998)=  29164.67669534575081402984961684101560034_dp
  t_lf(3999)=  29172.97049495459763239064686891582633029_dp
  t_lf(4000)=  29181.26454459469966006153530752279577605_dp
  t_lf(4001)=  29189.55884420355689508938985128171178958_dp
  t_lf(4002)=  29197.85339371870057380819235350426459958_dp
  t_lf(4003)=  29206.14819307769314742494958558191744218_dp
  t_lf(4004)=  29214.44324221812825862900483358780742295_dp
  t_lf(4005)=  29222.73854107763071822471389893554851122_dp
  t_lf(4006)=  29231.03408959385648178745633769138468717_dp
  t_lf(4007)=  29239.32988770449262634295281681327284216_dp
  t_lf(4008)=  29247.62593534725732706985950919131467694_dp
  t_lf(4009)=  29255.92223245989983402561049288865681059_dp
  t_lf(4010)=  29264.21877898020044889547916343069052453_dp
  t_lf(4011)=  29272.51557484597050176482971136325859711_dp
  t_lf(4012)=  29280.81261999505232791452976059776777441_dp
  t_lf(4013)=  29289.10991436531924463949530628276246843_dp
  t_lf(4014)=  29297.40745789467552809033913408778883643_dp
  t_lf(4015)=  29305.70525052105639013809394585641869594_dp
  t_lf(4016)=  29314.00329218242795526198145958125965545_dp
  t_lf(4017)=  29322.30158281678723746019879457480094058_dp
  t_lf(4018)=  29330.60012236216211718369349555618288420_dp
  t_lf(4019)=  29338.89891075661131829289859214558080779_dp
  t_lf(4020)=  29347.19794793822438503739913295500959702_dp
  t_lf(4021)=  29355.49723384512165905850167608713188208_dp
  t_lf(4022)=  29363.79676841545425641467826040223825598_dp
  t_lf(4023)=  29372.09655158740404462985642438810995358_dp
  t_lf(4024)=  29380.39658329918361976452688186812009332_dp
  t_lf(4025)=  29388.69686348903628350964050610982584444_dp
  t_lf(4026)=  29396.99739209523602030326631614959729076_dp
  t_lf(4027)=  29405.29816905608747446998220132866555304_dp
  t_lf(4028)=  29413.59919430992592738297016214249881748_dp
  t_lf(4029)=  29421.90046779511727464878788753877588064_dp
  t_lf(4030)=  29430.20198945005800331478853075956792066_dp
  t_lf(4031)=  29438.50375921317516909916058771080537605_dp
  t_lf(4032)=  29446.80577702292637364355982365684266660_dp
  t_lf(4033)=  29455.10804281779974178830523578008331541_dp
  t_lf(4034)=  29463.41055653631389887011108081533579235_dp
  t_lf(4035)=  29471.71331811701794804232703856597974281_dp
  t_lf(4036)=  29480.01632749849144761765862363427651666_dp
  t_lf(4037)=  29488.31958461934438843333999915140007513_dp
  t_lf(4038)=  29496.62308941821717123873138767413711561_dp
  t_lf(4039)=  29504.92684183378058410531331572485098374_dp
  t_lf(4040)=  29513.23084180473577985904996968936469025_dp
  t_lf(4041)=  29521.53508926981425353509398195403585166_dp
  t_lf(4042)=  29529.83958416777781985480500725861204897_dp
  t_lf(4043)=  29538.14432643741859072505449026561004985_dp
  t_lf(4044)=  29546.44931601755895275978906630009735990_dp
  t_lf(4045)=  29554.75455284705154482382507809601013272_dp
  t_lf(4046)=  29563.06003686477923559884673219665774209_dp
  t_lf(4047)=  29571.36576800965510117158045939798115543_dp
  t_lf(4048)=  29579.67174622062240264411808429358918712_dp
  t_lf(4049)=  29587.97797143665456376636144958073298551_dp
  t_lf(4050)=  29596.28444359675514859056118131633364018_dp
  t_lf(4051)=  29604.59116263995783914792232177208919998_dp
  t_lf(4052)=  29612.89812850532641314724959692769397328_dp
  t_lf(4053)=  29621.20534113195472169560512596144273747_dp
  t_lf(4054)=  29629.51280045896666704095142034810310839_dp
  t_lf(4055)=  29637.82050642551618033675256035505819652_dp
  t_lf(4056)=  29646.12845897078719942850647683948588913_dp
  t_lf(4057)=  29654.43665803399364666218130629188742165_dp
  t_lf(4058)=  29662.74510355437940671452882704474281195_dp
  t_lf(4059)=  29671.05379547121830444524802446959040025_dp
  t_lf(4060)=  29679.36273372381408277097187282153803236_dp
  t_lf(4061)=  29687.67191825150038056105046115724991546_dp
  t_lf(4062)=  29695.98134899364071055510363045095112976_dp
  t_lf(4063)=  29704.29102588962843730231632866308616540_dp
  t_lf(4064)=  29712.60094887888675512244993007809334063_dp
  t_lf(4065)=  29720.91111790086866608854280472144791601_dp
  t_lf(4066)=  29729.22153289505695803127346309181722312_dp
  t_lf(4067)=  29737.53219380096418256495964080199495303_dp
  t_lf(4068)=  29745.84310055813263313516672701237237925_dp
  t_lf(4069)=  29754.15425310613432308789897976319490821_dp
  t_lf(4070)=  29762.46565138457096376034701046687584888_dp
  t_lf(4071)=  29770.77729533307394259316505890932826833_dp
  t_lf(4072)=  29779.08918489130430126425161912976255668_dp
  t_lf(4073)=  29787.40131999895271384400701550181387453_dp
  t_lf(4074)=  29795.71370059573946497204156722534171824_dp
  t_lf(4075)=  29804.02632662141442805530801825791484340_dp
  t_lf(4076)=  29812.33919801575704348763194846798987202_dp
  t_lf(4077)=  29820.65231471857629689061392047824192543_dp
  t_lf(4078)=  29828.96567666971069737587715528754113114_dp
  t_lf(4079)=  29837.27928380902825582863456831382012262_dp
  t_lf(4080)=  29845.59313607642646321254903598767466878_dp
  t_lf(4081)=  29853.90723341183226889586080144811203558_dp
  t_lf(4082)=  29862.22157575520205899875596624753900600_dp
  t_lf(4083)=  29870.53616304652163476195005326299279468_dp
  t_lf(4084)=  29878.85099522580619093646066423489223234_dp
  t_lf(4085)=  29887.16607223310029419454329351335212055_dp
  t_lf(4086)=  29895.48139400847786156176439768548884558_dp
  t_lf(4087)=  29903.79696049204213887018585878527818307_dp
  t_lf(4088)=  29912.11277162392567923263501675053443508_dp
  t_lf(4089)=  29920.42882734429032153803448468959104610_dp
  t_lf(4090)=  29928.74512759332716896776599835340379639_dp
  t_lf(4091)=  29937.06167231125656753304258897719543686_dp
  t_lf(4092)=  29945.37846143832808463326340635954180222_dp
  t_lf(4093)=  29953.69549491482048763532555668609032940_dp
  t_lf(4094)=  29962.01277268104172247386735718002855130_dp
  t_lf(4095)=  29970.33029467732889227241744717210828869_dp
  t_lf(4096)=  29978.64806084404823598542423262960640750_dp
  t_lf(4097)=  29986.96607112159510706114017856619034977_dp
  t_lf(4098)=  29995.28432545039395212533550107338111400_dp
  t_lf(4099)=  30003.60282377089828968581584796929261232_dp
  t_lf(4100)=  30011.92156602359068885771859425169874833_dp
  t_lf(4101)=  30020.24055214898274810956241567036225118_dp
  t_lf(4102)=  30028.55978208761507403002484079807610567_dp
  t_lf(4103)=  30036.87925578005726011542251898114289910_dp
  t_lf(4104)=  30045.19897316690786557786897848817285946_dp
  t_lf(4105)=  30053.51893418879439417408468605124080840_dp
  t_lf(4106)=  30061.83913878637327305483425580572845267_dp
  t_lf(4107)=  30070.15958690032983163496569238471387270_dp
  t_lf(4108)=  30078.48027847137828048402658961067695536_dp
  t_lf(4109)=  30086.80121344026169023743224285168980777_dp
  t_lf(4110)=  30095.12239174775197052816066967127656307_dp
  t_lf(4111)=  30103.44381333464984893894956990087886290_dp
  t_lf(4112)=  30111.76547814178484997497029270147282463_dp
  t_lf(4113)=  30120.08738611001527405695391455647135957_dp
  t_lf(4114)=  30128.40953718022817653474456845173292127_dp
  t_lf(4115)=  30136.73193129333934672125520075040448706_dp
  t_lf(4116)=  30145.05456839029328694680096846057290509_dp
  t_lf(4117)=  30153.37744841206319163378552572240450338_dp
  t_lf(4118)=  30161.70057129965092639171548440873762565_dp
  t_lf(4119)=  30170.02393699408700713251836973907583900_dp
  t_lf(4120)=  30178.34754543643057920613942775172999716_dp
  t_lf(4121)=  30186.67139656776939655639267736259392708_dp
  t_lf(4122)=  30194.99549032921980089704163556182976337_dp
  t_lf(4123)=  30203.31982666192670090808518006170315385_dp
  t_lf(4124)=  30211.64440550706355145222404941006370939_dp
  t_lf(4125)=  30219.96922680583233281148351622462992732_dp
  t_lf(4126)=  30228.29429049946352994396780478342787390_dp
  t_lf(4127)=  30236.61959652921611176072185972656640981_dp
  t_lf(4128)=  30244.94514483637751042267610808412566539_dp
  t_lf(4129)=  30253.27093536226360065764989224440655065_dp
  t_lf(4130)=  30261.59696804821867909738928681625379221_dp
  t_lf(4131)=  30269.92324283561544363461504761873954727_dp
  t_lf(4132)=  30278.24975966585497280005647625129502070_dp
  t_lf(4133)=  30286.57651848036670515944701885751942130_dp
  t_lf(4134)=  30294.90351922060841873045745279649449995_dp
  t_lf(4135)=  30303.23076182806621041954254997660402803_dp
  t_lf(4136)=  30311.55824624425447547867714058871586074_dp
  t_lf(4137)=  30319.88597241071588698195753589824439715_dp
  t_lf(4138)=  30328.21394026902137532204430361918775931_dp
  t_lf(4139)=  30336.54214976077010772642242419784107906_dp
  t_lf(4140)=  30344.87060082758946779345489107963886737_dp
  t_lf(4141)=  30353.19929341113503504820585271958926411_dp
  t_lf(4142)=  30361.52822745309056451800942872514449587_dp
  t_lf(4143)=  30369.85740289516796632776036709021832775_dp
  t_lf(4144)=  30378.18681967910728531490274399052566092_dp
  t_lf(4145)=  30386.51647774667668066409294206359443054_dp
  t_lf(4146)=  30394.84637703967240556151317749179808799_dp
  t_lf(4147)=  30403.17651749991878686881188054369044750_dp
  t_lf(4148)=  30411.50689906926820481664726850790554192_dp
  t_lf(4149)=  30419.83752168960107271781048417502511987_dp
  t_lf(4150)=  30428.16838530282581669990470718622704217_dp
  t_lf(4151)=  30436.49948985087885545755667967332037054_dp
  t_lf(4152)=  30444.83083527572458002413712166305841742_dp
  t_lf(4153)=  30453.16242151935533356296654570951023246_dp
  t_lf(4154)=  30461.49424852379139117798301415187448899_dp
  t_lf(4155)=  30469.82631623108093974384841627154781317_dp
  t_lf(4156)=  30478.15862458330005775546987644162234174_dp
  t_lf(4157)=  30486.49117352255269519691293812439453403_dp
  t_lf(4158)=  30494.82396299097065342968320227802859675_dp
  t_lf(4159)=  30503.15699293071356510035313238234266367_dp
  t_lf(4160)=  30511.49026328396887406751077188588322932_dp
  t_lf(4161)=  30519.82377399295181534800715341213215113_dp
  t_lf(4162)=  30528.15752499990539508247921254195945952_dp
  t_lf(4163)=  30536.49151624710037052012505241240266457_dp
  t_lf(4164)=  30544.82574767683523002270843873862740294_dp
  t_lf(4165)=  30553.16021923143617308776943817661307658_dp
  t_lf(4166)=  30561.49493085325709039101814619881830929_dp
  t_lf(4167)=  30569.82988248467954384788848385392207064_dp
  t_lf(4168)=  30578.16507406811274669422907592481443681_dp
  t_lf(4169)=  30586.50050554599354358610825608643319402_dp
  t_lf(4170)=  30594.83617686078639071871027769691562771_dp
  t_lf(4171)=  30603.17208795498333596429984183196543646_dp
  t_lf(4172)=  30611.50823877110399902923208709342908981_dp
  t_lf(4173)=  30619.84462925169555162998521858894021112_dp
  t_lf(4174)=  30628.18125933933269768819298629023057778_dp
  t_lf(4175)=  30636.51812897661765354465425573342773243_dp
  t_lf(4176)=  30644.85523810618012819229694672546740172_dp
  t_lf(4177)=  30653.19258667067730352807364836674910931_dp
  t_lf(4178)=  30661.53017461279381462376625129146050532_dp
  t_lf(4179)=  30669.86800187524173001567697056369474796_dp
  t_lf(4180)=  30678.20606840076053201318316514969027157_dp
  t_lf(4181)=  30686.54437413211709702613339231433774040_dp
  t_lf(4182)=  30694.88291901210567591106216766362897680_dp
  t_lf(4183)=  30703.22170298354787433620093387407099973_dp
  t_lf(4184)=  30711.56072598929263316526277341535862473_dp
  t_lf(4185)=  30719.89998797221620885997843278389474669_dp
  t_lf(4186)=  30728.23948887522215390136125792217161771_dp
  t_lf(4187)=  30736.57922864124129722967867260268208641_dp
  t_lf(4188)=  30744.91920721323172470310786360501960496_dp
  t_lf(4189)=  30753.25942453417875957505336851125233668_dp
  t_lf(4190)=  30761.59988054709494299010429388762219127_dp
  t_lf(4191)=  30769.94057519502001449860892351022613867_dp
  t_lf(4192)=  30778.28150842102089258984450812868654741_dp
  t_lf(4193)=  30786.62268016819165524376006004501118382_dp
  t_lf(4194)=  30794.96409037965352050127000751498329928_dp
  t_lf(4195)=  30803.30573899855482705307659565660911195_dp
  t_lf(4196)=  30811.64762596807101484699895217448492882_dp
  t_lf(4197)=  30819.98975123140460571378676778052990617_dp
  t_lf(4198)=  30828.33211473178518401139657271046355158_dp
  t_lf(4199)=  30836.67471641246937728770862220178985026_dp
  t_lf(4200)=  30845.01755621674083696166243521298246067_dp
  t_lf(4201)=  30853.36063408791021902278906202514766283_dp
  t_lf(4202)=  30861.70394996931516474911818767677333584_dp
  t_lf(4203)=  30870.04750380432028144343820943935265624_dp
  t_lf(4204)=  30878.39129553631712318788745774679969662_dp
  t_lf(4205)=  30886.73532510872417161685476114474970747_dp
  t_lf(4206)=  30895.07959246498681670816758692715841032_dp
  t_lf(4207)=  30903.42409754857733759254602017718073618_dp
  t_lf(4208)=  30911.76884030299488338130087492721851454_dp
  t_lf(4209)=  30920.11382067176545401225426209937685111_dp
  t_lf(4210)=  30928.45903859844188111386096978245831279_dp
  t_lf(4211)=  30936.80449402660380888750904224515034197_dp
  t_lf(4212)=  30945.15018689985767500797797487732211653_dp
  t_lf(4213)=  30953.49611716183669154203297299243971608_dp
  t_lf(4214)=  30961.84228475620082588513375311413009914_dp
  t_lf(4215)=  30970.18868962663678171623639600897198443_dp
  t_lf(4216)=  30978.53533171685797997066679131576199788_dp
  t_lf(4217)=  30986.88221097060453983104424415889392598_dp
  t_lf(4218)=  30995.22932733164325973623384462019392895_dp
  t_lf(4219)=  31003.57668074376759840830623137967123316_dp
  t_lf(4220)=  31011.92427115079765589748341122126805405_dp
  t_lf(4221)=  31020.27209849658015464504932643492000794_dp
  t_lf(4222)=  31028.62016272498842056420389243116455775_dp
  t_lf(4223)=  31036.96846377992236413883925811925540616_dp
  t_lf(4224)=  31045.31700160530846154021707178435029714_dp
  t_lf(4225)=  31053.66577614509973576152556533393330963_dp
  t_lf(4226)=  31062.01478734327573777029529986830511845_dp
  t_lf(4227)=  31070.36403514384252767865244556482034942_dp
  t_lf(4228)=  31078.71351949083265593138849885066435858_dp
  t_lf(4229)=  31087.06324032830514451182536977443660963_dp
  t_lf(4230)=  31095.41319760034546816545480237273819896_dp
  t_lf(4231)=  31103.76339125106553564133112066444067355_dp
  t_lf(4232)=  31112.11382122460367095119632269243559549_dp
  t_lf(4233)=  31120.46448746512459464631657477052261819_dp
  t_lf(4234)=  31128.81538991681940511200918778178124893_dp
  t_lf(4235)=  31137.16652852390555987983918701438087490_dp
  t_lf(4236)=  31145.51790323062685695746461661140772374_dp
  t_lf(4237)=  31153.86951398125341617610974925301871731_dp
  t_lf(4238)=  31162.22136072008166055564540118216296236_dp
  t_lf(4239)=  31170.57344339143429768725558212933401472_dp
  t_lf(4240)=  31178.92576193966030113366973908742196640_dp
  t_lf(4241)=  31187.27831630913489184693988223481555736_dp
  t_lf(4242)=  31195.63110644425951960374191060355242852_dp
  t_lf(4243)=  31203.98413228946184445818048433962163993_dp
  t_lf(4244)=  31212.33739378919571821207681960457781362_dp
  t_lf(4245)=  31220.69089088794116590271881132152166717_dp
  t_lf(4246)=  31229.04462353020436730805291807432803180_dp
  t_lf(4247)=  31237.39859166051763846929727252685025440_dp
  t_lf(4248)=  31245.75279522343941323095550973878953314_dp
  t_lf(4249)=  31254.10723416355422479821083471707940464_dp
  t_lf(4250)=  31262.46190842547268731167987945608927066_dp
  t_lf(4251)=  31270.81681795383147743950592858678631662_dp
  t_lf(4252)=  31279.17196269329331598677112157430203549_dp
  t_lf(4253)=  31287.52734258854694952220726817521724202_dp
  t_lf(4254)=  31295.88295758430713202218494259039716739_dp
  t_lf(4255)=  31304.23880762531460653196055042646499832_dp
  t_lf(4256)=  31312.59489265633608684416109120908691442_dp
  t_lf(4257)=  31320.95121262216423919448636777424294217_dp
  t_lf(4258)=  31329.30776746761766397460842239966425373_dp
  t_lf(4259)=  31337.66455713754087746224800802771718060_dp
  t_lf(4260)=  31346.02158157680429356840793137329528118_dp
  t_lf(4261)=  31354.37884073030420560174313310583121180_dp
  t_lf(4262)=  31362.73633454296276805004739864344762765_dp
  t_lf(4263)=  31371.09406295972797837883662139961842442_dp
  t_lf(4264)=  31379.45202592557365884700856857859568032_dp
  t_lf(4265)=  31387.81022338549943833955912782536084207_dp
  t_lf(4266)=  31396.16865528453073421733504119906800668_dp
  t_lf(4267)=  31404.52732156771873418380316105594938646_dp
  t_lf(4268)=  31412.88622218014037816881629049853483061_dp
  t_lf(4269)=  31421.24535706689834022935569907288504978_dp
  t_lf(4270)=  31429.60472617312101046723043237443820846_dp
  t_lf(4271)=  31437.96432944396247696371356215610788742_dp
  t_lf(4272)=  31446.32416682460250773109555141953296862_dp
  t_lf(4273)=  31454.68423826024653268113493681195247031_dp
  t_lf(4274)=  31463.04454369612562561038655844714629231_dp
  t_lf(4275)=  31471.40508307749648620238759501933157253_dp
  t_lf(4276)=  31479.76585634964142204668168978391907809_dp
  t_lf(4277)=  31488.12686345786833067466148063869975013_dp
  t_lf(4278)=  31496.48810434751068161220987515343300574_dp
  t_lf(4279)=  31504.84957896392749844912043896503030769_dp
  t_lf(4280)=  31513.21128725250334092527729347965430203_dp
  t_lf(4281)=  31521.57322915864828703357494630216977357_dp
  t_lf(4282)=  31529.93540462779791513955850524757188430_dp
  t_lf(4283)=  31538.29781360541328611776475417836356498_dp
  t_lf(4284)=  31546.66045603698092550474459625644127579_dp
  t_lf(4285)=  31555.02333186801280566874739749796021201_dp
  t_lf(4286)=  31563.38644104404632799604779077496980333_dp
  t_lf(4287)=  31571.74978351064430509389552761842126421_dp
  t_lf(4288)=  31580.11335921339494301006899234353404637_dp
  t_lf(4289)=  31588.47716809791182346901302014055019645_dp
  t_lf(4290)=  31596.84121010983388612454168785168753427_dp
  t_lf(4291)=  31605.20548519482541082908677318870676493_dp
  t_lf(4292)=  31613.56999329857599991947260513501647684_dp
  t_lf(4293)=  31621.93473436680056051919805522173563743_dp
  t_lf(4294)=  31630.29970834523928685720644626869768921_dp
  t_lf(4295)=  31638.66491517965764260312418203909550641_dp
  t_lf(4296)=  31647.03035481584634321894892807041396400_dp
  t_lf(4297)=  31655.39602719962133832716820071455818937_dp
  t_lf(4298)=  31663.76193227682379409528924814674203795_dp
  t_lf(4299)=  31672.12806999332007563676113378583410924_dp
  t_lf(4300)=  31680.49444029500172942826995920854868469_dp
  t_lf(4301)=  31688.86104312778546574338819023619713664_dp
  t_lf(4302)=  31697.22787843761314110255907642576227372_dp
  t_lf(4303)=  31705.59494617045174073939718070690422913_dp
  t_lf(4304)=  31713.96224627229336108328606237323217252_dp
  t_lf(4305)=  31722.32977868915519225825418305986147318_dp
  t_lf(4306)=  31730.69754336707950059811013172000093597_dp
  t_lf(4307)=  31739.06554025213361117781829095115917617_dp
  t_lf(4308)=  31747.43376929040989036109609331660273499_dp
  t_lf(4309)=  31755.80223042802572836421404256002063598_dp
  t_lf(4310)=  31764.17092361112352183597970082103005044_dp
  t_lf(4311)=  31772.53984878587065645388686912627471846_dp
  t_lf(4312)=  31780.90900589845948953641121455550073693_dp
  t_lf(4313)=  31789.27839489510733267143362356422208751_dp
  t_lf(4314)=  31797.64801572205643436077258698448948328_dp
  t_lf(4315)=  31806.01786832557396268080694822292924395_dp
  t_lf(4316)=  31814.38795265195198795917037213070228615_dp
  t_lf(4317)=  31822.75826864750746546749891793342509183_dp
  t_lf(4318)=  31831.12881625858221813021312548047268732_dp
  t_lf(4319)=  31839.49959543154291924931604990252605788_dp
  t_lf(4320)=  31847.87060611278107524518870555381070547_dp
  t_lf(4321)=  31856.24184824871300841336440586127673757_dp
  t_lf(4322)=  31864.61332178577983969726351140707129535_dp
  t_lf(4323)=  31872.98502667044747147687012423312847625_dp
  t_lf(4324)=  31881.35696284920657037333229197762719921_dp
  t_lf(4325)=  31889.72913026857255006946731103252056596_dp
  t_lf(4326)=  31898.10152887508555414615374344939788879_dp
  t_lf(4327)=  31906.47415861531043893459178781767922963_dp
  t_lf(4328)=  31914.84701943583675638441366979463840947_dp
  t_lf(4329)=  31923.22011128327873694762574338108022665_dp
  t_lf(4330)=  31931.59343410427527247836401940973713483_dp
  t_lf(4331)=  31939.96698784548989914844486304567578367_dp
  t_lf(4332)=  31948.34077245361078037869262738929036959_dp
  t_lf(4333)=  31956.71478787535068978602601552288327305_dp
  t_lf(4334)=  31965.08903405744699414628498855146940989_dp
  t_lf(4335)=  31973.46351094666163637278006235736437631_dp
  t_lf(4336)=  31981.83821848978111851054586091640294462_dp
  t_lf(4337)=  31990.21315663361648474628081911135873719_dp
  t_lf(4338)=  31998.58832532500330443395495302537278090_dp
  t_lf(4339)=  32006.96372451080165513606764070502278166_dp
  t_lf(4340)=  32015.33935413789610568053738134915085937_dp
  t_lf(4341)=  32023.71521415319569923320552580578949652_dp
  t_lf(4342)=  32032.09130450363393638593599614555777106_dp
  t_lf(4343)=  32040.46762513616875826029303692581660725_dp
  t_lf(4344)=  32048.84417599778252962677906556574667159_dp
  t_lf(4345)=  32057.22095703548202203961471401841939945_dp
  t_lf(4346)=  32065.59796819629839698704317865194404137_dp
  t_lf(4347)=  32073.97520942728718905714101993796499434_dp
  t_lf(4348)=  32082.35268067552828911911757819222730933_dp
  t_lf(4349)=  32090.73038188812592752008519621869726644_dp
  t_lf(4350)=  32099.10831301220865729728246427589225560_dp
  t_lf(4351)=  32107.48647399492933740573272731171271745_dp
  t_lf(4352)=  32115.86486478346511596132011890125125653_dp
  t_lf(4353)=  32124.24348532501741349926541077085275848_dp
  t_lf(4354)=  32132.62233556681190624798399120118679706_dp
  t_lf(4355)=  32141.00141545609850941830830997234202605_dp
  t_lf(4356)=  32149.38072494015136050805715184503368920_dp
  t_lf(4357)=  32157.76026396626880262193412486400177110_dp
  t_lf(4358)=  32166.14003248177336780673777402264042966_dp
  t_lf(4359)=  32174.52003043401176040186575504191082198_dp
  t_lf(4360)=  32182.90025777035484040509552719172074098_dp
  t_lf(4361)=  32191.28071443819760685362404821927694966_dp
  t_lf(4362)=  32199.66140038495918122034897854650091982_dp
  t_lf(4363)=  32208.04231555808279082537392595751688790_dp
  t_lf(4364)=  32216.42345990503575226272028601754362512_dp
  t_lf(4365)=  32224.80483337330945484222825744631882588_dp
  t_lf(4366)=  32233.18643591041934404662963561252814724_dp
  t_lf(4367)=  32241.56826746390490500377501122067013582_dp
  t_lf(4368)=  32249.95032798132964597399802512843386613_dp
  t_lf(4369)=  32258.33261741028108185259935406106824812_dp
  t_lf(4370)=  32266.71513569837071768743312577945066198_dp
  t_lf(4371)=  32275.09788279323403221157848601068771861_dp
  t_lf(4372)=  32283.48085864253046139107906316417225588_dp
  t_lf(4373)=  32291.86406319394338198773310053214775174_dp
  t_lf(4374)=  32300.24749639518009513691704931206360824_dp
  t_lf(4375)=  32308.63115819397180994042543938841153852_dp
  t_lf(4376)=  32317.01504853807362707430986837438372898_dp
  t_lf(4377)=  32325.39916737526452241169997293865656865_dp
  t_lf(4378)=  32333.78351465334733066058926992994841172_dp
  t_lf(4379)=  32342.16809032014872901656877826179479961_dp
  t_lf(4380)=  32350.55289432351922083049135593229840867_dp
  t_lf(4381)=  32358.93792661133311929104970992851215848_dp
  t_lf(4382)=  32367.32318713148853112225106010267072513_dp
  t_lf(4383)=  32375.70867583190734029577146140776632205_dp
  t_lf(4384)=  32384.09439266053519175817281214303707028_dp
  t_lf(4385)=  32392.48033756534147517296559908586846865_dp
  t_lf(4386)=  32400.86651049431930867750045357546814450_dp
  t_lf(4387)=  32409.25291139548552265467161576552882909_dp
  t_lf(4388)=  32417.63954021688064351941542737801183201_dp
  t_lf(4389)=  32426.02639690656887751998699636823052028_dp
  t_lf(4390)=  32434.41348141263809455399819995265763819_dp
  t_lf(4391)=  32442.80079368319981199920021545538879158_dp
  t_lf(4392)=  32451.18833366638917855899379139703398744_dp
  t_lf(4393)=  32459.57610131036495812265049418104655070_dp
  t_lf(4394)=  32467.96409656330951364022818862720068309_dp
  t_lf(4395)=  32476.35231937342879101216403346016189462_dp
  t_lf(4396)=  32484.74076968895230299352829568292490274_dp
  t_lf(4397)=  32493.12944745813311311292231055038759518_dp
  t_lf(4398)=  32501.51835262924781960600393660755339576_dp
  t_lf(4399)=  32509.90748515059653936362387796987382398_dp
  t_lf(4400)=  32518.29684497050289189455626870012403483_dp
  t_lf(4401)=  32526.68643203731398330280693677701236182_dp
  t_lf(4402)=  32535.07624629940039027948278775552592970_dp
  t_lf(4403)=  32543.46628770515614410920577078787368313_dp
  t_lf(4404)=  32551.85655620299871469105491220687099113_dp
  t_lf(4405)=  32560.24705174136899457401992437078149721_dp
  t_lf(4406)=  32568.63777426873128300694991993005712215_dp
  t_lf(4407)=  32577.02872373357327000298078410216098813_dp
  t_lf(4408)=  32585.41990008440602041842477993078528297_dp
  t_lf(4409)=  32593.81130326976395804610598386035135594_dp
  t_lf(4410)=  32602.20293323820484972312517127576713102_dp
  t_lf(4411)=  32610.59478993830978945303779394108161005_dp
  t_lf(4412)=  32618.98687331868318254242871351898205346_dp
  t_lf(4413)=  32627.37918332795272975186737756609047801_dp
  t_lf(4414)=  32635.77171991476941146122714657679637449_dp
  t_lf(4415)=  32644.16448302780747184935250279097586969_dp
  t_lf(4416)=  32652.55747261576440308805789358845765092_dp
  t_lf(4417)=  32660.95068862736092955044198436556642473_dp
  t_lf(4418)=  32669.34413101134099203350111782656894872_dp
  t_lf(4419)=  32677.73779971647173199502579862542908478_dp
  t_lf(4420)=  32686.13169469154347580476404426101007070_dp
  t_lf(4421)=  32694.52581588536971900983546506180736395_dp
  t_lf(4422)=  32702.92016324678711061437995799451691827_dp
  t_lf(4423)=  32711.31473672465543737342492089430442194_dp
  t_lf(4424)=  32719.70953626785760810095491554360354230_dp
  t_lf(4425)=  32728.10456182529963799216772982069814091_dp
  t_lf(4426)=  32736.49981334591063295990081089929717828_dp
  t_lf(4427)=  32744.89529077864277398521206320585391930_dp
  t_lf(4428)=  32753.29099407247130148209902653257525714_dp
  t_lf(4429)=  32761.68692317639449967634047136097454499_dp
  t_lf(4430)=  32770.08307803943368099844447007350418490_dp
  t_lf(4431)=  32778.47945861063317049068702431932417045_dp
  t_lf(4432)=  32786.87606483906029022822535035468148757_dp
  t_lf(4433)=  32795.27289667380534375426994569875429268_dp
  t_lf(4434)=  32803.66995406398160052929958193221553329_dp
  t_lf(4435)=  32812.06723695872528039430338991825444969_dp
  t_lf(4436)=  32820.46474530719553804803422514442237303_dp
  t_lf(4437)=  32828.86247905857444753825752226850246258_dp
  t_lf(4438)=  32837.26043816206698676697986930270243019_dp
  t_lf(4439)=  32845.65862256690102200964155318789568507_dp
  t_lf(4440)=  32854.05703222232729244825734979345037528_dp
  t_lf(4441)=  32862.45566707761939471848985262844806031_dp
  t_lf(4442)=  32870.85452708207376747063965576686465501_dp
  t_lf(4443)=  32879.25361218500967594453672767262614945_dp
  t_lf(4444)=  32887.65292233576919655831733376042062104_dp
  t_lf(4445)=  32896.05245748371720151107088664480628081_dp
  t_lf(4446)=  32904.45221757824134339934112411356268093_dp
  t_lf(4447)=  32912.85220256875203984746603591144857860_dp
  t_lf(4448)=  32921.25241240468245815174098143761500522_dp
  t_lf(4449)=  32929.65284703548849993838946144393541184_dp
  t_lf(4450)=  32938.05350641064878583532602777251581291_dp
  t_lf(4451)=  32946.45439047966464015769583608869597220_dp
  t_lf(4452)=  32954.85549919206007560717536745100708854_dp
  t_lf(4453)=  32963.25683249738177798501886541187124586_dp
  t_lf(4454)=  32971.65839034519909091883505616237207486_dp
  t_lf(4455)=  32980.06017268510400060307874002125349461_dp
  t_lf(4456)=  32988.46217946671112055324186332247280608_dp
  t_lf(4457)=  32996.86441063965767637372870047720442097_dp
  t_lf(4458)=  33005.26686615360349053939979667521963613_dp
  t_lf(4459)=  33013.66954595823096719076934234711449392_dp
  t_lf(4460)=  33022.07245000324507694284067113298017473_dp
  t_lf(4461)=  33030.47557823837334170756459369486670157_dp
  t_lf(4462)=  33038.87893061336581952990530026983903597_dp
  t_lf(4463)=  33047.28250707799508943749858538762282657_dp
  t_lf(4464)=  33055.68630758205623630388716867184294146_dp
  t_lf(4465)=  33064.09033207536683572531790610672915634_dp
  t_lf(4466)=  33072.49458050776693891108570658195755355_dp
  t_lf(4467)=  33080.89905282911905758740898892707076560_dp
  t_lf(4468)=  33089.30374898930814891482153501373250859_dp
  t_lf(4469)=  33097.70866893824160041906561483897911783_dp
  t_lf(4470)=  33106.11381262584921493547127980569012823_dp
  t_lf(4471)=  33114.51918000208319556680674068776832686_dp
  t_lf(4472)=  33122.92477101691813065458476700705402220_dp
  t_lf(4473)=  33131.33058562035097876381006475685528744_dp
  t_lf(4474)=  33139.73662376240105368115260958321229168_dp
  t_lf(4475)=  33148.14288539311000942653193267968606913_dp
  t_lf(4476)=  33156.54937046254182527809737676462661211_dp
  t_lf(4477)=  33164.95607892078279081058935959158831787_dp
  t_lf(4478)=  33173.36301071794149094706670249387876589_dp
  t_lf(4479)=  33181.77016580414879102398510148320563279_dp
  t_lf(4480)=  33190.17754412955782186961183841008223614_dp
  t_lf(4481)=  33198.58514564434396489576184965012059162_dp
  t_lf(4482)=  33206.99297029870483720284028970563771575_dp
  t_lf(4483)=  33215.40101804286027669817674700618184016_dp
  t_lf(4484)=  33223.80928882705232722763628905470574446_dp
  t_lf(4485)=  33232.21778260154522372049253389822997253_dp
  t_lf(4486)=  33240.62649931662537734754796470300456929_dp
  t_lf(4487)=  33249.03543892260136069248672398444934921_dp
  t_lf(4488)=  33257.44460136980389293644514378158466045_dp
  t_lf(4489)=  33265.85398660858582505578528777431210595_dp
  t_lf(4490)=  33274.26359458932212503305680101982258016_dp
  t_lf(4491)=  33282.67342526240986308113238263165202459_dp
  t_lf(4492)=  33291.08347857826819688050221634152813207_dp
  t_lf(4493)=  33299.49375448733835682971271347020836606_dp
  t_lf(4494)=  33307.90425294008363130893494238905552507_dp
  t_lf(4495)=  33316.31497388698935195664813807918598152_dp
  t_lf(4496)=  33324.72591727856287895942370488971185974_dp
  t_lf(4497)=  33333.13708306533358635479514506093587971_dp
  t_lf(4498)=  33341.54847119785284734719936501240036635_dp
  t_lf(4499)=  33349.96008162669401963697483079949388174_dp
  t_lf(4500)=  33358.37191430245243076240206351593384920_dp
  t_lf(4501)=  33366.78396917574536345477198476292506202_dp
  t_lf(4502)=  33375.19624619721204100646764161919565891_dp
  t_lf(4503)=  33383.60874531751361265204485982948744907_dp
  t_lf(4504)=  33392.02146648733313896229739318247971976_dp
  t_lf(4505)=  33400.43440965737557725129215627260809171_dp
  t_lf(4506)=  33408.84757477836776699636014703385572791_dp
  t_lf(4507)=  33417.26096180105841527102868459739626899_dp
  t_lf(4508)=  33425.67457067621808219088060715900917770_dp
  t_lf(4509)=  33434.08840135463916637232609364652153418_dp
  t_lf(4510)=  33442.50245378713589040427279205220843528_dp
  t_lf(4511)=  33450.91672792454428633267995634015961342_dp
  t_lf(4512)=  33459.33122371772218115798231285414519567_dp
  t_lf(4513)=  33467.74594111754918234536939613754106021_dp
  t_lf(4514)=  33476.16088007492666334790611303345629869_dp
  t_lf(4515)=  33484.57604054077774914248031286039403879_dp
  t_lf(4516)=  33492.99142246604730177856316035662439645_dp
  t_lf(4517)=  33501.40702580170190593976812695500658377_dp
  t_lf(4518)=  33509.82285049872985451819443478931806499_dp
  t_lf(4519)=  33518.23889650814113420154080664328389252_dp
  t_lf(4520)=  33526.65516378096741107297539383450063154_dp
  t_lf(4521)=  33535.07165226826201622374777277736815194_dp
  t_lf(4522)=  33543.48836192109993137852891969203048831_dp
  t_lf(4523)=  33551.90529269057777453346509162023529472_dp
  t_lf(4524)=  33560.32244452781378560693156057400140377_dp
  t_lf(4525)=  33568.73981738394781210297216627908678798_dp
  t_lf(4526)=  33577.15741121014129478741067158252586433_dp
  t_lf(4527)=  33585.57522595757725337661992317200652577_dp
  t_lf(4528)=  33593.99326157746027223893483880463437225_dp
  t_lf(4529)=  33602.41151802101648610869526076373509239_dp
  t_lf(4530)=  33610.82999523949356581290473375482645859_dp
  t_lf(4531)=  33619.24869318416070401049128391579948569_dp
  t_lf(4532)=  33627.66761180630860094415629405173440894_dp
  t_lf(4533)=  33636.08675105724945020479758861169160307_dp
  t_lf(4534)=  33644.50611088831692450849286030331063228_dp
  t_lf(4535)=  33652.92569125086616148602958859117243481_dp
  t_lf(4536)=  33661.34549209627374948496761864668024706_dp
  t_lf(4537)=  33669.76551337593771338422058761074420537_dp
  t_lf(4538)=  33678.18575504127750042114240329586247341_dp
  t_lf(4539)=  33686.60621704373396603110499869132797487_dp
  t_lf(4540)=  33695.02689933476935969955360384430401269_dp
  t_lf(4541)=  33703.44780186586731082652579487045377532_dp
  t_lf(4542)=  33711.86892458853281460362059800072741915_dp
  t_lf(4543)=  33720.29026745429221790340394469585542580_dp
  t_lf(4544)=  33728.71183041469320518123679195711751931_dp
  t_lf(4545)=  33737.13361342130478438951224003110174751_dp
  t_lf(4546)=  33745.55561642571727290428799774748744560_dp
  t_lf(4547)=  33753.97783937954228346430056374242766911_dp
  t_lf(4548)=  33762.40028223441271012234750980592017586_dp
  t_lf(4549)=  33770.82294494198271420902427054968992012_dp
  t_lf(4550)=  33779.24582745392771030880186152260897030_dp
  t_lf(4551)=  33787.66892972194435224843196580360035076_dp
  t_lf(4552)=  33796.09225169775051909766584697735901894_dp
  t_lf(4553)=  33804.51579333308530118227356424612440580_dp
  t_lf(4554)=  33812.93955457970898610934998325120295919_dp
  t_lf(4555)=  33821.36353538940304480489409397101413063_dp
  t_lf(4556)=  33829.78773571397011756364816482816733306_dp
  t_lf(4557)=  33838.21215550523400011118327987651857543_dp
  t_lf(4558)=  33846.63679471503962967821782365035165680_dp
  t_lf(4559)=  33855.06165329525307108715549594182779223_dp
  t_lf(4560)=  33863.48673119776150285082945642969706538_dp
  t_lf(4561)=  33871.91202837447320328343921671201278307_dp
  t_lf(4562)=  33880.33754477731753662366691489828317645_dp
  t_lf(4563)=  33888.76328035824493916995962549218139077_dp
  t_lf(4564)=  33897.18923506922690542796437484466167351_dp
  t_lf(4565)=  33905.61540886225597427010254997914436090_dp
  t_lf(4566)=  33914.04180168934571510727040608538183280_dp
  t_lf(4567)=  33922.46841350253071407265239544674911999_dp
  t_lf(4568)=  33930.89524425386656021763405800706327886_dp
  t_lf(4569)=  33939.32229389542983171980123119767187437_dp
  t_lf(4570)=  33947.74956237931808210301235403350972212_dp
  t_lf(4571)=  33956.17704965764982646953065784815112693_dp
  t_lf(4572)=  33964.60475568256452774420305337262882387_dp
  t_lf(4573)=  33973.03268040622258293067254117099718752_dp
  t_lf(4574)=  33981.46082378080530937961098972733244786_dp
  t_lf(4575)=  33989.88918575851493106895914273413296359_dp
  t_lf(4576)=  33998.31776629157456489616073436095429433_dp
  t_lf(4577)=  34006.74656533222820698237760848463302766_dp
  t_lf(4578)=  34015.17558283274071898867275503866611131_dp
  t_lf(4579)=  34023.60481874539781444414819378926478058_dp
  t_lf(4580)=  34032.03427302250604508602465296933992967_dp
  t_lf(4581)=  34040.46394561639278721165000729924473973_dp
  t_lf(4582)=  34048.89383647940622804242345699454623883_dp
  t_lf(4583)=  34057.32394556391535209962244640646583610_dp
  t_lf(4584)=  34065.75427282230992759211933795996525866_dp
  t_lf(4585)=  34074.18481820700049281597487404780415192_dp
  t_lf(4586)=  34082.61558167041834256589547650630421871_dp
  t_lf(4587)=  34091.04656316501551455854145024006741750_dp
  t_lf(4588)=  34099.47776264326477586767317447855757501_dp
  t_lf(4589)=  34107.90918005765960937112238203731086362_dp
  t_lf(4590)=  34116.34081536071420020957564382063593147_dp
  t_lf(4591)=  34124.77266850496342225715719264104394764_dp
  t_lf(4592)=  34133.20473944296282460379823724335724123_dp
  t_lf(4593)=  34141.63702812728861804937993420852729131_dp
  t_lf(4594)=  34150.06953451053766160963720217369319328_dp
  t_lf(4595)=  34158.50225854532744903381057954097493127_dp
  t_lf(4596)=  34166.93520018429609533403334355796628036_dp
  t_lf(4597)=  34175.36835938010232332644112533791431619_dp
  t_lf(4598)=  34183.80173608542545018399127204719060490_dp
  t_lf(4599)=  34192.23533025296537400097922412191737861_dp
  t_lf(4600)=  34200.66914183544256036923919198455447971_dp
  t_lf(4601)=  34209.10317078559802896601643331492360375_dp
  t_lf(4602)=  34217.53741705619334015349844848858932225_dp
  t_lf(4603)=  34225.97188060001058158999242832877537385_dp
  t_lf(4604)=  34234.40656136985235485273630482611353957_dp
  t_lf(4605)=  34242.84145931854176207233077196354474519_dp
  t_lf(4606)=  34251.27657439892239257877966024166145493_dp
  t_lf(4607)=  34259.71190656385830955912606493274044351_dp
  t_lf(4608)=  34268.14745576623403672667164449970908403_dp
  t_lf(4609)=  34276.58322195895454500176652199935970372_dp
  t_lf(4610)=  34285.01920509494523920415723864731859446_dp
  t_lf(4611)=  34293.45540512715194475688022505563208954_dp
  t_lf(4612)=  34301.89182200854089440168827196239481756_dp
  t_lf(4613)=  34310.32845569209871492599749855665782139_dp
  t_lf(4614)=  34318.76530613083241390134233276095960131_dp
  t_lf(4615)=  34327.20237327776936643332603406826414165_dp
  t_lf(4616)=  34335.63965708595730192305430573990935976_dp
  t_lf(4617)=  34344.07715750846429084003955935640984182_dp
  t_lf(4618)=  34352.51487449837873150656341087366178637_dp
  t_lf(4619)=  34360.95280800880933689348500347330826407_dp
  t_lf(4620)=  34369.39095799288512142748276860778163958_dp
  t_lf(4621)=  34377.82932440375538780971725272788962405_dp
  t_lf(4622)=  34386.26790719458971384590265324379418769_dp
  t_lf(4623)=  34394.70670631857793928777472330889063226_dp
  t_lf(4624)=  34403.14572172893015268594272103046959160_dp
  t_lf(4625)=  34411.58495337887667825411309470117960169_dp
  t_lf(4626)=  34420.02440122166806274467261161124408512_dp
  t_lf(4627)=  34428.46406521057506233561865394316597212_dp
  t_lf(4628)=  34436.90394529888862952882442116831749308_dp
  t_lf(4629)=  34445.34404143991990005962679425840360757_dp
  t_lf(4630)=  34453.78435358700017981772463289434668172_dp
  t_lf(4631)=  34462.22488169348093177937529270070890849_dp
  t_lf(4632)=  34470.66562571273376295087716535538902267_dp
  t_lf(4633)=  34479.10658559815041132332606022204245193_dp
  t_lf(4634)=  34487.54776130314273283863326192652044503_dp
  t_lf(4635)=  34495.98915278114268836679311404864512490_dp
  t_lf(4636)=  34504.43075998560233069438799482687494632_dp
  t_lf(4637)=  34512.87258286999379152431856647590973218_dp
  t_lf(4638)=  34521.31462138780926848674719539607727694_dp
  t_lf(4639)=  34529.75687549256101216124245620847531979_dp
  t_lf(4640)=  34538.19934513778131311011264818135430188_dp
  t_lf(4641)=  34546.64203027702248892291626822115845245_dp
  t_lf(4642)=  34555.08493086385687127213740018603603844_dp
  t_lf(4643)=  34563.52804685187679298001399584052462363_dp
  t_lf(4644)=  34571.97137819469457509650703830755440091_dp
  t_lf(4645)=  34580.41492484594251398839859438793248898_dp
  t_lf(4646)=  34588.85868675927286843950677760811385082_dp
  t_lf(4647)=  34597.30266388835784676200565932437044415_dp
  t_lf(4648)=  34605.74685618688959391883818065547952403_dp
  t_lf(4649)=  34614.19126360858017865721013343680477965_dp
  t_lf(4650)=  34622.63588610716158065315329378618021441_dp
  t_lf(4651)=  34631.08072363638567766714580724636631152_dp
  t_lf(4652)=  34639.52577615002423271077793982007092440_dp
  t_lf(4653)=  34647.97104360186888122445132454165327882_dp
  t_lf(4654)=  34656.41652594573111826609984853469817617_dp
  t_lf(4655)=  34664.86222313544228571092034078669857681_dp
  t_lf(4656)=  34673.30813512485355946210123613115777050_dp
  t_lf(4657)=  34681.75426186783593667253740616355678611_dp
  t_lf(4658)=  34690.20060331828022297751936303086795460_dp
  t_lf(4659)=  34698.64715943009701973838505722467094306_dp
  t_lf(4660)=  34707.09393015721671129712250567548237131_dp
  t_lf(4661)=  34715.54091545358945224191150159068347986_dp
  t_lf(4662)=  34723.98811527318515468359267260046133473_dp
  t_lf(4663)=  34732.43552956999347554305216887550675439_dp
  t_lf(4664)=  34740.88315829802380384951027795687547117_dp
  t_lf(4665)=  34749.33100141130524804970227809245686679_dp
  t_lf(4666)=  34757.77905886388662332793985690594574489_dp
  t_lf(4667)=  34766.22733060983643893704143723311574347_dp
  t_lf(4668)=  34774.67581660324288554011976694658679496_dp
  t_lf(4669)=  34783.12451679821382256321514455420208242_dp
  t_lf(4670)=  34791.57343114887676555876266729762071561_dp
  t_lf(4671)=  34800.02255960937887357988190339682928419_dp
  t_lf(4672)=  34808.47190213388693656547740498301688617_dp
  t_lf(4673)=  34816.92145867658736273613849313668245489_dp
  t_lf(4674)=  34825.37122919168616600082676129998841829_dp
  t_lf(4675)=  34833.82121363340895337433975816227904940_dp
  t_lf(4676)=  34842.27141195600091240553932592538336163_dp
  t_lf(4677)=  34850.72182411372679861633308464085904874_dp
  t_lf(4678)=  34859.17245006087092295139756807474367615_dp
  t_lf(4679)=  34867.62328975173713923863153129669993349_dp
  t_lf(4680)=  34876.07434314064883166032796490971102085_dp
  t_lf(4681)=  34884.52561018194890223505336553373785435_dp
  t_lf(4682)=  34892.97709082999975831022282683202935793_dp
  t_lf(4683)=  34901.42878503918330006535953002211820339_dp
  t_lf(4684)=  34909.88069276390090802602722744497444416_dp
  t_lf(4685)=  34918.33281395857343058842432737536596054_dp
  t_lf(4686)=  34926.78514857764117155462820284422482621_dp
  t_lf(4687)=  34935.23769657556387767847836180977987490_dp
  t_lf(4688)=  34943.69045790682072622208713055842507936_dp
  t_lf(4689)=  34952.14343252591031252296651673878796723_dp
  t_lf(4690)=  34960.59662038735063757175993293327923392_dp
  t_lf(4691)=  34969.05002144567909560056747615058094444_dp
  t_lf(4692)=  34977.50363565545246168185347308010314594_dp
  t_lf(4693)=  34985.95746297124687933792501538544417162_dp
  t_lf(4694)=  34994.41150334765784816097022372836516593_dp
  t_lf(4695)=  35002.86575673930021144364499360777108963_dp
  t_lf(4696)=  35011.32022310080814382019699046971529201_dp
  t_lf(4697)=  35019.77490238683513891811567589454921579_dp
  t_lf(4698)=  35028.22979455205399702029716099605940678_dp
  t_lf(4699)=  35036.68489955115681273771269747480714634_dp
  t_lf(4700)=  35045.14021733885496269256963105394804846_dp
  t_lf(4701)=  35053.59574786987909321195365629059613677_dp
  t_lf(4702)=  35062.05149109897910803194122599934543951_dp
  t_lf(4703)=  35070.50744698092415601217098274690814564_dp
  t_lf(4704)=  35078.96361547050261886086309407800791364_dp
  t_lf(4705)=  35087.41999652252209887027538731271601158_dp
  t_lf(4706)=  35095.87659009180940666258519391437251568_dp
  t_lf(4707)=  35104.33339613321054894618582756513066238_dp
  t_lf(4708)=  35112.79041460159071628238663420303542182_dp
  t_lf(4709)=  35121.24764545183427086250556637043315919_dp
  t_lf(4710)=  35129.70508863884473429534324829844352435_dp
  t_lf(4711)=  35138.16274411754477540502751220624304229_dp
  t_lf(4712)=  35146.62061184287619803921740032704778274_dp
  t_lf(4713)=  35155.07869176979992888765564118497541263_dp
  t_lf(4714)=  35163.53698385329600531105862263845025968_dp
  t_lf(4715)=  35171.99548804836356318033289817652405103_dp
  t_lf(4716)=  35180.45420431002082472610727690445498147_dp
  t_lf(4717)=  35188.91313259330508639856956158415388846_dp
  t_lf(4718)=  35197.37227285327270673759701300370367597_dp
  t_lf(4719)=  35205.83162504499909425316963283812177888_dp
  t_lf(4720)=  35214.29118912357869531605537103090037041_dp
  t_lf(4721)=  35222.75096504412498205875637757266009637_dp
  t_lf(4722)=  35231.21095276177044028670543237952521444_dp
  t_lf(4723)=  35239.67115223166655739970170077960590131_dp
  t_lf(4724)=  35248.13156340898381032357497590129187485_dp
  t_lf(4725)=  35256.59218624891165345206758302195500697_dp
  t_lf(4726)=  35265.05302070665850659892313468016185385_dp
  t_lf(4727)=  35273.51406673745174296017133907864451534_dp
  t_lf(4728)=  35281.97532429653767708659807800910440172_dp
  t_lf(4729)=  35290.43679333918155286638998421346271472_dp
  t_lf(4730)=  35298.89847382066753151794276175945805526_dp
  t_lf(4731)=  35307.36036569629867959282250665155980496_dp
  t_lf(4732)=  35315.82246892139695698886929852104997503_dp
  t_lf(4733)=  35324.28478345130320497343234784186019687_dp
  t_lf(4734)=  35332.74730924137713421672599670136849694_dp
  t_lf(4735)=  35341.21004624699731283529588471789644658_dp
  t_lf(4736)=  35349.67299442356115444558460523913512972_dp
  t_lf(4737)=  35358.13615372648490622758619047820199032_dp
  t_lf(4738)=  35366.59952411120363699857877774652380382_dp
  t_lf(4739)=  35375.06310553317122529692482242528749558_dp
  t_lf(4740)=  35383.52689794786034747592823677983397298_dp
  t_lf(4741)=  35391.99090131076246580773784716412415254_dp
  t_lf(4742)=  35400.45511557738781659728657558531448841_dp
  t_lf(4743)=  35408.91954070326539830625576500157502357_dp
  t_lf(4744)=  35417.38417664394295968705408110959970191_dp
  t_lf(4745)=  35425.84902335498698792680043674182975072_dp
  t_lf(4746)=  35434.31408079198269680130039833726965440_dp
  t_lf(4747)=  35442.77934891053401483900554727395481726_dp
  t_lf(4748)=  35451.24482766626357349494528215566361658_dp
  t_lf(4749)=  35459.71051701481269533462056143038727468_dp
  t_lf(4750)=  35468.17641691184138222784909898341186771_dp
  t_lf(4751)=  35476.64252731302830355255153859366081047_dp
  t_lf(4752)=  35485.10884817407078440846814636822622316_dp
  t_lf(4753)=  35493.57537945068479384079557347681654433_dp
  t_lf(4754)=  35502.04212109860493307373325469519839230_dp
  t_lf(4755)=  35510.50907307358442375392902143464571696_dp
  t_lf(4756)=  35518.97623533139509620381352108296138929_dp
  t_lf(4757)=  35527.44360782782737768481304761183814613_dp
  t_lf(4758)=  35535.91119051869028067043040151520978176_dp
  t_lf(4759)=  35544.37898335981139112918341023384213300_dp
  t_lf(4760)=  35552.84698630703685681739075329275915661_dp
  t_lf(4761)=  35561.31519931623137558179474943022460025_dp
  t_lf(4762)=  35569.78362234327818367201077602993671604_dp
  t_lf(4763)=  35578.25225534407904406279300418187438931_dp
  t_lf(4764)=  35586.72109827455423478610614569189012743_dp
  t_lf(4765)=  35595.19015109064253727299292133571068375_dp
  t_lf(4766)=  35603.65941374830122470522697260951173071_dp
  t_lf(4767)=  35612.12888620350605037674095216671093319_dp
  t_lf(4768)=  35620.59856841225123606481954104910593652_dp
  t_lf(4769)=  35629.06846033054946041104715372000204329_dp
  t_lf(4770)=  35637.53856191443184731200010478756051668_dp
  t_lf(4771)=  35646.00887311994795431967302416828426445_dp
  t_lf(4772)=  35654.47939390316576105162932028337481637_dp
  t_lf(4773)=  35662.95012422017165761086550370467463627_dp
  t_lf(4774)=  35671.42106402707043301537919647208347908_dp
  t_lf(4775)=  35679.89221327998526363743066509073822151_dp
  t_lf(4776)=  35688.36357193505770165248772798390381412_dp
  t_lf(4777)=  35696.83513994844766349784390092647011259_dp
  t_lf(4778)=  35705.30691727633341834089965671421667976_dp
  t_lf(4779)=  35713.77890387491157655709668803562648021_dp
  t_lf(4780)=  35722.25109970039707821749507520603093008_dp
  t_lf(4781)=  35730.72350470902318158598327309828417176_dp
  t_lf(4782)=  35739.19611885704145162611084426002481261_dp
  t_lf(4783)=  35747.66894210072174851753387784491973779_dp
  t_lf(4784)=  35756.14197439635221618206304660412795878_dp
  t_lf(4785)=  35764.61521570023927081930426678460371303_dp
  t_lf(4786)=  35773.08866596870758945188193836280804991_dp
  t_lf(4787)=  35781.56232515810009848023475560594772908_dp
  t_lf(4788)=  35790.03619322477796224697409049804016689_dp
  t_lf(4789)=  35798.51027012512057161079496409494408386_dp
  t_lf(4790)=  35806.98455581552553252992963338102806468_dp
  t_lf(4791)=  35815.45905025240865465513383369040401469_dp
  t_lf(4792)=  35823.93375339220393993219572922766000108_dp
  t_lf(4793)=  35832.40866519136357121395763667681766316_dp
  t_lf(4794)=  35840.88378560635790088184059932284366568_dp
  t_lf(4795)=  35849.35911459367543947686190152749289844_dp
  t_lf(4796)=  35857.83465210982284434013562580058358252_dp
  t_lf(4797)=  35866.31039811132490826284636708903136013_dp
  t_lf(4798)=  35874.78635255472454814568623126913099611_dp
  t_lf(4799)=  35883.26251539658279366774525717270062349_dp
  t_lf(4800)=  35891.73888659347877596484541380482458388_dp
  t_lf(4801)=  35900.21546610200971631730833672007685275_dp
  t_lf(4802)=  35908.69225387879091484714697981530774824_dp
  t_lf(4803)=  35917.16924988045573922467137107036199330_dp
  t_lf(4804)=  35925.64645406365561338449867302349607217_dp
  t_lf(4805)=  35934.12386638506000625095776100580697423_dp
  t_lf(4806)=  35942.60148680135642047287854437870257714_dp
  t_lf(4807)=  35951.07931526925038116775626822036575609_dp
  t_lf(4808)=  35959.55735174546542467528104509131943193_dp
  t_lf(4809)=  35968.03559618674308732022287867561774568_dp
  t_lf(4810)=  35976.51404854984289418466245324289887566_dp
  t_lf(4811)=  35984.99270879154234788955797500756714240_dp
  t_lf(4812)=  35993.47157686863691738563836357475537097_dp
  t_lf(4813)=  36001.95065273794002675361310375848233695_dp
  t_lf(4814)=  36010.42993635628304401368908013559379558_dp
  t_lf(4815)=  36018.90942768051526994438472875968831163_dp
  t_lf(4816)=  36027.38912666750392691063185250231004466_dp
  t_lf(4817)=  36035.86903327413414770115545851426891988_dp
  t_lf(4818)=  36044.34914745730896437512198830805329402_dp
  t_lf(4819)=  36052.82946917394929711804632295296031969_dp
  t_lf(4820)=  36061.30999838099394310694795784781367620_dp
  t_lf(4821)=  36069.79073503539956538474675349199607403_dp
  t_lf(4822)=  36078.27167909414068174388868061402380105_dp
  t_lf(4823)=  36086.75283051420965361919198993806135715_dp
  t_lf(4824)=  36095.23418925261667498990424877264466102_dp
  t_lf(4825)=  36103.71575526638976129096069849248009506_dp
  t_lf(4826)=  36112.19752851257473833343439885354241756_dp
  t_lf(4827)=  36120.67950894823523123416863693383589362_dp
  t_lf(4828)=  36129.16169653045265335458209032713840739_dp
  t_lf(4829)=  36137.64409121632619524863724603484629039_dp
  t_lf(4830)=  36146.12669296297281361996258830170655615_dp
  t_lf(4831)=  36154.60950172752722028811908042479153766_dp
  t_lf(4832)=  36163.09251746714187116400147733156689679_dp
  t_lf(4833)=  36171.57574013898695523436501747235587608_dp
  t_lf(4834)=  36180.05916970025038355546805430493870352_dp
  t_lf(4835)=  36188.54280610813777825582119936447439687_dp
  t_lf(4836)=  36197.02664931987246154803356061042095153_dp
  t_lf(4837)=  36205.51069929269544474974667142368708644_dp
  t_lf(4838)=  36213.99495598386541731364671729190236822_dp
  t_lf(4839)=  36222.47941935065873586654567886847058107_dp
  t_lf(4840)=  36230.96408935036941325752202172200155701_dp
  t_lf(4841)=  36239.44896594030910761511157470682717154_dp
  t_lf(4842)=  36247.93404907780711141353925048262563664_dp
  t_lf(4843)=  36256.41933872021034054798227329173232761_dp
  t_lf(4844)=  36264.90483482488332341885559066653285194_dp
  t_lf(4845)=  36273.39053734920819002511015728644254536_dp
  t_lf(4846)=  36281.87644625058466106653479073440364948_dp
  t_lf(4847)=  36290.36256148643003705505231041660462254_dp
  t_lf(4848)=  36298.84888301417918743500068240627284368_dp
  t_lf(4849)=  36307.33541079128453971238990445293982642_dp
  t_lf(4850)=  36315.82214477521606859312537686255433998_dp
  t_lf(4851)=  36324.30908492346128513018851640125088015_dp
  t_lf(4852)=  36332.79623119352522587976538180649601572_dp
  t_lf(4853)=  36341.28358354293044206631409090376049196_dp
  t_lf(4854)=  36349.77114192921698875656182072482777941_dp
  t_lf(4855)=  36358.25890630994241404242219340537714448_dp
  t_lf(4856)=  36366.74687664268174823282386200459836595_dp
  t_lf(4857)=  36375.23505288502749305444112173833295916_dp
  t_lf(4858)=  36383.72343499458961086131738344962017500_dp
  t_lf(4859)=  36392.21202292899551385337235745658204370_dp
  t_lf(4860)=  36400.70081664589005330378380721733721300_dp
  t_lf(4861)=  36409.18981610293550879523474353511511482_dp
  t_lf(4862)=  36417.67902125781157746501694129397686583_dp
  t_lf(4863)=  36426.16843206821536325898167196656399314_dp
  t_lf(4864)=  36434.65804849186136619432855637011725922_dp
  t_lf(4865)=  36443.14787048648147163122345336566217037_dp
  t_lf(4866)=  36451.63789800982493955323631139777177313_dp
  t_lf(4867)=  36460.12813101965839385658992095871760559_dp
  t_lf(4868)=  36468.61856947376581164821051723113265975_dp
  t_lf(4869)=  36477.10921332994851255257119331756236130_dp
  t_lf(4870)=  36485.60006254602514802731909560349727101_dp
  t_lf(4871)=  36494.09111707983169068767738392269079433_dp
  t_lf(4872)=  36502.58237688922142363961295029979294107_dp
  t_lf(4873)=  36511.07384193206492982176090113560334608_dp
  t_lf(4874)=  36519.56551216625008135609681877458953515_dp
  t_lf(4875)=  36528.05738754968202890734782945275594238_dp
  t_lf(4876)=  36536.54946804028319105113351566651154941_dp
  t_lf(4877)=  36545.04175359599324365082772202989527198_dp
  t_lf(4878)=  36553.53424417476910924313231469840436199_dp
  t_lf(4879)=  36562.02693973458494643235396543275807585_dp
  t_lf(4880)=  36570.51984023343213929337504235524258450_dp
  t_lf(4881)=  36579.01294562931928678330970041484942302_dp
  t_lf(4882)=  36587.50625588027219216183627552526450423_dp
  t_lf(4883)=  36595.99977094433385242019709727191361322_dp
  t_lf(4884)=  36604.49349077956444771885684600074906852_dp
  t_lf(4885)=  36612.98741534404133083381059100229654727_dp
  t_lf(4886)=  36621.48154459585901661153265738969654328_dp
  t_lf(4887)=  36629.97587849312917143255748013909712932_dp
  t_lf(4888)=  36638.47041699398060268368361461480915134_dp
  t_lf(4889)=  36646.96516005655924823879208374014716959_dp
  t_lf(4890)=  36655.46010763902816594827025279787480938_dp
  t_lf(4891)=  36663.95525969956752313703243365167707470_dp
  t_lf(4892)=  36672.45061619637458611112843097211994842_dp
  t_lf(4893)=  36680.94617708766370967293125382715454133_dp
  t_lf(4894)=  36689.44194233166632664489522675840440341_dp
  t_lf(4895)=  36697.93791188663093740187574521026556957_dp
  t_lf(4896)=  36706.43408571082309941200193090927463035_dp
  t_lf(4897)=  36714.93046376252541678609345350628569843_dp
  t_lf(4898)=  36723.42704600003752983561279549376764313_dp
  t_lf(4899)=  36731.92383238167610463914424809501339905_dp
  t_lf(4900)=  36740.42082286577482261739093649126848692_dp
  t_lf(4901)=  36748.91801741068437011668118340676103574_dp
  t_lf(4902)=  36757.41541597477242800097553071037543859_dp
  t_lf(4903)=  36765.91301851642366125236574931628113979_dp
  t_lf(4904)=  36774.41082499403970858005717827423172043_dp
  t_lf(4905)=  36782.90883536603917203782574453351216113_dp
  t_lf(4906)=  36791.40704959085760664994102544265860702_dp
  t_lf(4907)=  36799.90546762694751004554672661012978732_dp
  t_lf(4908)=  36808.40408943277831210148995829909705225_dp
  t_lf(4909)=  36816.90291496683636459359070406246534016_dp
  t_lf(4910)=  36825.40194418762493085634288584216478902_dp
  t_lf(4911)=  36833.90117705366417545103844025968662553_dp
  t_lf(4912)=  36842.40061352349115384230583131280182408_dp
  t_lf(4913)=  36850.90025355565980208305443516642120369_dp
  t_lf(4914)=  36859.40009710874092650781624318365545571_dp
  t_lf(4915)=  36867.90014414132219343447633978633735778_dp
  t_lf(4916)=  36876.40039461200811887438362216260037197_dp
  t_lf(4917)=  36884.90084847942005825083323925259214746_dp
  t_lf(4918)=  36893.40150570219619612591223784206230515_dp
  t_lf(4919)=  36901.90236623899153593569991397742538293_dp
  t_lf(4920)=  36910.40343004847788973381437828498603357_dp
  t_lf(4921)=  36918.90469708934386794329685413134851400_dp
  t_lf(4922)=  36927.40616732029486911682523790164016369_dp
  t_lf(4923)=  36935.90784070005306970524846099708287585_dp
  t_lf(4924)=  36944.40971718735741383443320346367140902_dp
  t_lf(4925)=  36952.91179674096360309041451945928661528_dp
  t_lf(4926)=  36961.41407931964408631284194504750907885_dp
  t_lf(4927)=  36969.91656488218804939671266907272802536_dp
  t_lf(4928)=  36978.41925338740140510238335812288539380_dp
  t_lf(4929)=  36986.92214479410678287385223682337933448_dp
  t_lf(4930)=  36995.42523906114351866530303492829873683_dp
  t_lf(4931)=  37003.92853614736764477590242288329428756_dp
  t_lf(4932)=  37012.43203601165187969284256772803555750_dp
  t_lf(4933)=  37020.93573861288561794262045138538121426_dp
  t_lf(4934)=  37029.43964390997491995054560354912411701_dp
  t_lf(4935)=  37037.94375186184250190846791153248818379_dp
  t_lf(4936)=  37046.44806242742772565071717957547290432_dp
  t_lf(4937)=  37054.95257556568658853824612023068753157_dp
  t_lf(4938)=  37063.45729123559171335096847055451361199_dp
  t_lf(4939)=  37071.96220939613233818828393592330485129_dp
  t_lf(4940)=  37080.46733000631430637778167437290056421_dp
  t_lf(4941)=  37088.97265302516005639211404442401628177_dp
  t_lf(4942)=  37097.47817841170861177403234940610560807_dp
  t_lf(4943)=  37105.98390612501557106957632132808420556_dp
  t_lf(4944)=  37114.48983612415309776940909736589287835_dp
  t_lf(4945)=  37122.99596836820991025828945204427510915_dp
  t_lf(4946)=  37131.50230281629127177267305818337803706_dp
  t_lf(4947)=  37140.00883942751898036643455965987765102_dp
  t_lf(4948)=  37148.51557816103135888470224899730178154_dp
  t_lf(4949)=  37157.02251897598324494579715275110113015_dp
  t_lf(4950)=  37165.52966183154598093126833759082186271_dp
  t_lf(4951)=  37174.03700668690740398401625990448595402_dp
  t_lf(4952)=  37182.54455350127183601449599165901020745_dp
  t_lf(4953)=  37191.05230223386007371499216514521434415_dp
  t_lf(4954)=  37199.56025284390937858195748911670538138_dp
  t_lf(4955)=  37208.06840529067346694640669869870227579_dp
  t_lf(4956)=  37216.57675953342250001235781129570403199_dp
  t_lf(4957)=  37225.08531553144307390331257056582866483_dp
  t_lf(4958)=  37233.59407324403820971676797035468201131_dp
  t_lf(4959)=  37242.10303263052734358675076029277682786_dp
  t_lf(4960)=  37250.61219365024631675436684455783625555_dp
  t_lf(4961)=  37259.12155626254736564635749508680392174_dp
  t_lf(4962)=  37267.63112042679911196165431029206796445_dp
  t_lf(4963)=  37276.14088610238655276592486009231036728_dp
  t_lf(4964)=  37284.65085324871105059410096781053838953_dp
  t_lf(4965)=  37293.16102182519032356088158922026373968_dp
  t_lf(4966)=  37301.67139179125843547920225873548960200_dp
  t_lf(4967)=  37310.18196310636578598666308244116777598_dp
  t_lf(4968)=  37318.69273572997910067990726734812007892_dp
  t_lf(4969)=  37327.20370962158142125694218593010180285_dp
  t_lf(4970)=  37335.71488474067209566739498466074238088_dp
  t_lf(4971)=  37344.22626104676676827069475491455143485_dp
  t_lf(4972)=  37352.73783849939737000217329422904893896_dp
  t_lf(4973)=  37361.24961705811210854707649554438819419_dp
  t_lf(4974)=  37369.76159668247545852247841164261147845_dp
  t_lf(4975)=  37378.27377733206815166709005160093239093_dp
  t_lf(4976)=  37386.78615896648716703895497565219777948_dp
  t_lf(4977)=  37395.29874154534572122102376441096742154_dp
  t_lf(4978)=  37403.81152502827325853459944797548297753_dp
  t_lf(4979)=  37412.32450937491544126064598995420076639_dp
  t_lf(4980)=  37420.83769454493413986895193099055720333_dp
  t_lf(4981)=  37429.35108049800742325514130587124282805_dp
  t_lf(4982)=  37437.86466719382954898552395780150223610_dp
  t_lf(4983)=  37446.37845459211095354977738291587436739_dp
  t_lf(4984)=  37454.89244265257824262145224756436192605_dp
  t_lf(4985)=  37463.40663133497418132629373037229158620_dp
  t_lf(4986)=  37471.92102059905768451837085051711942501_dp
  t_lf(4987)=  37480.43561040460380706400595309717002121_dp
  t_lf(4988)=  37488.95040071140373413349653188579413299_dp
  t_lf(4989)=  37497.46539147926477150062157916971005184_dp
  t_lf(4990)=  37505.98058266801033584992466176237880907_dp
  t_lf(4991)=  37514.49597423747994509176593166217454213_dp
  t_lf(4992)=  37523.01156614752920868513528919086962242_dp
  t_lf(4993)=  37531.52735835802981796821892580058068125_dp
  t_lf(4994)=  37540.04335082886953649671148307683748416_dp
  t_lf(4995)=  37548.55954351995219038986607379186269608_dp
  t_lf(4996)=  37557.07593639119765868427442017550791347_dp
  t_lf(4997)=  37565.59252940254186369536937387160083909_dp
  t_lf(4998)=  37574.10932251393676138664209133474102736_dp
  t_lf(4999)=  37582.62631568535033174656614769685808280_dp
  t_lf(5000)=  37591.14350887676656917322088139413736307_dp
  t_lf(5001)=  37599.66090204818547286660627109324489481_dp
  t_lf(5002)=  37608.17849515962303722864165569116609372_dp
  t_lf(5003)=  37616.69628817111124227084061738543268741_dp

ENDIF

lfact = t_lf(n)

RETURN
END FUNCTION lfact

!***********************************************************************
